REM CommandInterpreter: $(COMSPEC)
if not exist .\release\nul mkdir release
if not exist .\release\obj\nul mkdir release\obj
if not exist .\release\obj\libexpat\nul mkdir release\obj\libexpat
if not exist .\release\obj\libexpatw\nul mkdir release\obj\libexpatw
if not exist .\release\obj\libexpat_static\nul mkdir release\obj\libexpat_static
if not exist .\release\obj\libexpatw_static\nul mkdir release\obj\libexpatw_static
if not exist .\release\obj\examples\nul mkdir release\obj\examples
if not exist .\release\obj\xmlwf\nul mkdir release\obj\xmlwf
