/*
DO NOT EDIT!

This file is autogenerated by astropy/wcs/setup_package.py.  To edit
its contents, edit astropy/wcs/docstrings.py
*/

#ifndef __DOCSTRINGS_H__
#define __DOCSTRINGS_H__

extern char doc_DistortionLookupTable[371];
extern char doc_InconsistentAxisTypes[100];
extern char doc_InvalidCoordinate[76];
extern char doc_InvalidSubimageSpecification[77];
extern char doc_InvalidTabularParameters[76];
extern char doc_InvalidTransform[107];
extern char doc_K[202];
extern char doc_M[56];
extern char doc_NoSolution[68];
extern char doc_NoWcsKeywordsFound[76];
extern char doc_NonseparableSubimageCoordinateSystem[88];
extern char doc_SingularMatrix[70];
extern char doc_Sip[1040];
extern char doc_Tabprm[234];
extern char doc_Wcs[513];
extern char doc_WcsError[39];
extern char doc_Wcsprm[2390];
extern char doc_Wtbarr[202];
extern char doc_a[264];
extern char doc_a_order[60];
extern char doc_all_pix2world[1275];
extern char doc_alt[236];
extern char doc_ap[276];
extern char doc_ap_order[61];
extern char doc_axis_types[916];
extern char doc_b[273];
extern char doc_b_order[60];
extern char doc_bounds_check[469];
extern char doc_bp[276];
extern char doc_bp_order[61];
extern char doc_cd[1059];
extern char doc_cdelt[308];
extern char doc_cdfix[480];
extern char doc_cel_offset[176];
extern char doc_celfix[171];
extern char doc_cname[76];
extern char doc_colax[91];
extern char doc_colnum[290];
extern char doc_compare[1271];
extern char doc_convert[121];
extern char doc_coord[277];
extern char doc_copy[40];
extern char doc_cpdis1[106];
extern char doc_cpdis2[106];
extern char doc_crder[123];
extern char doc_crota[1055];
extern char doc_crpix[88];
extern char doc_crval[93];
extern char doc_crval_tabprm[94];
extern char doc_csyer[132];
extern char doc_ctype[228];
extern char doc_cubeface[1236];
extern char doc_cunit[1323];
extern char doc_cylfix[162];
extern char doc_data[78];
extern char doc_data_wtbarr[52];
extern char doc_dateavg[152];
extern char doc_dateobs[133];
extern char doc_datfix[512];
extern char doc_delta[238];
extern char doc_det2im[58];
extern char doc_det2im1[103];
extern char doc_det2im2[103];
extern char doc_dims[98];
extern char doc_equinox[243];
extern char doc_extlev[73];
extern char doc_extnam[74];
extern char doc_extrema[439];
extern char doc_extver[73];
extern char doc_find_all_wcs[866];
extern char doc_fix[1686];
extern char doc_get_cdelt[453];
extern char doc_get_offset[206];
extern char doc_get_pc[392];
extern char doc_get_ps[433];
extern char doc_get_pv[716];
extern char doc_has_cd[911];
extern char doc_has_cdi_ja[103];
extern char doc_has_crota[652];
extern char doc_has_crotaia[107];
extern char doc_has_pc[213];
extern char doc_has_pci_ja[103];
extern char doc_i[41];
extern char doc_imgpix_matrix[173];
extern char doc_is_unity[109];
extern char doc_kind[130];
extern char doc_lat[86];
extern char doc_latpole[75];
extern char doc_lattyp[236];
extern char doc_lng[87];
extern char doc_lngtyp[237];
extern char doc_lonpole[77];
extern char doc_m[59];
extern char doc_map[967];
extern char doc_mix[4806];
extern char doc_mjdavg[181];
extern char doc_mjdobs[181];
extern char doc_name[74];
extern char doc_naxis[761];
extern char doc_nc[172];
extern char doc_ndim[67];
extern char doc_obsgeo[187];
extern char doc_p0[244];
extern char doc_p2s[2006];
extern char doc_p4_pix2foc[649];
extern char doc_pc[1126];
extern char doc_phi0[289];
extern char doc_pix2foc[633];
extern char doc_piximg_matrix[128];
extern char doc_print_contents[210];
extern char doc_print_contents_tabprm[211];
extern char doc_radesys[77];
extern char doc_restfrq[94];
extern char doc_restwav[94];
extern char doc_row[40];
extern char doc_s2p[1910];
extern char doc_sense[232];
extern char doc_set[1206];
extern char doc_set_ps[414];
extern char doc_set_pv[409];
extern char doc_set_tabprm[325];
extern char doc_sip[84];
extern char doc_sip_foc2pix[649];
extern char doc_sip_pix2foc[649];
extern char doc_spcfix[272];
extern char doc_spec[68];
extern char doc_specsys[145];
extern char doc_sptr[1370];
extern char doc_ssysobs[253];
extern char doc_ssyssrc[150];
extern char doc_sub[3440];
extern char doc_tab[112];
extern char doc_theta0[289];
extern char doc_to_header[2224];
extern char doc_ttype[108];
extern char doc_unitfix[1047];
extern char doc_velangl[186];
extern char doc_velosys[316];
extern char doc_velref[55];
extern char doc_wcs[83];
extern char doc_zsource[98];

#endif

