.. Known projects

.. PROJECTNAME placeholders
.. _PROJECTNAME: http://neuroimaging.scipy.org
.. _`PROJECTNAME github`: https://github.com/nipy
.. _`PROJECTNAME mailing list`: http://projects.scipy.org/mailman/listinfo/nipy-devel

.. numpy
.. _numpy: http://www.numpy.org/
.. _`numpy github`: https://github.com/numpy/numpy
.. _`numpy mailing list`: http://mail.python.org/mailman/listinfo/numpy-discussion

.. scipy
.. _scipy: https://www.scipy.org/
.. _`scipy github`: https://github.com/scipy/scipy
.. _`scipy mailing list`: http://mail.python.org/mailman/listinfo/scipy-dev

.. ipython
.. _ipython: http://ipython.scipy.org
.. _`ipython github`: https://github.com/ipython/ipython
.. _`ipython mailing list`: http://mail.python.org/mailman/listinfo/IPython-dev

.. pip
.. _pip: https://pip.pypa.io/en/latest/

.. virtualenv
.. _virtualenv: https://pypi.python.org/pypi/virtualenv
.. _virtualenvwrapper: https://pypi.python.org/pypi/virtualenvwrapper
.. _virtualenvwrapper-csh: https://bitbucket.org/embray/virtualenvwrapper-csh
.. _virtualenvwrapper-win: https://github.com/davidmarble/virtualenvwrapper-win
.. _venv: https://docs.python.org/dev/library/venv.html

.. conda
.. _conda: https://conda.io/docs/

.. pytest
.. _py.test: https://pytest.org/en/latest/

.. pychecker
.. _pychecker: https://pypi.python.org/pypi/PyChecker
