/* Generated by Cython 0.29.10 */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_10"
#define CYTHON_HEX_VERSION 0x001D0AF0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX < 0x030800A4
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
  #define PyObject_Unicode             PyObject_Str
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : (Py_INCREF(func), func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__astropy__io__ascii__cparser
#define __PYX_HAVE_API__astropy__io__ascii__cparser
/* Early includes */
#include <string.h>
#include <stdio.h>
#include "numpy/arrayobject.h"
#include "numpy/ufuncobject.h"
#include "src/tokenizer.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "astropy/io/ascii/cparser.pyx",
  "stringsource",
  "__init__.pxd",
  "type.pxd",
};

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":776
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":777
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":778
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":779
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":783
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":784
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":785
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":786
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":790
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":791
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":800
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */
typedef npy_long __pyx_t_5numpy_int_t;

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":801
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":802
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":804
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":805
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":806
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":808
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":809
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":811
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":812
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":813
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;
/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);


/*--- Type declarations ---*/
struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString;
struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser;
struct __pyx_obj_7astropy_2io_5ascii_7cparser_FastWriter;
struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines;
struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel;
struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr;
struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr;
struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr;
struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk;
struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr;

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":815
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":816
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":817
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":819
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */
typedef npy_cdouble __pyx_t_5numpy_complex_t;

/* "astropy/io/ascii/cparser.pyx":116
 *     ]))
 * 
 * cdef class FileString:             # <<<<<<<<<<<<<<
 *     """
 *     A wrapper class for a memory-mapped file pointer.
 */
struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString {
  PyObject_HEAD
  PyObject *fhandle;
  PyObject *mmap;
  void const *mmap_ptr;
  Py_buffer buf;
};


/* "astropy/io/ascii/cparser.pyx":165
 *             ptr = tmp
 * 
 * cdef class CParser:             # <<<<<<<<<<<<<<
 *     """
 *     A fast Cython parser class which uses underlying C code
 */
struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser {
  PyObject_HEAD
  struct __pyx_vtabstruct_7astropy_2io_5ascii_7cparser_CParser *__pyx_vtab;
  tokenizer_t *tokenizer;
  PyObject *names;
  PyObject *header_names;
  int data_start;
  PyObject *data_end;
  PyObject *include_names;
  PyObject *exclude_names;
  PyObject *fill_values;
  PyObject *fill_empty;
  PyObject *fill_include_names;
  PyObject *fill_exclude_names;
  PyObject *fill_names;
  int fill_extra_cols;
  PyObject *source_bytes;
  char *source_ptr;
  PyObject *parallel;
  PyObject *use_cols;
  int width;
  PyObject *source;
  PyObject *header_start;
};


/* "astropy/io/ascii/cparser.pyx":856
 *     reconvert_queue.put(reconvert_cols) # return to the queue for other processes
 * 
 * cdef class FastWriter:             # <<<<<<<<<<<<<<
 *     """
 *     A fast Cython writing class for writing tables
 */
struct __pyx_obj_7astropy_2io_5ascii_7cparser_FastWriter {
  PyObject_HEAD
  struct __pyx_vtabstruct_7astropy_2io_5ascii_7cparser_FastWriter *__pyx_vtab;
  PyObject *table;
  PyObject *use_names;
  PyObject *fill_values;
  PyObject *fill_cols;
  PyObject *col_iters;
  PyObject *formats;
  PyObject *format_funcs;
  PyObject *types;
  PyObject *line_comments;
  PyObject *quotechar;
  PyObject *expchar;
  PyObject *delimiter;
  int strip_whitespace;
  PyObject *comment;
};


/* "astropy/io/ascii/cparser.pyx":151
 *         return self.mmap[i]
 * 
 *     def splitlines(self):             # <<<<<<<<<<<<<<
 *         """
 *         Return a generator yielding lines from the memory map.
 */
struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines {
  PyObject_HEAD
  size_t __pyx_v_line_len;
  size_t __pyx_v_map_len;
  char *__pyx_v_ptr;
  struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString *__pyx_v_self;
  char *__pyx_v_tmp;
};


/* "astropy/io/ascii/cparser.pyx":400
 *                                   try_string, num_rows)
 * 
 *     def _read_parallel(self, try_int, try_float, try_string):             # <<<<<<<<<<<<<<
 *         cdef size_t source_len = <size_t>len(self.source)
 *         self.tokenizer.source_pos = 0
 */
struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel {
  PyObject_HEAD
  PyObject *__pyx_v_col_chunks;
  struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self;
};


/* "astropy/io/ascii/cparser.pyx":418
 * 
 *         if offset == source_len: # no data
 *             return (dict((name, np.array([], dtype=np.int_)) for name in             # <<<<<<<<<<<<<<
 *                          self.names),
 *                     self._get_comments(self.tokenizer))
 */
struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr {
  PyObject_HEAD
  struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel *__pyx_outer_scope;
  PyObject *__pyx_v_name;
};


/* "astropy/io/ascii/cparser.pyx":525
 * 
 *             if self.data_end < 0: # no data
 *                 chunks = [dict((name, []) for name in self.names)]             # <<<<<<<<<<<<<<
 *             else:
 *                 line_no = 0
 */
struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr {
  PyObject_HEAD
  struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel *__pyx_outer_scope;
  PyObject *__pyx_v_name;
};


/* "astropy/io/ascii/cparser.pyx":541
 *         for name in self.get_names():
 *             col_chunks = [chunk.pop(name) for chunk in chunks]
 *             if any(isinstance(col_chunk, ma.masked_array) for col_chunk in col_chunks):             # <<<<<<<<<<<<<<
 *                 ret[name] = ma.concatenate(col_chunks)
 *             else:
 */
struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr {
  PyObject_HEAD
  struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel *__pyx_outer_scope;
  PyObject *__pyx_v_col_chunk;
};


/* "astropy/io/ascii/cparser.pyx":819
 *     return parser
 * 
 * def _read_chunk(CParser self, start, end, try_int,             # <<<<<<<<<<<<<<
 *                 try_float, try_string, queue, reconvert_queue, i):
 *     cdef tokenizer_t *chunk_tokenizer = self.tokenizer
 */
struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk {
  PyObject_HEAD
  struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self;
};


/* "astropy/io/ascii/cparser.pyx":832
 *         err = (chunk_tokenizer.code, chunk_tokenizer.num_rows)
 *     if chunk_tokenizer.num_rows == 0: # no data
 *         data = dict((name, np.array([], np.int_)) for name in self.get_names())             # <<<<<<<<<<<<<<
 *         line_comments = self._get_comments(chunk_tokenizer)
 *     else:
 */
struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr {
  PyObject_HEAD
  struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk *__pyx_outer_scope;
  PyObject *__pyx_v_name;
};



/* "astropy/io/ascii/cparser.pyx":165
 *             ptr = tmp
 * 
 * cdef class CParser:             # <<<<<<<<<<<<<<
 *     """
 *     A fast Cython parser class which uses underlying C code
 */

struct __pyx_vtabstruct_7astropy_2io_5ascii_7cparser_CParser {
  PyObject *(*get_error)(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *, PyObject *, PyObject *, PyObject *);
  PyObject *(*raise_error)(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *, PyObject *);
  PyObject *(*setup_tokenizer)(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_set_fill_values)(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *);
  PyObject *(*_get_comments)(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *, tokenizer_t *);
  PyObject *(*_convert_data)(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *, tokenizer_t *, PyObject *, PyObject *, PyObject *, PyObject *);
  PyArrayObject *(*_convert_int)(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *, tokenizer_t *, int, int);
  PyArrayObject *(*_convert_float)(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *, tokenizer_t *, int, int);
  PyObject *(*_convert_str)(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *, tokenizer_t *, int, int);
};
static struct __pyx_vtabstruct_7astropy_2io_5ascii_7cparser_CParser *__pyx_vtabptr_7astropy_2io_5ascii_7cparser_CParser;


/* "astropy/io/ascii/cparser.pyx":856
 *     reconvert_queue.put(reconvert_cols) # return to the queue for other processes
 * 
 * cdef class FastWriter:             # <<<<<<<<<<<<<<
 *     """
 *     A fast Cython writing class for writing tables
 */

struct __pyx_vtabstruct_7astropy_2io_5ascii_7cparser_FastWriter {
  PyObject *(*_write_comments)(struct __pyx_obj_7astropy_2io_5ascii_7cparser_FastWriter *, PyObject *);
};
static struct __pyx_vtabstruct_7astropy_2io_5ascii_7cparser_FastWriter *__pyx_vtabptr_7astropy_2io_5ascii_7cparser_FastWriter;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* IncludeStringH.proto */
#include <string.h>

/* decode_c_string_utf16.proto */
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 0;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16LE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = -1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16BE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}

/* decode_c_string.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* StrEquals.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyString_Equals __Pyx_PyUnicode_Equals
#else
#define __Pyx_PyString_Equals __Pyx_PyBytes_Equals
#endif

/* UnicodeAsUCS4.proto */
static CYTHON_INLINE Py_UCS4 __Pyx_PyUnicode_AsPy_UCS4(PyObject*);

/* object_ord.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyObject_Ord(c)\
    (likely(PyUnicode_Check(c)) ? (long)__Pyx_PyUnicode_AsPy_UCS4(c) : __Pyx__PyObject_Ord(c))
#else
#define __Pyx_PyObject_Ord(c) __Pyx__PyObject_Ord(c)
#endif
static long __Pyx__PyObject_Ord(PyObject* c);

/* CallableCheck.proto */
#if CYTHON_USE_TYPE_SLOTS && PY_MAJOR_VERSION >= 3
#define __Pyx_PyCallable_Check(obj)   ((obj)->ob_type->tp_call != NULL)
#else
#define __Pyx_PyCallable_Check(obj)   PyCallable_Check(obj)
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* PyObjectLookupSpecial.proto */
#if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name) {
    PyObject *res;
    PyTypeObject *tp = Py_TYPE(obj);
#if PY_MAJOR_VERSION < 3
    if (unlikely(PyInstance_Check(obj)))
        return __Pyx_PyObject_GetAttrStr(obj, attr_name);
#endif
    res = _PyType_Lookup(tp, attr_name);
    if (likely(res)) {
        descrgetfunc f = Py_TYPE(res)->tp_descr_get;
        if (!f) {
            Py_INCREF(res);
        } else {
            res = f(res, obj, (PyObject *)tp);
        }
    } else {
        PyErr_SetObject(PyExc_AttributeError, attr_name);
    }
    return res;
}
#else
#define __Pyx_PyObject_LookupSpecial(o,n) __Pyx_PyObject_GetAttrStr(o,n)
#endif

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* StringJoin.proto */
#if PY_MAJOR_VERSION < 3
#define __Pyx_PyString_Join __Pyx_PyBytes_Join
#define __Pyx_PyBaseString_Join(s, v) (PyUnicode_CheckExact(s) ? PyUnicode_Join(s, v) : __Pyx_PyBytes_Join(s, v))
#else
#define __Pyx_PyString_Join PyUnicode_Join
#define __Pyx_PyBaseString_Join PyUnicode_Join
#endif
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION < 3
    #define __Pyx_PyBytes_Join _PyString_Join
    #else
    #define __Pyx_PyBytes_Join _PyBytes_Join
    #endif
#else
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values);
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* append.proto */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_SubtractObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname);

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* ListExtend.proto */
static CYTHON_INLINE int __Pyx_PyList_Extend(PyObject* L, PyObject* v) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject* none = _PyList_Extend((PyListObject*)L, v);
    if (unlikely(!none))
        return -1;
    Py_DECREF(none);
    return 0;
#else
    return PyList_SetSlice(L, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, v);
#endif
}

/* PyDictContains.proto */
static CYTHON_INLINE int __Pyx_PyDict_ContainsTF(PyObject* item, PyObject* dict, int eq) {
    int result = PyDict_Contains(dict, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* SliceObject.proto */
#define __Pyx_PyObject_DelSlice(obj, cstart, cstop, py_start, py_stop, py_slice, has_cstart, has_cstop, wraparound)\
    __Pyx_PyObject_SetSlice(obj, (PyObject*)NULL, cstart, cstop, py_start, py_stop, py_slice, has_cstart, has_cstop, wraparound)
static CYTHON_INLINE int __Pyx_PyObject_SetSlice(
        PyObject* obj, PyObject* value, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* pyfrozenset_new.proto */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it);

/* PySetContains.proto */
static CYTHON_INLINE int __Pyx_PySet_ContainsTF(PyObject* key, PyObject* set, int eq);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* pop.proto */
static CYTHON_INLINE PyObject* __Pyx__PyObject_Pop(PyObject* L);
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE PyObject* __Pyx_PyList_Pop(PyObject* L);
#define __Pyx_PyObject_Pop(L) (likely(PyList_CheckExact(L)) ?\
    __Pyx_PyList_Pop(L) : __Pyx__PyObject_Pop(L))
#else
#define __Pyx_PyList_Pop(L)  __Pyx__PyObject_Pop(L)
#define __Pyx_PyObject_Pop(L)  __Pyx__PyObject_Pop(L)
#endif

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod0.proto */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CallUnboundCMethod0(cfunc, self)\
    (likely((cfunc)->func) ?\
        (likely((cfunc)->flag == METH_NOARGS) ?  (*((cfunc)->func))(self, NULL) :\
         (PY_VERSION_HEX >= 0x030600B1 && likely((cfunc)->flag == METH_FASTCALL) ?\
            (PY_VERSION_HEX >= 0x030700A0 ?\
                (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0) :\
                (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL)) :\
          (PY_VERSION_HEX >= 0x030700A0 && (cfunc)->flag == (METH_FASTCALL | METH_KEYWORDS) ?\
            (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL) :\
            (likely((cfunc)->flag == (METH_VARARGS | METH_KEYWORDS)) ?  ((*(PyCFunctionWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, __pyx_empty_tuple, NULL)) :\
               ((cfunc)->flag == METH_VARARGS ?  (*((cfunc)->func))(self, __pyx_empty_tuple) :\
               __Pyx__CallUnboundCMethod0(cfunc, self)))))) :\
        __Pyx__CallUnboundCMethod0(cfunc, self))
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* CallUnboundCMethod1.proto */
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#else
#define __Pyx_CallUnboundCMethod1(cfunc, self, arg)  __Pyx__CallUnboundCMethod1(cfunc, self, arg)
#endif

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* IterNext.proto */
#define __Pyx_PyIter_Next(obj) __Pyx_PyIter_Next2(obj, NULL)
static CYTHON_INLINE PyObject *__Pyx_PyIter_Next2(PyObject *, PyObject *);

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* SliceTupleAndList.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_GetSlice(PyObject* src, Py_ssize_t start, Py_ssize_t stop);
static CYTHON_INLINE PyObject* __Pyx_PyTuple_GetSlice(PyObject* src, Py_ssize_t start, Py_ssize_t stop);
#else
#define __Pyx_PyList_GetSlice(seq, start, stop)   PySequence_GetSlice(seq, start, stop)
#define __Pyx_PyTuple_GetSlice(seq, start, stop)  PySequence_GetSlice(seq, start, stop)
#endif

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunction.proto */
#define __Pyx_CyFunction_USED 1
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_Check(obj)  (__Pyx_TypeCheck(obj, __pyx_CyFunctionType))
#define __Pyx_CyFunction_NewEx(ml, flags, qualname, self, module, globals, code)\
    __Pyx_CyFunction_New(__pyx_CyFunctionType, ml, flags, qualname, self, module, globals, code)
static PyObject *__Pyx_CyFunction_New(PyTypeObject *, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_err_code(err_code value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_char(char value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__NPY_TYPES(enum NPY_TYPES value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* CoroutineBase.proto */
typedef PyObject *(*__pyx_coroutine_body_t)(PyObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) (Py_TYPE(obj) == __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static PyObject *__pyx_f_7astropy_2io_5ascii_7cparser_7CParser_get_error(CYTHON_UNUSED struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self, PyObject *__pyx_v_code, PyObject *__pyx_v_num_rows, PyObject *__pyx_v_msg); /* proto*/
static PyObject *__pyx_f_7astropy_2io_5ascii_7cparser_7CParser_raise_error(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self, PyObject *__pyx_v_msg); /* proto*/
static PyObject *__pyx_f_7astropy_2io_5ascii_7cparser_7CParser_setup_tokenizer(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self, PyObject *__pyx_v_source, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_7astropy_2io_5ascii_7cparser_7CParser__set_fill_values(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_7astropy_2io_5ascii_7cparser_7CParser__get_comments(CYTHON_UNUSED struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self, tokenizer_t *__pyx_v_t); /* proto*/
static PyObject *__pyx_f_7astropy_2io_5ascii_7cparser_7CParser__convert_data(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self, tokenizer_t *__pyx_v_t, PyObject *__pyx_v_try_int, PyObject *__pyx_v_try_float, PyObject *__pyx_v_try_string, PyObject *__pyx_v_num_rows); /* proto*/
static PyArrayObject *__pyx_f_7astropy_2io_5ascii_7cparser_7CParser__convert_int(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self, tokenizer_t *__pyx_v_t, int __pyx_v_i, int __pyx_v_nrows); /* proto*/
static PyArrayObject *__pyx_f_7astropy_2io_5ascii_7cparser_7CParser__convert_float(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self, tokenizer_t *__pyx_v_t, int __pyx_v_i, int __pyx_v_nrows); /* proto*/
static PyObject *__pyx_f_7astropy_2io_5ascii_7cparser_7CParser__convert_str(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self, tokenizer_t *__pyx_v_t, int __pyx_v_i, int __pyx_v_nrows); /* proto*/
static PyObject *__pyx_f_7astropy_2io_5ascii_7cparser_10FastWriter__write_comments(struct __pyx_obj_7astropy_2io_5ascii_7cparser_FastWriter *__pyx_v_self, PyObject *__pyx_v_output); /* proto*/

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;
static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *, char *, char *, int *); /*proto*/

/* Module declarations from 'libc' */

/* Module declarations from 'astropy.io.ascii.cparser' */
static PyTypeObject *__pyx_ptype_7astropy_2io_5ascii_7cparser_FileString = 0;
static PyTypeObject *__pyx_ptype_7astropy_2io_5ascii_7cparser_CParser = 0;
static PyTypeObject *__pyx_ptype_7astropy_2io_5ascii_7cparser_FastWriter = 0;
static PyTypeObject *__pyx_ptype_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines = 0;
static PyTypeObject *__pyx_ptype_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel = 0;
static PyTypeObject *__pyx_ptype_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr = 0;
static PyTypeObject *__pyx_ptype_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr = 0;
static PyTypeObject *__pyx_ptype_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr = 0;
static PyTypeObject *__pyx_ptype_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk = 0;
static PyTypeObject *__pyx_ptype_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr = 0;
#define __Pyx_MODULE_NAME "astropy.io.ascii.cparser"
extern int __pyx_module_is_main_astropy__io__ascii__cparser;
int __pyx_module_is_main_astropy__io__ascii__cparser = 0;

/* Implementation of 'astropy.io.ascii.cparser' */
static PyObject *__pyx_builtin_ImportError;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_open;
static PyObject *__pyx_builtin_IOError;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_NotImplementedError;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_chr;
static PyObject *__pyx_builtin_AttributeError;
static PyObject *__pyx_builtin_OSError;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_IndexError;
static PyObject *__pyx_builtin_RuntimeError;
static const char __pyx_k_0[] = "0";
static const char __pyx_k_A[] = "A";
static const char __pyx_k_E[] = "E";
static const char __pyx_k_N[] = "N";
static const char __pyx_k_S[] = "S";
static const char __pyx_k_U[] = "U";
static const char __pyx_k_e[] = "e";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_l[] = "l";
static const char __pyx_k_r[] = "r";
static const char __pyx_k_w[] = "w";
static const char __pyx_k__3[] = ",";
static const char __pyx_k__4[] = "\"";
static const char __pyx_k__5[] = "";
static const char __pyx_k_el[] = "el";
static const char __pyx_k_ma[] = "ma";
static const char __pyx_k_np[] = "np";
static const char __pyx_k_nt[] = "nt";
static const char __pyx_k_os[] = "os";
static const char __pyx_k_0_1[] = "{0}: {1}";
static const char __pyx_k_PY2[] = "PY2";
static const char __pyx_k_PY3[] = "PY3";
static const char __pyx_k__12[] = "\000";
static const char __pyx_k__13[] = "\n";
static const char __pyx_k__14[] = "\r";
static const char __pyx_k__17[] = "\001";
static const char __pyx_k__20[] = "# ";
static const char __pyx_k__22[] = " ";
static const char __pyx_k_chr[] = "chr";
static const char __pyx_k_col[] = "col";
static const char __pyx_k_csv[] = "csv";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_end[] = "end";
static const char __pyx_k_err[] = "err";
static const char __pyx_k_get[] = "get";
static const char __pyx_k_int[] = "int_";
static const char __pyx_k_pop[] = "pop";
static const char __pyx_k_put[] = "put";
static const char __pyx_k_six[] = "six";
static const char __pyx_k_str[] = "str";
static const char __pyx_k_FORT[] = "FORT";
static const char __pyx_k_Full[] = "Full";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_ceil[] = "ceil";
static const char __pyx_k_copy[] = "copy";
static const char __pyx_k_core[] = "core";
static const char __pyx_k_data[] = "data";
static const char __pyx_k_exit[] = "__exit__";
static const char __pyx_k_info[] = "info";
static const char __pyx_k_join[] = "join";
static const char __pyx_k_kind[] = "kind";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mask[] = "mask";
static const char __pyx_k_math[] = "math";
static const char __pyx_k_meta[] = "meta";
static const char __pyx_k_mmap[] = "mmap";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_open[] = "open";
static const char __pyx_k_read[] = "read";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_warn[] = "warn";
static const char __pyx_k_Empty[] = "Empty";
static const char __pyx_k_Queue[] = "Queue";
static const char __pyx_k_array[] = "array";
static const char __pyx_k_ascii[] = "ascii";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_col_0[] = "col{0}";
static const char __pyx_k_dtype[] = "dtype";
static const char __pyx_k_empty[] = "empty";
static const char __pyx_k_enter[] = "__enter__";
static const char __pyx_k_float[] = "float_";
static const char __pyx_k_fname[] = "fname";
static const char __pyx_k_force[] = "force";
static const char __pyx_k_names[] = "names";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_queue[] = "queue";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_strip[] = "strip";
static const char __pyx_k_table[] = "table";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_upper[] = "upper";
static const char __pyx_k_write[] = "write";
static const char __pyx_k_access[] = "access";
static const char __pyx_k_append[] = "append";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_extern[] = "extern";
static const char __pyx_k_fileno[] = "fileno";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_kwargs[] = "kwargs";
static const char __pyx_k_lambda[] = "<lambda>";
static const char __pyx_k_masked[] = "masked";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_output[] = "output";
static const char __pyx_k_parser[] = "parser";
static const char __pyx_k_pprint[] = "pprint";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_source[] = "source";
static const char __pyx_k_target[] = "target";
static const char __pyx_k_tolist[] = "tolist";
static const char __pyx_k_writer[] = "writer";
static const char __pyx_k_CParser[] = "CParser";
static const char __pyx_k_IOError[] = "IOError";
static const char __pyx_k_IntType[] = "IntType";
static const char __pyx_k_OSError[] = "OSError";
static const char __pyx_k_Process[] = "Process";
static const char __pyx_k_columns[] = "columns";
static const char __pyx_k_comment[] = "comment";
static const char __pyx_k_expchar[] = "expchar";
static const char __pyx_k_formats[] = "formats";
static const char __pyx_k_genexpr[] = "genexpr";
static const char __pyx_k_linesep[] = "linesep";
static const char __pyx_k_prepare[] = "__prepare__";
static const char __pyx_k_quoting[] = "quoting";
static const char __pyx_k_replace[] = "replace";
static const char __pyx_k_src_ptr[] = "src_ptr";
static const char __pyx_k_try_int[] = "try_int";
static const char __pyx_k_colnames[] = "colnames";
static const char __pyx_k_comments[] = "comments";
static const char __pyx_k_data_end[] = "data_end";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_no_error[] = "no error";
static const char __pyx_k_parallel[] = "parallel";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_use_cols[] = "use_cols";
static const char __pyx_k_warnings[] = "warnings";
static const char __pyx_k_writerow[] = "writerow";
static const char __pyx_k_CsvWriter[] = "CsvWriter";
static const char __pyx_k_ERR_CODES[] = "ERR_CODES";
static const char __pyx_k_FloatType[] = "FloatType";
static const char __pyx_k_MaskError[] = "MaskError";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_cpu_count[] = "cpu_count";
static const char __pyx_k_delimiter[] = "delimiter";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_get_names[] = "get_names";
static const char __pyx_k_iteritems[] = "iteritems";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_quotechar[] = "quotechar";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_terminate[] = "terminate";
static const char __pyx_k_try_float[] = "try_float";
static const char __pyx_k_writerows[] = "writerows";
static const char __pyx_k_FastWriter[] = "FastWriter";
static const char __pyx_k_FileString[] = "FileString";
static const char __pyx_k_IndexError[] = "IndexError";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_data_start[] = "data_start";
static const char __pyx_k_escapechar[] = "escapechar";
static const char __pyx_k_fill_empty[] = "fill_empty";
static const char __pyx_k_fill_names[] = "fill_names";
static const char __pyx_k_itervalues[] = "itervalues";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_read_chunk[] = "_read_chunk";
static const char __pyx_k_splitlines[] = "splitlines";
static const char __pyx_k_startswith[] = "startswith";
static const char __pyx_k_try_string[] = "try_string";
static const char __pyx_k_utils_data[] = "utils.data";
static const char __pyx_k_ACCESS_READ[] = "ACCESS_READ";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_concatenate[] = "concatenate";
static const char __pyx_k_doublequote[] = "doublequote";
static const char __pyx_k_fast_reader[] = "fast_reader";
static const char __pyx_k_fast_writer[] = "fast_writer";
static const char __pyx_k_fill_values[] = "fill_values";
static const char __pyx_k_CParserError[] = "CParserError";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_copy_cparser[] = "_copy_cparser";
static const char __pyx_k_format_funcs[] = "_format_funcs";
static const char __pyx_k_header_start[] = "header_start";
static const char __pyx_k_masked_array[] = "masked_array";
static const char __pyx_k_output_types[] = "output_types";
static const char __pyx_k_source_bytes[] = "source_bytes";
static const char __pyx_k_string_types[] = "string_types";
static const char __pyx_k_write_header[] = "_write_header";
static const char __pyx_k_QUOTE_MINIMAL[] = "QUOTE_MINIMAL";
static const char __pyx_k_exclude_names[] = "exclude_names";
static const char __pyx_k_header_output[] = "header_output";
static const char __pyx_k_include_names[] = "include_names";
static const char __pyx_k_line_comments[] = "line_comments";
static const char __pyx_k_read_parallel[] = "_read_parallel";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_unknown_error[] = "unknown error";
static const char __pyx_k_AstropyWarning[] = "AstropyWarning";
static const char __pyx_k_AttributeError[] = "AttributeError";
static const char __pyx_k_ParameterError[] = "ParameterError";
static const char __pyx_k_exponent_style[] = "exponent_style";
static const char __pyx_k_lineterminator[] = "lineterminator";
static const char __pyx_k_overflow_error[] = "overflow error";
static const char __pyx_k_reconvert_cols[] = "reconvert_cols";
static const char __pyx_k_chunk_tokenizer[] = "chunk_tokenizer";
static const char __pyx_k_fill_extra_cols[] = "fill_extra_cols";
static const char __pyx_k_get_fill_values[] = "get_fill_values";
static const char __pyx_k_multiprocessing[] = "multiprocessing";
static const char __pyx_k_reconvert_queue[] = "reconvert_queue";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_setup_tokenizer[] = "setup_tokenizer";
static const char __pyx_k_FastOptionsError[] = "FastOptionsError";
static const char __pyx_k_strip_whitespace[] = "strip_whitespace";
static const char __pyx_k_utils_exceptions[] = "utils.exceptions";
static const char __pyx_k_difference_update[] = "difference_update";
static const char __pyx_k_strip_line_fields[] = "strip_line_fields";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_fill_exclude_names[] = "fill_exclude_names";
static const char __pyx_k_fill_include_names[] = "fill_include_names";
static const char __pyx_k_use_fast_converter[] = "use_fast_converter";
static const char __pyx_k_NotImplementedError[] = "NotImplementedError";
static const char __pyx_k_intersection_update[] = "intersection_update";
static const char __pyx_k_get_auto_format_func[] = "get_auto_format_func";
static const char __pyx_k_get_readable_fileobj[] = "get_readable_fileobj";
static const char __pyx_k_FileString_splitlines[] = "FileString.splitlines";
static const char __pyx_k_invalid_line_supplied[] = "invalid line supplied";
static const char __pyx_k_strip_line_whitespace[] = "strip_line_whitespace";
static const char __pyx_k_type_conversion_error[] = "type conversion error";
static const char __pyx_k_InconsistentTableError[] = "InconsistentTableError";
static const char __pyx_k_strip_whitespace_lines[] = "strip_whitespace_lines";
static const char __pyx_k_strip_whitespace_fields[] = "strip_whitespace_fields";
static const char __pyx_k_astropy_io_ascii_cparser[] = "astropy.io.ascii.cparser";
static const char __pyx_k_read_chunk_locals_genexpr[] = "_read_chunk.<locals>.genexpr";
static const char __pyx_k_Column_0_failed_to_convert[] = "Column {0} failed to convert";
static const char __pyx_k_File_0_could_not_be_opened[] = "File \"{0}\" could not be opened";
static const char __pyx_k_ndarray_is_not_C_contiguous[] = "ndarray is not C contiguous";
static const char __pyx_k_astropy_io_ascii_cparser_pyx[] = "astropy/io/ascii/cparser.pyx";
static const char __pyx_k_read_parallel_locals_genexpr[] = "_read_parallel.<locals>.genexpr";
static const char __pyx_k_An_instance_of_this_class_is_th[] = "\n    An instance of this class is thrown when an error occurs\n    during C parsing.\n    ";
static const char __pyx_k_Input_table_must_be_a_file_like[] = "Input \"table\" must be a file-like object, a string (filename or data), or an iterable";
static const char __pyx_k_an_error_occurred_while_parsing[] = "an error occurred while parsing table data";
static const char __pyx_k_fast_reader_cannot_be_False_for[] = "fast_reader cannot be False for fast readers";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_shared_semaphore_implementation[] = "shared semaphore implementation required but not available";
static const char __pyx_k_unknown_dtype_code_in_numpy_pxd[] = "unknown dtype code in numpy.pxd (%d)";
static const char __pyx_k_Format_of_fill_values_must_be_ba[] = "Format of fill_values must be (<bad>, <fill>, <optional col1>, ...)";
static const char __pyx_k_Format_string_allocated_too_shor[] = "Format string allocated too short, see comment in numpy.pxd";
static const char __pyx_k_Invalid_parameter_in_fast_reader[] = "Invalid parameter in fast_reader dict";
static const char __pyx_k_Multiprocessing_is_not_yet_suppo[] = "Multiprocessing is not yet supported on Windows";
static const char __pyx_k_No_data_lines_found_C_reader_can[] = "No data lines found, C reader cannot autogenerate column names";
static const char __pyx_k_Non_native_byte_order_not_suppor[] = "Non-native byte order not supported";
static const char __pyx_k_OverflowError_converting_to_0_in[] = "OverflowError converting to {0} in column {1}, reverting to String.";
static const char __pyx_k_an_error_occurred_while_advancin[] = "an error occurred while advancing to the first header line";
static const char __pyx_k_an_error_occurred_while_tokenizi[] = "an error occurred while tokenizing the header line";
static const char __pyx_k_fast_reader_exponent_style_requi[] = "fast_reader: exponent_style requires use_fast_converter";
static const char __pyx_k_ndarray_is_not_Fortran_contiguou[] = "ndarray is not Fortran contiguous";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static const char __pyx_k_not_enough_columns_found_in_line[] = "not enough columns found in line {0} of data";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static const char __pyx_k_too_many_columns_found_in_line_0[] = "too many columns found in line {0} of data";
static const char __pyx_k_Format_string_allocated_too_shor_2[] = "Format string allocated too short.";
static const char __pyx_k_OverflowError_converting_to_0_in_2[] = "OverflowError converting to {0} in column {1}, possibly resulting in degraded precision.";
static const char __pyx_k_an_error_occurred_while_advancin_2[] = "an error occurred while advancing to the first line of data";
static const char __pyx_k_an_error_occurred_while_tokenizi_2[] = "an error occurred while tokenizing the first line of data";
static PyObject *__pyx_kp_s_0;
static PyObject *__pyx_kp_s_0_1;
static PyObject *__pyx_n_s_A;
static PyObject *__pyx_n_s_ACCESS_READ;
static PyObject *__pyx_kp_s_An_instance_of_this_class_is_th;
static PyObject *__pyx_n_s_AstropyWarning;
static PyObject *__pyx_n_s_AttributeError;
static PyObject *__pyx_n_s_CParser;
static PyObject *__pyx_n_s_CParserError;
static PyObject *__pyx_kp_s_Column_0_failed_to_convert;
static PyObject *__pyx_n_s_CsvWriter;
static PyObject *__pyx_n_s_E;
static PyObject *__pyx_n_s_ERR_CODES;
static PyObject *__pyx_n_s_Empty;
static PyObject *__pyx_n_s_FORT;
static PyObject *__pyx_n_s_FastOptionsError;
static PyObject *__pyx_n_s_FastWriter;
static PyObject *__pyx_n_s_FileString;
static PyObject *__pyx_n_s_FileString_splitlines;
static PyObject *__pyx_kp_s_File_0_could_not_be_opened;
static PyObject *__pyx_n_s_FloatType;
static PyObject *__pyx_kp_s_Format_of_fill_values_must_be_ba;
static PyObject *__pyx_kp_u_Format_string_allocated_too_shor;
static PyObject *__pyx_kp_u_Format_string_allocated_too_shor_2;
static PyObject *__pyx_n_s_Full;
static PyObject *__pyx_n_s_IOError;
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_n_s_InconsistentTableError;
static PyObject *__pyx_n_s_IndexError;
static PyObject *__pyx_kp_s_Input_table_must_be_a_file_like;
static PyObject *__pyx_n_s_IntType;
static PyObject *__pyx_kp_s_Invalid_parameter_in_fast_reader;
static PyObject *__pyx_n_s_MaskError;
static PyObject *__pyx_kp_s_Multiprocessing_is_not_yet_suppo;
static PyObject *__pyx_n_s_N;
static PyObject *__pyx_kp_s_No_data_lines_found_C_reader_can;
static PyObject *__pyx_kp_u_Non_native_byte_order_not_suppor;
static PyObject *__pyx_n_s_NotImplementedError;
static PyObject *__pyx_n_s_OSError;
static PyObject *__pyx_kp_s_OverflowError_converting_to_0_in;
static PyObject *__pyx_kp_s_OverflowError_converting_to_0_in_2;
static PyObject *__pyx_n_s_PY2;
static PyObject *__pyx_n_s_PY3;
static PyObject *__pyx_n_s_ParameterError;
static PyObject *__pyx_n_s_Process;
static PyObject *__pyx_n_s_QUOTE_MINIMAL;
static PyObject *__pyx_n_s_Queue;
static PyObject *__pyx_n_s_RuntimeError;
static PyObject *__pyx_n_s_S;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_U;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_kp_s__12;
static PyObject *__pyx_kp_s__13;
static PyObject *__pyx_kp_s__14;
static PyObject *__pyx_kp_s__17;
static PyObject *__pyx_kp_s__20;
static PyObject *__pyx_kp_s__22;
static PyObject *__pyx_kp_s__3;
static PyObject *__pyx_kp_s__4;
static PyObject *__pyx_kp_b__5;
static PyObject *__pyx_kp_s__5;
static PyObject *__pyx_n_s_access;
static PyObject *__pyx_kp_s_an_error_occurred_while_advancin;
static PyObject *__pyx_kp_s_an_error_occurred_while_advancin_2;
static PyObject *__pyx_kp_s_an_error_occurred_while_parsing;
static PyObject *__pyx_kp_s_an_error_occurred_while_tokenizi;
static PyObject *__pyx_kp_s_an_error_occurred_while_tokenizi_2;
static PyObject *__pyx_n_s_append;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_array;
static PyObject *__pyx_n_s_ascii;
static PyObject *__pyx_n_s_astropy_io_ascii_cparser;
static PyObject *__pyx_kp_s_astropy_io_ascii_cparser_pyx;
static PyObject *__pyx_n_s_ceil;
static PyObject *__pyx_n_s_chr;
static PyObject *__pyx_n_s_chunk_tokenizer;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_col;
static PyObject *__pyx_kp_s_col_0;
static PyObject *__pyx_n_s_colnames;
static PyObject *__pyx_n_s_columns;
static PyObject *__pyx_n_s_comment;
static PyObject *__pyx_n_s_comments;
static PyObject *__pyx_n_s_concatenate;
static PyObject *__pyx_n_s_copy;
static PyObject *__pyx_n_s_copy_cparser;
static PyObject *__pyx_n_s_core;
static PyObject *__pyx_n_s_cpu_count;
static PyObject *__pyx_n_s_csv;
static PyObject *__pyx_n_s_data;
static PyObject *__pyx_n_s_data_end;
static PyObject *__pyx_n_s_data_start;
static PyObject *__pyx_n_s_delimiter;
static PyObject *__pyx_n_s_difference_update;
static PyObject *__pyx_n_s_doc;
static PyObject *__pyx_n_s_doublequote;
static PyObject *__pyx_n_s_dtype;
static PyObject *__pyx_n_s_e;
static PyObject *__pyx_n_s_el;
static PyObject *__pyx_n_s_empty;
static PyObject *__pyx_n_s_encode;
static PyObject *__pyx_n_s_end;
static PyObject *__pyx_n_s_enter;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_err;
static PyObject *__pyx_n_s_escapechar;
static PyObject *__pyx_n_s_exclude_names;
static PyObject *__pyx_n_s_exit;
static PyObject *__pyx_n_s_expchar;
static PyObject *__pyx_n_s_exponent_style;
static PyObject *__pyx_n_s_extern;
static PyObject *__pyx_n_s_fast_reader;
static PyObject *__pyx_kp_s_fast_reader_cannot_be_False_for;
static PyObject *__pyx_kp_s_fast_reader_exponent_style_requi;
static PyObject *__pyx_n_s_fast_writer;
static PyObject *__pyx_n_s_fileno;
static PyObject *__pyx_n_s_fill_empty;
static PyObject *__pyx_n_s_fill_exclude_names;
static PyObject *__pyx_n_s_fill_extra_cols;
static PyObject *__pyx_n_s_fill_include_names;
static PyObject *__pyx_n_s_fill_names;
static PyObject *__pyx_n_s_fill_values;
static PyObject *__pyx_n_s_float;
static PyObject *__pyx_n_s_fname;
static PyObject *__pyx_n_s_force;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_format_funcs;
static PyObject *__pyx_n_s_formats;
static PyObject *__pyx_n_s_genexpr;
static PyObject *__pyx_n_s_get;
static PyObject *__pyx_n_s_get_auto_format_func;
static PyObject *__pyx_n_s_get_fill_values;
static PyObject *__pyx_n_s_get_names;
static PyObject *__pyx_n_s_get_readable_fileobj;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_header_output;
static PyObject *__pyx_n_s_header_start;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_include_names;
static PyObject *__pyx_n_s_info;
static PyObject *__pyx_n_s_int;
static PyObject *__pyx_n_s_intersection_update;
static PyObject *__pyx_kp_s_invalid_line_supplied;
static PyObject *__pyx_n_s_iteritems;
static PyObject *__pyx_n_s_itervalues;
static PyObject *__pyx_n_s_join;
static PyObject *__pyx_n_s_kind;
static PyObject *__pyx_n_s_kwargs;
static PyObject *__pyx_n_s_l;
static PyObject *__pyx_n_s_lambda;
static PyObject *__pyx_n_s_line_comments;
static PyObject *__pyx_n_s_linesep;
static PyObject *__pyx_n_s_lineterminator;
static PyObject *__pyx_n_s_ma;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_mask;
static PyObject *__pyx_n_s_masked;
static PyObject *__pyx_n_s_masked_array;
static PyObject *__pyx_n_s_math;
static PyObject *__pyx_n_s_meta;
static PyObject *__pyx_n_s_metaclass;
static PyObject *__pyx_n_s_mmap;
static PyObject *__pyx_n_s_module;
static PyObject *__pyx_n_s_multiprocessing;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_n_s_names;
static PyObject *__pyx_kp_u_ndarray_is_not_C_contiguous;
static PyObject *__pyx_kp_u_ndarray_is_not_Fortran_contiguou;
static PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
static PyObject *__pyx_kp_s_no_error;
static PyObject *__pyx_kp_s_not_enough_columns_found_in_line;
static PyObject *__pyx_n_s_np;
static PyObject *__pyx_n_s_nt;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_kp_s_numpy_core_multiarray_failed_to;
static PyObject *__pyx_kp_s_numpy_core_umath_failed_to_impor;
static PyObject *__pyx_n_s_open;
static PyObject *__pyx_n_s_os;
static PyObject *__pyx_n_s_output;
static PyObject *__pyx_n_s_output_types;
static PyObject *__pyx_kp_s_overflow_error;
static PyObject *__pyx_n_s_parallel;
static PyObject *__pyx_n_s_parser;
static PyObject *__pyx_n_s_pop;
static PyObject *__pyx_n_s_pprint;
static PyObject *__pyx_n_s_prepare;
static PyObject *__pyx_n_s_put;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_qualname;
static PyObject *__pyx_n_s_queue;
static PyObject *__pyx_n_s_quotechar;
static PyObject *__pyx_n_s_quoting;
static PyObject *__pyx_n_s_r;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_read;
static PyObject *__pyx_n_s_read_chunk;
static PyObject *__pyx_n_s_read_chunk_locals_genexpr;
static PyObject *__pyx_n_s_read_parallel;
static PyObject *__pyx_n_s_read_parallel_locals_genexpr;
static PyObject *__pyx_n_s_reconvert_cols;
static PyObject *__pyx_n_s_reconvert_queue;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_replace;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_setup_tokenizer;
static PyObject *__pyx_kp_s_shared_semaphore_implementation;
static PyObject *__pyx_n_s_six;
static PyObject *__pyx_n_s_source;
static PyObject *__pyx_n_s_source_bytes;
static PyObject *__pyx_n_s_splitlines;
static PyObject *__pyx_n_s_src_ptr;
static PyObject *__pyx_n_s_start;
static PyObject *__pyx_n_s_startswith;
static PyObject *__pyx_n_s_str;
static PyObject *__pyx_n_s_string_types;
static PyObject *__pyx_n_s_strip;
static PyObject *__pyx_n_s_strip_line_fields;
static PyObject *__pyx_n_s_strip_line_whitespace;
static PyObject *__pyx_n_s_strip_whitespace;
static PyObject *__pyx_n_s_strip_whitespace_fields;
static PyObject *__pyx_n_s_strip_whitespace_lines;
static PyObject *__pyx_n_s_table;
static PyObject *__pyx_n_s_target;
static PyObject *__pyx_n_s_terminate;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_n_s_tolist;
static PyObject *__pyx_kp_s_too_many_columns_found_in_line_0;
static PyObject *__pyx_n_s_try_float;
static PyObject *__pyx_n_s_try_int;
static PyObject *__pyx_n_s_try_string;
static PyObject *__pyx_kp_s_type_conversion_error;
static PyObject *__pyx_kp_u_unknown_dtype_code_in_numpy_pxd;
static PyObject *__pyx_kp_s_unknown_error;
static PyObject *__pyx_n_s_upper;
static PyObject *__pyx_n_s_use_cols;
static PyObject *__pyx_n_s_use_fast_converter;
static PyObject *__pyx_n_s_utils_data;
static PyObject *__pyx_n_s_utils_exceptions;
static PyObject *__pyx_n_s_w;
static PyObject *__pyx_n_s_warn;
static PyObject *__pyx_n_s_warnings;
static PyObject *__pyx_n_s_write;
static PyObject *__pyx_n_s_write_header;
static PyObject *__pyx_n_s_writer;
static PyObject *__pyx_n_s_writerow;
static PyObject *__pyx_n_s_writerows;
static PyObject *__pyx_lambda_funcdef_7astropy_2io_5ascii_7cparser_lambda(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_line); /* proto */
static PyObject *__pyx_lambda_funcdef_7astropy_2io_5ascii_7cparser_lambda1(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_line); /* proto */
static int __pyx_pf_7astropy_2io_5ascii_7cparser_10FileString___cinit__(struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString *__pyx_v_self, PyObject *__pyx_v_fname); /* proto */
static void __pyx_pf_7astropy_2io_5ascii_7cparser_10FileString_2__dealloc__(struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_pf_7astropy_2io_5ascii_7cparser_10FileString_4__len__(struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_10FileString_6__getitem__(struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_10FileString_8splitlines(struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_10FileString_11__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_10FileString_13__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7astropy_2io_5ascii_7cparser_7CParser___cinit__(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self, PyObject *__pyx_v_source, PyObject *__pyx_v_strip_line_whitespace, PyObject *__pyx_v_strip_line_fields, PyObject *__pyx_v_delimiter, PyObject *__pyx_v_comment, PyObject *__pyx_v_quotechar, PyObject *__pyx_v_header_start, PyObject *__pyx_v_data_start, PyObject *__pyx_v_data_end, PyObject *__pyx_v_names, PyObject *__pyx_v_include_names, PyObject *__pyx_v_exclude_names, PyObject *__pyx_v_fill_values, PyObject *__pyx_v_fill_include_names, PyObject *__pyx_v_fill_exclude_names, PyObject *__pyx_v_fill_extra_cols, PyObject *__pyx_v_fast_reader); /* proto */
static void __pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_2__dealloc__(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_4setup_tokenizer(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self, PyObject *__pyx_v_source); /* proto */
static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_6read_header(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_8read(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self, PyObject *__pyx_v_try_int, PyObject *__pyx_v_try_float, PyObject *__pyx_v_try_string); /* proto */
static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_14_read_parallel_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_14_read_parallel_3genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_14_read_parallel_6genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_10_read_parallel(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self, PyObject *__pyx_v_try_int, PyObject *__pyx_v_try_float, PyObject *__pyx_v_try_string); /* proto */
static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_12get_names(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_14set_names(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self, PyObject *__pyx_v_names); /* proto */
static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_16get_header_names(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_18__reduce__(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_5width___get__(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self); /* proto */
static int __pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_5width_2__set__(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_6source___get__(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self); /* proto */
static int __pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_6source_2__set__(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_6source_4__del__(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_12header_start___get__(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self); /* proto */
static int __pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_12header_start_2__set__(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_12header_start_4__del__(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser__copy_cparser(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_src_ptr, PyObject *__pyx_v_source_bytes, PyObject *__pyx_v_use_cols, PyObject *__pyx_v_fill_names, PyObject *__pyx_v_fill_values, PyObject *__pyx_v_strip_whitespace_lines, PyObject *__pyx_v_strip_whitespace_fields, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_11_read_chunk_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_2_read_chunk(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self, PyObject *__pyx_v_start, PyObject *__pyx_v_end, PyObject *__pyx_v_try_int, PyObject *__pyx_v_try_float, PyObject *__pyx_v_try_string, PyObject *__pyx_v_queue, PyObject *__pyx_v_reconvert_queue, PyObject *__pyx_v_i); /* proto */
static int __pyx_pf_7astropy_2io_5ascii_7cparser_10FastWriter___cinit__(struct __pyx_obj_7astropy_2io_5ascii_7cparser_FastWriter *__pyx_v_self, PyObject *__pyx_v_table, PyObject *__pyx_v_delimiter, PyObject *__pyx_v_comment, PyObject *__pyx_v_quotechar, CYTHON_UNUSED PyObject *__pyx_v_expchar, PyObject *__pyx_v_formats, PyObject *__pyx_v_strip_whitespace, CYTHON_UNUSED PyObject *__pyx_v_names, PyObject *__pyx_v_include_names, PyObject *__pyx_v_exclude_names, PyObject *__pyx_v_fill_values, PyObject *__pyx_v_fill_include_names, PyObject *__pyx_v_fill_exclude_names, PyObject *__pyx_v_fast_writer); /* proto */
static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_10FastWriter_2_write_header(struct __pyx_obj_7astropy_2io_5ascii_7cparser_FastWriter *__pyx_v_self, PyObject *__pyx_v_output, PyObject *__pyx_v_writer, PyObject *__pyx_v_header_output, PyObject *__pyx_v_output_types); /* proto */
static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_10FastWriter_4write(struct __pyx_obj_7astropy_2io_5ascii_7cparser_FastWriter *__pyx_v_self, PyObject *__pyx_v_output, PyObject *__pyx_v_header_output, PyObject *__pyx_v_output_types); /* proto */
static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_10FastWriter_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7astropy_2io_5ascii_7cparser_FastWriter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_10FastWriter_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7astropy_2io_5ascii_7cparser_FastWriter *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_4get_fill_values(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_fill_values, PyObject *__pyx_v_read); /* proto */
static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info); /* proto */
static PyObject *__pyx_tp_new_7astropy_2io_5ascii_7cparser_FileString(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7astropy_2io_5ascii_7cparser_CParser(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7astropy_2io_5ascii_7cparser_FastWriter(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyList_Type_pop = {0, &__pyx_n_s_pop, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PySet_Type_difference_update = {0, &__pyx_n_s_difference_update, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PySet_Type_intersection_update = {0, &__pyx_n_s_intersection_update, 0, 0, 0};
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_k__21;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_slice__19;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__29;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__32;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_tuple__35;
static PyObject *__pyx_tuple__37;
static PyObject *__pyx_codeobj__34;
static PyObject *__pyx_codeobj__36;
static PyObject *__pyx_codeobj__38;
/* Late includes */

/* "astropy/io/ascii/cparser.pyx":110
 *     "no error",
 *     "invalid line supplied",
 *     lambda line: "too many columns found in line {0} of data".format(line),             # <<<<<<<<<<<<<<
 *     lambda line: "not enough columns found in line {0} of data".format(line),
 *     "type conversion error",
 */

/* Python wrapper */
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_6lambda(PyObject *__pyx_self, PyObject *__pyx_v_line); /*proto*/
static PyMethodDef __pyx_mdef_7astropy_2io_5ascii_7cparser_6lambda = {"lambda", (PyCFunction)__pyx_pw_7astropy_2io_5ascii_7cparser_6lambda, METH_O, 0};
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_6lambda(PyObject *__pyx_self, PyObject *__pyx_v_line) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_7astropy_2io_5ascii_7cparser_lambda(__pyx_self, ((PyObject *)__pyx_v_line));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_7astropy_2io_5ascii_7cparser_lambda(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_line) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("lambda", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_too_many_columns_found_in_line_0, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_line) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_line);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("astropy.io.ascii.cparser.lambda", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "astropy/io/ascii/cparser.pyx":111
 *     "invalid line supplied",
 *     lambda line: "too many columns found in line {0} of data".format(line),
 *     lambda line: "not enough columns found in line {0} of data".format(line),             # <<<<<<<<<<<<<<
 *     "type conversion error",
 *     "overflow error"
 */

/* Python wrapper */
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_7lambda1(PyObject *__pyx_self, PyObject *__pyx_v_line); /*proto*/
static PyMethodDef __pyx_mdef_7astropy_2io_5ascii_7cparser_7lambda1 = {"lambda1", (PyCFunction)__pyx_pw_7astropy_2io_5ascii_7cparser_7lambda1, METH_O, 0};
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_7lambda1(PyObject *__pyx_self, PyObject *__pyx_v_line) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda1 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_7astropy_2io_5ascii_7cparser_lambda1(__pyx_self, ((PyObject *)__pyx_v_line));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_7astropy_2io_5ascii_7cparser_lambda1(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_line) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("lambda1", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_not_enough_columns_found_in_line, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_line) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_line);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("astropy.io.ascii.cparser.lambda1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "astropy/io/ascii/cparser.pyx":126
 *         Py_buffer buf
 * 
 *     def __cinit__(self, fname):             # <<<<<<<<<<<<<<
 *         self.fhandle = open(fname, 'r')
 *         if not self.fhandle:
 */

/* Python wrapper */
static int __pyx_pw_7astropy_2io_5ascii_7cparser_10FileString_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7astropy_2io_5ascii_7cparser_10FileString_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_fname = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fname,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fname)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 126, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_fname = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 126, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("astropy.io.ascii.cparser.FileString.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7astropy_2io_5ascii_7cparser_10FileString___cinit__(((struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString *)__pyx_v_self), __pyx_v_fname);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7astropy_2io_5ascii_7cparser_10FileString___cinit__(struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString *__pyx_v_self, PyObject *__pyx_v_fname) {
  Py_ssize_t __pyx_v_buf_len;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  int __pyx_t_9;
  void *__pyx_t_10;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "astropy/io/ascii/cparser.pyx":127
 * 
 *     def __cinit__(self, fname):
 *         self.fhandle = open(fname, 'r')             # <<<<<<<<<<<<<<
 *         if not self.fhandle:
 *             raise IOError('File "{0}" could not be opened'.format(fname))
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_fname);
  __Pyx_GIVEREF(__pyx_v_fname);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_fname);
  __Pyx_INCREF(__pyx_n_s_r);
  __Pyx_GIVEREF(__pyx_n_s_r);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_r);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_open, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->fhandle);
  __Pyx_DECREF(__pyx_v_self->fhandle);
  __pyx_v_self->fhandle = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "astropy/io/ascii/cparser.pyx":128
 *     def __cinit__(self, fname):
 *         self.fhandle = open(fname, 'r')
 *         if not self.fhandle:             # <<<<<<<<<<<<<<
 *             raise IOError('File "{0}" could not be opened'.format(fname))
 *         self.mmap = mmap.mmap(self.fhandle.fileno(), 0, access=mmap.ACCESS_READ)
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_self->fhandle); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 128, __pyx_L1_error)
  __pyx_t_4 = ((!__pyx_t_3) != 0);
  if (unlikely(__pyx_t_4)) {

    /* "astropy/io/ascii/cparser.pyx":129
 *         self.fhandle = open(fname, 'r')
 *         if not self.fhandle:
 *             raise IOError('File "{0}" could not be opened'.format(fname))             # <<<<<<<<<<<<<<
 *         self.mmap = mmap.mmap(self.fhandle.fileno(), 0, access=mmap.ACCESS_READ)
 *         cdef Py_ssize_t buf_len = len(self.mmap)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_File_0_could_not_be_opened, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v_fname) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_fname);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IOError, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 129, __pyx_L1_error)

    /* "astropy/io/ascii/cparser.pyx":128
 *     def __cinit__(self, fname):
 *         self.fhandle = open(fname, 'r')
 *         if not self.fhandle:             # <<<<<<<<<<<<<<
 *             raise IOError('File "{0}" could not be opened'.format(fname))
 *         self.mmap = mmap.mmap(self.fhandle.fileno(), 0, access=mmap.ACCESS_READ)
 */
  }

  /* "astropy/io/ascii/cparser.pyx":130
 *         if not self.fhandle:
 *             raise IOError('File "{0}" could not be opened'.format(fname))
 *         self.mmap = mmap.mmap(self.fhandle.fileno(), 0, access=mmap.ACCESS_READ)             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t buf_len = len(self.mmap)
 *         if six.PY2:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_mmap); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_mmap); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->fhandle, __pyx_n_s_fileno); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_0);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_mmap); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_ACCESS_READ); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_access, __pyx_t_7) < 0) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_7);
  __Pyx_GOTREF(__pyx_v_self->mmap);
  __Pyx_DECREF(__pyx_v_self->mmap);
  __pyx_v_self->mmap = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "astropy/io/ascii/cparser.pyx":131
 *             raise IOError('File "{0}" could not be opened'.format(fname))
 *         self.mmap = mmap.mmap(self.fhandle.fileno(), 0, access=mmap.ACCESS_READ)
 *         cdef Py_ssize_t buf_len = len(self.mmap)             # <<<<<<<<<<<<<<
 *         if six.PY2:
 *             PyObject_AsReadBuffer(self.mmap, &self.mmap_ptr, &buf_len)
 */
  __pyx_t_7 = __pyx_v_self->mmap;
  __Pyx_INCREF(__pyx_t_7);
  __pyx_t_8 = PyObject_Length(__pyx_t_7); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 131, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_buf_len = __pyx_t_8;

  /* "astropy/io/ascii/cparser.pyx":132
 *         self.mmap = mmap.mmap(self.fhandle.fileno(), 0, access=mmap.ACCESS_READ)
 *         cdef Py_ssize_t buf_len = len(self.mmap)
 *         if six.PY2:             # <<<<<<<<<<<<<<
 *             PyObject_AsReadBuffer(self.mmap, &self.mmap_ptr, &buf_len)
 *         else:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_six); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_PY2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 132, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "astropy/io/ascii/cparser.pyx":133
 *         cdef Py_ssize_t buf_len = len(self.mmap)
 *         if six.PY2:
 *             PyObject_AsReadBuffer(self.mmap, &self.mmap_ptr, &buf_len)             # <<<<<<<<<<<<<<
 *         else:
 *             PyObject_GetBuffer(self.mmap, &self.buf, PyBUF_SIMPLE)
 */
    __pyx_t_1 = __pyx_v_self->mmap;
    __Pyx_INCREF(__pyx_t_1);
    (void)(PyObject_AsReadBuffer(__pyx_t_1, (&__pyx_v_self->mmap_ptr), (&__pyx_v_buf_len)));
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "astropy/io/ascii/cparser.pyx":132
 *         self.mmap = mmap.mmap(self.fhandle.fileno(), 0, access=mmap.ACCESS_READ)
 *         cdef Py_ssize_t buf_len = len(self.mmap)
 *         if six.PY2:             # <<<<<<<<<<<<<<
 *             PyObject_AsReadBuffer(self.mmap, &self.mmap_ptr, &buf_len)
 *         else:
 */
    goto __pyx_L4;
  }

  /* "astropy/io/ascii/cparser.pyx":135
 *             PyObject_AsReadBuffer(self.mmap, &self.mmap_ptr, &buf_len)
 *         else:
 *             PyObject_GetBuffer(self.mmap, &self.buf, PyBUF_SIMPLE)             # <<<<<<<<<<<<<<
 *             self.mmap_ptr = self.buf.buf
 * 
 */
  /*else*/ {
    __pyx_t_1 = __pyx_v_self->mmap;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_9 = PyObject_GetBuffer(__pyx_t_1, (&__pyx_v_self->buf), PyBUF_SIMPLE); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 135, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "astropy/io/ascii/cparser.pyx":136
 *         else:
 *             PyObject_GetBuffer(self.mmap, &self.buf, PyBUF_SIMPLE)
 *             self.mmap_ptr = self.buf.buf             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
    __pyx_t_10 = __pyx_v_self->buf.buf;
    __pyx_v_self->mmap_ptr = __pyx_t_10;
  }
  __pyx_L4:;

  /* "astropy/io/ascii/cparser.pyx":126
 *         Py_buffer buf
 * 
 *     def __cinit__(self, fname):             # <<<<<<<<<<<<<<
 *         self.fhandle = open(fname, 'r')
 *         if not self.fhandle:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("astropy.io.ascii.cparser.FileString.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "astropy/io/ascii/cparser.pyx":138
 *             self.mmap_ptr = self.buf.buf
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.mmap:
 *             if not six.PY2: # free buffer memory to prevent a resource leak
 */

/* Python wrapper */
static void __pyx_pw_7astropy_2io_5ascii_7cparser_10FileString_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_7astropy_2io_5ascii_7cparser_10FileString_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_7astropy_2io_5ascii_7cparser_10FileString_2__dealloc__(((struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_7astropy_2io_5ascii_7cparser_10FileString_2__dealloc__(struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "astropy/io/ascii/cparser.pyx":139
 * 
 *     def __dealloc__(self):
 *         if self.mmap:             # <<<<<<<<<<<<<<
 *             if not six.PY2: # free buffer memory to prevent a resource leak
 *                 PyBuffer_Release(&self.buf)
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_self->mmap); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 139, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "astropy/io/ascii/cparser.pyx":140
 *     def __dealloc__(self):
 *         if self.mmap:
 *             if not six.PY2: # free buffer memory to prevent a resource leak             # <<<<<<<<<<<<<<
 *                 PyBuffer_Release(&self.buf)
 *             self.mmap.close()
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_six); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_PY2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 140, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = ((!__pyx_t_1) != 0);
    if (__pyx_t_4) {

      /* "astropy/io/ascii/cparser.pyx":141
 *         if self.mmap:
 *             if not six.PY2: # free buffer memory to prevent a resource leak
 *                 PyBuffer_Release(&self.buf)             # <<<<<<<<<<<<<<
 *             self.mmap.close()
 *             self.fhandle.close()
 */
      PyBuffer_Release((&__pyx_v_self->buf));

      /* "astropy/io/ascii/cparser.pyx":140
 *     def __dealloc__(self):
 *         if self.mmap:
 *             if not six.PY2: # free buffer memory to prevent a resource leak             # <<<<<<<<<<<<<<
 *                 PyBuffer_Release(&self.buf)
 *             self.mmap.close()
 */
    }

    /* "astropy/io/ascii/cparser.pyx":142
 *             if not six.PY2: # free buffer memory to prevent a resource leak
 *                 PyBuffer_Release(&self.buf)
 *             self.mmap.close()             # <<<<<<<<<<<<<<
 *             self.fhandle.close()
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->mmap, __pyx_n_s_close); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "astropy/io/ascii/cparser.pyx":143
 *                 PyBuffer_Release(&self.buf)
 *             self.mmap.close()
 *             self.fhandle.close()             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->fhandle, __pyx_n_s_close); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 143, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 143, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "astropy/io/ascii/cparser.pyx":139
 * 
 *     def __dealloc__(self):
 *         if self.mmap:             # <<<<<<<<<<<<<<
 *             if not six.PY2: # free buffer memory to prevent a resource leak
 *                 PyBuffer_Release(&self.buf)
 */
  }

  /* "astropy/io/ascii/cparser.pyx":138
 *             self.mmap_ptr = self.buf.buf
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.mmap:
 *             if not six.PY2: # free buffer memory to prevent a resource leak
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("astropy.io.ascii.cparser.FileString.__dealloc__", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "astropy/io/ascii/cparser.pyx":145
 *             self.fhandle.close()
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return len(self.mmap)
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_7astropy_2io_5ascii_7cparser_10FileString_5__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_7astropy_2io_5ascii_7cparser_10FileString_5__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7astropy_2io_5ascii_7cparser_10FileString_4__len__(((struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_7astropy_2io_5ascii_7cparser_10FileString_4__len__(struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "astropy/io/ascii/cparser.pyx":146
 * 
 *     def __len__(self):
 *         return len(self.mmap)             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, i):
 */
  __pyx_t_1 = __pyx_v_self->mmap;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "astropy/io/ascii/cparser.pyx":145
 *             self.fhandle.close()
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return len(self.mmap)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("astropy.io.ascii.cparser.FileString.__len__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "astropy/io/ascii/cparser.pyx":148
 *         return len(self.mmap)
 * 
 *     def __getitem__(self, i):             # <<<<<<<<<<<<<<
 *         return self.mmap[i]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_10FileString_7__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_i); /*proto*/
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_10FileString_7__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7astropy_2io_5ascii_7cparser_10FileString_6__getitem__(((struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString *)__pyx_v_self), ((PyObject *)__pyx_v_i));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_10FileString_6__getitem__(struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "astropy/io/ascii/cparser.pyx":149
 * 
 *     def __getitem__(self, i):
 *         return self.mmap[i]             # <<<<<<<<<<<<<<
 * 
 *     def splitlines(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_self->mmap, __pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "astropy/io/ascii/cparser.pyx":148
 *         return len(self.mmap)
 * 
 *     def __getitem__(self, i):             # <<<<<<<<<<<<<<
 *         return self.mmap[i]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("astropy.io.ascii.cparser.FileString.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_7astropy_2io_5ascii_7cparser_10FileString_10generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "astropy/io/ascii/cparser.pyx":151
 *         return self.mmap[i]
 * 
 *     def splitlines(self):             # <<<<<<<<<<<<<<
 *         """
 *         Return a generator yielding lines from the memory map.
 */

/* Python wrapper */
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_10FileString_9splitlines(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_7astropy_2io_5ascii_7cparser_10FileString_8splitlines[] = "\n        Return a generator yielding lines from the memory map.\n        ";
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_10FileString_9splitlines(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("splitlines (wrapper)", 0);
  __pyx_r = __pyx_pf_7astropy_2io_5ascii_7cparser_10FileString_8splitlines(((struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_10FileString_8splitlines(struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString *__pyx_v_self) {
  struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("splitlines", 0);
  __pyx_cur_scope = (struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines *)__pyx_tp_new_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines(__pyx_ptype_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 151, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_7astropy_2io_5ascii_7cparser_10FileString_10generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_splitlines, __pyx_n_s_FileString_splitlines, __pyx_n_s_astropy_io_ascii_cparser); if (unlikely(!gen)) __PYX_ERR(0, 151, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("astropy.io.ascii.cparser.FileString.splitlines", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_7astropy_2io_5ascii_7cparser_10FileString_10generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines *__pyx_cur_scope = ((struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("splitlines", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 151, __pyx_L1_error)

  /* "astropy/io/ascii/cparser.pyx":155
 *         Return a generator yielding lines from the memory map.
 *         """
 *         cdef char *ptr = <char *>self.mmap_ptr             # <<<<<<<<<<<<<<
 *         cdef char *tmp
 *         cdef size_t line_len
 */
  __pyx_cur_scope->__pyx_v_ptr = ((char *)__pyx_cur_scope->__pyx_v_self->mmap_ptr);

  /* "astropy/io/ascii/cparser.pyx":158
 *         cdef char *tmp
 *         cdef size_t line_len
 *         cdef size_t map_len = len(self.mmap)             # <<<<<<<<<<<<<<
 * 
 *         while ptr:
 */
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_self->mmap;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 158, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_cur_scope->__pyx_v_map_len = __pyx_t_2;

  /* "astropy/io/ascii/cparser.pyx":160
 *         cdef size_t map_len = len(self.mmap)
 * 
 *         while ptr:             # <<<<<<<<<<<<<<
 *             tmp = get_line(ptr, &line_len, map_len)
 *             yield ptr[:line_len].decode('ascii')
 */
  while (1) {
    __pyx_t_3 = (__pyx_cur_scope->__pyx_v_ptr != 0);
    if (!__pyx_t_3) break;

    /* "astropy/io/ascii/cparser.pyx":161
 * 
 *         while ptr:
 *             tmp = get_line(ptr, &line_len, map_len)             # <<<<<<<<<<<<<<
 *             yield ptr[:line_len].decode('ascii')
 *             ptr = tmp
 */
    __pyx_cur_scope->__pyx_v_tmp = get_line(__pyx_cur_scope->__pyx_v_ptr, (&__pyx_cur_scope->__pyx_v_line_len), __pyx_cur_scope->__pyx_v_map_len);

    /* "astropy/io/ascii/cparser.pyx":162
 *         while ptr:
 *             tmp = get_line(ptr, &line_len, map_len)
 *             yield ptr[:line_len].decode('ascii')             # <<<<<<<<<<<<<<
 *             ptr = tmp
 * 
 */
    __pyx_t_1 = __Pyx_decode_c_string(__pyx_cur_scope->__pyx_v_ptr, 0, __pyx_cur_scope->__pyx_v_line_len, NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 162, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 162, __pyx_L1_error)

    /* "astropy/io/ascii/cparser.pyx":163
 *             tmp = get_line(ptr, &line_len, map_len)
 *             yield ptr[:line_len].decode('ascii')
 *             ptr = tmp             # <<<<<<<<<<<<<<
 * 
 * cdef class CParser:
 */
    __pyx_cur_scope->__pyx_v_ptr = __pyx_cur_scope->__pyx_v_tmp;
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "astropy/io/ascii/cparser.pyx":151
 *         return self.mmap[i]
 * 
 *     def splitlines(self):             # <<<<<<<<<<<<<<
 *         """
 *         Return a generator yielding lines from the memory map.
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("splitlines", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_10FileString_12__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_10FileString_12__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7astropy_2io_5ascii_7cparser_10FileString_11__reduce_cython__(((struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_10FileString_11__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("astropy.io.ascii.cparser.FileString.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_10FileString_14__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_10FileString_14__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7astropy_2io_5ascii_7cparser_10FileString_13__setstate_cython__(((struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_10FileString_13__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("astropy.io.ascii.cparser.FileString.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "astropy/io/ascii/cparser.pyx":195
 *         object header_start
 * 
 *     def __cinit__(self, source, strip_line_whitespace, strip_line_fields,             # <<<<<<<<<<<<<<
 *                   delimiter=',',
 *                   comment=None,
 */

/* Python wrapper */
static int __pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_source = 0;
  PyObject *__pyx_v_strip_line_whitespace = 0;
  PyObject *__pyx_v_strip_line_fields = 0;
  PyObject *__pyx_v_delimiter = 0;
  PyObject *__pyx_v_comment = 0;
  PyObject *__pyx_v_quotechar = 0;
  PyObject *__pyx_v_header_start = 0;
  PyObject *__pyx_v_data_start = 0;
  PyObject *__pyx_v_data_end = 0;
  PyObject *__pyx_v_names = 0;
  PyObject *__pyx_v_include_names = 0;
  PyObject *__pyx_v_exclude_names = 0;
  PyObject *__pyx_v_fill_values = 0;
  PyObject *__pyx_v_fill_include_names = 0;
  PyObject *__pyx_v_fill_exclude_names = 0;
  PyObject *__pyx_v_fill_extra_cols = 0;
  PyObject *__pyx_v_fast_reader = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_source,&__pyx_n_s_strip_line_whitespace,&__pyx_n_s_strip_line_fields,&__pyx_n_s_delimiter,&__pyx_n_s_comment,&__pyx_n_s_quotechar,&__pyx_n_s_header_start,&__pyx_n_s_data_start,&__pyx_n_s_data_end,&__pyx_n_s_names,&__pyx_n_s_include_names,&__pyx_n_s_exclude_names,&__pyx_n_s_fill_values,&__pyx_n_s_fill_include_names,&__pyx_n_s_fill_exclude_names,&__pyx_n_s_fill_extra_cols,&__pyx_n_s_fast_reader,0};
    PyObject* values[17] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    values[3] = ((PyObject *)__pyx_kp_s__3);

    /* "astropy/io/ascii/cparser.pyx":197
 *     def __cinit__(self, source, strip_line_whitespace, strip_line_fields,
 *                   delimiter=',',
 *                   comment=None,             # <<<<<<<<<<<<<<
 *                   quotechar='"',
 *                   header_start=0,
 */
    values[4] = ((PyObject *)Py_None);
    values[5] = ((PyObject *)__pyx_kp_s__4);
    values[6] = ((PyObject *)__pyx_int_0);
    values[7] = ((PyObject *)__pyx_int_1);

    /* "astropy/io/ascii/cparser.pyx":201
 *                   header_start=0,
 *                   data_start=1,
 *                   data_end=None,             # <<<<<<<<<<<<<<
 *                   names=None,
 *                   include_names=None,
 */
    values[8] = ((PyObject *)Py_None);

    /* "astropy/io/ascii/cparser.pyx":202
 *                   data_start=1,
 *                   data_end=None,
 *                   names=None,             # <<<<<<<<<<<<<<
 *                   include_names=None,
 *                   exclude_names=None,
 */
    values[9] = ((PyObject *)Py_None);

    /* "astropy/io/ascii/cparser.pyx":203
 *                   data_end=None,
 *                   names=None,
 *                   include_names=None,             # <<<<<<<<<<<<<<
 *                   exclude_names=None,
 *                   fill_values=('', '0'),
 */
    values[10] = ((PyObject *)Py_None);

    /* "astropy/io/ascii/cparser.pyx":204
 *                   names=None,
 *                   include_names=None,
 *                   exclude_names=None,             # <<<<<<<<<<<<<<
 *                   fill_values=('', '0'),
 *                   fill_include_names=None,
 */
    values[11] = ((PyObject *)Py_None);

    /* "astropy/io/ascii/cparser.pyx":205
 *                   include_names=None,
 *                   exclude_names=None,
 *                   fill_values=('', '0'),             # <<<<<<<<<<<<<<
 *                   fill_include_names=None,
 *                   fill_exclude_names=None,
 */
    values[12] = ((PyObject *)__pyx_tuple__6);

    /* "astropy/io/ascii/cparser.pyx":206
 *                   exclude_names=None,
 *                   fill_values=('', '0'),
 *                   fill_include_names=None,             # <<<<<<<<<<<<<<
 *                   fill_exclude_names=None,
 *                   fill_extra_cols=0,
 */
    values[13] = ((PyObject *)Py_None);

    /* "astropy/io/ascii/cparser.pyx":207
 *                   fill_values=('', '0'),
 *                   fill_include_names=None,
 *                   fill_exclude_names=None,             # <<<<<<<<<<<<<<
 *                   fill_extra_cols=0,
 *                   fast_reader=True):
 */
    values[14] = ((PyObject *)Py_None);
    values[15] = ((PyObject *)__pyx_int_0);

    /* "astropy/io/ascii/cparser.pyx":209
 *                   fill_exclude_names=None,
 *                   fill_extra_cols=0,
 *                   fast_reader=True):             # <<<<<<<<<<<<<<
 * 
 *         # Handle fast_reader parameter (True or dict specifying options)
 */
    values[16] = ((PyObject *)Py_True);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 17: values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_source)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strip_line_whitespace)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 17, 1); __PYX_ERR(0, 195, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strip_line_fields)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 17, 2); __PYX_ERR(0, 195, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_delimiter);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_comment);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_quotechar);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_header_start);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data_start);
          if (value) { values[7] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data_end);
          if (value) { values[8] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_names);
          if (value) { values[9] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_include_names);
          if (value) { values[10] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_exclude_names);
          if (value) { values[11] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fill_values);
          if (value) { values[12] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fill_include_names);
          if (value) { values[13] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fill_exclude_names);
          if (value) { values[14] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fill_extra_cols);
          if (value) { values[15] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 16:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fast_reader);
          if (value) { values[16] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 195, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case 17: values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_source = values[0];
    __pyx_v_strip_line_whitespace = values[1];
    __pyx_v_strip_line_fields = values[2];
    __pyx_v_delimiter = values[3];
    __pyx_v_comment = values[4];
    __pyx_v_quotechar = values[5];
    __pyx_v_header_start = values[6];
    __pyx_v_data_start = values[7];
    __pyx_v_data_end = values[8];
    __pyx_v_names = values[9];
    __pyx_v_include_names = values[10];
    __pyx_v_exclude_names = values[11];
    __pyx_v_fill_values = values[12];
    __pyx_v_fill_include_names = values[13];
    __pyx_v_fill_exclude_names = values[14];
    __pyx_v_fill_extra_cols = values[15];
    __pyx_v_fast_reader = values[16];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 17, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 195, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("astropy.io.ascii.cparser.CParser.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7astropy_2io_5ascii_7cparser_7CParser___cinit__(((struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *)__pyx_v_self), __pyx_v_source, __pyx_v_strip_line_whitespace, __pyx_v_strip_line_fields, __pyx_v_delimiter, __pyx_v_comment, __pyx_v_quotechar, __pyx_v_header_start, __pyx_v_data_start, __pyx_v_data_end, __pyx_v_names, __pyx_v_include_names, __pyx_v_exclude_names, __pyx_v_fill_values, __pyx_v_fill_include_names, __pyx_v_fill_exclude_names, __pyx_v_fill_extra_cols, __pyx_v_fast_reader);

  /* "astropy/io/ascii/cparser.pyx":195
 *         object header_start
 * 
 *     def __cinit__(self, source, strip_line_whitespace, strip_line_fields,             # <<<<<<<<<<<<<<
 *                   delimiter=',',
 *                   comment=None,
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7astropy_2io_5ascii_7cparser_7CParser___cinit__(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self, PyObject *__pyx_v_source, PyObject *__pyx_v_strip_line_whitespace, PyObject *__pyx_v_strip_line_fields, PyObject *__pyx_v_delimiter, PyObject *__pyx_v_comment, PyObject *__pyx_v_quotechar, PyObject *__pyx_v_header_start, PyObject *__pyx_v_data_start, PyObject *__pyx_v_data_end, PyObject *__pyx_v_names, PyObject *__pyx_v_include_names, PyObject *__pyx_v_exclude_names, PyObject *__pyx_v_fill_values, PyObject *__pyx_v_fill_include_names, PyObject *__pyx_v_fill_exclude_names, PyObject *__pyx_v_fill_extra_cols, PyObject *__pyx_v_fast_reader) {
  PyObject *__pyx_v_expchar = NULL;
  PyObject *__pyx_v_use_fast_converter = NULL;
  PyObject *__pyx_v_parallel = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  long __pyx_t_7;
  long __pyx_t_8;
  long __pyx_t_9;
  long __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_t_14;
  __Pyx_RefNannySetupContext("__cinit__", 0);
  __Pyx_INCREF(__pyx_v_comment);
  __Pyx_INCREF(__pyx_v_fast_reader);

  /* "astropy/io/ascii/cparser.pyx":212
 * 
 *         # Handle fast_reader parameter (True or dict specifying options)
 *         if fast_reader is True or fast_reader == 'force':             # <<<<<<<<<<<<<<
 *             fast_reader = {}
 *         elif fast_reader is False: # shouldn't happen
 */
  __pyx_t_2 = (__pyx_v_fast_reader == Py_True);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_v_fast_reader, __pyx_n_s_force, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 212, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "astropy/io/ascii/cparser.pyx":213
 *         # Handle fast_reader parameter (True or dict specifying options)
 *         if fast_reader is True or fast_reader == 'force':
 *             fast_reader = {}             # <<<<<<<<<<<<<<
 *         elif fast_reader is False: # shouldn't happen
 *             raise core.ParameterError("fast_reader cannot be False for fast readers")
 */
    __pyx_t_4 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF_SET(__pyx_v_fast_reader, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "astropy/io/ascii/cparser.pyx":212
 * 
 *         # Handle fast_reader parameter (True or dict specifying options)
 *         if fast_reader is True or fast_reader == 'force':             # <<<<<<<<<<<<<<
 *             fast_reader = {}
 *         elif fast_reader is False: # shouldn't happen
 */
    goto __pyx_L3;
  }

  /* "astropy/io/ascii/cparser.pyx":214
 *         if fast_reader is True or fast_reader == 'force':
 *             fast_reader = {}
 *         elif fast_reader is False: # shouldn't happen             # <<<<<<<<<<<<<<
 *             raise core.ParameterError("fast_reader cannot be False for fast readers")
 *         expchar = fast_reader.pop('exponent_style', 'E').upper()
 */
  __pyx_t_1 = (__pyx_v_fast_reader == Py_False);
  __pyx_t_3 = (__pyx_t_1 != 0);
  if (unlikely(__pyx_t_3)) {

    /* "astropy/io/ascii/cparser.pyx":215
 *             fast_reader = {}
 *         elif fast_reader is False: # shouldn't happen
 *             raise core.ParameterError("fast_reader cannot be False for fast readers")             # <<<<<<<<<<<<<<
 *         expchar = fast_reader.pop('exponent_style', 'E').upper()
 *         # parallel and use_fast_reader are False by default, but only the latter
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_core); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 215, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_ParameterError); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 215, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_5, __pyx_kp_s_fast_reader_cannot_be_False_for) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_kp_s_fast_reader_cannot_be_False_for);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 215, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 215, __pyx_L1_error)

    /* "astropy/io/ascii/cparser.pyx":214
 *         if fast_reader is True or fast_reader == 'force':
 *             fast_reader = {}
 *         elif fast_reader is False: # shouldn't happen             # <<<<<<<<<<<<<<
 *             raise core.ParameterError("fast_reader cannot be False for fast readers")
 *         expchar = fast_reader.pop('exponent_style', 'E').upper()
 */
  }
  __pyx_L3:;

  /* "astropy/io/ascii/cparser.pyx":216
 *         elif fast_reader is False: # shouldn't happen
 *             raise core.ParameterError("fast_reader cannot be False for fast readers")
 *         expchar = fast_reader.pop('exponent_style', 'E').upper()             # <<<<<<<<<<<<<<
 *         # parallel and use_fast_reader are False by default, but only the latter
 *         # supports Fortran double precision notation
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_fast_reader, __pyx_n_s_pop); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_upper); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_expchar = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "astropy/io/ascii/cparser.pyx":219
 *         # parallel and use_fast_reader are False by default, but only the latter
 *         # supports Fortran double precision notation
 *         if expchar == 'E':             # <<<<<<<<<<<<<<
 *             use_fast_converter = fast_reader.pop('use_fast_converter', False)
 *         else:
 */
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_v_expchar, __pyx_n_s_E, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 219, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "astropy/io/ascii/cparser.pyx":220
 *         # supports Fortran double precision notation
 *         if expchar == 'E':
 *             use_fast_converter = fast_reader.pop('use_fast_converter', False)             # <<<<<<<<<<<<<<
 *         else:
 *             use_fast_converter = fast_reader.pop('use_fast_converter', True)
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_fast_reader, __pyx_n_s_pop); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 220, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 220, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_use_fast_converter = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "astropy/io/ascii/cparser.pyx":219
 *         # parallel and use_fast_reader are False by default, but only the latter
 *         # supports Fortran double precision notation
 *         if expchar == 'E':             # <<<<<<<<<<<<<<
 *             use_fast_converter = fast_reader.pop('use_fast_converter', False)
 *         else:
 */
    goto __pyx_L6;
  }

  /* "astropy/io/ascii/cparser.pyx":222
 *             use_fast_converter = fast_reader.pop('use_fast_converter', False)
 *         else:
 *             use_fast_converter = fast_reader.pop('use_fast_converter', True)             # <<<<<<<<<<<<<<
 *             if not use_fast_converter:
 *                 raise core.FastOptionsError("fast_reader: exponent_style requires use_fast_converter")
 */
  /*else*/ {
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_fast_reader, __pyx_n_s_pop); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 222, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 222, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_use_fast_converter = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "astropy/io/ascii/cparser.pyx":223
 *         else:
 *             use_fast_converter = fast_reader.pop('use_fast_converter', True)
 *             if not use_fast_converter:             # <<<<<<<<<<<<<<
 *                 raise core.FastOptionsError("fast_reader: exponent_style requires use_fast_converter")
 *             if expchar.startswith('FORT'):
 */
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_use_fast_converter); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 223, __pyx_L1_error)
    __pyx_t_1 = ((!__pyx_t_3) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "astropy/io/ascii/cparser.pyx":224
 *             use_fast_converter = fast_reader.pop('use_fast_converter', True)
 *             if not use_fast_converter:
 *                 raise core.FastOptionsError("fast_reader: exponent_style requires use_fast_converter")             # <<<<<<<<<<<<<<
 *             if expchar.startswith('FORT'):
 *                 expchar = 'A'
 */
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_core); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 224, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_FastOptionsError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 224, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_kp_s_fast_reader_exponent_style_requi) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_kp_s_fast_reader_exponent_style_requi);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 224, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(0, 224, __pyx_L1_error)

      /* "astropy/io/ascii/cparser.pyx":223
 *         else:
 *             use_fast_converter = fast_reader.pop('use_fast_converter', True)
 *             if not use_fast_converter:             # <<<<<<<<<<<<<<
 *                 raise core.FastOptionsError("fast_reader: exponent_style requires use_fast_converter")
 *             if expchar.startswith('FORT'):
 */
    }

    /* "astropy/io/ascii/cparser.pyx":225
 *             if not use_fast_converter:
 *                 raise core.FastOptionsError("fast_reader: exponent_style requires use_fast_converter")
 *             if expchar.startswith('FORT'):             # <<<<<<<<<<<<<<
 *                 expchar = 'A'
 *         parallel = fast_reader.pop('parallel', False)
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_expchar, __pyx_n_s_startswith); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 225, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_n_s_FORT) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_n_s_FORT);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 225, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 225, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_1) {

      /* "astropy/io/ascii/cparser.pyx":226
 *                 raise core.FastOptionsError("fast_reader: exponent_style requires use_fast_converter")
 *             if expchar.startswith('FORT'):
 *                 expchar = 'A'             # <<<<<<<<<<<<<<
 *         parallel = fast_reader.pop('parallel', False)
 * 
 */
      __Pyx_INCREF(__pyx_n_s_A);
      __Pyx_DECREF_SET(__pyx_v_expchar, __pyx_n_s_A);

      /* "astropy/io/ascii/cparser.pyx":225
 *             if not use_fast_converter:
 *                 raise core.FastOptionsError("fast_reader: exponent_style requires use_fast_converter")
 *             if expchar.startswith('FORT'):             # <<<<<<<<<<<<<<
 *                 expchar = 'A'
 *         parallel = fast_reader.pop('parallel', False)
 */
    }
  }
  __pyx_L6:;

  /* "astropy/io/ascii/cparser.pyx":227
 *             if expchar.startswith('FORT'):
 *                 expchar = 'A'
 *         parallel = fast_reader.pop('parallel', False)             # <<<<<<<<<<<<<<
 * 
 *         if fast_reader:
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_fast_reader, __pyx_n_s_pop); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_parallel = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "astropy/io/ascii/cparser.pyx":229
 *         parallel = fast_reader.pop('parallel', False)
 * 
 *         if fast_reader:             # <<<<<<<<<<<<<<
 *             raise core.FastOptionsError("Invalid parameter in fast_reader dict")
 *         if parallel and os.name == 'nt':
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_fast_reader); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 229, __pyx_L1_error)
  if (unlikely(__pyx_t_1)) {

    /* "astropy/io/ascii/cparser.pyx":230
 * 
 *         if fast_reader:
 *             raise core.FastOptionsError("Invalid parameter in fast_reader dict")             # <<<<<<<<<<<<<<
 *         if parallel and os.name == 'nt':
 *             raise NotImplementedError("Multiprocessing is not yet supported on Windows")
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_core); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 230, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_FastOptionsError); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 230, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_4, __pyx_kp_s_Invalid_parameter_in_fast_reader) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_kp_s_Invalid_parameter_in_fast_reader);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 230, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 230, __pyx_L1_error)

    /* "astropy/io/ascii/cparser.pyx":229
 *         parallel = fast_reader.pop('parallel', False)
 * 
 *         if fast_reader:             # <<<<<<<<<<<<<<
 *             raise core.FastOptionsError("Invalid parameter in fast_reader dict")
 *         if parallel and os.name == 'nt':
 */
  }

  /* "astropy/io/ascii/cparser.pyx":231
 *         if fast_reader:
 *             raise core.FastOptionsError("Invalid parameter in fast_reader dict")
 *         if parallel and os.name == 'nt':             # <<<<<<<<<<<<<<
 *             raise NotImplementedError("Multiprocessing is not yet supported on Windows")
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_parallel); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 231, __pyx_L1_error)
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L11_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_os); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_t_6, __pyx_n_s_nt, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_1 = __pyx_t_3;
  __pyx_L11_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "astropy/io/ascii/cparser.pyx":232
 *             raise core.FastOptionsError("Invalid parameter in fast_reader dict")
 *         if parallel and os.name == 'nt':
 *             raise NotImplementedError("Multiprocessing is not yet supported on Windows")             # <<<<<<<<<<<<<<
 * 
 *         if comment is None:
 */
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_NotImplementedError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 232, __pyx_L1_error)

    /* "astropy/io/ascii/cparser.pyx":231
 *         if fast_reader:
 *             raise core.FastOptionsError("Invalid parameter in fast_reader dict")
 *         if parallel and os.name == 'nt':             # <<<<<<<<<<<<<<
 *             raise NotImplementedError("Multiprocessing is not yet supported on Windows")
 * 
 */
  }

  /* "astropy/io/ascii/cparser.pyx":234
 *             raise NotImplementedError("Multiprocessing is not yet supported on Windows")
 * 
 *         if comment is None:             # <<<<<<<<<<<<<<
 *             comment = '\x00' # tokenizer ignores all comments if comment='\x00'
 *         self.tokenizer = create_tokenizer(ord(delimiter), ord(comment),
 */
  __pyx_t_1 = (__pyx_v_comment == Py_None);
  __pyx_t_3 = (__pyx_t_1 != 0);
  if (__pyx_t_3) {

    /* "astropy/io/ascii/cparser.pyx":235
 * 
 *         if comment is None:
 *             comment = '\x00' # tokenizer ignores all comments if comment='\x00'             # <<<<<<<<<<<<<<
 *         self.tokenizer = create_tokenizer(ord(delimiter), ord(comment),
 *                                           ord(quotechar), ord(expchar),
 */
    __Pyx_INCREF(__pyx_kp_s__12);
    __Pyx_DECREF_SET(__pyx_v_comment, __pyx_kp_s__12);

    /* "astropy/io/ascii/cparser.pyx":234
 *             raise NotImplementedError("Multiprocessing is not yet supported on Windows")
 * 
 *         if comment is None:             # <<<<<<<<<<<<<<
 *             comment = '\x00' # tokenizer ignores all comments if comment='\x00'
 *         self.tokenizer = create_tokenizer(ord(delimiter), ord(comment),
 */
  }

  /* "astropy/io/ascii/cparser.pyx":236
 *         if comment is None:
 *             comment = '\x00' # tokenizer ignores all comments if comment='\x00'
 *         self.tokenizer = create_tokenizer(ord(delimiter), ord(comment),             # <<<<<<<<<<<<<<
 *                                           ord(quotechar), ord(expchar),
 *                                           fill_extra_cols,
 */
  __pyx_t_7 = __Pyx_PyObject_Ord(__pyx_v_delimiter); if (unlikely(__pyx_t_7 == ((long)(long)(Py_UCS4)-1))) __PYX_ERR(0, 236, __pyx_L1_error)
  __pyx_t_8 = __Pyx_PyObject_Ord(__pyx_v_comment); if (unlikely(__pyx_t_8 == ((long)(long)(Py_UCS4)-1))) __PYX_ERR(0, 236, __pyx_L1_error)

  /* "astropy/io/ascii/cparser.pyx":237
 *             comment = '\x00' # tokenizer ignores all comments if comment='\x00'
 *         self.tokenizer = create_tokenizer(ord(delimiter), ord(comment),
 *                                           ord(quotechar), ord(expchar),             # <<<<<<<<<<<<<<
 *                                           fill_extra_cols,
 *                                           strip_line_whitespace,
 */
  __pyx_t_9 = __Pyx_PyObject_Ord(__pyx_v_quotechar); if (unlikely(__pyx_t_9 == ((long)(long)(Py_UCS4)-1))) __PYX_ERR(0, 237, __pyx_L1_error)
  __pyx_t_10 = __Pyx_PyObject_Ord(__pyx_v_expchar); if (unlikely(__pyx_t_10 == ((long)(long)(Py_UCS4)-1))) __PYX_ERR(0, 237, __pyx_L1_error)

  /* "astropy/io/ascii/cparser.pyx":238
 *         self.tokenizer = create_tokenizer(ord(delimiter), ord(comment),
 *                                           ord(quotechar), ord(expchar),
 *                                           fill_extra_cols,             # <<<<<<<<<<<<<<
 *                                           strip_line_whitespace,
 *                                           strip_line_fields,
 */
  __pyx_t_11 = __Pyx_PyInt_As_int(__pyx_v_fill_extra_cols); if (unlikely((__pyx_t_11 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 238, __pyx_L1_error)

  /* "astropy/io/ascii/cparser.pyx":239
 *                                           ord(quotechar), ord(expchar),
 *                                           fill_extra_cols,
 *                                           strip_line_whitespace,             # <<<<<<<<<<<<<<
 *                                           strip_line_fields,
 *                                           use_fast_converter)
 */
  __pyx_t_12 = __Pyx_PyInt_As_int(__pyx_v_strip_line_whitespace); if (unlikely((__pyx_t_12 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 239, __pyx_L1_error)

  /* "astropy/io/ascii/cparser.pyx":240
 *                                           fill_extra_cols,
 *                                           strip_line_whitespace,
 *                                           strip_line_fields,             # <<<<<<<<<<<<<<
 *                                           use_fast_converter)
 *         self.source = None
 */
  __pyx_t_13 = __Pyx_PyInt_As_int(__pyx_v_strip_line_fields); if (unlikely((__pyx_t_13 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 240, __pyx_L1_error)

  /* "astropy/io/ascii/cparser.pyx":241
 *                                           strip_line_whitespace,
 *                                           strip_line_fields,
 *                                           use_fast_converter)             # <<<<<<<<<<<<<<
 *         self.source = None
 *         if source is not None:
 */
  __pyx_t_14 = __Pyx_PyInt_As_int(__pyx_v_use_fast_converter); if (unlikely((__pyx_t_14 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 241, __pyx_L1_error)

  /* "astropy/io/ascii/cparser.pyx":236
 *         if comment is None:
 *             comment = '\x00' # tokenizer ignores all comments if comment='\x00'
 *         self.tokenizer = create_tokenizer(ord(delimiter), ord(comment),             # <<<<<<<<<<<<<<
 *                                           ord(quotechar), ord(expchar),
 *                                           fill_extra_cols,
 */
  __pyx_v_self->tokenizer = create_tokenizer(__pyx_t_7, __pyx_t_8, __pyx_t_9, __pyx_t_10, __pyx_t_11, __pyx_t_12, __pyx_t_13, __pyx_t_14);

  /* "astropy/io/ascii/cparser.pyx":242
 *                                           strip_line_fields,
 *                                           use_fast_converter)
 *         self.source = None             # <<<<<<<<<<<<<<
 *         if source is not None:
 *             self.setup_tokenizer(source)
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->source);
  __Pyx_DECREF(__pyx_v_self->source);
  __pyx_v_self->source = Py_None;

  /* "astropy/io/ascii/cparser.pyx":243
 *                                           use_fast_converter)
 *         self.source = None
 *         if source is not None:             # <<<<<<<<<<<<<<
 *             self.setup_tokenizer(source)
 *         self.header_start = header_start
 */
  __pyx_t_3 = (__pyx_v_source != Py_None);
  __pyx_t_1 = (__pyx_t_3 != 0);
  if (__pyx_t_1) {

    /* "astropy/io/ascii/cparser.pyx":244
 *         self.source = None
 *         if source is not None:
 *             self.setup_tokenizer(source)             # <<<<<<<<<<<<<<
 *         self.header_start = header_start
 *         self.data_start = data_start
 */
    __pyx_t_6 = ((struct __pyx_vtabstruct_7astropy_2io_5ascii_7cparser_CParser *)__pyx_v_self->__pyx_vtab)->setup_tokenizer(__pyx_v_self, __pyx_v_source, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 244, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "astropy/io/ascii/cparser.pyx":243
 *                                           use_fast_converter)
 *         self.source = None
 *         if source is not None:             # <<<<<<<<<<<<<<
 *             self.setup_tokenizer(source)
 *         self.header_start = header_start
 */
  }

  /* "astropy/io/ascii/cparser.pyx":245
 *         if source is not None:
 *             self.setup_tokenizer(source)
 *         self.header_start = header_start             # <<<<<<<<<<<<<<
 *         self.data_start = data_start
 *         self.data_end = data_end
 */
  __Pyx_INCREF(__pyx_v_header_start);
  __Pyx_GIVEREF(__pyx_v_header_start);
  __Pyx_GOTREF(__pyx_v_self->header_start);
  __Pyx_DECREF(__pyx_v_self->header_start);
  __pyx_v_self->header_start = __pyx_v_header_start;

  /* "astropy/io/ascii/cparser.pyx":246
 *             self.setup_tokenizer(source)
 *         self.header_start = header_start
 *         self.data_start = data_start             # <<<<<<<<<<<<<<
 *         self.data_end = data_end
 *         self.names = names
 */
  __pyx_t_14 = __Pyx_PyInt_As_int(__pyx_v_data_start); if (unlikely((__pyx_t_14 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 246, __pyx_L1_error)
  __pyx_v_self->data_start = __pyx_t_14;

  /* "astropy/io/ascii/cparser.pyx":247
 *         self.header_start = header_start
 *         self.data_start = data_start
 *         self.data_end = data_end             # <<<<<<<<<<<<<<
 *         self.names = names
 *         self.include_names = include_names
 */
  __Pyx_INCREF(__pyx_v_data_end);
  __Pyx_GIVEREF(__pyx_v_data_end);
  __Pyx_GOTREF(__pyx_v_self->data_end);
  __Pyx_DECREF(__pyx_v_self->data_end);
  __pyx_v_self->data_end = __pyx_v_data_end;

  /* "astropy/io/ascii/cparser.pyx":248
 *         self.data_start = data_start
 *         self.data_end = data_end
 *         self.names = names             # <<<<<<<<<<<<<<
 *         self.include_names = include_names
 *         self.exclude_names = exclude_names
 */
  __Pyx_INCREF(__pyx_v_names);
  __Pyx_GIVEREF(__pyx_v_names);
  __Pyx_GOTREF(__pyx_v_self->names);
  __Pyx_DECREF(__pyx_v_self->names);
  __pyx_v_self->names = __pyx_v_names;

  /* "astropy/io/ascii/cparser.pyx":249
 *         self.data_end = data_end
 *         self.names = names
 *         self.include_names = include_names             # <<<<<<<<<<<<<<
 *         self.exclude_names = exclude_names
 *         self.fill_values = fill_values
 */
  __Pyx_INCREF(__pyx_v_include_names);
  __Pyx_GIVEREF(__pyx_v_include_names);
  __Pyx_GOTREF(__pyx_v_self->include_names);
  __Pyx_DECREF(__pyx_v_self->include_names);
  __pyx_v_self->include_names = __pyx_v_include_names;

  /* "astropy/io/ascii/cparser.pyx":250
 *         self.names = names
 *         self.include_names = include_names
 *         self.exclude_names = exclude_names             # <<<<<<<<<<<<<<
 *         self.fill_values = fill_values
 *         self.fill_include_names = fill_include_names
 */
  __Pyx_INCREF(__pyx_v_exclude_names);
  __Pyx_GIVEREF(__pyx_v_exclude_names);
  __Pyx_GOTREF(__pyx_v_self->exclude_names);
  __Pyx_DECREF(__pyx_v_self->exclude_names);
  __pyx_v_self->exclude_names = __pyx_v_exclude_names;

  /* "astropy/io/ascii/cparser.pyx":251
 *         self.include_names = include_names
 *         self.exclude_names = exclude_names
 *         self.fill_values = fill_values             # <<<<<<<<<<<<<<
 *         self.fill_include_names = fill_include_names
 *         self.fill_exclude_names = fill_exclude_names
 */
  __Pyx_INCREF(__pyx_v_fill_values);
  __Pyx_GIVEREF(__pyx_v_fill_values);
  __Pyx_GOTREF(__pyx_v_self->fill_values);
  __Pyx_DECREF(__pyx_v_self->fill_values);
  __pyx_v_self->fill_values = __pyx_v_fill_values;

  /* "astropy/io/ascii/cparser.pyx":252
 *         self.exclude_names = exclude_names
 *         self.fill_values = fill_values
 *         self.fill_include_names = fill_include_names             # <<<<<<<<<<<<<<
 *         self.fill_exclude_names = fill_exclude_names
 *         self.fill_names = None
 */
  __Pyx_INCREF(__pyx_v_fill_include_names);
  __Pyx_GIVEREF(__pyx_v_fill_include_names);
  __Pyx_GOTREF(__pyx_v_self->fill_include_names);
  __Pyx_DECREF(__pyx_v_self->fill_include_names);
  __pyx_v_self->fill_include_names = __pyx_v_fill_include_names;

  /* "astropy/io/ascii/cparser.pyx":253
 *         self.fill_values = fill_values
 *         self.fill_include_names = fill_include_names
 *         self.fill_exclude_names = fill_exclude_names             # <<<<<<<<<<<<<<
 *         self.fill_names = None
 *         self.fill_extra_cols = fill_extra_cols
 */
  __Pyx_INCREF(__pyx_v_fill_exclude_names);
  __Pyx_GIVEREF(__pyx_v_fill_exclude_names);
  __Pyx_GOTREF(__pyx_v_self->fill_exclude_names);
  __Pyx_DECREF(__pyx_v_self->fill_exclude_names);
  __pyx_v_self->fill_exclude_names = __pyx_v_fill_exclude_names;

  /* "astropy/io/ascii/cparser.pyx":254
 *         self.fill_include_names = fill_include_names
 *         self.fill_exclude_names = fill_exclude_names
 *         self.fill_names = None             # <<<<<<<<<<<<<<
 *         self.fill_extra_cols = fill_extra_cols
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->fill_names);
  __Pyx_DECREF(__pyx_v_self->fill_names);
  __pyx_v_self->fill_names = Py_None;

  /* "astropy/io/ascii/cparser.pyx":255
 *         self.fill_exclude_names = fill_exclude_names
 *         self.fill_names = None
 *         self.fill_extra_cols = fill_extra_cols             # <<<<<<<<<<<<<<
 * 
 *         # parallel=True indicates that we should use the CPU count
 */
  __pyx_t_14 = __Pyx_PyInt_As_int(__pyx_v_fill_extra_cols); if (unlikely((__pyx_t_14 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 255, __pyx_L1_error)
  __pyx_v_self->fill_extra_cols = __pyx_t_14;

  /* "astropy/io/ascii/cparser.pyx":258
 * 
 *         # parallel=True indicates that we should use the CPU count
 *         if parallel is True:             # <<<<<<<<<<<<<<
 *             parallel = multiprocessing.cpu_count()
 *         # If parallel = 1 or 0, don't use multiprocessing
 */
  __pyx_t_1 = (__pyx_v_parallel == Py_True);
  __pyx_t_3 = (__pyx_t_1 != 0);
  if (__pyx_t_3) {

    /* "astropy/io/ascii/cparser.pyx":259
 *         # parallel=True indicates that we should use the CPU count
 *         if parallel is True:
 *             parallel = multiprocessing.cpu_count()             # <<<<<<<<<<<<<<
 *         # If parallel = 1 or 0, don't use multiprocessing
 *         elif parallel is not False and parallel < 2:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_multiprocessing); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 259, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_cpu_count); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 259, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_6 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 259, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_parallel, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "astropy/io/ascii/cparser.pyx":258
 * 
 *         # parallel=True indicates that we should use the CPU count
 *         if parallel is True:             # <<<<<<<<<<<<<<
 *             parallel = multiprocessing.cpu_count()
 *         # If parallel = 1 or 0, don't use multiprocessing
 */
    goto __pyx_L15;
  }

  /* "astropy/io/ascii/cparser.pyx":261
 *             parallel = multiprocessing.cpu_count()
 *         # If parallel = 1 or 0, don't use multiprocessing
 *         elif parallel is not False and parallel < 2:             # <<<<<<<<<<<<<<
 *             parallel = False
 *         self.parallel = parallel
 */
  __pyx_t_1 = (__pyx_v_parallel != Py_False);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_3 = __pyx_t_2;
    goto __pyx_L16_bool_binop_done;
  }
  __pyx_t_6 = PyObject_RichCompare(__pyx_v_parallel, __pyx_int_2, Py_LT); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 261, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_3 = __pyx_t_2;
  __pyx_L16_bool_binop_done:;
  if (__pyx_t_3) {

    /* "astropy/io/ascii/cparser.pyx":262
 *         # If parallel = 1 or 0, don't use multiprocessing
 *         elif parallel is not False and parallel < 2:
 *             parallel = False             # <<<<<<<<<<<<<<
 *         self.parallel = parallel
 * 
 */
    __Pyx_INCREF(Py_False);
    __Pyx_DECREF_SET(__pyx_v_parallel, Py_False);

    /* "astropy/io/ascii/cparser.pyx":261
 *             parallel = multiprocessing.cpu_count()
 *         # If parallel = 1 or 0, don't use multiprocessing
 *         elif parallel is not False and parallel < 2:             # <<<<<<<<<<<<<<
 *             parallel = False
 *         self.parallel = parallel
 */
  }
  __pyx_L15:;

  /* "astropy/io/ascii/cparser.pyx":263
 *         elif parallel is not False and parallel < 2:
 *             parallel = False
 *         self.parallel = parallel             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __Pyx_INCREF(__pyx_v_parallel);
  __Pyx_GIVEREF(__pyx_v_parallel);
  __Pyx_GOTREF(__pyx_v_self->parallel);
  __Pyx_DECREF(__pyx_v_self->parallel);
  __pyx_v_self->parallel = __pyx_v_parallel;

  /* "astropy/io/ascii/cparser.pyx":195
 *         object header_start
 * 
 *     def __cinit__(self, source, strip_line_whitespace, strip_line_fields,             # <<<<<<<<<<<<<<
 *                   delimiter=',',
 *                   comment=None,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("astropy.io.ascii.cparser.CParser.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_expchar);
  __Pyx_XDECREF(__pyx_v_use_fast_converter);
  __Pyx_XDECREF(__pyx_v_parallel);
  __Pyx_XDECREF(__pyx_v_comment);
  __Pyx_XDECREF(__pyx_v_fast_reader);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "astropy/io/ascii/cparser.pyx":265
 *         self.parallel = parallel
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.tokenizer:
 *             delete_tokenizer(self.tokenizer) # perform C memory cleanup
 */

/* Python wrapper */
static void __pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_2__dealloc__(((struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_2__dealloc__(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "astropy/io/ascii/cparser.pyx":266
 * 
 *     def __dealloc__(self):
 *         if self.tokenizer:             # <<<<<<<<<<<<<<
 *             delete_tokenizer(self.tokenizer) # perform C memory cleanup
 * 
 */
  __pyx_t_1 = (__pyx_v_self->tokenizer != 0);
  if (__pyx_t_1) {

    /* "astropy/io/ascii/cparser.pyx":267
 *     def __dealloc__(self):
 *         if self.tokenizer:
 *             delete_tokenizer(self.tokenizer) # perform C memory cleanup             # <<<<<<<<<<<<<<
 * 
 *     cdef get_error(self, code, num_rows, msg):
 */
    delete_tokenizer(__pyx_v_self->tokenizer);

    /* "astropy/io/ascii/cparser.pyx":266
 * 
 *     def __dealloc__(self):
 *         if self.tokenizer:             # <<<<<<<<<<<<<<
 *             delete_tokenizer(self.tokenizer) # perform C memory cleanup
 * 
 */
  }

  /* "astropy/io/ascii/cparser.pyx":265
 *         self.parallel = parallel
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.tokenizer:
 *             delete_tokenizer(self.tokenizer) # perform C memory cleanup
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "astropy/io/ascii/cparser.pyx":269
 *             delete_tokenizer(self.tokenizer) # perform C memory cleanup
 * 
 *     cdef get_error(self, code, num_rows, msg):             # <<<<<<<<<<<<<<
 *         err_msg = ERR_CODES.get(code, "unknown error")
 * 
 */

static PyObject *__pyx_f_7astropy_2io_5ascii_7cparser_7CParser_get_error(CYTHON_UNUSED struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self, PyObject *__pyx_v_code, PyObject *__pyx_v_num_rows, PyObject *__pyx_v_msg) {
  PyObject *__pyx_v_err_msg = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("get_error", 0);

  /* "astropy/io/ascii/cparser.pyx":270
 * 
 *     cdef get_error(self, code, num_rows, msg):
 *         err_msg = ERR_CODES.get(code, "unknown error")             # <<<<<<<<<<<<<<
 * 
 *         # error code is lambda function taking current line as input
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ERR_CODES); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_get); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_code, __pyx_kp_s_unknown_error};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 270, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_code, __pyx_kp_s_unknown_error};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 270, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 270, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_code);
    __Pyx_GIVEREF(__pyx_v_code);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_code);
    __Pyx_INCREF(__pyx_kp_s_unknown_error);
    __Pyx_GIVEREF(__pyx_kp_s_unknown_error);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_kp_s_unknown_error);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 270, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_err_msg = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "astropy/io/ascii/cparser.pyx":273
 * 
 *         # error code is lambda function taking current line as input
 *         if callable(err_msg):             # <<<<<<<<<<<<<<
 *             err_msg = err_msg(num_rows + 1)
 * 
 */
  __pyx_t_6 = __Pyx_PyCallable_Check(__pyx_v_err_msg); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 273, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 != 0);
  if (__pyx_t_7) {

    /* "astropy/io/ascii/cparser.pyx":274
 *         # error code is lambda function taking current line as input
 *         if callable(err_msg):
 *             err_msg = err_msg(num_rows + 1)             # <<<<<<<<<<<<<<
 * 
 *         return CParserError("{0}: {1}".format(msg, err_msg))
 */
    __pyx_t_3 = __Pyx_PyInt_AddObjC(__pyx_v_num_rows, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_err_msg);
    __pyx_t_5 = __pyx_v_err_msg; __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_err_msg, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "astropy/io/ascii/cparser.pyx":273
 * 
 *         # error code is lambda function taking current line as input
 *         if callable(err_msg):             # <<<<<<<<<<<<<<
 *             err_msg = err_msg(num_rows + 1)
 * 
 */
  }

  /* "astropy/io/ascii/cparser.pyx":276
 *             err_msg = err_msg(num_rows + 1)
 * 
 *         return CParserError("{0}: {1}".format(msg, err_msg))             # <<<<<<<<<<<<<<
 * 
 *     cdef raise_error(self, msg):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_CParserError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_0_1, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_msg, __pyx_v_err_msg};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 276, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_msg, __pyx_v_err_msg};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 276, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  {
    __pyx_t_9 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__pyx_t_8) {
      __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_8); __pyx_t_8 = NULL;
    }
    __Pyx_INCREF(__pyx_v_msg);
    __Pyx_GIVEREF(__pyx_v_msg);
    PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_4, __pyx_v_msg);
    __Pyx_INCREF(__pyx_v_err_msg);
    __Pyx_GIVEREF(__pyx_v_err_msg);
    PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_4, __pyx_v_err_msg);
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_9, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "astropy/io/ascii/cparser.pyx":269
 *             delete_tokenizer(self.tokenizer) # perform C memory cleanup
 * 
 *     cdef get_error(self, code, num_rows, msg):             # <<<<<<<<<<<<<<
 *         err_msg = ERR_CODES.get(code, "unknown error")
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("astropy.io.ascii.cparser.CParser.get_error", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_err_msg);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "astropy/io/ascii/cparser.pyx":278
 *         return CParserError("{0}: {1}".format(msg, err_msg))
 * 
 *     cdef raise_error(self, msg):             # <<<<<<<<<<<<<<
 *         raise self.get_error(self.tokenizer.code, self.tokenizer.num_rows, msg)
 * 
 */

static PyObject *__pyx_f_7astropy_2io_5ascii_7cparser_7CParser_raise_error(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self, PyObject *__pyx_v_msg) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("raise_error", 0);

  /* "astropy/io/ascii/cparser.pyx":279
 * 
 *     cdef raise_error(self, msg):
 *         raise self.get_error(self.tokenizer.code, self.tokenizer.num_rows, msg)             # <<<<<<<<<<<<<<
 * 
 *     cpdef setup_tokenizer(self, source):
 */
  __pyx_t_1 = __Pyx_PyInt_From_err_code(__pyx_v_self->tokenizer->code); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->tokenizer->num_rows); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = ((struct __pyx_vtabstruct_7astropy_2io_5ascii_7cparser_CParser *)__pyx_v_self->__pyx_vtab)->get_error(__pyx_v_self, __pyx_t_1, __pyx_t_2, __pyx_v_msg); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_Raise(__pyx_t_3, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __PYX_ERR(0, 279, __pyx_L1_error)

  /* "astropy/io/ascii/cparser.pyx":278
 *         return CParserError("{0}: {1}".format(msg, err_msg))
 * 
 *     cdef raise_error(self, msg):             # <<<<<<<<<<<<<<
 *         raise self.get_error(self.tokenizer.code, self.tokenizer.num_rows, msg)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("astropy.io.ascii.cparser.CParser.raise_error", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "astropy/io/ascii/cparser.pyx":281
 *         raise self.get_error(self.tokenizer.code, self.tokenizer.num_rows, msg)
 * 
 *     cpdef setup_tokenizer(self, source):             # <<<<<<<<<<<<<<
 *         cdef FileString fstring
 * 
 */

static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_5setup_tokenizer(PyObject *__pyx_v_self, PyObject *__pyx_v_source); /*proto*/
static PyObject *__pyx_f_7astropy_2io_5ascii_7cparser_7CParser_setup_tokenizer(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self, PyObject *__pyx_v_source, int __pyx_skip_dispatch) {
  struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString *__pyx_v_fstring = 0;
  PyObject *__pyx_v_file_obj = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char *__pyx_t_15;
  __Pyx_RefNannySetupContext("setup_tokenizer", 0);
  __Pyx_INCREF(__pyx_v_source);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_setup_tokenizer); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 281, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_5setup_tokenizer)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_source) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_source);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 281, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "astropy/io/ascii/cparser.pyx":284
 *         cdef FileString fstring
 * 
 *         if isinstance(source, six.string_types): # filename or data             # <<<<<<<<<<<<<<
 *             if '\n' not in source and '\r' not in source: # filename
 *                 fstring = FileString(source)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_six); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_string_types); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = PyObject_IsInstance(__pyx_v_source, __pyx_t_2); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 284, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (__pyx_t_6) {

    /* "astropy/io/ascii/cparser.pyx":285
 * 
 *         if isinstance(source, six.string_types): # filename or data
 *             if '\n' not in source and '\r' not in source: # filename             # <<<<<<<<<<<<<<
 *                 fstring = FileString(source)
 *                 self.tokenizer.source = <char *>fstring.mmap_ptr
 */
    __pyx_t_5 = (__Pyx_PySequence_ContainsTF(__pyx_kp_s__13, __pyx_v_source, Py_NE)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 285, __pyx_L1_error)
    __pyx_t_7 = (__pyx_t_5 != 0);
    if (__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_7 = (__Pyx_PySequence_ContainsTF(__pyx_kp_s__14, __pyx_v_source, Py_NE)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 285, __pyx_L1_error)
    __pyx_t_5 = (__pyx_t_7 != 0);
    __pyx_t_6 = __pyx_t_5;
    __pyx_L5_bool_binop_done:;
    if (__pyx_t_6) {

      /* "astropy/io/ascii/cparser.pyx":286
 *         if isinstance(source, six.string_types): # filename or data
 *             if '\n' not in source and '\r' not in source: # filename
 *                 fstring = FileString(source)             # <<<<<<<<<<<<<<
 *                 self.tokenizer.source = <char *>fstring.mmap_ptr
 *                 self.source_ptr = <char *>fstring.mmap_ptr
 */
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7astropy_2io_5ascii_7cparser_FileString), __pyx_v_source); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 286, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_v_fstring = ((struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString *)__pyx_t_2);
      __pyx_t_2 = 0;

      /* "astropy/io/ascii/cparser.pyx":287
 *             if '\n' not in source and '\r' not in source: # filename
 *                 fstring = FileString(source)
 *                 self.tokenizer.source = <char *>fstring.mmap_ptr             # <<<<<<<<<<<<<<
 *                 self.source_ptr = <char *>fstring.mmap_ptr
 *                 self.source = fstring
 */
      __pyx_v_self->tokenizer->source = ((char *)__pyx_v_fstring->mmap_ptr);

      /* "astropy/io/ascii/cparser.pyx":288
 *                 fstring = FileString(source)
 *                 self.tokenizer.source = <char *>fstring.mmap_ptr
 *                 self.source_ptr = <char *>fstring.mmap_ptr             # <<<<<<<<<<<<<<
 *                 self.source = fstring
 *                 self.tokenizer.source_len = <size_t>len(fstring)
 */
      __pyx_v_self->source_ptr = ((char *)__pyx_v_fstring->mmap_ptr);

      /* "astropy/io/ascii/cparser.pyx":289
 *                 self.tokenizer.source = <char *>fstring.mmap_ptr
 *                 self.source_ptr = <char *>fstring.mmap_ptr
 *                 self.source = fstring             # <<<<<<<<<<<<<<
 *                 self.tokenizer.source_len = <size_t>len(fstring)
 *                 return
 */
      __Pyx_INCREF(((PyObject *)__pyx_v_fstring));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_fstring));
      __Pyx_GOTREF(__pyx_v_self->source);
      __Pyx_DECREF(__pyx_v_self->source);
      __pyx_v_self->source = ((PyObject *)__pyx_v_fstring);

      /* "astropy/io/ascii/cparser.pyx":290
 *                 self.source_ptr = <char *>fstring.mmap_ptr
 *                 self.source = fstring
 *                 self.tokenizer.source_len = <size_t>len(fstring)             # <<<<<<<<<<<<<<
 *                 return
 *             # Otherwise, source is the actual data so we leave it be
 */
      __pyx_t_8 = PyObject_Length(((PyObject *)__pyx_v_fstring)); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 290, __pyx_L1_error)
      __pyx_v_self->tokenizer->source_len = ((size_t)__pyx_t_8);

      /* "astropy/io/ascii/cparser.pyx":291
 *                 self.source = fstring
 *                 self.tokenizer.source_len = <size_t>len(fstring)
 *                 return             # <<<<<<<<<<<<<<
 *             # Otherwise, source is the actual data so we leave it be
 *         elif hasattr(source, 'read'): # file-like object
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "astropy/io/ascii/cparser.pyx":285
 * 
 *         if isinstance(source, six.string_types): # filename or data
 *             if '\n' not in source and '\r' not in source: # filename             # <<<<<<<<<<<<<<
 *                 fstring = FileString(source)
 *                 self.tokenizer.source = <char *>fstring.mmap_ptr
 */
    }

    /* "astropy/io/ascii/cparser.pyx":284
 *         cdef FileString fstring
 * 
 *         if isinstance(source, six.string_types): # filename or data             # <<<<<<<<<<<<<<
 *             if '\n' not in source and '\r' not in source: # filename
 *                 fstring = FileString(source)
 */
    goto __pyx_L3;
  }

  /* "astropy/io/ascii/cparser.pyx":293
 *                 return
 *             # Otherwise, source is the actual data so we leave it be
 *         elif hasattr(source, 'read'): # file-like object             # <<<<<<<<<<<<<<
 *             with get_readable_fileobj(source) as file_obj:
 *                 source = file_obj.read()
 */
  __pyx_t_6 = __Pyx_HasAttr(__pyx_v_source, __pyx_n_s_read); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 293, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_6 != 0);
  if (__pyx_t_5) {

    /* "astropy/io/ascii/cparser.pyx":294
 *             # Otherwise, source is the actual data so we leave it be
 *         elif hasattr(source, 'read'): # file-like object
 *             with get_readable_fileobj(source) as file_obj:             # <<<<<<<<<<<<<<
 *                 source = file_obj.read()
 *         elif isinstance(source, FileString):
 */
    /*with:*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_get_readable_fileobj); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 294, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_source) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_source);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 294, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_9 = __Pyx_PyObject_LookupSpecial(__pyx_t_2, __pyx_n_s_exit); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 294, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_3 = __Pyx_PyObject_LookupSpecial(__pyx_t_2, __pyx_n_s_enter); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 294, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 294, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __pyx_t_1;
      __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      /*try:*/ {
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
          __Pyx_XGOTREF(__pyx_t_10);
          __Pyx_XGOTREF(__pyx_t_11);
          __Pyx_XGOTREF(__pyx_t_12);
          /*try:*/ {
            __pyx_v_file_obj = __pyx_t_3;
            __pyx_t_3 = 0;

            /* "astropy/io/ascii/cparser.pyx":295
 *         elif hasattr(source, 'read'): # file-like object
 *             with get_readable_fileobj(source) as file_obj:
 *                 source = file_obj.read()             # <<<<<<<<<<<<<<
 *         elif isinstance(source, FileString):
 *             self.tokenizer.source = <char *>((<FileString>source).mmap_ptr)
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_file_obj, __pyx_n_s_read); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 295, __pyx_L11_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_1 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_1)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_1);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 295, __pyx_L11_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF_SET(__pyx_v_source, __pyx_t_3);
            __pyx_t_3 = 0;

            /* "astropy/io/ascii/cparser.pyx":294
 *             # Otherwise, source is the actual data so we leave it be
 *         elif hasattr(source, 'read'): # file-like object
 *             with get_readable_fileobj(source) as file_obj:             # <<<<<<<<<<<<<<
 *                 source = file_obj.read()
 *         elif isinstance(source, FileString):
 */
          }
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          goto __pyx_L16_try_end;
          __pyx_L11_error:;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          /*except:*/ {
            __Pyx_AddTraceback("astropy.io.ascii.cparser.CParser.setup_tokenizer", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_2, &__pyx_t_1) < 0) __PYX_ERR(0, 294, __pyx_L13_except_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_4 = PyTuple_Pack(3, __pyx_t_3, __pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 294, __pyx_L13_except_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_4, NULL);
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 294, __pyx_L13_except_error)
            __Pyx_GOTREF(__pyx_t_13);
            __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_13);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            if (__pyx_t_5 < 0) __PYX_ERR(0, 294, __pyx_L13_except_error)
            __pyx_t_6 = ((!(__pyx_t_5 != 0)) != 0);
            if (__pyx_t_6) {
              __Pyx_GIVEREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_2);
              __Pyx_XGIVEREF(__pyx_t_1);
              __Pyx_ErrRestoreWithState(__pyx_t_3, __pyx_t_2, __pyx_t_1);
              __pyx_t_3 = 0; __pyx_t_2 = 0; __pyx_t_1 = 0; 
              __PYX_ERR(0, 294, __pyx_L13_except_error)
            }
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            goto __pyx_L12_exception_handled;
          }
          __pyx_L13_except_error:;
          __Pyx_XGIVEREF(__pyx_t_10);
          __Pyx_XGIVEREF(__pyx_t_11);
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
          goto __pyx_L1_error;
          __pyx_L12_exception_handled:;
          __Pyx_XGIVEREF(__pyx_t_10);
          __Pyx_XGIVEREF(__pyx_t_11);
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
          __pyx_L16_try_end:;
        }
      }
      /*finally:*/ {
        /*normal exit:*/{
          if (__pyx_t_9) {
            __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_tuple__15, NULL);
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
            if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 294, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          }
          goto __pyx_L10;
        }
        __pyx_L10:;
      }
      goto __pyx_L20;
      __pyx_L7_error:;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L1_error;
      __pyx_L20:;
    }

    /* "astropy/io/ascii/cparser.pyx":293
 *                 return
 *             # Otherwise, source is the actual data so we leave it be
 *         elif hasattr(source, 'read'): # file-like object             # <<<<<<<<<<<<<<
 *             with get_readable_fileobj(source) as file_obj:
 *                 source = file_obj.read()
 */
    goto __pyx_L3;
  }

  /* "astropy/io/ascii/cparser.pyx":296
 *             with get_readable_fileobj(source) as file_obj:
 *                 source = file_obj.read()
 *         elif isinstance(source, FileString):             # <<<<<<<<<<<<<<
 *             self.tokenizer.source = <char *>((<FileString>source).mmap_ptr)
 *             self.source = source
 */
  __pyx_t_6 = __Pyx_TypeCheck(__pyx_v_source, __pyx_ptype_7astropy_2io_5ascii_7cparser_FileString); 
  __pyx_t_5 = (__pyx_t_6 != 0);
  if (__pyx_t_5) {

    /* "astropy/io/ascii/cparser.pyx":297
 *                 source = file_obj.read()
 *         elif isinstance(source, FileString):
 *             self.tokenizer.source = <char *>((<FileString>source).mmap_ptr)             # <<<<<<<<<<<<<<
 *             self.source = source
 *             self.tokenizer.source_len = <size_t>len(source)
 */
    __pyx_v_self->tokenizer->source = ((char *)((struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString *)__pyx_v_source)->mmap_ptr);

    /* "astropy/io/ascii/cparser.pyx":298
 *         elif isinstance(source, FileString):
 *             self.tokenizer.source = <char *>((<FileString>source).mmap_ptr)
 *             self.source = source             # <<<<<<<<<<<<<<
 *             self.tokenizer.source_len = <size_t>len(source)
 *             return
 */
    __Pyx_INCREF(__pyx_v_source);
    __Pyx_GIVEREF(__pyx_v_source);
    __Pyx_GOTREF(__pyx_v_self->source);
    __Pyx_DECREF(__pyx_v_self->source);
    __pyx_v_self->source = __pyx_v_source;

    /* "astropy/io/ascii/cparser.pyx":299
 *             self.tokenizer.source = <char *>((<FileString>source).mmap_ptr)
 *             self.source = source
 *             self.tokenizer.source_len = <size_t>len(source)             # <<<<<<<<<<<<<<
 *             return
 *         else:
 */
    __pyx_t_8 = PyObject_Length(__pyx_v_source); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 299, __pyx_L1_error)
    __pyx_v_self->tokenizer->source_len = ((size_t)__pyx_t_8);

    /* "astropy/io/ascii/cparser.pyx":300
 *             self.source = source
 *             self.tokenizer.source_len = <size_t>len(source)
 *             return             # <<<<<<<<<<<<<<
 *         else:
 *             try:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "astropy/io/ascii/cparser.pyx":296
 *             with get_readable_fileobj(source) as file_obj:
 *                 source = file_obj.read()
 *         elif isinstance(source, FileString):             # <<<<<<<<<<<<<<
 *             self.tokenizer.source = <char *>((<FileString>source).mmap_ptr)
 *             self.source = source
 */
  }

  /* "astropy/io/ascii/cparser.pyx":302
 *             return
 *         else:
 *             try:             # <<<<<<<<<<<<<<
 *                 source = '\n'.join(source) # iterable sequence of lines
 *             except TypeError:
 */
  /*else*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_12, &__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_11);
      /*try:*/ {

        /* "astropy/io/ascii/cparser.pyx":303
 *         else:
 *             try:
 *                 source = '\n'.join(source) # iterable sequence of lines             # <<<<<<<<<<<<<<
 *             except TypeError:
 *                 raise TypeError('Input "table" must be a file-like object, a '
 */
        __pyx_t_1 = __Pyx_PyString_Join(__pyx_kp_s__13, __pyx_v_source); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 303, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF_SET(__pyx_v_source, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "astropy/io/ascii/cparser.pyx":302
 *             return
 *         else:
 *             try:             # <<<<<<<<<<<<<<
 *                 source = '\n'.join(source) # iterable sequence of lines
 *             except TypeError:
 */
      }
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      goto __pyx_L26_try_end;
      __pyx_L21_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "astropy/io/ascii/cparser.pyx":304
 *             try:
 *                 source = '\n'.join(source) # iterable sequence of lines
 *             except TypeError:             # <<<<<<<<<<<<<<
 *                 raise TypeError('Input "table" must be a file-like object, a '
 *                                 'string (filename or data), or an iterable')
 */
      __pyx_t_14 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
      if (__pyx_t_14) {
        __Pyx_AddTraceback("astropy.io.ascii.cparser.CParser.setup_tokenizer", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3) < 0) __PYX_ERR(0, 304, __pyx_L23_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_3);

        /* "astropy/io/ascii/cparser.pyx":305
 *                 source = '\n'.join(source) # iterable sequence of lines
 *             except TypeError:
 *                 raise TypeError('Input "table" must be a file-like object, a '             # <<<<<<<<<<<<<<
 *                                 'string (filename or data), or an iterable')
 *         # Create a reference to the Python object so its char * pointer remains valid
 */
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 305, __pyx_L23_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(0, 305, __pyx_L23_except_error)
      }
      goto __pyx_L23_except_error;
      __pyx_L23_except_error:;

      /* "astropy/io/ascii/cparser.pyx":302
 *             return
 *         else:
 *             try:             # <<<<<<<<<<<<<<
 *                 source = '\n'.join(source) # iterable sequence of lines
 *             except TypeError:
 */
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_12, __pyx_t_11);
      goto __pyx_L1_error;
      __pyx_L26_try_end:;
    }
  }
  __pyx_L3:;

  /* "astropy/io/ascii/cparser.pyx":308
 *                                 'string (filename or data), or an iterable')
 *         # Create a reference to the Python object so its char * pointer remains valid
 *         self.source = source             # <<<<<<<<<<<<<<
 * 
 *         # encode in ASCII for char * handling
 */
  __Pyx_INCREF(__pyx_v_source);
  __Pyx_GIVEREF(__pyx_v_source);
  __Pyx_GOTREF(__pyx_v_self->source);
  __Pyx_DECREF(__pyx_v_self->source);
  __pyx_v_self->source = __pyx_v_source;

  /* "astropy/io/ascii/cparser.pyx":311
 * 
 *         # encode in ASCII for char * handling
 *         self.source_bytes = self.source.encode('ascii')             # <<<<<<<<<<<<<<
 *         self.tokenizer.source = self.source_bytes
 *         self.tokenizer.source_len = <size_t>len(self.source_bytes)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->source, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_1, __pyx_n_s_ascii) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_n_s_ascii);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->source_bytes);
  __Pyx_DECREF(__pyx_v_self->source_bytes);
  __pyx_v_self->source_bytes = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "astropy/io/ascii/cparser.pyx":312
 *         # encode in ASCII for char * handling
 *         self.source_bytes = self.source.encode('ascii')
 *         self.tokenizer.source = self.source_bytes             # <<<<<<<<<<<<<<
 *         self.tokenizer.source_len = <size_t>len(self.source_bytes)
 * 
 */
  if (unlikely(__pyx_v_self->source_bytes == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 312, __pyx_L1_error)
  }
  __pyx_t_15 = __Pyx_PyBytes_AsWritableString(__pyx_v_self->source_bytes); if (unlikely((!__pyx_t_15) && PyErr_Occurred())) __PYX_ERR(0, 312, __pyx_L1_error)
  __pyx_v_self->tokenizer->source = __pyx_t_15;

  /* "astropy/io/ascii/cparser.pyx":313
 *         self.source_bytes = self.source.encode('ascii')
 *         self.tokenizer.source = self.source_bytes
 *         self.tokenizer.source_len = <size_t>len(self.source_bytes)             # <<<<<<<<<<<<<<
 * 
 *     def read_header(self):
 */
  __pyx_t_3 = __pyx_v_self->source_bytes;
  __Pyx_INCREF(__pyx_t_3);
  if (unlikely(__pyx_t_3 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 313, __pyx_L1_error)
  }
  __pyx_t_8 = PyBytes_GET_SIZE(__pyx_t_3); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 313, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_self->tokenizer->source_len = ((size_t)__pyx_t_8);

  /* "astropy/io/ascii/cparser.pyx":281
 *         raise self.get_error(self.tokenizer.code, self.tokenizer.num_rows, msg)
 * 
 *     cpdef setup_tokenizer(self, source):             # <<<<<<<<<<<<<<
 *         cdef FileString fstring
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("astropy.io.ascii.cparser.CParser.setup_tokenizer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_fstring);
  __Pyx_XDECREF(__pyx_v_file_obj);
  __Pyx_XDECREF(__pyx_v_source);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_5setup_tokenizer(PyObject *__pyx_v_self, PyObject *__pyx_v_source); /*proto*/
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_5setup_tokenizer(PyObject *__pyx_v_self, PyObject *__pyx_v_source) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setup_tokenizer (wrapper)", 0);
  __pyx_r = __pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_4setup_tokenizer(((struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *)__pyx_v_self), ((PyObject *)__pyx_v_source));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_4setup_tokenizer(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self, PyObject *__pyx_v_source) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setup_tokenizer", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7astropy_2io_5ascii_7cparser_7CParser_setup_tokenizer(__pyx_v_self, __pyx_v_source, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("astropy.io.ascii.cparser.CParser.setup_tokenizer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "astropy/io/ascii/cparser.pyx":315
 *         self.tokenizer.source_len = <size_t>len(self.source_bytes)
 * 
 *     def read_header(self):             # <<<<<<<<<<<<<<
 *         self.tokenizer.source_pos = 0
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_7read_header(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_7read_header(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("read_header (wrapper)", 0);
  __pyx_r = __pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_6read_header(((struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_6read_header(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self) {
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_c = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  __Pyx_RefNannySetupContext("read_header", 0);

  /* "astropy/io/ascii/cparser.pyx":316
 * 
 *     def read_header(self):
 *         self.tokenizer.source_pos = 0             # <<<<<<<<<<<<<<
 * 
 *         # header_start is a valid line number
 */
  __pyx_v_self->tokenizer->source_pos = 0;

  /* "astropy/io/ascii/cparser.pyx":319
 * 
 *         # header_start is a valid line number
 *         if self.header_start is not None and self.header_start >= 0:             # <<<<<<<<<<<<<<
 *             if skip_lines(self.tokenizer, self.header_start, 1) != 0:
 *                 self.raise_error("an error occurred while advancing to the "
 */
  __pyx_t_2 = (__pyx_v_self->header_start != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_self->header_start, __pyx_int_0, Py_GE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 319, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 319, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "astropy/io/ascii/cparser.pyx":320
 *         # header_start is a valid line number
 *         if self.header_start is not None and self.header_start >= 0:
 *             if skip_lines(self.tokenizer, self.header_start, 1) != 0:             # <<<<<<<<<<<<<<
 *                 self.raise_error("an error occurred while advancing to the "
 *                                  "first header line")
 */
    __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_v_self->header_start); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 320, __pyx_L1_error)
    __pyx_t_1 = ((skip_lines(__pyx_v_self->tokenizer, __pyx_t_5, 1) != 0) != 0);
    if (__pyx_t_1) {

      /* "astropy/io/ascii/cparser.pyx":321
 *         if self.header_start is not None and self.header_start >= 0:
 *             if skip_lines(self.tokenizer, self.header_start, 1) != 0:
 *                 self.raise_error("an error occurred while advancing to the "             # <<<<<<<<<<<<<<
 *                                  "first header line")
 *             if tokenize(self.tokenizer, -1, 1, 0) != 0:
 */
      __pyx_t_4 = ((struct __pyx_vtabstruct_7astropy_2io_5ascii_7cparser_CParser *)__pyx_v_self->__pyx_vtab)->raise_error(__pyx_v_self, __pyx_kp_s_an_error_occurred_while_advancin); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 321, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "astropy/io/ascii/cparser.pyx":320
 *         # header_start is a valid line number
 *         if self.header_start is not None and self.header_start >= 0:
 *             if skip_lines(self.tokenizer, self.header_start, 1) != 0:             # <<<<<<<<<<<<<<
 *                 self.raise_error("an error occurred while advancing to the "
 *                                  "first header line")
 */
    }

    /* "astropy/io/ascii/cparser.pyx":323
 *                 self.raise_error("an error occurred while advancing to the "
 *                                  "first header line")
 *             if tokenize(self.tokenizer, -1, 1, 0) != 0:             # <<<<<<<<<<<<<<
 *                 self.raise_error("an error occurred while tokenizing the header line")
 *             self.header_names = []
 */
    __pyx_t_1 = ((tokenize(__pyx_v_self->tokenizer, -1, 1, 0) != 0) != 0);
    if (__pyx_t_1) {

      /* "astropy/io/ascii/cparser.pyx":324
 *                                  "first header line")
 *             if tokenize(self.tokenizer, -1, 1, 0) != 0:
 *                 self.raise_error("an error occurred while tokenizing the header line")             # <<<<<<<<<<<<<<
 *             self.header_names = []
 *             name = ''
 */
      __pyx_t_4 = ((struct __pyx_vtabstruct_7astropy_2io_5ascii_7cparser_CParser *)__pyx_v_self->__pyx_vtab)->raise_error(__pyx_v_self, __pyx_kp_s_an_error_occurred_while_tokenizi); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 324, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "astropy/io/ascii/cparser.pyx":323
 *                 self.raise_error("an error occurred while advancing to the "
 *                                  "first header line")
 *             if tokenize(self.tokenizer, -1, 1, 0) != 0:             # <<<<<<<<<<<<<<
 *                 self.raise_error("an error occurred while tokenizing the header line")
 *             self.header_names = []
 */
    }

    /* "astropy/io/ascii/cparser.pyx":325
 *             if tokenize(self.tokenizer, -1, 1, 0) != 0:
 *                 self.raise_error("an error occurred while tokenizing the header line")
 *             self.header_names = []             # <<<<<<<<<<<<<<
 *             name = ''
 * 
 */
    __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 325, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __Pyx_GOTREF(__pyx_v_self->header_names);
    __Pyx_DECREF(__pyx_v_self->header_names);
    __pyx_v_self->header_names = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "astropy/io/ascii/cparser.pyx":326
 *                 self.raise_error("an error occurred while tokenizing the header line")
 *             self.header_names = []
 *             name = ''             # <<<<<<<<<<<<<<
 * 
 *             for i in range(self.tokenizer.output_len[0]): # header is in first col string
 */
    __Pyx_INCREF(__pyx_kp_s__5);
    __pyx_v_name = __pyx_kp_s__5;

    /* "astropy/io/ascii/cparser.pyx":328
 *             name = ''
 * 
 *             for i in range(self.tokenizer.output_len[0]): # header is in first col string             # <<<<<<<<<<<<<<
 *                 c = self.tokenizer.output_cols[0][i] # next char in header string
 *                 if not c: # zero byte -- field terminator
 */
    __pyx_t_4 = __Pyx_PyInt_From_int((__pyx_v_self->tokenizer->output_len[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
      __pyx_t_4 = __pyx_t_6; __Pyx_INCREF(__pyx_t_4); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 328, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 328, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_7); __Pyx_INCREF(__pyx_t_6); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 328, __pyx_L1_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_4, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 328, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        } else {
          if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_7); __Pyx_INCREF(__pyx_t_6); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 328, __pyx_L1_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_4, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 328, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        }
      } else {
        __pyx_t_6 = __pyx_t_8(__pyx_t_4);
        if (unlikely(!__pyx_t_6)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 328, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_6);
      }
      __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "astropy/io/ascii/cparser.pyx":329
 * 
 *             for i in range(self.tokenizer.output_len[0]): # header is in first col string
 *                 c = self.tokenizer.output_cols[0][i] # next char in header string             # <<<<<<<<<<<<<<
 *                 if not c: # zero byte -- field terminator
 *                     if name:
 */
      __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 329, __pyx_L1_error)
      __pyx_t_6 = __Pyx_PyInt_From_char(((__pyx_v_self->tokenizer->output_cols[0])[__pyx_t_9])); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 329, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_c, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "astropy/io/ascii/cparser.pyx":330
 *             for i in range(self.tokenizer.output_len[0]): # header is in first col string
 *                 c = self.tokenizer.output_cols[0][i] # next char in header string
 *                 if not c: # zero byte -- field terminator             # <<<<<<<<<<<<<<
 *                     if name:
 *                         # replace empty placeholder with ''
 */
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_c); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 330, __pyx_L1_error)
      __pyx_t_3 = ((!__pyx_t_1) != 0);
      if (__pyx_t_3) {

        /* "astropy/io/ascii/cparser.pyx":331
 *                 c = self.tokenizer.output_cols[0][i] # next char in header string
 *                 if not c: # zero byte -- field terminator
 *                     if name:             # <<<<<<<<<<<<<<
 *                         # replace empty placeholder with ''
 *                         self.header_names.append(name.replace('\x01', ''))
 */
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_name); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 331, __pyx_L1_error)
        if (__pyx_t_3) {

          /* "astropy/io/ascii/cparser.pyx":333
 *                     if name:
 *                         # replace empty placeholder with ''
 *                         self.header_names.append(name.replace('\x01', ''))             # <<<<<<<<<<<<<<
 *                         name = ''
 *                     else:
 */
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_name, __pyx_n_s_replace); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 333, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_tuple__18, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 333, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_11 = __Pyx_PyObject_Append(__pyx_v_self->header_names, __pyx_t_10); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 333, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "astropy/io/ascii/cparser.pyx":334
 *                         # replace empty placeholder with ''
 *                         self.header_names.append(name.replace('\x01', ''))
 *                         name = ''             # <<<<<<<<<<<<<<
 *                     else:
 *                         break # end of string
 */
          __Pyx_INCREF(__pyx_kp_s__5);
          __Pyx_DECREF_SET(__pyx_v_name, __pyx_kp_s__5);

          /* "astropy/io/ascii/cparser.pyx":331
 *                 c = self.tokenizer.output_cols[0][i] # next char in header string
 *                 if not c: # zero byte -- field terminator
 *                     if name:             # <<<<<<<<<<<<<<
 *                         # replace empty placeholder with ''
 *                         self.header_names.append(name.replace('\x01', ''))
 */
          goto __pyx_L11;
        }

        /* "astropy/io/ascii/cparser.pyx":336
 *                         name = ''
 *                     else:
 *                         break # end of string             # <<<<<<<<<<<<<<
 *                 else:
 *                     name += chr(c)
 */
        /*else*/ {
          goto __pyx_L9_break;
        }
        __pyx_L11:;

        /* "astropy/io/ascii/cparser.pyx":330
 *             for i in range(self.tokenizer.output_len[0]): # header is in first col string
 *                 c = self.tokenizer.output_cols[0][i] # next char in header string
 *                 if not c: # zero byte -- field terminator             # <<<<<<<<<<<<<<
 *                     if name:
 *                         # replace empty placeholder with ''
 */
        goto __pyx_L10;
      }

      /* "astropy/io/ascii/cparser.pyx":338
 *                         break # end of string
 *                 else:
 *                     name += chr(c)             # <<<<<<<<<<<<<<
 *             self.width = <int>len(self.header_names)
 * 
 */
      /*else*/ {
        __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_chr, __pyx_v_c); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 338, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_6 = PyNumber_InPlaceAdd(__pyx_v_name, __pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 338, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF_SET(__pyx_v_name, __pyx_t_6);
        __pyx_t_6 = 0;
      }
      __pyx_L10:;

      /* "astropy/io/ascii/cparser.pyx":328
 *             name = ''
 * 
 *             for i in range(self.tokenizer.output_len[0]): # header is in first col string             # <<<<<<<<<<<<<<
 *                 c = self.tokenizer.output_cols[0][i] # next char in header string
 *                 if not c: # zero byte -- field terminator
 */
    }
    __pyx_L9_break:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "astropy/io/ascii/cparser.pyx":339
 *                 else:
 *                     name += chr(c)
 *             self.width = <int>len(self.header_names)             # <<<<<<<<<<<<<<
 * 
 *         else:
 */
    __pyx_t_4 = __pyx_v_self->header_names;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_7 = PyObject_Length(__pyx_t_4); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 339, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_self->width = ((int)__pyx_t_7);

    /* "astropy/io/ascii/cparser.pyx":319
 * 
 *         # header_start is a valid line number
 *         if self.header_start is not None and self.header_start >= 0:             # <<<<<<<<<<<<<<
 *             if skip_lines(self.tokenizer, self.header_start, 1) != 0:
 *                 self.raise_error("an error occurred while advancing to the "
 */
    goto __pyx_L3;
  }

  /* "astropy/io/ascii/cparser.pyx":343
 *         else:
 *             # Get number of columns from first data row
 *             if tokenize(self.tokenizer, -1, 1, 0) != 0:             # <<<<<<<<<<<<<<
 *                 self.raise_error("an error occurred while tokenizing the first line of data")
 *             self.width = 0
 */
  /*else*/ {
    __pyx_t_3 = ((tokenize(__pyx_v_self->tokenizer, -1, 1, 0) != 0) != 0);
    if (__pyx_t_3) {

      /* "astropy/io/ascii/cparser.pyx":344
 *             # Get number of columns from first data row
 *             if tokenize(self.tokenizer, -1, 1, 0) != 0:
 *                 self.raise_error("an error occurred while tokenizing the first line of data")             # <<<<<<<<<<<<<<
 *             self.width = 0
 *             for i in range(self.tokenizer.output_len[0]): # header is in first col string
 */
      __pyx_t_4 = ((struct __pyx_vtabstruct_7astropy_2io_5ascii_7cparser_CParser *)__pyx_v_self->__pyx_vtab)->raise_error(__pyx_v_self, __pyx_kp_s_an_error_occurred_while_tokenizi_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 344, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "astropy/io/ascii/cparser.pyx":343
 *         else:
 *             # Get number of columns from first data row
 *             if tokenize(self.tokenizer, -1, 1, 0) != 0:             # <<<<<<<<<<<<<<
 *                 self.raise_error("an error occurred while tokenizing the first line of data")
 *             self.width = 0
 */
    }

    /* "astropy/io/ascii/cparser.pyx":345
 *             if tokenize(self.tokenizer, -1, 1, 0) != 0:
 *                 self.raise_error("an error occurred while tokenizing the first line of data")
 *             self.width = 0             # <<<<<<<<<<<<<<
 *             for i in range(self.tokenizer.output_len[0]): # header is in first col string
 *                 # zero byte -- field terminator
 */
    __pyx_v_self->width = 0;

    /* "astropy/io/ascii/cparser.pyx":346
 *                 self.raise_error("an error occurred while tokenizing the first line of data")
 *             self.width = 0
 *             for i in range(self.tokenizer.output_len[0]): # header is in first col string             # <<<<<<<<<<<<<<
 *                 # zero byte -- field terminator
 *                 if not self.tokenizer.output_cols[0][i]:
 */
    __pyx_t_4 = __Pyx_PyInt_From_int((__pyx_v_self->tokenizer->output_len[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 346, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 346, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
      __pyx_t_4 = __pyx_t_6; __Pyx_INCREF(__pyx_t_4); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 346, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 346, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_7); __Pyx_INCREF(__pyx_t_6); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 346, __pyx_L1_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_4, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 346, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        } else {
          if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_7); __Pyx_INCREF(__pyx_t_6); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 346, __pyx_L1_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_4, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 346, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        }
      } else {
        __pyx_t_6 = __pyx_t_8(__pyx_t_4);
        if (unlikely(!__pyx_t_6)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 346, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_6);
      }
      __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "astropy/io/ascii/cparser.pyx":348
 *             for i in range(self.tokenizer.output_len[0]): # header is in first col string
 *                 # zero byte -- field terminator
 *                 if not self.tokenizer.output_cols[0][i]:             # <<<<<<<<<<<<<<
 *                     # ends valid field
 *                     if i > 0 and self.tokenizer.output_cols[0][i - 1]:
 */
      __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 348, __pyx_L1_error)
      __pyx_t_3 = ((!(((__pyx_v_self->tokenizer->output_cols[0])[__pyx_t_9]) != 0)) != 0);
      if (__pyx_t_3) {

        /* "astropy/io/ascii/cparser.pyx":350
 *                 if not self.tokenizer.output_cols[0][i]:
 *                     # ends valid field
 *                     if i > 0 and self.tokenizer.output_cols[0][i - 1]:             # <<<<<<<<<<<<<<
 *                         self.width += 1
 *                     else: # end of line
 */
        __pyx_t_6 = PyObject_RichCompare(__pyx_v_i, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 350, __pyx_L1_error)
        __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 350, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (__pyx_t_1) {
        } else {
          __pyx_t_3 = __pyx_t_1;
          goto __pyx_L17_bool_binop_done;
        }
        __pyx_t_6 = __Pyx_PyInt_SubtractObjC(__pyx_v_i, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 350, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 350, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_1 = (((__pyx_v_self->tokenizer->output_cols[0])[__pyx_t_9]) != 0);
        __pyx_t_3 = __pyx_t_1;
        __pyx_L17_bool_binop_done:;
        if (__pyx_t_3) {

          /* "astropy/io/ascii/cparser.pyx":351
 *                     # ends valid field
 *                     if i > 0 and self.tokenizer.output_cols[0][i - 1]:
 *                         self.width += 1             # <<<<<<<<<<<<<<
 *                     else: # end of line
 *                         break
 */
          __pyx_v_self->width = (__pyx_v_self->width + 1);

          /* "astropy/io/ascii/cparser.pyx":350
 *                 if not self.tokenizer.output_cols[0][i]:
 *                     # ends valid field
 *                     if i > 0 and self.tokenizer.output_cols[0][i - 1]:             # <<<<<<<<<<<<<<
 *                         self.width += 1
 *                     else: # end of line
 */
          goto __pyx_L16;
        }

        /* "astropy/io/ascii/cparser.pyx":353
 *                         self.width += 1
 *                     else: # end of line
 *                         break             # <<<<<<<<<<<<<<
 *             if self.width == 0: # no data
 *                 raise core.InconsistentTableError('No data lines found, C reader '
 */
        /*else*/ {
          goto __pyx_L14_break;
        }
        __pyx_L16:;

        /* "astropy/io/ascii/cparser.pyx":348
 *             for i in range(self.tokenizer.output_len[0]): # header is in first col string
 *                 # zero byte -- field terminator
 *                 if not self.tokenizer.output_cols[0][i]:             # <<<<<<<<<<<<<<
 *                     # ends valid field
 *                     if i > 0 and self.tokenizer.output_cols[0][i - 1]:
 */
      }

      /* "astropy/io/ascii/cparser.pyx":346
 *                 self.raise_error("an error occurred while tokenizing the first line of data")
 *             self.width = 0
 *             for i in range(self.tokenizer.output_len[0]): # header is in first col string             # <<<<<<<<<<<<<<
 *                 # zero byte -- field terminator
 *                 if not self.tokenizer.output_cols[0][i]:
 */
    }
    __pyx_L14_break:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "astropy/io/ascii/cparser.pyx":354
 *                     else: # end of line
 *                         break
 *             if self.width == 0: # no data             # <<<<<<<<<<<<<<
 *                 raise core.InconsistentTableError('No data lines found, C reader '
 *                                             'cannot autogenerate column names')
 */
    __pyx_t_3 = ((__pyx_v_self->width == 0) != 0);
    if (unlikely(__pyx_t_3)) {

      /* "astropy/io/ascii/cparser.pyx":355
 *                         break
 *             if self.width == 0: # no data
 *                 raise core.InconsistentTableError('No data lines found, C reader '             # <<<<<<<<<<<<<<
 *                                             'cannot autogenerate column names')
 *             # auto-generate names
 */
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_core); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 355, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_InconsistentTableError); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 355, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
        }
      }
      __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_6, __pyx_kp_s_No_data_lines_found_C_reader_can) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_kp_s_No_data_lines_found_C_reader_can);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 355, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(0, 355, __pyx_L1_error)

      /* "astropy/io/ascii/cparser.pyx":354
 *                     else: # end of line
 *                         break
 *             if self.width == 0: # no data             # <<<<<<<<<<<<<<
 *                 raise core.InconsistentTableError('No data lines found, C reader '
 *                                             'cannot autogenerate column names')
 */
    }

    /* "astropy/io/ascii/cparser.pyx":358
 *                                             'cannot autogenerate column names')
 *             # auto-generate names
 *             self.header_names = ['col{0}'.format(i + 1) for i in range(self.width)]             # <<<<<<<<<<<<<<
 * 
 *         if self.names:
 */
    __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 358, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_self->width); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 358, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 358, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
      __pyx_t_10 = __pyx_t_6; __Pyx_INCREF(__pyx_t_10); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_10 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 358, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_8 = Py_TYPE(__pyx_t_10)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 358, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_10))) {
          if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_10)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyList_GET_ITEM(__pyx_t_10, __pyx_t_7); __Pyx_INCREF(__pyx_t_6); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 358, __pyx_L1_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_10, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 358, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        } else {
          if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_10)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_10, __pyx_t_7); __Pyx_INCREF(__pyx_t_6); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 358, __pyx_L1_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_10, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 358, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        }
      } else {
        __pyx_t_6 = __pyx_t_8(__pyx_t_10);
        if (unlikely(!__pyx_t_6)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 358, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_6);
      }
      __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_col_0, __pyx_n_s_format); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 358, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_13 = __Pyx_PyInt_AddObjC(__pyx_v_i, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 358, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_14 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
        __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_12);
        if (likely(__pyx_t_14)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
          __Pyx_INCREF(__pyx_t_14);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_12, function);
        }
      }
      __pyx_t_6 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_14, __pyx_t_13) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_t_13);
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 358, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 358, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_GIVEREF(__pyx_t_4);
    __Pyx_GOTREF(__pyx_v_self->header_names);
    __Pyx_DECREF(__pyx_v_self->header_names);
    __pyx_v_self->header_names = __pyx_t_4;
    __pyx_t_4 = 0;
  }
  __pyx_L3:;

  /* "astropy/io/ascii/cparser.pyx":360
 *             self.header_names = ['col{0}'.format(i + 1) for i in range(self.width)]
 * 
 *         if self.names:             # <<<<<<<<<<<<<<
 *             self.width = <int>len(self.names)
 *         else:
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_self->names); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 360, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "astropy/io/ascii/cparser.pyx":361
 * 
 *         if self.names:
 *             self.width = <int>len(self.names)             # <<<<<<<<<<<<<<
 *         else:
 *             self.names = self.header_names
 */
    __pyx_t_4 = __pyx_v_self->names;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_7 = PyObject_Length(__pyx_t_4); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 361, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_self->width = ((int)__pyx_t_7);

    /* "astropy/io/ascii/cparser.pyx":360
 *             self.header_names = ['col{0}'.format(i + 1) for i in range(self.width)]
 * 
 *         if self.names:             # <<<<<<<<<<<<<<
 *             self.width = <int>len(self.names)
 *         else:
 */
    goto __pyx_L22;
  }

  /* "astropy/io/ascii/cparser.pyx":363
 *             self.width = <int>len(self.names)
 *         else:
 *             self.names = self.header_names             # <<<<<<<<<<<<<<
 * 
 *         # self.use_cols should only contain columns included in output
 */
  /*else*/ {
    __pyx_t_4 = __pyx_v_self->header_names;
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __Pyx_GOTREF(__pyx_v_self->names);
    __Pyx_DECREF(__pyx_v_self->names);
    __pyx_v_self->names = __pyx_t_4;
    __pyx_t_4 = 0;
  }
  __pyx_L22:;

  /* "astropy/io/ascii/cparser.pyx":366
 * 
 *         # self.use_cols should only contain columns included in output
 *         self.use_cols = set(self.names)             # <<<<<<<<<<<<<<
 *         if self.include_names is not None:
 *             self.use_cols.intersection_update(self.include_names)
 */
  __pyx_t_4 = PySet_New(__pyx_v_self->names); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_GOTREF(__pyx_v_self->use_cols);
  __Pyx_DECREF(__pyx_v_self->use_cols);
  __pyx_v_self->use_cols = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "astropy/io/ascii/cparser.pyx":367
 *         # self.use_cols should only contain columns included in output
 *         self.use_cols = set(self.names)
 *         if self.include_names is not None:             # <<<<<<<<<<<<<<
 *             self.use_cols.intersection_update(self.include_names)
 *         if self.exclude_names is not None:
 */
  __pyx_t_3 = (__pyx_v_self->include_names != Py_None);
  __pyx_t_1 = (__pyx_t_3 != 0);
  if (__pyx_t_1) {

    /* "astropy/io/ascii/cparser.pyx":368
 *         self.use_cols = set(self.names)
 *         if self.include_names is not None:
 *             self.use_cols.intersection_update(self.include_names)             # <<<<<<<<<<<<<<
 *         if self.exclude_names is not None:
 *             self.use_cols.difference_update(self.exclude_names)
 */
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->use_cols, __pyx_n_s_intersection_update); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 368, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_10);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_10, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_6, __pyx_v_self->include_names) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_v_self->include_names);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 368, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "astropy/io/ascii/cparser.pyx":367
 *         # self.use_cols should only contain columns included in output
 *         self.use_cols = set(self.names)
 *         if self.include_names is not None:             # <<<<<<<<<<<<<<
 *             self.use_cols.intersection_update(self.include_names)
 *         if self.exclude_names is not None:
 */
  }

  /* "astropy/io/ascii/cparser.pyx":369
 *         if self.include_names is not None:
 *             self.use_cols.intersection_update(self.include_names)
 *         if self.exclude_names is not None:             # <<<<<<<<<<<<<<
 *             self.use_cols.difference_update(self.exclude_names)
 * 
 */
  __pyx_t_1 = (__pyx_v_self->exclude_names != Py_None);
  __pyx_t_3 = (__pyx_t_1 != 0);
  if (__pyx_t_3) {

    /* "astropy/io/ascii/cparser.pyx":370
 *             self.use_cols.intersection_update(self.include_names)
 *         if self.exclude_names is not None:
 *             self.use_cols.difference_update(self.exclude_names)             # <<<<<<<<<<<<<<
 * 
 *         self.width = <int>len(self.names)
 */
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->use_cols, __pyx_n_s_difference_update); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 370, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_10);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_10, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_6, __pyx_v_self->exclude_names) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_v_self->exclude_names);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 370, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "astropy/io/ascii/cparser.pyx":369
 *         if self.include_names is not None:
 *             self.use_cols.intersection_update(self.include_names)
 *         if self.exclude_names is not None:             # <<<<<<<<<<<<<<
 *             self.use_cols.difference_update(self.exclude_names)
 * 
 */
  }

  /* "astropy/io/ascii/cparser.pyx":372
 *             self.use_cols.difference_update(self.exclude_names)
 * 
 *         self.width = <int>len(self.names)             # <<<<<<<<<<<<<<
 * 
 *     def read(self, try_int, try_float, try_string):
 */
  __pyx_t_4 = __pyx_v_self->names;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_7 = PyObject_Length(__pyx_t_4); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 372, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_self->width = ((int)__pyx_t_7);

  /* "astropy/io/ascii/cparser.pyx":315
 *         self.tokenizer.source_len = <size_t>len(self.source_bytes)
 * 
 *     def read_header(self):             # <<<<<<<<<<<<<<
 *         self.tokenizer.source_pos = 0
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("astropy.io.ascii.cparser.CParser.read_header", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_c);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "astropy/io/ascii/cparser.pyx":374
 *         self.width = <int>len(self.names)
 * 
 *     def read(self, try_int, try_float, try_string):             # <<<<<<<<<<<<<<
 *         if self.parallel:
 *             return self._read_parallel(try_int, try_float, try_string)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_9read(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_9read(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_try_int = 0;
  PyObject *__pyx_v_try_float = 0;
  PyObject *__pyx_v_try_string = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("read (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_try_int,&__pyx_n_s_try_float,&__pyx_n_s_try_string,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_try_int)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_try_float)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("read", 1, 3, 3, 1); __PYX_ERR(0, 374, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_try_string)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("read", 1, 3, 3, 2); __PYX_ERR(0, 374, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "read") < 0)) __PYX_ERR(0, 374, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_try_int = values[0];
    __pyx_v_try_float = values[1];
    __pyx_v_try_string = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("read", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 374, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("astropy.io.ascii.cparser.CParser.read", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_8read(((struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *)__pyx_v_self), __pyx_v_try_int, __pyx_v_try_float, __pyx_v_try_string);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_8read(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self, PyObject *__pyx_v_try_int, PyObject *__pyx_v_try_float, PyObject *__pyx_v_try_string) {
  int __pyx_v_data_end;
  int __pyx_v_num_rows;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  long __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  __Pyx_RefNannySetupContext("read", 0);

  /* "astropy/io/ascii/cparser.pyx":375
 * 
 *     def read(self, try_int, try_float, try_string):
 *         if self.parallel:             # <<<<<<<<<<<<<<
 *             return self._read_parallel(try_int, try_float, try_string)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_self->parallel); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 375, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "astropy/io/ascii/cparser.pyx":376
 *     def read(self, try_int, try_float, try_string):
 *         if self.parallel:
 *             return self._read_parallel(try_int, try_float, try_string)             # <<<<<<<<<<<<<<
 * 
 *         # Read in a single process
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_read_parallel); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 376, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_try_int, __pyx_v_try_float, __pyx_v_try_string};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 376, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_try_int, __pyx_v_try_float, __pyx_v_try_string};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 376, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 376, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_INCREF(__pyx_v_try_int);
      __Pyx_GIVEREF(__pyx_v_try_int);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_try_int);
      __Pyx_INCREF(__pyx_v_try_float);
      __Pyx_GIVEREF(__pyx_v_try_float);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_try_float);
      __Pyx_INCREF(__pyx_v_try_string);
      __Pyx_GIVEREF(__pyx_v_try_string);
      PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_5, __pyx_v_try_string);
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 376, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "astropy/io/ascii/cparser.pyx":375
 * 
 *     def read(self, try_int, try_float, try_string):
 *         if self.parallel:             # <<<<<<<<<<<<<<
 *             return self._read_parallel(try_int, try_float, try_string)
 * 
 */
  }

  /* "astropy/io/ascii/cparser.pyx":379
 * 
 *         # Read in a single process
 *         self.tokenizer.source_pos = 0             # <<<<<<<<<<<<<<
 *         if skip_lines(self.tokenizer, self.data_start, 0) != 0:
 *             self.raise_error("an error occurred while advancing to the first "
 */
  __pyx_v_self->tokenizer->source_pos = 0;

  /* "astropy/io/ascii/cparser.pyx":380
 *         # Read in a single process
 *         self.tokenizer.source_pos = 0
 *         if skip_lines(self.tokenizer, self.data_start, 0) != 0:             # <<<<<<<<<<<<<<
 *             self.raise_error("an error occurred while advancing to the first "
 *                              "line of data")
 */
  __pyx_t_1 = ((skip_lines(__pyx_v_self->tokenizer, __pyx_v_self->data_start, 0) != 0) != 0);
  if (__pyx_t_1) {

    /* "astropy/io/ascii/cparser.pyx":381
 *         self.tokenizer.source_pos = 0
 *         if skip_lines(self.tokenizer, self.data_start, 0) != 0:
 *             self.raise_error("an error occurred while advancing to the first "             # <<<<<<<<<<<<<<
 *                              "line of data")
 * 
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_7astropy_2io_5ascii_7cparser_CParser *)__pyx_v_self->__pyx_vtab)->raise_error(__pyx_v_self, __pyx_kp_s_an_error_occurred_while_advancin_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 381, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "astropy/io/ascii/cparser.pyx":380
 *         # Read in a single process
 *         self.tokenizer.source_pos = 0
 *         if skip_lines(self.tokenizer, self.data_start, 0) != 0:             # <<<<<<<<<<<<<<
 *             self.raise_error("an error occurred while advancing to the first "
 *                              "line of data")
 */
  }

  /* "astropy/io/ascii/cparser.pyx":384
 *                              "line of data")
 * 
 *         cdef int data_end = -1 # keep reading data until the end             # <<<<<<<<<<<<<<
 *         if self.data_end is not None and self.data_end >= 0:
 *             data_end = max(self.data_end - self.data_start, 0) # read nothing if data_end < 0
 */
  __pyx_v_data_end = -1;

  /* "astropy/io/ascii/cparser.pyx":385
 * 
 *         cdef int data_end = -1 # keep reading data until the end
 *         if self.data_end is not None and self.data_end >= 0:             # <<<<<<<<<<<<<<
 *             data_end = max(self.data_end - self.data_start, 0) # read nothing if data_end < 0
 * 
 */
  __pyx_t_7 = (__pyx_v_self->data_end != Py_None);
  __pyx_t_8 = (__pyx_t_7 != 0);
  if (__pyx_t_8) {
  } else {
    __pyx_t_1 = __pyx_t_8;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_self->data_end, __pyx_int_0, Py_GE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 385, __pyx_L1_error)
  __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __pyx_t_8;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_1) {

    /* "astropy/io/ascii/cparser.pyx":386
 *         cdef int data_end = -1 # keep reading data until the end
 *         if self.data_end is not None and self.data_end >= 0:
 *             data_end = max(self.data_end - self.data_start, 0) # read nothing if data_end < 0             # <<<<<<<<<<<<<<
 * 
 *         if tokenize(self.tokenizer, data_end, 0, <int>len(self.names)) != 0:
 */
    __pyx_t_9 = 0;
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->data_start); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyNumber_Subtract(__pyx_v_self->data_end, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_PyInt_From_long(__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = PyObject_RichCompare(__pyx_t_6, __pyx_t_3, Py_GT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 386, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 386, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_1) {
      __pyx_t_4 = __Pyx_PyInt_From_long(__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 386, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_2 = __pyx_t_4;
      __pyx_t_4 = 0;
    } else {
      __Pyx_INCREF(__pyx_t_3);
      __pyx_t_2 = __pyx_t_3;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 386, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_data_end = __pyx_t_5;

    /* "astropy/io/ascii/cparser.pyx":385
 * 
 *         cdef int data_end = -1 # keep reading data until the end
 *         if self.data_end is not None and self.data_end >= 0:             # <<<<<<<<<<<<<<
 *             data_end = max(self.data_end - self.data_start, 0) # read nothing if data_end < 0
 * 
 */
  }

  /* "astropy/io/ascii/cparser.pyx":388
 *             data_end = max(self.data_end - self.data_start, 0) # read nothing if data_end < 0
 * 
 *         if tokenize(self.tokenizer, data_end, 0, <int>len(self.names)) != 0:             # <<<<<<<<<<<<<<
 *             self.raise_error("an error occurred while parsing table data")
 *         elif self.tokenizer.num_rows == 0: # no data
 */
  __pyx_t_2 = __pyx_v_self->names;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_10 = PyObject_Length(__pyx_t_2); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 388, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = ((tokenize(__pyx_v_self->tokenizer, __pyx_v_data_end, 0, ((int)__pyx_t_10)) != 0) != 0);
  if (__pyx_t_1) {

    /* "astropy/io/ascii/cparser.pyx":389
 * 
 *         if tokenize(self.tokenizer, data_end, 0, <int>len(self.names)) != 0:
 *             self.raise_error("an error occurred while parsing table data")             # <<<<<<<<<<<<<<
 *         elif self.tokenizer.num_rows == 0: # no data
 *             return ([np.array([], dtype=np.int_)] * self.width,
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_7astropy_2io_5ascii_7cparser_CParser *)__pyx_v_self->__pyx_vtab)->raise_error(__pyx_v_self, __pyx_kp_s_an_error_occurred_while_parsing); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "astropy/io/ascii/cparser.pyx":388
 *             data_end = max(self.data_end - self.data_start, 0) # read nothing if data_end < 0
 * 
 *         if tokenize(self.tokenizer, data_end, 0, <int>len(self.names)) != 0:             # <<<<<<<<<<<<<<
 *             self.raise_error("an error occurred while parsing table data")
 *         elif self.tokenizer.num_rows == 0: # no data
 */
    goto __pyx_L8;
  }

  /* "astropy/io/ascii/cparser.pyx":390
 *         if tokenize(self.tokenizer, data_end, 0, <int>len(self.names)) != 0:
 *             self.raise_error("an error occurred while parsing table data")
 *         elif self.tokenizer.num_rows == 0: # no data             # <<<<<<<<<<<<<<
 *             return ([np.array([], dtype=np.int_)] * self.width,
 *                     self._get_comments(self.tokenizer))
 */
  __pyx_t_1 = ((__pyx_v_self->tokenizer->num_rows == 0) != 0);
  if (__pyx_t_1) {

    /* "astropy/io/ascii/cparser.pyx":391
 *             self.raise_error("an error occurred while parsing table data")
 *         elif self.tokenizer.num_rows == 0: # no data
 *             return ([np.array([], dtype=np.int_)] * self.width,             # <<<<<<<<<<<<<<
 *                     self._get_comments(self.tokenizer))
 *         self._set_fill_values()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 391, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 391, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 391, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 391, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 391, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 391, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_int); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 391, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_11) < 0) __PYX_ERR(0, 391, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 391, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyList_New(1 * ((__pyx_v_self->width<0) ? 0:__pyx_v_self->width)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 391, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    { Py_ssize_t __pyx_temp;
      for (__pyx_temp=0; __pyx_temp < __pyx_v_self->width; __pyx_temp++) {
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_GIVEREF(__pyx_t_11);
        PyList_SET_ITEM(__pyx_t_2, __pyx_temp, __pyx_t_11);
      }
    }
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

    /* "astropy/io/ascii/cparser.pyx":392
 *         elif self.tokenizer.num_rows == 0: # no data
 *             return ([np.array([], dtype=np.int_)] * self.width,
 *                     self._get_comments(self.tokenizer))             # <<<<<<<<<<<<<<
 *         self._set_fill_values()
 *         cdef int num_rows = self.tokenizer.num_rows
 */
    __pyx_t_11 = ((struct __pyx_vtabstruct_7astropy_2io_5ascii_7cparser_CParser *)__pyx_v_self->__pyx_vtab)->_get_comments(__pyx_v_self, __pyx_v_self->tokenizer); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);

    /* "astropy/io/ascii/cparser.pyx":391
 *             self.raise_error("an error occurred while parsing table data")
 *         elif self.tokenizer.num_rows == 0: # no data
 *             return ([np.array([], dtype=np.int_)] * self.width,             # <<<<<<<<<<<<<<
 *                     self._get_comments(self.tokenizer))
 *         self._set_fill_values()
 */
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 391, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_11);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_11);
    __pyx_t_2 = 0;
    __pyx_t_11 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "astropy/io/ascii/cparser.pyx":390
 *         if tokenize(self.tokenizer, data_end, 0, <int>len(self.names)) != 0:
 *             self.raise_error("an error occurred while parsing table data")
 *         elif self.tokenizer.num_rows == 0: # no data             # <<<<<<<<<<<<<<
 *             return ([np.array([], dtype=np.int_)] * self.width,
 *                     self._get_comments(self.tokenizer))
 */
  }
  __pyx_L8:;

  /* "astropy/io/ascii/cparser.pyx":393
 *             return ([np.array([], dtype=np.int_)] * self.width,
 *                     self._get_comments(self.tokenizer))
 *         self._set_fill_values()             # <<<<<<<<<<<<<<
 *         cdef int num_rows = self.tokenizer.num_rows
 *         if self.data_end is not None and self.data_end < 0: # negative indexing
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_7astropy_2io_5ascii_7cparser_CParser *)__pyx_v_self->__pyx_vtab)->_set_fill_values(__pyx_v_self); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "astropy/io/ascii/cparser.pyx":394
 *                     self._get_comments(self.tokenizer))
 *         self._set_fill_values()
 *         cdef int num_rows = self.tokenizer.num_rows             # <<<<<<<<<<<<<<
 *         if self.data_end is not None and self.data_end < 0: # negative indexing
 *             num_rows += self.data_end
 */
  __pyx_t_5 = __pyx_v_self->tokenizer->num_rows;
  __pyx_v_num_rows = __pyx_t_5;

  /* "astropy/io/ascii/cparser.pyx":395
 *         self._set_fill_values()
 *         cdef int num_rows = self.tokenizer.num_rows
 *         if self.data_end is not None and self.data_end < 0: # negative indexing             # <<<<<<<<<<<<<<
 *             num_rows += self.data_end
 *         return self._convert_data(self.tokenizer, try_int, try_float,
 */
  __pyx_t_8 = (__pyx_v_self->data_end != Py_None);
  __pyx_t_7 = (__pyx_t_8 != 0);
  if (__pyx_t_7) {
  } else {
    __pyx_t_1 = __pyx_t_7;
    goto __pyx_L10_bool_binop_done;
  }
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_self->data_end, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 395, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 395, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __pyx_t_7;
  __pyx_L10_bool_binop_done:;
  if (__pyx_t_1) {

    /* "astropy/io/ascii/cparser.pyx":396
 *         cdef int num_rows = self.tokenizer.num_rows
 *         if self.data_end is not None and self.data_end < 0: # negative indexing
 *             num_rows += self.data_end             # <<<<<<<<<<<<<<
 *         return self._convert_data(self.tokenizer, try_int, try_float,
 *                                   try_string, num_rows)
 */
    __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_num_rows); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 396, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_11 = PyNumber_InPlaceAdd(__pyx_t_4, __pyx_v_self->data_end); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 396, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_11); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 396, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_v_num_rows = __pyx_t_5;

    /* "astropy/io/ascii/cparser.pyx":395
 *         self._set_fill_values()
 *         cdef int num_rows = self.tokenizer.num_rows
 *         if self.data_end is not None and self.data_end < 0: # negative indexing             # <<<<<<<<<<<<<<
 *             num_rows += self.data_end
 *         return self._convert_data(self.tokenizer, try_int, try_float,
 */
  }

  /* "astropy/io/ascii/cparser.pyx":397
 *         if self.data_end is not None and self.data_end < 0: # negative indexing
 *             num_rows += self.data_end
 *         return self._convert_data(self.tokenizer, try_int, try_float,             # <<<<<<<<<<<<<<
 *                                   try_string, num_rows)
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "astropy/io/ascii/cparser.pyx":398
 *             num_rows += self.data_end
 *         return self._convert_data(self.tokenizer, try_int, try_float,
 *                                   try_string, num_rows)             # <<<<<<<<<<<<<<
 * 
 *     def _read_parallel(self, try_int, try_float, try_string):
 */
  __pyx_t_11 = __Pyx_PyInt_From_int(__pyx_v_num_rows); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);

  /* "astropy/io/ascii/cparser.pyx":397
 *         if self.data_end is not None and self.data_end < 0: # negative indexing
 *             num_rows += self.data_end
 *         return self._convert_data(self.tokenizer, try_int, try_float,             # <<<<<<<<<<<<<<
 *                                   try_string, num_rows)
 * 
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_7astropy_2io_5ascii_7cparser_CParser *)__pyx_v_self->__pyx_vtab)->_convert_data(__pyx_v_self, __pyx_v_self->tokenizer, __pyx_v_try_int, __pyx_v_try_float, __pyx_v_try_string, __pyx_t_11); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "astropy/io/ascii/cparser.pyx":374
 *         self.width = <int>len(self.names)
 * 
 *     def read(self, try_int, try_float, try_string):             # <<<<<<<<<<<<<<
 *         if self.parallel:
 *             return self._read_parallel(try_int, try_float, try_string)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("astropy.io.ascii.cparser.CParser.read", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "astropy/io/ascii/cparser.pyx":400
 *                                   try_string, num_rows)
 * 
 *     def _read_parallel(self, try_int, try_float, try_string):             # <<<<<<<<<<<<<<
 *         cdef size_t source_len = <size_t>len(self.source)
 *         self.tokenizer.source_pos = 0
 */

/* Python wrapper */
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_11_read_parallel(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_11_read_parallel(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_try_int = 0;
  PyObject *__pyx_v_try_float = 0;
  PyObject *__pyx_v_try_string = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_read_parallel (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_try_int,&__pyx_n_s_try_float,&__pyx_n_s_try_string,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_try_int)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_try_float)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_read_parallel", 1, 3, 3, 1); __PYX_ERR(0, 400, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_try_string)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_read_parallel", 1, 3, 3, 2); __PYX_ERR(0, 400, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_read_parallel") < 0)) __PYX_ERR(0, 400, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_try_int = values[0];
    __pyx_v_try_float = values[1];
    __pyx_v_try_string = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_read_parallel", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 400, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("astropy.io.ascii.cparser.CParser._read_parallel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_10_read_parallel(((struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *)__pyx_v_self), __pyx_v_try_int, __pyx_v_try_float, __pyx_v_try_string);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_7astropy_2io_5ascii_7cparser_7CParser_14_read_parallel_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "astropy/io/ascii/cparser.pyx":418
 * 
 *         if offset == source_len: # no data
 *             return (dict((name, np.array([], dtype=np.int_)) for name in             # <<<<<<<<<<<<<<
 *                          self.names),
 *                     self._get_comments(self.tokenizer))
 */

static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_14_read_parallel_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr *)__pyx_tp_new_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr(__pyx_ptype_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 418, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_7astropy_2io_5ascii_7cparser_7CParser_14_read_parallel_2generator1, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_read_parallel_locals_genexpr, __pyx_n_s_astropy_io_ascii_cparser); if (unlikely(!gen)) __PYX_ERR(0, 418, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("astropy.io.ascii.cparser.CParser._read_parallel.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_7astropy_2io_5ascii_7cparser_7CParser_14_read_parallel_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr *__pyx_cur_scope = ((struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 418, __pyx_L1_error)
  __pyx_r = PyDict_New(); if (unlikely(!__pyx_r)) __PYX_ERR(0, 418, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);

  /* "astropy/io/ascii/cparser.pyx":419
 *         if offset == source_len: # no data
 *             return (dict((name, np.array([], dtype=np.int_)) for name in
 *                          self.names),             # <<<<<<<<<<<<<<
 *                     self._get_comments(self.tokenizer))
 * 
 */
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 419, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->names)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->names)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->names; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->names); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 419, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 419, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 419, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 419, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 419, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 419, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 419, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_name);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_name, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;

    /* "astropy/io/ascii/cparser.pyx":418
 * 
 *         if offset == source_len: # no data
 *             return (dict((name, np.array([], dtype=np.int_)) for name in             # <<<<<<<<<<<<<<
 *                          self.names),
 *                     self._get_comments(self.tokenizer))
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_array); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_int); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_dtype, __pyx_t_8) < 0) __PYX_ERR(0, 418, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, __pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(PyDict_SetItem(__pyx_r, (PyObject*)__pyx_cur_scope->__pyx_v_name, (PyObject*)__pyx_t_8))) __PYX_ERR(0, 418, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_7astropy_2io_5ascii_7cparser_7CParser_14_read_parallel_5generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "astropy/io/ascii/cparser.pyx":525
 * 
 *             if self.data_end < 0: # no data
 *                 chunks = [dict((name, []) for name in self.names)]             # <<<<<<<<<<<<<<
 *             else:
 *                 line_no = 0
 */

static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_14_read_parallel_3genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr *)__pyx_tp_new_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr(__pyx_ptype_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 525, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_7astropy_2io_5ascii_7cparser_7CParser_14_read_parallel_5generator2, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_read_parallel_locals_genexpr, __pyx_n_s_astropy_io_ascii_cparser); if (unlikely(!gen)) __PYX_ERR(0, 525, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("astropy.io.ascii.cparser.CParser._read_parallel.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_7astropy_2io_5ascii_7cparser_7CParser_14_read_parallel_5generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr *__pyx_cur_scope = ((struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 525, __pyx_L1_error)
  __pyx_r = PyDict_New(); if (unlikely(!__pyx_r)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 525, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->names)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->names)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->names; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->names); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 525, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 525, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 525, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 525, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 525, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 525, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 525, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_name);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_name, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 525, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (unlikely(PyDict_SetItem(__pyx_r, (PyObject*)__pyx_cur_scope->__pyx_v_name, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 525, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_7astropy_2io_5ascii_7cparser_7CParser_14_read_parallel_8generator3(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "astropy/io/ascii/cparser.pyx":541
 *         for name in self.get_names():
 *             col_chunks = [chunk.pop(name) for chunk in chunks]
 *             if any(isinstance(col_chunk, ma.masked_array) for col_chunk in col_chunks):             # <<<<<<<<<<<<<<
 *                 ret[name] = ma.concatenate(col_chunks)
 *             else:
 */

static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_14_read_parallel_6genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr *)__pyx_tp_new_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr(__pyx_ptype_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 541, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_7astropy_2io_5ascii_7cparser_7CParser_14_read_parallel_8generator3, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_read_parallel_locals_genexpr, __pyx_n_s_astropy_io_ascii_cparser); if (unlikely(!gen)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("astropy.io.ascii.cparser.CParser._read_parallel.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_7astropy_2io_5ascii_7cparser_7CParser_14_read_parallel_8generator3(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr *__pyx_cur_scope = ((struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 541, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_col_chunks)) { __Pyx_RaiseClosureNameError("col_chunks"); __PYX_ERR(0, 541, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_col_chunks == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 541, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_col_chunks; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 541, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_col_chunk);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_col_chunk, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ma); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_masked_array); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_5 = PyObject_IsInstance(__pyx_cur_scope->__pyx_v_col_chunk, __pyx_t_4); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_6 = (__pyx_t_5 != 0);
    if (__pyx_t_6) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_True);
      __pyx_r = Py_True;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "astropy/io/ascii/cparser.pyx":400
 *                                   try_string, num_rows)
 * 
 *     def _read_parallel(self, try_int, try_float, try_string):             # <<<<<<<<<<<<<<
 *         cdef size_t source_len = <size_t>len(self.source)
 *         self.tokenizer.source_pos = 0
 */

static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_10_read_parallel(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self, PyObject *__pyx_v_try_int, PyObject *__pyx_v_try_float, PyObject *__pyx_v_try_string) {
  struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel *__pyx_cur_scope;
  size_t __pyx_v_source_len;
  PyObject *__pyx_v_line_comments = 0;
  int __pyx_v_N;
  PyObject *__pyx_v_queue = NULL;
  size_t __pyx_v_offset;
  long __pyx_v_chunksize;
  PyObject *__pyx_v_chunkindices = 0;
  PyObject *__pyx_v_reconvert_queue = NULL;
  int __pyx_v_i;
  size_t __pyx_v_index;
  PyObject *__pyx_v_processes = 0;
  PyObject *__pyx_v_process = NULL;
  PyObject *__pyx_v_chunks = 0;
  PyObject *__pyx_v_comments_chunks = 0;
  PyObject *__pyx_v_failed_procs = 0;
  PyObject *__pyx_v_queue_ret = NULL;
  PyObject *__pyx_v_err = NULL;
  PyObject *__pyx_v_proc = NULL;
  PyObject *__pyx_v_comments = NULL;
  PyObject *__pyx_v_data = NULL;
  PyObject *__pyx_v_chunk = NULL;
  PyObject *__pyx_v_line_no = NULL;
  PyObject *__pyx_v_seen_str = NULL;
  PyObject *__pyx_v_seen_numeric = NULL;
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_v_reconvert_cols = NULL;
  PyObject *__pyx_v_reconverted = NULL;
  PyObject *__pyx_v_col = NULL;
  PyObject *__pyx_v_num_rows = NULL;
  PyObject *__pyx_v_ret = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  size_t __pyx_t_11;
  long __pyx_t_12;
  int __pyx_t_13;
  int __pyx_t_14;
  int __pyx_t_15;
  int __pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *(*__pyx_t_19)(PyObject *);
  int __pyx_t_20;
  PyObject *(*__pyx_t_21)(PyObject *);
  Py_ssize_t __pyx_t_22;
  Py_ssize_t __pyx_t_23;
  __Pyx_RefNannySetupContext("_read_parallel", 0);
  __pyx_cur_scope = (struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel *)__pyx_tp_new_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel(__pyx_ptype_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 400, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);

  /* "astropy/io/ascii/cparser.pyx":401
 * 
 *     def _read_parallel(self, try_int, try_float, try_string):
 *         cdef size_t source_len = <size_t>len(self.source)             # <<<<<<<<<<<<<<
 *         self.tokenizer.source_pos = 0
 * 
 */
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_self->source;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 401, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_source_len = ((size_t)__pyx_t_2);

  /* "astropy/io/ascii/cparser.pyx":402
 *     def _read_parallel(self, try_int, try_float, try_string):
 *         cdef size_t source_len = <size_t>len(self.source)
 *         self.tokenizer.source_pos = 0             # <<<<<<<<<<<<<<
 * 
 *         if skip_lines(self.tokenizer, self.data_start, 0) != 0:
 */
  __pyx_cur_scope->__pyx_v_self->tokenizer->source_pos = 0;

  /* "astropy/io/ascii/cparser.pyx":404
 *         self.tokenizer.source_pos = 0
 * 
 *         if skip_lines(self.tokenizer, self.data_start, 0) != 0:             # <<<<<<<<<<<<<<
 *             self.raise_error("an error occurred while advancing to the first "
 *                              "line of data")
 */
  __pyx_t_3 = ((skip_lines(__pyx_cur_scope->__pyx_v_self->tokenizer, __pyx_cur_scope->__pyx_v_self->data_start, 0) != 0) != 0);
  if (__pyx_t_3) {

    /* "astropy/io/ascii/cparser.pyx":405
 * 
 *         if skip_lines(self.tokenizer, self.data_start, 0) != 0:
 *             self.raise_error("an error occurred while advancing to the first "             # <<<<<<<<<<<<<<
 *                              "line of data")
 * 
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_7astropy_2io_5ascii_7cparser_CParser *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->raise_error(__pyx_cur_scope->__pyx_v_self, __pyx_kp_s_an_error_occurred_while_advancin_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "astropy/io/ascii/cparser.pyx":404
 *         self.tokenizer.source_pos = 0
 * 
 *         if skip_lines(self.tokenizer, self.data_start, 0) != 0:             # <<<<<<<<<<<<<<
 *             self.raise_error("an error occurred while advancing to the first "
 *                              "line of data")
 */
  }

  /* "astropy/io/ascii/cparser.pyx":408
 *                              "line of data")
 * 
 *         cdef list line_comments = self._get_comments(self.tokenizer)             # <<<<<<<<<<<<<<
 *         cdef int N = self.parallel
 *         try:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_7astropy_2io_5ascii_7cparser_CParser *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_get_comments(__pyx_cur_scope->__pyx_v_self, __pyx_cur_scope->__pyx_v_self->tokenizer); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 408, __pyx_L1_error)
  __pyx_v_line_comments = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "astropy/io/ascii/cparser.pyx":409
 * 
 *         cdef list line_comments = self._get_comments(self.tokenizer)
 *         cdef int N = self.parallel             # <<<<<<<<<<<<<<
 *         try:
 *             queue = multiprocessing.Queue()
 */
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_cur_scope->__pyx_v_self->parallel); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 409, __pyx_L1_error)
  __pyx_v_N = __pyx_t_4;

  /* "astropy/io/ascii/cparser.pyx":410
 *         cdef list line_comments = self._get_comments(self.tokenizer)
 *         cdef int N = self.parallel
 *         try:             # <<<<<<<<<<<<<<
 *             queue = multiprocessing.Queue()
 *         except (ImportError, NotImplementedError, AttributeError, OSError):
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_7);
    /*try:*/ {

      /* "astropy/io/ascii/cparser.pyx":411
 *         cdef int N = self.parallel
 *         try:
 *             queue = multiprocessing.Queue()             # <<<<<<<<<<<<<<
 *         except (ImportError, NotImplementedError, AttributeError, OSError):
 *             self.raise_error("shared semaphore implementation required "
 */
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_multiprocessing); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 411, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Queue); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 411, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
        }
      }
      __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 411, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_queue = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "astropy/io/ascii/cparser.pyx":410
 *         cdef list line_comments = self._get_comments(self.tokenizer)
 *         cdef int N = self.parallel
 *         try:             # <<<<<<<<<<<<<<
 *             queue = multiprocessing.Queue()
 *         except (ImportError, NotImplementedError, AttributeError, OSError):
 */
    }
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    goto __pyx_L9_try_end;
    __pyx_L4_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "astropy/io/ascii/cparser.pyx":412
 *         try:
 *             queue = multiprocessing.Queue()
 *         except (ImportError, NotImplementedError, AttributeError, OSError):             # <<<<<<<<<<<<<<
 *             self.raise_error("shared semaphore implementation required "
 *                              "but not available")
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ImportError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_NotImplementedError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_OSError);
    if (__pyx_t_4) {
      __Pyx_AddTraceback("astropy.io.ascii.cparser.CParser._read_parallel", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_9, &__pyx_t_8) < 0) __PYX_ERR(0, 412, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_GOTREF(__pyx_t_8);

      /* "astropy/io/ascii/cparser.pyx":413
 *             queue = multiprocessing.Queue()
 *         except (ImportError, NotImplementedError, AttributeError, OSError):
 *             self.raise_error("shared semaphore implementation required "             # <<<<<<<<<<<<<<
 *                              "but not available")
 *         cdef size_t offset = self.tokenizer.source_pos
 */
      __pyx_t_10 = ((struct __pyx_vtabstruct_7astropy_2io_5ascii_7cparser_CParser *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->raise_error(__pyx_cur_scope->__pyx_v_self, __pyx_kp_s_shared_semaphore_implementation); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 413, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L5_exception_handled;
    }
    goto __pyx_L6_except_error;
    __pyx_L6_except_error:;

    /* "astropy/io/ascii/cparser.pyx":410
 *         cdef list line_comments = self._get_comments(self.tokenizer)
 *         cdef int N = self.parallel
 *         try:             # <<<<<<<<<<<<<<
 *             queue = multiprocessing.Queue()
 *         except (ImportError, NotImplementedError, AttributeError, OSError):
 */
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
    goto __pyx_L1_error;
    __pyx_L5_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
    __pyx_L9_try_end:;
  }

  /* "astropy/io/ascii/cparser.pyx":415
 *             self.raise_error("shared semaphore implementation required "
 *                              "but not available")
 *         cdef size_t offset = self.tokenizer.source_pos             # <<<<<<<<<<<<<<
 * 
 *         if offset == source_len: # no data
 */
  __pyx_t_11 = __pyx_cur_scope->__pyx_v_self->tokenizer->source_pos;
  __pyx_v_offset = __pyx_t_11;

  /* "astropy/io/ascii/cparser.pyx":417
 *         cdef size_t offset = self.tokenizer.source_pos
 * 
 *         if offset == source_len: # no data             # <<<<<<<<<<<<<<
 *             return (dict((name, np.array([], dtype=np.int_)) for name in
 *                          self.names),
 */
  __pyx_t_3 = ((__pyx_v_offset == __pyx_v_source_len) != 0);
  if (__pyx_t_3) {

    /* "astropy/io/ascii/cparser.pyx":418
 * 
 *         if offset == source_len: # no data
 *             return (dict((name, np.array([], dtype=np.int_)) for name in             # <<<<<<<<<<<<<<
 *                          self.names),
 *                     self._get_comments(self.tokenizer))
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_8 = __pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_14_read_parallel_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_Generator_Next(__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "astropy/io/ascii/cparser.pyx":420
 *             return (dict((name, np.array([], dtype=np.int_)) for name in
 *                          self.names),
 *                     self._get_comments(self.tokenizer))             # <<<<<<<<<<<<<<
 * 
 *         cdef long chunksize = math.ceil((source_len - offset) / float(N))
 */
    __pyx_t_8 = ((struct __pyx_vtabstruct_7astropy_2io_5ascii_7cparser_CParser *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_get_comments(__pyx_cur_scope->__pyx_v_self, __pyx_cur_scope->__pyx_v_self->tokenizer); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 420, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);

    /* "astropy/io/ascii/cparser.pyx":418
 * 
 *         if offset == source_len: # no data
 *             return (dict((name, np.array([], dtype=np.int_)) for name in             # <<<<<<<<<<<<<<
 *                          self.names),
 *                     self._get_comments(self.tokenizer))
 */
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_8);
    __pyx_t_9 = 0;
    __pyx_t_8 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "astropy/io/ascii/cparser.pyx":417
 *         cdef size_t offset = self.tokenizer.source_pos
 * 
 *         if offset == source_len: # no data             # <<<<<<<<<<<<<<
 *             return (dict((name, np.array([], dtype=np.int_)) for name in
 *                          self.names),
 */
  }

  /* "astropy/io/ascii/cparser.pyx":422
 *                     self._get_comments(self.tokenizer))
 * 
 *         cdef long chunksize = math.ceil((source_len - offset) / float(N))             # <<<<<<<<<<<<<<
 *         cdef list chunkindices = [offset]
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_math); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_ceil); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_11 = (__pyx_v_source_len - __pyx_v_offset);
  if (unlikely(((double)__pyx_v_N) == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(0, 422, __pyx_L1_error)
  }
  __pyx_t_8 = PyFloat_FromDouble((__pyx_t_11 / ((double)__pyx_v_N))); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_10 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
    if (likely(__pyx_t_10)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_9, function);
    }
  }
  __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_12 = __Pyx_PyInt_As_long(__pyx_t_1); if (unlikely((__pyx_t_12 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 422, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_chunksize = __pyx_t_12;

  /* "astropy/io/ascii/cparser.pyx":423
 * 
 *         cdef long chunksize = math.ceil((source_len - offset) / float(N))
 *         cdef list chunkindices = [offset]             # <<<<<<<<<<<<<<
 * 
 *         # This queue is used to signal processes to reconvert if necessary
 */
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_offset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = PyList_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_1);
  PyList_SET_ITEM(__pyx_t_9, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_chunkindices = ((PyObject*)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "astropy/io/ascii/cparser.pyx":426
 * 
 *         # This queue is used to signal processes to reconvert if necessary
 *         reconvert_queue = multiprocessing.Queue()             # <<<<<<<<<<<<<<
 * 
 *         cdef int i
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_multiprocessing); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 426, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_Queue); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 426, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
    }
  }
  __pyx_t_9 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 426, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_v_reconvert_queue = __pyx_t_9;
  __pyx_t_9 = 0;

  /* "astropy/io/ascii/cparser.pyx":431
 *         cdef size_t index
 * 
 *         for i in range(1, N):             # <<<<<<<<<<<<<<
 *             index = max(offset + chunksize * i, chunkindices[i - 1])
 *             while index < source_len and self.source[index] != '\n':
 */
  __pyx_t_4 = __pyx_v_N;
  __pyx_t_13 = __pyx_t_4;
  for (__pyx_t_14 = 1; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
    __pyx_v_i = __pyx_t_14;

    /* "astropy/io/ascii/cparser.pyx":432
 * 
 *         for i in range(1, N):
 *             index = max(offset + chunksize * i, chunkindices[i - 1])             # <<<<<<<<<<<<<<
 *             while index < source_len and self.source[index] != '\n':
 *                 index += 1
 */
    __pyx_t_12 = (__pyx_v_i - 1);
    __pyx_t_9 = __Pyx_GetItemInt_List(__pyx_v_chunkindices, __pyx_t_12, long, 1, __Pyx_PyInt_From_long, 1, 1, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 432, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_11 = (__pyx_v_offset + (__pyx_v_chunksize * __pyx_v_i));
    __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_t_11); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 432, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_10 = PyObject_RichCompare(__pyx_t_9, __pyx_t_1, Py_GT); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 432, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 432, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (__pyx_t_3) {
      __Pyx_INCREF(__pyx_t_9);
      __pyx_t_8 = __pyx_t_9;
    } else {
      __pyx_t_10 = __Pyx_PyInt_FromSize_t(__pyx_t_11); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 432, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_8 = __pyx_t_10;
      __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_11 = __Pyx_PyInt_As_size_t(__pyx_t_8); if (unlikely((__pyx_t_11 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 432, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_v_index = __pyx_t_11;

    /* "astropy/io/ascii/cparser.pyx":433
 *         for i in range(1, N):
 *             index = max(offset + chunksize * i, chunkindices[i - 1])
 *             while index < source_len and self.source[index] != '\n':             # <<<<<<<<<<<<<<
 *                 index += 1
 *             if index < source_len:
 */
    while (1) {
      __pyx_t_15 = ((__pyx_v_index < __pyx_v_source_len) != 0);
      if (__pyx_t_15) {
      } else {
        __pyx_t_3 = __pyx_t_15;
        goto __pyx_L17_bool_binop_done;
      }
      __pyx_t_8 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_self->source, __pyx_v_index, size_t, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 433, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_15 = (__Pyx_PyString_Equals(__pyx_t_8, __pyx_kp_s__13, Py_NE)); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 433, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_3 = __pyx_t_15;
      __pyx_L17_bool_binop_done:;
      if (!__pyx_t_3) break;

      /* "astropy/io/ascii/cparser.pyx":434
 *             index = max(offset + chunksize * i, chunkindices[i - 1])
 *             while index < source_len and self.source[index] != '\n':
 *                 index += 1             # <<<<<<<<<<<<<<
 *             if index < source_len:
 *                 chunkindices.append(index + 1)
 */
      __pyx_v_index = (__pyx_v_index + 1);
    }

    /* "astropy/io/ascii/cparser.pyx":435
 *             while index < source_len and self.source[index] != '\n':
 *                 index += 1
 *             if index < source_len:             # <<<<<<<<<<<<<<
 *                 chunkindices.append(index + 1)
 *             else:
 */
    __pyx_t_3 = ((__pyx_v_index < __pyx_v_source_len) != 0);
    if (__pyx_t_3) {

      /* "astropy/io/ascii/cparser.pyx":436
 *                 index += 1
 *             if index < source_len:
 *                 chunkindices.append(index + 1)             # <<<<<<<<<<<<<<
 *             else:
 *                 N = i
 */
      __pyx_t_8 = __Pyx_PyInt_FromSize_t((__pyx_v_index + 1)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 436, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_chunkindices, __pyx_t_8); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 436, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "astropy/io/ascii/cparser.pyx":435
 *             while index < source_len and self.source[index] != '\n':
 *                 index += 1
 *             if index < source_len:             # <<<<<<<<<<<<<<
 *                 chunkindices.append(index + 1)
 *             else:
 */
      goto __pyx_L19;
    }

    /* "astropy/io/ascii/cparser.pyx":438
 *                 chunkindices.append(index + 1)
 *             else:
 *                 N = i             # <<<<<<<<<<<<<<
 *                 break
 * 
 */
    /*else*/ {
      __pyx_v_N = __pyx_v_i;

      /* "astropy/io/ascii/cparser.pyx":439
 *             else:
 *                 N = i
 *                 break             # <<<<<<<<<<<<<<
 * 
 *         self._set_fill_values()
 */
      goto __pyx_L14_break;
    }
    __pyx_L19:;
  }
  __pyx_L14_break:;

  /* "astropy/io/ascii/cparser.pyx":441
 *                 break
 * 
 *         self._set_fill_values()             # <<<<<<<<<<<<<<
 *         chunkindices.append(source_len)
 *         cdef list processes = []
 */
  __pyx_t_8 = ((struct __pyx_vtabstruct_7astropy_2io_5ascii_7cparser_CParser *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_set_fill_values(__pyx_cur_scope->__pyx_v_self); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "astropy/io/ascii/cparser.pyx":442
 * 
 *         self._set_fill_values()
 *         chunkindices.append(source_len)             # <<<<<<<<<<<<<<
 *         cdef list processes = []
 * 
 */
  __pyx_t_8 = __Pyx_PyInt_FromSize_t(__pyx_v_source_len); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_chunkindices, __pyx_t_8); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 442, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "astropy/io/ascii/cparser.pyx":443
 *         self._set_fill_values()
 *         chunkindices.append(source_len)
 *         cdef list processes = []             # <<<<<<<<<<<<<<
 * 
 *         for i in range(N):
 */
  __pyx_t_8 = PyList_New(0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 443, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_v_processes = ((PyObject*)__pyx_t_8);
  __pyx_t_8 = 0;

  /* "astropy/io/ascii/cparser.pyx":445
 *         cdef list processes = []
 * 
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             process = multiprocessing.Process(target=_read_chunk, args=(self,
 *                 chunkindices[i], chunkindices[i + 1],
 */
  __pyx_t_4 = __pyx_v_N;
  __pyx_t_13 = __pyx_t_4;
  for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
    __pyx_v_i = __pyx_t_14;

    /* "astropy/io/ascii/cparser.pyx":446
 * 
 *         for i in range(N):
 *             process = multiprocessing.Process(target=_read_chunk, args=(self,             # <<<<<<<<<<<<<<
 *                 chunkindices[i], chunkindices[i + 1],
 *                 try_int, try_float, try_string, queue, reconvert_queue, i))
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_multiprocessing); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 446, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Process); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 446, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 446, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_read_chunk); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 446, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_target, __pyx_t_10) < 0) __PYX_ERR(0, 446, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

    /* "astropy/io/ascii/cparser.pyx":447
 *         for i in range(N):
 *             process = multiprocessing.Process(target=_read_chunk, args=(self,
 *                 chunkindices[i], chunkindices[i + 1],             # <<<<<<<<<<<<<<
 *                 try_int, try_float, try_string, queue, reconvert_queue, i))
 *             processes.append(process)
 */
    __pyx_t_10 = __Pyx_GetItemInt_List(__pyx_v_chunkindices, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 447, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_12 = (__pyx_v_i + 1);
    __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_chunkindices, __pyx_t_12, long, 1, __Pyx_PyInt_From_long, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 447, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "astropy/io/ascii/cparser.pyx":448
 *             process = multiprocessing.Process(target=_read_chunk, args=(self,
 *                 chunkindices[i], chunkindices[i + 1],
 *                 try_int, try_float, try_string, queue, reconvert_queue, i))             # <<<<<<<<<<<<<<
 *             processes.append(process)
 *             process.start()
 */
    if (unlikely(!__pyx_v_queue)) { __Pyx_RaiseUnboundLocalError("queue"); __PYX_ERR(0, 448, __pyx_L1_error) }
    __pyx_t_17 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 448, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);

    /* "astropy/io/ascii/cparser.pyx":446
 * 
 *         for i in range(N):
 *             process = multiprocessing.Process(target=_read_chunk, args=(self,             # <<<<<<<<<<<<<<
 *                 chunkindices[i], chunkindices[i + 1],
 *                 try_int, try_float, try_string, queue, reconvert_queue, i))
 */
    __pyx_t_18 = PyTuple_New(9); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 446, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_v_self));
    __Pyx_GIVEREF(((PyObject *)__pyx_cur_scope->__pyx_v_self));
    PyTuple_SET_ITEM(__pyx_t_18, 0, ((PyObject *)__pyx_cur_scope->__pyx_v_self));
    __Pyx_GIVEREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_18, 1, __pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_18, 2, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_try_int);
    __Pyx_GIVEREF(__pyx_v_try_int);
    PyTuple_SET_ITEM(__pyx_t_18, 3, __pyx_v_try_int);
    __Pyx_INCREF(__pyx_v_try_float);
    __Pyx_GIVEREF(__pyx_v_try_float);
    PyTuple_SET_ITEM(__pyx_t_18, 4, __pyx_v_try_float);
    __Pyx_INCREF(__pyx_v_try_string);
    __Pyx_GIVEREF(__pyx_v_try_string);
    PyTuple_SET_ITEM(__pyx_t_18, 5, __pyx_v_try_string);
    __Pyx_INCREF(__pyx_v_queue);
    __Pyx_GIVEREF(__pyx_v_queue);
    PyTuple_SET_ITEM(__pyx_t_18, 6, __pyx_v_queue);
    __Pyx_INCREF(__pyx_v_reconvert_queue);
    __Pyx_GIVEREF(__pyx_v_reconvert_queue);
    PyTuple_SET_ITEM(__pyx_t_18, 7, __pyx_v_reconvert_queue);
    __Pyx_GIVEREF(__pyx_t_17);
    PyTuple_SET_ITEM(__pyx_t_18, 8, __pyx_t_17);
    __pyx_t_10 = 0;
    __pyx_t_1 = 0;
    __pyx_t_17 = 0;
    if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_args, __pyx_t_18) < 0) __PYX_ERR(0, 446, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __pyx_t_18 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_empty_tuple, __pyx_t_8); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 446, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF_SET(__pyx_v_process, __pyx_t_18);
    __pyx_t_18 = 0;

    /* "astropy/io/ascii/cparser.pyx":449
 *                 chunkindices[i], chunkindices[i + 1],
 *                 try_int, try_float, try_string, queue, reconvert_queue, i))
 *             processes.append(process)             # <<<<<<<<<<<<<<
 *             process.start()
 * 
 */
    __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_processes, __pyx_v_process); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 449, __pyx_L1_error)

    /* "astropy/io/ascii/cparser.pyx":450
 *                 try_int, try_float, try_string, queue, reconvert_queue, i))
 *             processes.append(process)
 *             process.start()             # <<<<<<<<<<<<<<
 * 
 *         cdef list chunks = [None] * N
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_process, __pyx_n_s_start); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 450, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_18 = (__pyx_t_9) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_9) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 450, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  }

  /* "astropy/io/ascii/cparser.pyx":452
 *             process.start()
 * 
 *         cdef list chunks = [None] * N             # <<<<<<<<<<<<<<
 *         cdef list comments_chunks = [None] * N
 *         cdef dict failed_procs = {}
 */
  __pyx_t_18 = PyList_New(1 * ((__pyx_v_N<0) ? 0:__pyx_v_N)); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  { Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < __pyx_v_N; __pyx_temp++) {
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      PyList_SET_ITEM(__pyx_t_18, __pyx_temp, Py_None);
    }
  }
  __pyx_v_chunks = ((PyObject*)__pyx_t_18);
  __pyx_t_18 = 0;

  /* "astropy/io/ascii/cparser.pyx":453
 * 
 *         cdef list chunks = [None] * N
 *         cdef list comments_chunks = [None] * N             # <<<<<<<<<<<<<<
 *         cdef dict failed_procs = {}
 * 
 */
  __pyx_t_18 = PyList_New(1 * ((__pyx_v_N<0) ? 0:__pyx_v_N)); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  { Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < __pyx_v_N; __pyx_temp++) {
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      PyList_SET_ITEM(__pyx_t_18, __pyx_temp, Py_None);
    }
  }
  __pyx_v_comments_chunks = ((PyObject*)__pyx_t_18);
  __pyx_t_18 = 0;

  /* "astropy/io/ascii/cparser.pyx":454
 *         cdef list chunks = [None] * N
 *         cdef list comments_chunks = [None] * N
 *         cdef dict failed_procs = {}             # <<<<<<<<<<<<<<
 * 
 *         for i in range(N):
 */
  __pyx_t_18 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_v_failed_procs = ((PyObject*)__pyx_t_18);
  __pyx_t_18 = 0;

  /* "astropy/io/ascii/cparser.pyx":456
 *         cdef dict failed_procs = {}
 * 
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             queue_ret, err, proc = queue.get()
 *             if isinstance(err, Exception):
 */
  __pyx_t_4 = __pyx_v_N;
  __pyx_t_13 = __pyx_t_4;
  for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
    __pyx_v_i = __pyx_t_14;

    /* "astropy/io/ascii/cparser.pyx":457
 * 
 *         for i in range(N):
 *             queue_ret, err, proc = queue.get()             # <<<<<<<<<<<<<<
 *             if isinstance(err, Exception):
 *                 for process in processes:
 */
    if (unlikely(!__pyx_v_queue)) { __Pyx_RaiseUnboundLocalError("queue"); __PYX_ERR(0, 457, __pyx_L1_error) }
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_queue, __pyx_n_s_get); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 457, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_18 = (__pyx_t_9) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_9) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 457, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_18))) || (PyList_CheckExact(__pyx_t_18))) {
      PyObject* sequence = __pyx_t_18;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 457, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_17 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_8 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_17 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_17);
      #else
      __pyx_t_8 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 457, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 457, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_17 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 457, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      #endif
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_1 = PyObject_GetIter(__pyx_t_18); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 457, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
      __pyx_t_19 = Py_TYPE(__pyx_t_1)->tp_iternext;
      index = 0; __pyx_t_8 = __pyx_t_19(__pyx_t_1); if (unlikely(!__pyx_t_8)) goto __pyx_L24_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_8);
      index = 1; __pyx_t_9 = __pyx_t_19(__pyx_t_1); if (unlikely(!__pyx_t_9)) goto __pyx_L24_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      index = 2; __pyx_t_17 = __pyx_t_19(__pyx_t_1); if (unlikely(!__pyx_t_17)) goto __pyx_L24_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_17);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_19(__pyx_t_1), 3) < 0) __PYX_ERR(0, 457, __pyx_L1_error)
      __pyx_t_19 = NULL;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L25_unpacking_done;
      __pyx_L24_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_19 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 457, __pyx_L1_error)
      __pyx_L25_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_queue_ret, __pyx_t_8);
    __pyx_t_8 = 0;
    __Pyx_XDECREF_SET(__pyx_v_err, __pyx_t_9);
    __pyx_t_9 = 0;
    __Pyx_XDECREF_SET(__pyx_v_proc, __pyx_t_17);
    __pyx_t_17 = 0;

    /* "astropy/io/ascii/cparser.pyx":458
 *         for i in range(N):
 *             queue_ret, err, proc = queue.get()
 *             if isinstance(err, Exception):             # <<<<<<<<<<<<<<
 *                 for process in processes:
 *                     process.terminate()
 */
    __pyx_t_3 = __Pyx_PyException_Check(__pyx_v_err); 
    __pyx_t_15 = (__pyx_t_3 != 0);
    if (__pyx_t_15) {

      /* "astropy/io/ascii/cparser.pyx":459
 *             queue_ret, err, proc = queue.get()
 *             if isinstance(err, Exception):
 *                 for process in processes:             # <<<<<<<<<<<<<<
 *                     process.terminate()
 *                 raise err
 */
      __pyx_t_18 = __pyx_v_processes; __Pyx_INCREF(__pyx_t_18); __pyx_t_2 = 0;
      for (;;) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_18)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_17 = PyList_GET_ITEM(__pyx_t_18, __pyx_t_2); __Pyx_INCREF(__pyx_t_17); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 459, __pyx_L1_error)
        #else
        __pyx_t_17 = PySequence_ITEM(__pyx_t_18, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 459, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_17);
        #endif
        __Pyx_XDECREF_SET(__pyx_v_process, __pyx_t_17);
        __pyx_t_17 = 0;

        /* "astropy/io/ascii/cparser.pyx":460
 *             if isinstance(err, Exception):
 *                 for process in processes:
 *                     process.terminate()             # <<<<<<<<<<<<<<
 *                 raise err
 *             elif err is not None: # err is (error code, error line)
 */
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_process, __pyx_n_s_terminate); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 460, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_8 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_9);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_9, function);
          }
        }
        __pyx_t_17 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 460, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_17);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

        /* "astropy/io/ascii/cparser.pyx":459
 *             queue_ret, err, proc = queue.get()
 *             if isinstance(err, Exception):
 *                 for process in processes:             # <<<<<<<<<<<<<<
 *                     process.terminate()
 *                 raise err
 */
      }
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

      /* "astropy/io/ascii/cparser.pyx":461
 *                 for process in processes:
 *                     process.terminate()
 *                 raise err             # <<<<<<<<<<<<<<
 *             elif err is not None: # err is (error code, error line)
 *                 failed_procs[proc] = err
 */
      __Pyx_Raise(__pyx_v_err, 0, 0, 0);
      __PYX_ERR(0, 461, __pyx_L1_error)

      /* "astropy/io/ascii/cparser.pyx":458
 *         for i in range(N):
 *             queue_ret, err, proc = queue.get()
 *             if isinstance(err, Exception):             # <<<<<<<<<<<<<<
 *                 for process in processes:
 *                     process.terminate()
 */
    }

    /* "astropy/io/ascii/cparser.pyx":462
 *                     process.terminate()
 *                 raise err
 *             elif err is not None: # err is (error code, error line)             # <<<<<<<<<<<<<<
 *                 failed_procs[proc] = err
 *             comments, data = queue_ret
 */
    __pyx_t_15 = (__pyx_v_err != Py_None);
    __pyx_t_3 = (__pyx_t_15 != 0);
    if (__pyx_t_3) {

      /* "astropy/io/ascii/cparser.pyx":463
 *                 raise err
 *             elif err is not None: # err is (error code, error line)
 *                 failed_procs[proc] = err             # <<<<<<<<<<<<<<
 *             comments, data = queue_ret
 *             comments_chunks[proc] = comments
 */
      if (unlikely(PyDict_SetItem(__pyx_v_failed_procs, __pyx_v_proc, __pyx_v_err) < 0)) __PYX_ERR(0, 463, __pyx_L1_error)

      /* "astropy/io/ascii/cparser.pyx":462
 *                     process.terminate()
 *                 raise err
 *             elif err is not None: # err is (error code, error line)             # <<<<<<<<<<<<<<
 *                 failed_procs[proc] = err
 *             comments, data = queue_ret
 */
    }

    /* "astropy/io/ascii/cparser.pyx":464
 *             elif err is not None: # err is (error code, error line)
 *                 failed_procs[proc] = err
 *             comments, data = queue_ret             # <<<<<<<<<<<<<<
 *             comments_chunks[proc] = comments
 *             chunks[proc] = data
 */
    if ((likely(PyTuple_CheckExact(__pyx_v_queue_ret))) || (PyList_CheckExact(__pyx_v_queue_ret))) {
      PyObject* sequence = __pyx_v_queue_ret;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 464, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_18 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_17 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_18 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_17 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_18);
      __Pyx_INCREF(__pyx_t_17);
      #else
      __pyx_t_18 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 464, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      __pyx_t_17 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 464, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      #endif
    } else {
      Py_ssize_t index = -1;
      __pyx_t_9 = PyObject_GetIter(__pyx_v_queue_ret); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 464, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_19 = Py_TYPE(__pyx_t_9)->tp_iternext;
      index = 0; __pyx_t_18 = __pyx_t_19(__pyx_t_9); if (unlikely(!__pyx_t_18)) goto __pyx_L29_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_18);
      index = 1; __pyx_t_17 = __pyx_t_19(__pyx_t_9); if (unlikely(!__pyx_t_17)) goto __pyx_L29_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_17);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_19(__pyx_t_9), 2) < 0) __PYX_ERR(0, 464, __pyx_L1_error)
      __pyx_t_19 = NULL;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L30_unpacking_done;
      __pyx_L29_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_19 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 464, __pyx_L1_error)
      __pyx_L30_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_comments, __pyx_t_18);
    __pyx_t_18 = 0;
    __Pyx_XDECREF_SET(__pyx_v_data, __pyx_t_17);
    __pyx_t_17 = 0;

    /* "astropy/io/ascii/cparser.pyx":465
 *                 failed_procs[proc] = err
 *             comments, data = queue_ret
 *             comments_chunks[proc] = comments             # <<<<<<<<<<<<<<
 *             chunks[proc] = data
 *         for chunk in comments_chunks:
 */
    if (unlikely(PyObject_SetItem(__pyx_v_comments_chunks, __pyx_v_proc, __pyx_v_comments) < 0)) __PYX_ERR(0, 465, __pyx_L1_error)

    /* "astropy/io/ascii/cparser.pyx":466
 *             comments, data = queue_ret
 *             comments_chunks[proc] = comments
 *             chunks[proc] = data             # <<<<<<<<<<<<<<
 *         for chunk in comments_chunks:
 *             line_comments.extend(chunk)
 */
    if (unlikely(PyObject_SetItem(__pyx_v_chunks, __pyx_v_proc, __pyx_v_data) < 0)) __PYX_ERR(0, 466, __pyx_L1_error)
  }

  /* "astropy/io/ascii/cparser.pyx":467
 *             comments_chunks[proc] = comments
 *             chunks[proc] = data
 *         for chunk in comments_chunks:             # <<<<<<<<<<<<<<
 *             line_comments.extend(chunk)
 * 
 */
  __pyx_t_17 = __pyx_v_comments_chunks; __Pyx_INCREF(__pyx_t_17); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_17)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_18 = PyList_GET_ITEM(__pyx_t_17, __pyx_t_2); __Pyx_INCREF(__pyx_t_18); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 467, __pyx_L1_error)
    #else
    __pyx_t_18 = PySequence_ITEM(__pyx_t_17, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 467, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_chunk, __pyx_t_18);
    __pyx_t_18 = 0;

    /* "astropy/io/ascii/cparser.pyx":468
 *             chunks[proc] = data
 *         for chunk in comments_chunks:
 *             line_comments.extend(chunk)             # <<<<<<<<<<<<<<
 * 
 *         if failed_procs:
 */
    if (unlikely(__pyx_v_line_comments == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "extend");
      __PYX_ERR(0, 468, __pyx_L1_error)
    }
    __pyx_t_16 = __Pyx_PyList_Extend(__pyx_v_line_comments, __pyx_v_chunk); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 468, __pyx_L1_error)

    /* "astropy/io/ascii/cparser.pyx":467
 *             comments_chunks[proc] = comments
 *             chunks[proc] = data
 *         for chunk in comments_chunks:             # <<<<<<<<<<<<<<
 *             line_comments.extend(chunk)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

  /* "astropy/io/ascii/cparser.pyx":470
 *             line_comments.extend(chunk)
 * 
 *         if failed_procs:             # <<<<<<<<<<<<<<
 *             # find the line number of the error
 *             line_no = 0
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_failed_procs); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 470, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "astropy/io/ascii/cparser.pyx":472
 *         if failed_procs:
 *             # find the line number of the error
 *             line_no = 0             # <<<<<<<<<<<<<<
 *             for i in range(N):
 *                 # ignore errors after data_end
 */
    __Pyx_INCREF(__pyx_int_0);
    __pyx_v_line_no = __pyx_int_0;

    /* "astropy/io/ascii/cparser.pyx":473
 *             # find the line number of the error
 *             line_no = 0
 *             for i in range(N):             # <<<<<<<<<<<<<<
 *                 # ignore errors after data_end
 *                 if i in failed_procs and self.data_end is None or line_no < self.data_end:
 */
    __pyx_t_4 = __pyx_v_N;
    __pyx_t_13 = __pyx_t_4;
    for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
      __pyx_v_i = __pyx_t_14;

      /* "astropy/io/ascii/cparser.pyx":475
 *             for i in range(N):
 *                 # ignore errors after data_end
 *                 if i in failed_procs and self.data_end is None or line_no < self.data_end:             # <<<<<<<<<<<<<<
 *                     for process in processes:
 *                         process.terminate()
 */
      __pyx_t_17 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 475, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __pyx_t_15 = (__Pyx_PyDict_ContainsTF(__pyx_t_17, __pyx_v_failed_procs, Py_EQ)); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 475, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      __pyx_t_20 = (__pyx_t_15 != 0);
      if (!__pyx_t_20) {
        goto __pyx_L38_next_or;
      } else {
      }
      __pyx_t_20 = (__pyx_cur_scope->__pyx_v_self->data_end == Py_None);
      __pyx_t_15 = (__pyx_t_20 != 0);
      if (!__pyx_t_15) {
      } else {
        __pyx_t_3 = __pyx_t_15;
        goto __pyx_L37_bool_binop_done;
      }
      __pyx_L38_next_or:;
      __pyx_t_17 = PyObject_RichCompare(__pyx_v_line_no, __pyx_cur_scope->__pyx_v_self->data_end, Py_LT); __Pyx_XGOTREF(__pyx_t_17); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 475, __pyx_L1_error)
      __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_17); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 475, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      __pyx_t_3 = __pyx_t_15;
      __pyx_L37_bool_binop_done:;
      if (__pyx_t_3) {

        /* "astropy/io/ascii/cparser.pyx":476
 *                 # ignore errors after data_end
 *                 if i in failed_procs and self.data_end is None or line_no < self.data_end:
 *                     for process in processes:             # <<<<<<<<<<<<<<
 *                         process.terminate()
 *                     raise self.get_error(failed_procs[i][0], failed_procs[i][1] + line_no,
 */
        __pyx_t_17 = __pyx_v_processes; __Pyx_INCREF(__pyx_t_17); __pyx_t_2 = 0;
        for (;;) {
          if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_17)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_18 = PyList_GET_ITEM(__pyx_t_17, __pyx_t_2); __Pyx_INCREF(__pyx_t_18); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 476, __pyx_L1_error)
          #else
          __pyx_t_18 = PySequence_ITEM(__pyx_t_17, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 476, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_18);
          #endif
          __Pyx_XDECREF_SET(__pyx_v_process, __pyx_t_18);
          __pyx_t_18 = 0;

          /* "astropy/io/ascii/cparser.pyx":477
 *                 if i in failed_procs and self.data_end is None or line_no < self.data_end:
 *                     for process in processes:
 *                         process.terminate()             # <<<<<<<<<<<<<<
 *                     raise self.get_error(failed_procs[i][0], failed_procs[i][1] + line_no,
 *                                          "an error occurred while parsing table data")
 */
          __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_process, __pyx_n_s_terminate); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 477, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_8 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
            __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_9);
            if (likely(__pyx_t_8)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
              __Pyx_INCREF(__pyx_t_8);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_9, function);
            }
          }
          __pyx_t_18 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 477, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_18);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

          /* "astropy/io/ascii/cparser.pyx":476
 *                 # ignore errors after data_end
 *                 if i in failed_procs and self.data_end is None or line_no < self.data_end:
 *                     for process in processes:             # <<<<<<<<<<<<<<
 *                         process.terminate()
 *                     raise self.get_error(failed_procs[i][0], failed_procs[i][1] + line_no,
 */
        }
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

        /* "astropy/io/ascii/cparser.pyx":478
 *                     for process in processes:
 *                         process.terminate()
 *                     raise self.get_error(failed_procs[i][0], failed_procs[i][1] + line_no,             # <<<<<<<<<<<<<<
 *                                          "an error occurred while parsing table data")
 *                 line_no += len(chunks[i][self.names[0]])
 */
        __pyx_t_17 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 478, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_17);
        __pyx_t_18 = __Pyx_PyDict_GetItem(__pyx_v_failed_procs, __pyx_t_17); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 478, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_18);
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        __pyx_t_17 = __Pyx_GetItemInt(__pyx_t_18, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 478, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_17);
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
        __pyx_t_18 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 478, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_18);
        __pyx_t_9 = __Pyx_PyDict_GetItem(__pyx_v_failed_procs, __pyx_t_18); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 478, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
        __pyx_t_18 = __Pyx_GetItemInt(__pyx_t_9, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 478, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_18);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_9 = PyNumber_Add(__pyx_t_18, __pyx_v_line_no); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 478, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
        __pyx_t_18 = ((struct __pyx_vtabstruct_7astropy_2io_5ascii_7cparser_CParser *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->get_error(__pyx_cur_scope->__pyx_v_self, __pyx_t_17, __pyx_t_9, __pyx_kp_s_an_error_occurred_while_parsing); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 478, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_18);
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_Raise(__pyx_t_18, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
        __PYX_ERR(0, 478, __pyx_L1_error)

        /* "astropy/io/ascii/cparser.pyx":475
 *             for i in range(N):
 *                 # ignore errors after data_end
 *                 if i in failed_procs and self.data_end is None or line_no < self.data_end:             # <<<<<<<<<<<<<<
 *                     for process in processes:
 *                         process.terminate()
 */
      }

      /* "astropy/io/ascii/cparser.pyx":480
 *                     raise self.get_error(failed_procs[i][0], failed_procs[i][1] + line_no,
 *                                          "an error occurred while parsing table data")
 *                 line_no += len(chunks[i][self.names[0]])             # <<<<<<<<<<<<<<
 * 
 *         seen_str = {}
 */
      __pyx_t_18 = __Pyx_GetItemInt_List(__pyx_v_chunks, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 480, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      __pyx_t_9 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_self->names, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 480, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_17 = __Pyx_PyObject_GetItem(__pyx_t_18, __pyx_t_9); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 480, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_2 = PyObject_Length(__pyx_t_17); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 480, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      __pyx_t_17 = PyInt_FromSsize_t(__pyx_t_2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 480, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __pyx_t_9 = PyNumber_InPlaceAdd(__pyx_v_line_no, __pyx_t_17); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 480, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      __Pyx_DECREF_SET(__pyx_v_line_no, __pyx_t_9);
      __pyx_t_9 = 0;
    }

    /* "astropy/io/ascii/cparser.pyx":470
 *             line_comments.extend(chunk)
 * 
 *         if failed_procs:             # <<<<<<<<<<<<<<
 *             # find the line number of the error
 *             line_no = 0
 */
  }

  /* "astropy/io/ascii/cparser.pyx":482
 *                 line_no += len(chunks[i][self.names[0]])
 * 
 *         seen_str = {}             # <<<<<<<<<<<<<<
 *         seen_numeric = {}
 *         for name in self.names:
 */
  __pyx_t_9 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_v_seen_str = ((PyObject*)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "astropy/io/ascii/cparser.pyx":483
 * 
 *         seen_str = {}
 *         seen_numeric = {}             # <<<<<<<<<<<<<<
 *         for name in self.names:
 *             seen_str[name] = False
 */
  __pyx_t_9 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_v_seen_numeric = ((PyObject*)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "astropy/io/ascii/cparser.pyx":484
 *         seen_str = {}
 *         seen_numeric = {}
 *         for name in self.names:             # <<<<<<<<<<<<<<
 *             seen_str[name] = False
 *             seen_numeric[name] = False
 */
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_v_self->names)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_v_self->names)) {
    __pyx_t_9 = __pyx_cur_scope->__pyx_v_self->names; __Pyx_INCREF(__pyx_t_9); __pyx_t_2 = 0;
    __pyx_t_21 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_9 = PyObject_GetIter(__pyx_cur_scope->__pyx_v_self->names); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 484, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_21 = Py_TYPE(__pyx_t_9)->tp_iternext; if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 484, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_21)) {
      if (likely(PyList_CheckExact(__pyx_t_9))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_9)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_17 = PyList_GET_ITEM(__pyx_t_9, __pyx_t_2); __Pyx_INCREF(__pyx_t_17); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 484, __pyx_L1_error)
        #else
        __pyx_t_17 = PySequence_ITEM(__pyx_t_9, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 484, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_17);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_9)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_17 = PyTuple_GET_ITEM(__pyx_t_9, __pyx_t_2); __Pyx_INCREF(__pyx_t_17); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 484, __pyx_L1_error)
        #else
        __pyx_t_17 = PySequence_ITEM(__pyx_t_9, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 484, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_17);
        #endif
      }
    } else {
      __pyx_t_17 = __pyx_t_21(__pyx_t_9);
      if (unlikely(!__pyx_t_17)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 484, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_17);
    }
    __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_17);
    __pyx_t_17 = 0;

    /* "astropy/io/ascii/cparser.pyx":485
 *         seen_numeric = {}
 *         for name in self.names:
 *             seen_str[name] = False             # <<<<<<<<<<<<<<
 *             seen_numeric[name] = False
 * 
 */
    if (unlikely(PyDict_SetItem(__pyx_v_seen_str, __pyx_v_name, Py_False) < 0)) __PYX_ERR(0, 485, __pyx_L1_error)

    /* "astropy/io/ascii/cparser.pyx":486
 *         for name in self.names:
 *             seen_str[name] = False
 *             seen_numeric[name] = False             # <<<<<<<<<<<<<<
 * 
 *         for chunk in chunks:
 */
    if (unlikely(PyDict_SetItem(__pyx_v_seen_numeric, __pyx_v_name, Py_False) < 0)) __PYX_ERR(0, 486, __pyx_L1_error)

    /* "astropy/io/ascii/cparser.pyx":484
 *         seen_str = {}
 *         seen_numeric = {}
 *         for name in self.names:             # <<<<<<<<<<<<<<
 *             seen_str[name] = False
 *             seen_numeric[name] = False
 */
  }
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "astropy/io/ascii/cparser.pyx":488
 *             seen_numeric[name] = False
 * 
 *         for chunk in chunks:             # <<<<<<<<<<<<<<
 *             for name in chunk:
 *                 if chunk[name].dtype.kind in ('S', 'U'):
 */
  __pyx_t_9 = __pyx_v_chunks; __Pyx_INCREF(__pyx_t_9); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_9)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_17 = PyList_GET_ITEM(__pyx_t_9, __pyx_t_2); __Pyx_INCREF(__pyx_t_17); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 488, __pyx_L1_error)
    #else
    __pyx_t_17 = PySequence_ITEM(__pyx_t_9, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 488, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_chunk, __pyx_t_17);
    __pyx_t_17 = 0;

    /* "astropy/io/ascii/cparser.pyx":489
 * 
 *         for chunk in chunks:
 *             for name in chunk:             # <<<<<<<<<<<<<<
 *                 if chunk[name].dtype.kind in ('S', 'U'):
 *                     # string values in column
 */
    if (likely(PyList_CheckExact(__pyx_v_chunk)) || PyTuple_CheckExact(__pyx_v_chunk)) {
      __pyx_t_17 = __pyx_v_chunk; __Pyx_INCREF(__pyx_t_17); __pyx_t_22 = 0;
      __pyx_t_21 = NULL;
    } else {
      __pyx_t_22 = -1; __pyx_t_17 = PyObject_GetIter(__pyx_v_chunk); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 489, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __pyx_t_21 = Py_TYPE(__pyx_t_17)->tp_iternext; if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 489, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_21)) {
        if (likely(PyList_CheckExact(__pyx_t_17))) {
          if (__pyx_t_22 >= PyList_GET_SIZE(__pyx_t_17)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_18 = PyList_GET_ITEM(__pyx_t_17, __pyx_t_22); __Pyx_INCREF(__pyx_t_18); __pyx_t_22++; if (unlikely(0 < 0)) __PYX_ERR(0, 489, __pyx_L1_error)
          #else
          __pyx_t_18 = PySequence_ITEM(__pyx_t_17, __pyx_t_22); __pyx_t_22++; if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 489, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_18);
          #endif
        } else {
          if (__pyx_t_22 >= PyTuple_GET_SIZE(__pyx_t_17)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_18 = PyTuple_GET_ITEM(__pyx_t_17, __pyx_t_22); __Pyx_INCREF(__pyx_t_18); __pyx_t_22++; if (unlikely(0 < 0)) __PYX_ERR(0, 489, __pyx_L1_error)
          #else
          __pyx_t_18 = PySequence_ITEM(__pyx_t_17, __pyx_t_22); __pyx_t_22++; if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 489, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_18);
          #endif
        }
      } else {
        __pyx_t_18 = __pyx_t_21(__pyx_t_17);
        if (unlikely(!__pyx_t_18)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 489, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_18);
      }
      __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_18);
      __pyx_t_18 = 0;

      /* "astropy/io/ascii/cparser.pyx":490
 *         for chunk in chunks:
 *             for name in chunk:
 *                 if chunk[name].dtype.kind in ('S', 'U'):             # <<<<<<<<<<<<<<
 *                     # string values in column
 *                     seen_str[name] = True
 */
      __pyx_t_18 = __Pyx_PyObject_GetItem(__pyx_v_chunk, __pyx_v_name); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 490, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_18, __pyx_n_s_dtype); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 490, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
      __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_kind); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 490, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_15 = (__Pyx_PyString_Equals(__pyx_t_18, __pyx_n_s_S, Py_EQ)); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 490, __pyx_L1_error)
      if (!__pyx_t_15) {
      } else {
        __pyx_t_3 = __pyx_t_15;
        goto __pyx_L49_bool_binop_done;
      }
      __pyx_t_15 = (__Pyx_PyString_Equals(__pyx_t_18, __pyx_n_s_U, Py_EQ)); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 490, __pyx_L1_error)
      __pyx_t_3 = __pyx_t_15;
      __pyx_L49_bool_binop_done:;
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
      __pyx_t_15 = (__pyx_t_3 != 0);
      if (__pyx_t_15) {

        /* "astropy/io/ascii/cparser.pyx":492
 *                 if chunk[name].dtype.kind in ('S', 'U'):
 *                     # string values in column
 *                     seen_str[name] = True             # <<<<<<<<<<<<<<
 *                 elif len(chunk[name]) > 0: # ignore empty chunk columns
 *                     seen_numeric[name] = True
 */
        if (unlikely(PyDict_SetItem(__pyx_v_seen_str, __pyx_v_name, Py_True) < 0)) __PYX_ERR(0, 492, __pyx_L1_error)

        /* "astropy/io/ascii/cparser.pyx":490
 *         for chunk in chunks:
 *             for name in chunk:
 *                 if chunk[name].dtype.kind in ('S', 'U'):             # <<<<<<<<<<<<<<
 *                     # string values in column
 *                     seen_str[name] = True
 */
        goto __pyx_L48;
      }

      /* "astropy/io/ascii/cparser.pyx":493
 *                     # string values in column
 *                     seen_str[name] = True
 *                 elif len(chunk[name]) > 0: # ignore empty chunk columns             # <<<<<<<<<<<<<<
 *                     seen_numeric[name] = True
 * 
 */
      __pyx_t_18 = __Pyx_PyObject_GetItem(__pyx_v_chunk, __pyx_v_name); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 493, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      __pyx_t_23 = PyObject_Length(__pyx_t_18); if (unlikely(__pyx_t_23 == ((Py_ssize_t)-1))) __PYX_ERR(0, 493, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
      __pyx_t_15 = ((__pyx_t_23 > 0) != 0);
      if (__pyx_t_15) {

        /* "astropy/io/ascii/cparser.pyx":494
 *                     seen_str[name] = True
 *                 elif len(chunk[name]) > 0: # ignore empty chunk columns
 *                     seen_numeric[name] = True             # <<<<<<<<<<<<<<
 * 
 *         reconvert_cols = []
 */
        if (unlikely(PyDict_SetItem(__pyx_v_seen_numeric, __pyx_v_name, Py_True) < 0)) __PYX_ERR(0, 494, __pyx_L1_error)

        /* "astropy/io/ascii/cparser.pyx":493
 *                     # string values in column
 *                     seen_str[name] = True
 *                 elif len(chunk[name]) > 0: # ignore empty chunk columns             # <<<<<<<<<<<<<<
 *                     seen_numeric[name] = True
 * 
 */
      }
      __pyx_L48:;

      /* "astropy/io/ascii/cparser.pyx":489
 * 
 *         for chunk in chunks:
 *             for name in chunk:             # <<<<<<<<<<<<<<
 *                 if chunk[name].dtype.kind in ('S', 'U'):
 *                     # string values in column
 */
    }
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

    /* "astropy/io/ascii/cparser.pyx":488
 *             seen_numeric[name] = False
 * 
 *         for chunk in chunks:             # <<<<<<<<<<<<<<
 *             for name in chunk:
 *                 if chunk[name].dtype.kind in ('S', 'U'):
 */
  }
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "astropy/io/ascii/cparser.pyx":496
 *                     seen_numeric[name] = True
 * 
 *         reconvert_cols = []             # <<<<<<<<<<<<<<
 * 
 *         for i, name in enumerate(self.names):
 */
  __pyx_t_9 = PyList_New(0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_v_reconvert_cols = ((PyObject*)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "astropy/io/ascii/cparser.pyx":498
 *         reconvert_cols = []
 * 
 *         for i, name in enumerate(self.names):             # <<<<<<<<<<<<<<
 *             if seen_str[name] and seen_numeric[name]:
 *                 # Reconvert to str to avoid conversion issues, e.g.
 */
  __pyx_t_4 = 0;
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_v_self->names)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_v_self->names)) {
    __pyx_t_9 = __pyx_cur_scope->__pyx_v_self->names; __Pyx_INCREF(__pyx_t_9); __pyx_t_2 = 0;
    __pyx_t_21 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_9 = PyObject_GetIter(__pyx_cur_scope->__pyx_v_self->names); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 498, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_21 = Py_TYPE(__pyx_t_9)->tp_iternext; if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 498, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_21)) {
      if (likely(PyList_CheckExact(__pyx_t_9))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_9)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_17 = PyList_GET_ITEM(__pyx_t_9, __pyx_t_2); __Pyx_INCREF(__pyx_t_17); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 498, __pyx_L1_error)
        #else
        __pyx_t_17 = PySequence_ITEM(__pyx_t_9, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 498, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_17);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_9)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_17 = PyTuple_GET_ITEM(__pyx_t_9, __pyx_t_2); __Pyx_INCREF(__pyx_t_17); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 498, __pyx_L1_error)
        #else
        __pyx_t_17 = PySequence_ITEM(__pyx_t_9, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 498, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_17);
        #endif
      }
    } else {
      __pyx_t_17 = __pyx_t_21(__pyx_t_9);
      if (unlikely(!__pyx_t_17)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 498, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_17);
    }
    __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_17);
    __pyx_t_17 = 0;
    __pyx_v_i = __pyx_t_4;
    __pyx_t_4 = (__pyx_t_4 + 1);

    /* "astropy/io/ascii/cparser.pyx":499
 * 
 *         for i, name in enumerate(self.names):
 *             if seen_str[name] and seen_numeric[name]:             # <<<<<<<<<<<<<<
 *                 # Reconvert to str to avoid conversion issues, e.g.
 *                 # 5 (int) -> 5.0 (float) -> 5.0 (string)
 */
    __pyx_t_17 = __Pyx_PyDict_GetItem(__pyx_v_seen_str, __pyx_v_name); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 499, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_17); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 499, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    if (__pyx_t_3) {
    } else {
      __pyx_t_15 = __pyx_t_3;
      goto __pyx_L54_bool_binop_done;
    }
    __pyx_t_17 = __Pyx_PyDict_GetItem(__pyx_v_seen_numeric, __pyx_v_name); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 499, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_17); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 499, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __pyx_t_15 = __pyx_t_3;
    __pyx_L54_bool_binop_done:;
    if (__pyx_t_15) {

      /* "astropy/io/ascii/cparser.pyx":502
 *                 # Reconvert to str to avoid conversion issues, e.g.
 *                 # 5 (int) -> 5.0 (float) -> 5.0 (string)
 *                 reconvert_cols.append(i)             # <<<<<<<<<<<<<<
 * 
 *         reconvert_queue.put(reconvert_cols)
 */
      __pyx_t_17 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 502, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_reconvert_cols, __pyx_t_17); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 502, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

      /* "astropy/io/ascii/cparser.pyx":499
 * 
 *         for i, name in enumerate(self.names):
 *             if seen_str[name] and seen_numeric[name]:             # <<<<<<<<<<<<<<
 *                 # Reconvert to str to avoid conversion issues, e.g.
 *                 # 5 (int) -> 5.0 (float) -> 5.0 (string)
 */
    }

    /* "astropy/io/ascii/cparser.pyx":498
 *         reconvert_cols = []
 * 
 *         for i, name in enumerate(self.names):             # <<<<<<<<<<<<<<
 *             if seen_str[name] and seen_numeric[name]:
 *                 # Reconvert to str to avoid conversion issues, e.g.
 */
  }
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "astropy/io/ascii/cparser.pyx":504
 *                 reconvert_cols.append(i)
 * 
 *         reconvert_queue.put(reconvert_cols)             # <<<<<<<<<<<<<<
 *         for process in processes:
 *             process.join() # wait for each process to finish
 */
  __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_v_reconvert_queue, __pyx_n_s_put); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __pyx_t_18 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_17))) {
    __pyx_t_18 = PyMethod_GET_SELF(__pyx_t_17);
    if (likely(__pyx_t_18)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_17);
      __Pyx_INCREF(__pyx_t_18);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_17, function);
    }
  }
  __pyx_t_9 = (__pyx_t_18) ? __Pyx_PyObject_Call2Args(__pyx_t_17, __pyx_t_18, __pyx_v_reconvert_cols) : __Pyx_PyObject_CallOneArg(__pyx_t_17, __pyx_v_reconvert_cols);
  __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
  if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "astropy/io/ascii/cparser.pyx":505
 * 
 *         reconvert_queue.put(reconvert_cols)
 *         for process in processes:             # <<<<<<<<<<<<<<
 *             process.join() # wait for each process to finish
 *         try:
 */
  __pyx_t_9 = __pyx_v_processes; __Pyx_INCREF(__pyx_t_9); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_9)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_17 = PyList_GET_ITEM(__pyx_t_9, __pyx_t_2); __Pyx_INCREF(__pyx_t_17); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 505, __pyx_L1_error)
    #else
    __pyx_t_17 = PySequence_ITEM(__pyx_t_9, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 505, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_process, __pyx_t_17);
    __pyx_t_17 = 0;

    /* "astropy/io/ascii/cparser.pyx":506
 *         reconvert_queue.put(reconvert_cols)
 *         for process in processes:
 *             process.join() # wait for each process to finish             # <<<<<<<<<<<<<<
 *         try:
 *             while True:
 */
    __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_v_process, __pyx_n_s_join); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 506, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_18))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_18);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_18);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_18, function);
      }
    }
    __pyx_t_17 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_18, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_18);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 506, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

    /* "astropy/io/ascii/cparser.pyx":505
 * 
 *         reconvert_queue.put(reconvert_cols)
 *         for process in processes:             # <<<<<<<<<<<<<<
 *             process.join() # wait for each process to finish
 *         try:
 */
  }
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "astropy/io/ascii/cparser.pyx":507
 *         for process in processes:
 *             process.join() # wait for each process to finish
 *         try:             # <<<<<<<<<<<<<<
 *             while True:
 *                 reconverted, proc, col = queue.get(False)
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_6, &__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_5);
    /*try:*/ {

      /* "astropy/io/ascii/cparser.pyx":508
 *             process.join() # wait for each process to finish
 *         try:
 *             while True:             # <<<<<<<<<<<<<<
 *                 reconverted, proc, col = queue.get(False)
 *                 chunks[proc][self.names[col]] = reconverted
 */
      while (1) {

        /* "astropy/io/ascii/cparser.pyx":509
 *         try:
 *             while True:
 *                 reconverted, proc, col = queue.get(False)             # <<<<<<<<<<<<<<
 *                 chunks[proc][self.names[col]] = reconverted
 *         except Queue.Empty:
 */
        if (unlikely(!__pyx_v_queue)) { __Pyx_RaiseUnboundLocalError("queue"); __PYX_ERR(0, 509, __pyx_L58_error) }
        __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_v_queue, __pyx_n_s_get); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 509, __pyx_L58_error)
        __Pyx_GOTREF(__pyx_t_17);
        __pyx_t_18 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_17))) {
          __pyx_t_18 = PyMethod_GET_SELF(__pyx_t_17);
          if (likely(__pyx_t_18)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_17);
            __Pyx_INCREF(__pyx_t_18);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_17, function);
          }
        }
        __pyx_t_9 = (__pyx_t_18) ? __Pyx_PyObject_Call2Args(__pyx_t_17, __pyx_t_18, Py_False) : __Pyx_PyObject_CallOneArg(__pyx_t_17, Py_False);
        __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
        if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 509, __pyx_L58_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        if ((likely(PyTuple_CheckExact(__pyx_t_9))) || (PyList_CheckExact(__pyx_t_9))) {
          PyObject* sequence = __pyx_t_9;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 3)) {
            if (size > 3) __Pyx_RaiseTooManyValuesError(3);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 509, __pyx_L58_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_17 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_18 = PyTuple_GET_ITEM(sequence, 1); 
            __pyx_t_8 = PyTuple_GET_ITEM(sequence, 2); 
          } else {
            __pyx_t_17 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_18 = PyList_GET_ITEM(sequence, 1); 
            __pyx_t_8 = PyList_GET_ITEM(sequence, 2); 
          }
          __Pyx_INCREF(__pyx_t_17);
          __Pyx_INCREF(__pyx_t_18);
          __Pyx_INCREF(__pyx_t_8);
          #else
          __pyx_t_17 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 509, __pyx_L58_error)
          __Pyx_GOTREF(__pyx_t_17);
          __pyx_t_18 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 509, __pyx_L58_error)
          __Pyx_GOTREF(__pyx_t_18);
          __pyx_t_8 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 509, __pyx_L58_error)
          __Pyx_GOTREF(__pyx_t_8);
          #endif
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_1 = PyObject_GetIter(__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 509, __pyx_L58_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __pyx_t_19 = Py_TYPE(__pyx_t_1)->tp_iternext;
          index = 0; __pyx_t_17 = __pyx_t_19(__pyx_t_1); if (unlikely(!__pyx_t_17)) goto __pyx_L66_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_17);
          index = 1; __pyx_t_18 = __pyx_t_19(__pyx_t_1); if (unlikely(!__pyx_t_18)) goto __pyx_L66_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_18);
          index = 2; __pyx_t_8 = __pyx_t_19(__pyx_t_1); if (unlikely(!__pyx_t_8)) goto __pyx_L66_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_8);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_19(__pyx_t_1), 3) < 0) __PYX_ERR(0, 509, __pyx_L58_error)
          __pyx_t_19 = NULL;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          goto __pyx_L67_unpacking_done;
          __pyx_L66_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_19 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 509, __pyx_L58_error)
          __pyx_L67_unpacking_done:;
        }
        __Pyx_XDECREF_SET(__pyx_v_reconverted, __pyx_t_17);
        __pyx_t_17 = 0;
        __Pyx_XDECREF_SET(__pyx_v_proc, __pyx_t_18);
        __pyx_t_18 = 0;
        __Pyx_XDECREF_SET(__pyx_v_col, __pyx_t_8);
        __pyx_t_8 = 0;

        /* "astropy/io/ascii/cparser.pyx":510
 *             while True:
 *                 reconverted, proc, col = queue.get(False)
 *                 chunks[proc][self.names[col]] = reconverted             # <<<<<<<<<<<<<<
 *         except Queue.Empty:
 *             pass
 */
        __pyx_t_9 = __Pyx_PyObject_GetItem(__pyx_v_chunks, __pyx_v_proc); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 510, __pyx_L58_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_8 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_v_self->names, __pyx_v_col); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 510, __pyx_L58_error)
        __Pyx_GOTREF(__pyx_t_8);
        if (unlikely(PyObject_SetItem(__pyx_t_9, __pyx_t_8, __pyx_v_reconverted) < 0)) __PYX_ERR(0, 510, __pyx_L58_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }

      /* "astropy/io/ascii/cparser.pyx":507
 *         for process in processes:
 *             process.join() # wait for each process to finish
 *         try:             # <<<<<<<<<<<<<<
 *             while True:
 *                 reconverted, proc, col = queue.get(False)
 */
    }
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L63_try_end;
    __pyx_L58_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "astropy/io/ascii/cparser.pyx":511
 *                 reconverted, proc, col = queue.get(False)
 *                 chunks[proc][self.names[col]] = reconverted
 *         except Queue.Empty:             # <<<<<<<<<<<<<<
 *             pass
 * 
 */
    __Pyx_ErrFetch(&__pyx_t_8, &__pyx_t_9, &__pyx_t_18);
    __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_n_s_Queue); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 511, __pyx_L60_except_error)
    __Pyx_GOTREF(__pyx_t_17);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_17, __pyx_n_s_Empty); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 511, __pyx_L60_except_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __pyx_t_4 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_8, __pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_ErrRestore(__pyx_t_8, __pyx_t_9, __pyx_t_18);
    __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_18 = 0;
    if (__pyx_t_4) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L59_exception_handled;
    }
    goto __pyx_L60_except_error;
    __pyx_L60_except_error:;

    /* "astropy/io/ascii/cparser.pyx":507
 *         for process in processes:
 *             process.join() # wait for each process to finish
 *         try:             # <<<<<<<<<<<<<<
 *             while True:
 *                 reconverted, proc, col = queue.get(False)
 */
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_6, __pyx_t_5);
    goto __pyx_L1_error;
    __pyx_L59_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_6, __pyx_t_5);
    __pyx_L63_try_end:;
  }

  /* "astropy/io/ascii/cparser.pyx":514
 *             pass
 * 
 *         if self.data_end is not None:             # <<<<<<<<<<<<<<
 *             if self.data_end < 0:
 *                 # e.g. if data_end = -1, cut the last row
 */
  __pyx_t_15 = (__pyx_cur_scope->__pyx_v_self->data_end != Py_None);
  __pyx_t_3 = (__pyx_t_15 != 0);
  if (__pyx_t_3) {

    /* "astropy/io/ascii/cparser.pyx":515
 * 
 *         if self.data_end is not None:
 *             if self.data_end < 0:             # <<<<<<<<<<<<<<
 *                 # e.g. if data_end = -1, cut the last row
 *                 num_rows = 0
 */
    __pyx_t_18 = PyObject_RichCompare(__pyx_cur_scope->__pyx_v_self->data_end, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_18); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 515, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_18); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 515, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    if (__pyx_t_3) {

      /* "astropy/io/ascii/cparser.pyx":517
 *             if self.data_end < 0:
 *                 # e.g. if data_end = -1, cut the last row
 *                 num_rows = 0             # <<<<<<<<<<<<<<
 *                 for chunk in chunks:
 *                     num_rows += len(chunk[self.names[0]])
 */
      __Pyx_INCREF(__pyx_int_0);
      __pyx_v_num_rows = __pyx_int_0;

      /* "astropy/io/ascii/cparser.pyx":518
 *                 # e.g. if data_end = -1, cut the last row
 *                 num_rows = 0
 *                 for chunk in chunks:             # <<<<<<<<<<<<<<
 *                     num_rows += len(chunk[self.names[0]])
 *                 self.data_end += num_rows
 */
      __pyx_t_18 = __pyx_v_chunks; __Pyx_INCREF(__pyx_t_18); __pyx_t_2 = 0;
      for (;;) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_18)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_9 = PyList_GET_ITEM(__pyx_t_18, __pyx_t_2); __Pyx_INCREF(__pyx_t_9); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 518, __pyx_L1_error)
        #else
        __pyx_t_9 = PySequence_ITEM(__pyx_t_18, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 518, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
        __Pyx_XDECREF_SET(__pyx_v_chunk, __pyx_t_9);
        __pyx_t_9 = 0;

        /* "astropy/io/ascii/cparser.pyx":519
 *                 num_rows = 0
 *                 for chunk in chunks:
 *                     num_rows += len(chunk[self.names[0]])             # <<<<<<<<<<<<<<
 *                 self.data_end += num_rows
 *             else:
 */
        __pyx_t_9 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_self->names, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 519, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_8 = __Pyx_PyObject_GetItem(__pyx_v_chunk, __pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 519, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_22 = PyObject_Length(__pyx_t_8); if (unlikely(__pyx_t_22 == ((Py_ssize_t)-1))) __PYX_ERR(0, 519, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_8 = PyInt_FromSsize_t(__pyx_t_22); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 519, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_9 = PyNumber_InPlaceAdd(__pyx_v_num_rows, __pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 519, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF_SET(__pyx_v_num_rows, __pyx_t_9);
        __pyx_t_9 = 0;

        /* "astropy/io/ascii/cparser.pyx":518
 *                 # e.g. if data_end = -1, cut the last row
 *                 num_rows = 0
 *                 for chunk in chunks:             # <<<<<<<<<<<<<<
 *                     num_rows += len(chunk[self.names[0]])
 *                 self.data_end += num_rows
 */
      }
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

      /* "astropy/io/ascii/cparser.pyx":520
 *                 for chunk in chunks:
 *                     num_rows += len(chunk[self.names[0]])
 *                 self.data_end += num_rows             # <<<<<<<<<<<<<<
 *             else:
 *                 self.data_end -= self.data_start # ignore header
 */
      __pyx_t_18 = PyNumber_InPlaceAdd(__pyx_cur_scope->__pyx_v_self->data_end, __pyx_v_num_rows); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 520, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      __Pyx_GIVEREF(__pyx_t_18);
      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_self->data_end);
      __Pyx_DECREF(__pyx_cur_scope->__pyx_v_self->data_end);
      __pyx_cur_scope->__pyx_v_self->data_end = __pyx_t_18;
      __pyx_t_18 = 0;

      /* "astropy/io/ascii/cparser.pyx":515
 * 
 *         if self.data_end is not None:
 *             if self.data_end < 0:             # <<<<<<<<<<<<<<
 *                 # e.g. if data_end = -1, cut the last row
 *                 num_rows = 0
 */
      goto __pyx_L69;
    }

    /* "astropy/io/ascii/cparser.pyx":522
 *                 self.data_end += num_rows
 *             else:
 *                 self.data_end -= self.data_start # ignore header             # <<<<<<<<<<<<<<
 * 
 *             if self.data_end < 0: # no data
 */
    /*else*/ {
      __pyx_t_18 = __Pyx_PyInt_From_int(__pyx_cur_scope->__pyx_v_self->data_start); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 522, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      __pyx_t_9 = PyNumber_InPlaceSubtract(__pyx_cur_scope->__pyx_v_self->data_end, __pyx_t_18); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 522, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
      __Pyx_GIVEREF(__pyx_t_9);
      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_self->data_end);
      __Pyx_DECREF(__pyx_cur_scope->__pyx_v_self->data_end);
      __pyx_cur_scope->__pyx_v_self->data_end = __pyx_t_9;
      __pyx_t_9 = 0;
    }
    __pyx_L69:;

    /* "astropy/io/ascii/cparser.pyx":524
 *                 self.data_end -= self.data_start # ignore header
 * 
 *             if self.data_end < 0: # no data             # <<<<<<<<<<<<<<
 *                 chunks = [dict((name, []) for name in self.names)]
 *             else:
 */
    __pyx_t_9 = PyObject_RichCompare(__pyx_cur_scope->__pyx_v_self->data_end, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 524, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 524, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (__pyx_t_3) {

      /* "astropy/io/ascii/cparser.pyx":525
 * 
 *             if self.data_end < 0: # no data
 *                 chunks = [dict((name, []) for name in self.names)]             # <<<<<<<<<<<<<<
 *             else:
 *                 line_no = 0
 */
      __pyx_t_9 = __pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_14_read_parallel_3genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 525, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_18 = __Pyx_Generator_Next(__pyx_t_9); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 525, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = PyList_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 525, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_GIVEREF(__pyx_t_18);
      PyList_SET_ITEM(__pyx_t_9, 0, __pyx_t_18);
      __pyx_t_18 = 0;
      __Pyx_DECREF_SET(__pyx_v_chunks, ((PyObject*)__pyx_t_9));
      __pyx_t_9 = 0;

      /* "astropy/io/ascii/cparser.pyx":524
 *                 self.data_end -= self.data_start # ignore header
 * 
 *             if self.data_end < 0: # no data             # <<<<<<<<<<<<<<
 *                 chunks = [dict((name, []) for name in self.names)]
 *             else:
 */
      goto __pyx_L72;
    }

    /* "astropy/io/ascii/cparser.pyx":527
 *                 chunks = [dict((name, []) for name in self.names)]
 *             else:
 *                 line_no = 0             # <<<<<<<<<<<<<<
 *                 for i, chunk in enumerate(chunks):
 *                     num_rows = len(chunk[self.names[0]])
 */
    /*else*/ {
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_XDECREF_SET(__pyx_v_line_no, __pyx_int_0);

      /* "astropy/io/ascii/cparser.pyx":528
 *             else:
 *                 line_no = 0
 *                 for i, chunk in enumerate(chunks):             # <<<<<<<<<<<<<<
 *                     num_rows = len(chunk[self.names[0]])
 *                     if line_no + num_rows > self.data_end:
 */
      __pyx_t_4 = 0;
      __pyx_t_9 = __pyx_v_chunks; __Pyx_INCREF(__pyx_t_9); __pyx_t_2 = 0;
      for (;;) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_9)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_18 = PyList_GET_ITEM(__pyx_t_9, __pyx_t_2); __Pyx_INCREF(__pyx_t_18); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 528, __pyx_L1_error)
        #else
        __pyx_t_18 = PySequence_ITEM(__pyx_t_9, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 528, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_18);
        #endif
        __Pyx_XDECREF_SET(__pyx_v_chunk, __pyx_t_18);
        __pyx_t_18 = 0;
        __pyx_v_i = __pyx_t_4;
        __pyx_t_4 = (__pyx_t_4 + 1);

        /* "astropy/io/ascii/cparser.pyx":529
 *                 line_no = 0
 *                 for i, chunk in enumerate(chunks):
 *                     num_rows = len(chunk[self.names[0]])             # <<<<<<<<<<<<<<
 *                     if line_no + num_rows > self.data_end:
 *                         for name in self.names:
 */
        __pyx_t_18 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_self->names, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 529, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_18);
        __pyx_t_8 = __Pyx_PyObject_GetItem(__pyx_v_chunk, __pyx_t_18); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 529, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
        __pyx_t_22 = PyObject_Length(__pyx_t_8); if (unlikely(__pyx_t_22 == ((Py_ssize_t)-1))) __PYX_ERR(0, 529, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_8 = PyInt_FromSsize_t(__pyx_t_22); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 529, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_XDECREF_SET(__pyx_v_num_rows, __pyx_t_8);
        __pyx_t_8 = 0;

        /* "astropy/io/ascii/cparser.pyx":530
 *                 for i, chunk in enumerate(chunks):
 *                     num_rows = len(chunk[self.names[0]])
 *                     if line_no + num_rows > self.data_end:             # <<<<<<<<<<<<<<
 *                         for name in self.names:
 *                             # truncate columns
 */
        __pyx_t_8 = PyNumber_Add(__pyx_v_line_no, __pyx_v_num_rows); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 530, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_18 = PyObject_RichCompare(__pyx_t_8, __pyx_cur_scope->__pyx_v_self->data_end, Py_GT); __Pyx_XGOTREF(__pyx_t_18); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 530, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_18); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 530, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
        if (__pyx_t_3) {

          /* "astropy/io/ascii/cparser.pyx":531
 *                     num_rows = len(chunk[self.names[0]])
 *                     if line_no + num_rows > self.data_end:
 *                         for name in self.names:             # <<<<<<<<<<<<<<
 *                             # truncate columns
 *                             chunk[name] = chunk[name][:self.data_end - line_no]
 */
          if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_v_self->names)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_v_self->names)) {
            __pyx_t_18 = __pyx_cur_scope->__pyx_v_self->names; __Pyx_INCREF(__pyx_t_18); __pyx_t_22 = 0;
            __pyx_t_21 = NULL;
          } else {
            __pyx_t_22 = -1; __pyx_t_18 = PyObject_GetIter(__pyx_cur_scope->__pyx_v_self->names); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 531, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_18);
            __pyx_t_21 = Py_TYPE(__pyx_t_18)->tp_iternext; if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 531, __pyx_L1_error)
          }
          for (;;) {
            if (likely(!__pyx_t_21)) {
              if (likely(PyList_CheckExact(__pyx_t_18))) {
                if (__pyx_t_22 >= PyList_GET_SIZE(__pyx_t_18)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_8 = PyList_GET_ITEM(__pyx_t_18, __pyx_t_22); __Pyx_INCREF(__pyx_t_8); __pyx_t_22++; if (unlikely(0 < 0)) __PYX_ERR(0, 531, __pyx_L1_error)
                #else
                __pyx_t_8 = PySequence_ITEM(__pyx_t_18, __pyx_t_22); __pyx_t_22++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 531, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_8);
                #endif
              } else {
                if (__pyx_t_22 >= PyTuple_GET_SIZE(__pyx_t_18)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_18, __pyx_t_22); __Pyx_INCREF(__pyx_t_8); __pyx_t_22++; if (unlikely(0 < 0)) __PYX_ERR(0, 531, __pyx_L1_error)
                #else
                __pyx_t_8 = PySequence_ITEM(__pyx_t_18, __pyx_t_22); __pyx_t_22++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 531, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_8);
                #endif
              }
            } else {
              __pyx_t_8 = __pyx_t_21(__pyx_t_18);
              if (unlikely(!__pyx_t_8)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else __PYX_ERR(0, 531, __pyx_L1_error)
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_8);
            }
            __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_8);
            __pyx_t_8 = 0;

            /* "astropy/io/ascii/cparser.pyx":533
 *                         for name in self.names:
 *                             # truncate columns
 *                             chunk[name] = chunk[name][:self.data_end - line_no]             # <<<<<<<<<<<<<<
 *                         del chunks[i + 1:]
 *                         break
 */
            __pyx_t_8 = __Pyx_PyObject_GetItem(__pyx_v_chunk, __pyx_v_name); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 533, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_8);
            __pyx_t_1 = PyNumber_Subtract(__pyx_cur_scope->__pyx_v_self->data_end, __pyx_v_line_no); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 533, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_17 = __Pyx_PyObject_GetSlice(__pyx_t_8, 0, 0, NULL, &__pyx_t_1, NULL, 0, 0, 1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 533, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_17);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (unlikely(PyObject_SetItem(__pyx_v_chunk, __pyx_v_name, __pyx_t_17) < 0)) __PYX_ERR(0, 533, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

            /* "astropy/io/ascii/cparser.pyx":531
 *                     num_rows = len(chunk[self.names[0]])
 *                     if line_no + num_rows > self.data_end:
 *                         for name in self.names:             # <<<<<<<<<<<<<<
 *                             # truncate columns
 *                             chunk[name] = chunk[name][:self.data_end - line_no]
 */
          }
          __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

          /* "astropy/io/ascii/cparser.pyx":534
 *                             # truncate columns
 *                             chunk[name] = chunk[name][:self.data_end - line_no]
 *                         del chunks[i + 1:]             # <<<<<<<<<<<<<<
 *                         break
 *                     line_no += num_rows
 */
          if (__Pyx_PyObject_DelSlice(__pyx_v_chunks, (__pyx_v_i + 1), 0, NULL, NULL, NULL, 1, 0, 1) < 0) __PYX_ERR(0, 534, __pyx_L1_error)

          /* "astropy/io/ascii/cparser.pyx":535
 *                             chunk[name] = chunk[name][:self.data_end - line_no]
 *                         del chunks[i + 1:]
 *                         break             # <<<<<<<<<<<<<<
 *                     line_no += num_rows
 * 
 */
          goto __pyx_L74_break;

          /* "astropy/io/ascii/cparser.pyx":530
 *                 for i, chunk in enumerate(chunks):
 *                     num_rows = len(chunk[self.names[0]])
 *                     if line_no + num_rows > self.data_end:             # <<<<<<<<<<<<<<
 *                         for name in self.names:
 *                             # truncate columns
 */
        }

        /* "astropy/io/ascii/cparser.pyx":536
 *                         del chunks[i + 1:]
 *                         break
 *                     line_no += num_rows             # <<<<<<<<<<<<<<
 * 
 *         ret = {}
 */
        __pyx_t_18 = PyNumber_InPlaceAdd(__pyx_v_line_no, __pyx_v_num_rows); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 536, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_18);
        __Pyx_DECREF_SET(__pyx_v_line_no, __pyx_t_18);
        __pyx_t_18 = 0;

        /* "astropy/io/ascii/cparser.pyx":528
 *             else:
 *                 line_no = 0
 *                 for i, chunk in enumerate(chunks):             # <<<<<<<<<<<<<<
 *                     num_rows = len(chunk[self.names[0]])
 *                     if line_no + num_rows > self.data_end:
 */
      }
      __pyx_L74_break:;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __pyx_L72:;

    /* "astropy/io/ascii/cparser.pyx":514
 *             pass
 * 
 *         if self.data_end is not None:             # <<<<<<<<<<<<<<
 *             if self.data_end < 0:
 *                 # e.g. if data_end = -1, cut the last row
 */
  }

  /* "astropy/io/ascii/cparser.pyx":538
 *                     line_no += num_rows
 * 
 *         ret = {}             # <<<<<<<<<<<<<<
 *         for name in self.get_names():
 *             col_chunks = [chunk.pop(name) for chunk in chunks]
 */
  __pyx_t_9 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_v_ret = ((PyObject*)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "astropy/io/ascii/cparser.pyx":539
 * 
 *         ret = {}
 *         for name in self.get_names():             # <<<<<<<<<<<<<<
 *             col_chunks = [chunk.pop(name) for chunk in chunks]
 *             if any(isinstance(col_chunk, ma.masked_array) for col_chunk in col_chunks):
 */
  __pyx_t_18 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_get_names); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_17 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_18))) {
    __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_18);
    if (likely(__pyx_t_17)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_18);
      __Pyx_INCREF(__pyx_t_17);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_18, function);
    }
  }
  __pyx_t_9 = (__pyx_t_17) ? __Pyx_PyObject_CallOneArg(__pyx_t_18, __pyx_t_17) : __Pyx_PyObject_CallNoArg(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
  if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  if (likely(PyList_CheckExact(__pyx_t_9)) || PyTuple_CheckExact(__pyx_t_9)) {
    __pyx_t_18 = __pyx_t_9; __Pyx_INCREF(__pyx_t_18); __pyx_t_2 = 0;
    __pyx_t_21 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_18 = PyObject_GetIter(__pyx_t_9); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 539, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __pyx_t_21 = Py_TYPE(__pyx_t_18)->tp_iternext; if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 539, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  for (;;) {
    if (likely(!__pyx_t_21)) {
      if (likely(PyList_CheckExact(__pyx_t_18))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_18)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_9 = PyList_GET_ITEM(__pyx_t_18, __pyx_t_2); __Pyx_INCREF(__pyx_t_9); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 539, __pyx_L1_error)
        #else
        __pyx_t_9 = PySequence_ITEM(__pyx_t_18, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 539, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_18)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_9 = PyTuple_GET_ITEM(__pyx_t_18, __pyx_t_2); __Pyx_INCREF(__pyx_t_9); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 539, __pyx_L1_error)
        #else
        __pyx_t_9 = PySequence_ITEM(__pyx_t_18, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 539, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
      }
    } else {
      __pyx_t_9 = __pyx_t_21(__pyx_t_18);
      if (unlikely(!__pyx_t_9)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 539, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_9);
    }
    __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "astropy/io/ascii/cparser.pyx":540
 *         ret = {}
 *         for name in self.get_names():
 *             col_chunks = [chunk.pop(name) for chunk in chunks]             # <<<<<<<<<<<<<<
 *             if any(isinstance(col_chunk, ma.masked_array) for col_chunk in col_chunks):
 *                 ret[name] = ma.concatenate(col_chunks)
 */
    __pyx_t_9 = PyList_New(0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_17 = __pyx_v_chunks; __Pyx_INCREF(__pyx_t_17); __pyx_t_22 = 0;
    for (;;) {
      if (__pyx_t_22 >= PyList_GET_SIZE(__pyx_t_17)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_1 = PyList_GET_ITEM(__pyx_t_17, __pyx_t_22); __Pyx_INCREF(__pyx_t_1); __pyx_t_22++; if (unlikely(0 < 0)) __PYX_ERR(0, 540, __pyx_L1_error)
      #else
      __pyx_t_1 = PySequence_ITEM(__pyx_t_17, __pyx_t_22); __pyx_t_22++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 540, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      #endif
      __Pyx_XDECREF_SET(__pyx_v_chunk, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_chunk, __pyx_n_s_pop); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 540, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_10 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_10, __pyx_v_name) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_v_name);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 540, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_9, (PyObject*)__pyx_t_1))) __PYX_ERR(0, 540, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_col_chunks);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_col_chunks, ((PyObject*)__pyx_t_9));
    __Pyx_GIVEREF(__pyx_t_9);
    __pyx_t_9 = 0;

    /* "astropy/io/ascii/cparser.pyx":541
 *         for name in self.get_names():
 *             col_chunks = [chunk.pop(name) for chunk in chunks]
 *             if any(isinstance(col_chunk, ma.masked_array) for col_chunk in col_chunks):             # <<<<<<<<<<<<<<
 *                 ret[name] = ma.concatenate(col_chunks)
 *             else:
 */
    __pyx_t_9 = __pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_14_read_parallel_6genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_17 = __Pyx_Generator_Next(__pyx_t_9); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_17); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    if (__pyx_t_3) {

      /* "astropy/io/ascii/cparser.pyx":542
 *             col_chunks = [chunk.pop(name) for chunk in chunks]
 *             if any(isinstance(col_chunk, ma.masked_array) for col_chunk in col_chunks):
 *                 ret[name] = ma.concatenate(col_chunks)             # <<<<<<<<<<<<<<
 *             else:
 *                 ret[name] = np.concatenate(col_chunks)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_ma); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 542, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_concatenate); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 542, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_17 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_9, __pyx_cur_scope->__pyx_v_col_chunks) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_cur_scope->__pyx_v_col_chunks);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 542, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(PyDict_SetItem(__pyx_v_ret, __pyx_v_name, __pyx_t_17) < 0)) __PYX_ERR(0, 542, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

      /* "astropy/io/ascii/cparser.pyx":541
 *         for name in self.get_names():
 *             col_chunks = [chunk.pop(name) for chunk in chunks]
 *             if any(isinstance(col_chunk, ma.masked_array) for col_chunk in col_chunks):             # <<<<<<<<<<<<<<
 *                 ret[name] = ma.concatenate(col_chunks)
 *             else:
 */
      goto __pyx_L82;
    }

    /* "astropy/io/ascii/cparser.pyx":544
 *                 ret[name] = ma.concatenate(col_chunks)
 *             else:
 *                 ret[name] = np.concatenate(col_chunks)             # <<<<<<<<<<<<<<
 * 
 *         for process in processes:
 */
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 544, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_concatenate); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 544, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
        }
      }
      __pyx_t_17 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_1, __pyx_cur_scope->__pyx_v_col_chunks) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_cur_scope->__pyx_v_col_chunks);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 544, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(PyDict_SetItem(__pyx_v_ret, __pyx_v_name, __pyx_t_17) < 0)) __PYX_ERR(0, 544, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    }
    __pyx_L82:;

    /* "astropy/io/ascii/cparser.pyx":539
 * 
 *         ret = {}
 *         for name in self.get_names():             # <<<<<<<<<<<<<<
 *             col_chunks = [chunk.pop(name) for chunk in chunks]
 *             if any(isinstance(col_chunk, ma.masked_array) for col_chunk in col_chunks):
 */
  }
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "astropy/io/ascii/cparser.pyx":546
 *                 ret[name] = np.concatenate(col_chunks)
 * 
 *         for process in processes:             # <<<<<<<<<<<<<<
 *             process.terminate()
 * 
 */
  __pyx_t_18 = __pyx_v_processes; __Pyx_INCREF(__pyx_t_18); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_18)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_17 = PyList_GET_ITEM(__pyx_t_18, __pyx_t_2); __Pyx_INCREF(__pyx_t_17); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 546, __pyx_L1_error)
    #else
    __pyx_t_17 = PySequence_ITEM(__pyx_t_18, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 546, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_process, __pyx_t_17);
    __pyx_t_17 = 0;

    /* "astropy/io/ascii/cparser.pyx":547
 * 
 *         for process in processes:
 *             process.terminate()             # <<<<<<<<<<<<<<
 * 
 *         return ret, line_comments
 */
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_process, __pyx_n_s_terminate); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 547, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_17 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 547, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

    /* "astropy/io/ascii/cparser.pyx":546
 *                 ret[name] = np.concatenate(col_chunks)
 * 
 *         for process in processes:             # <<<<<<<<<<<<<<
 *             process.terminate()
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "astropy/io/ascii/cparser.pyx":549
 *             process.terminate()
 * 
 *         return ret, line_comments             # <<<<<<<<<<<<<<
 * 
 *     cdef _set_fill_values(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_18 = PyTuple_New(2); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_INCREF(__pyx_v_ret);
  __Pyx_GIVEREF(__pyx_v_ret);
  PyTuple_SET_ITEM(__pyx_t_18, 0, __pyx_v_ret);
  __Pyx_INCREF(__pyx_v_line_comments);
  __Pyx_GIVEREF(__pyx_v_line_comments);
  PyTuple_SET_ITEM(__pyx_t_18, 1, __pyx_v_line_comments);
  __pyx_r = __pyx_t_18;
  __pyx_t_18 = 0;
  goto __pyx_L0;

  /* "astropy/io/ascii/cparser.pyx":400
 *                                   try_string, num_rows)
 * 
 *     def _read_parallel(self, try_int, try_float, try_string):             # <<<<<<<<<<<<<<
 *         cdef size_t source_len = <size_t>len(self.source)
 *         self.tokenizer.source_pos = 0
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_AddTraceback("astropy.io.ascii.cparser.CParser._read_parallel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_line_comments);
  __Pyx_XDECREF(__pyx_v_queue);
  __Pyx_XDECREF(__pyx_v_chunkindices);
  __Pyx_XDECREF(__pyx_v_reconvert_queue);
  __Pyx_XDECREF(__pyx_v_processes);
  __Pyx_XDECREF(__pyx_v_process);
  __Pyx_XDECREF(__pyx_v_chunks);
  __Pyx_XDECREF(__pyx_v_comments_chunks);
  __Pyx_XDECREF(__pyx_v_failed_procs);
  __Pyx_XDECREF(__pyx_v_queue_ret);
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XDECREF(__pyx_v_proc);
  __Pyx_XDECREF(__pyx_v_comments);
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF(__pyx_v_chunk);
  __Pyx_XDECREF(__pyx_v_line_no);
  __Pyx_XDECREF(__pyx_v_seen_str);
  __Pyx_XDECREF(__pyx_v_seen_numeric);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_reconvert_cols);
  __Pyx_XDECREF(__pyx_v_reconverted);
  __Pyx_XDECREF(__pyx_v_col);
  __Pyx_XDECREF(__pyx_v_num_rows);
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "astropy/io/ascii/cparser.pyx":551
 *         return ret, line_comments
 * 
 *     cdef _set_fill_values(self):             # <<<<<<<<<<<<<<
 *         if self.fill_names is None:
 *             self.fill_names = set(self.names)
 */

static PyObject *__pyx_f_7astropy_2io_5ascii_7cparser_7CParser__set_fill_values(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *(*__pyx_t_7)(PyObject *);
  __Pyx_RefNannySetupContext("_set_fill_values", 0);

  /* "astropy/io/ascii/cparser.pyx":552
 * 
 *     cdef _set_fill_values(self):
 *         if self.fill_names is None:             # <<<<<<<<<<<<<<
 *             self.fill_names = set(self.names)
 *             if self.fill_include_names is not None:
 */
  __pyx_t_1 = (__pyx_v_self->fill_names == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "astropy/io/ascii/cparser.pyx":553
 *     cdef _set_fill_values(self):
 *         if self.fill_names is None:
 *             self.fill_names = set(self.names)             # <<<<<<<<<<<<<<
 *             if self.fill_include_names is not None:
 *                 self.fill_names.intersection_update(self.fill_include_names)
 */
    __pyx_t_3 = PySet_New(__pyx_v_self->names); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 553, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->fill_names);
    __Pyx_DECREF(__pyx_v_self->fill_names);
    __pyx_v_self->fill_names = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "astropy/io/ascii/cparser.pyx":554
 *         if self.fill_names is None:
 *             self.fill_names = set(self.names)
 *             if self.fill_include_names is not None:             # <<<<<<<<<<<<<<
 *                 self.fill_names.intersection_update(self.fill_include_names)
 *             if self.fill_exclude_names is not None:
 */
    __pyx_t_2 = (__pyx_v_self->fill_include_names != Py_None);
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "astropy/io/ascii/cparser.pyx":555
 *             self.fill_names = set(self.names)
 *             if self.fill_include_names is not None:
 *                 self.fill_names.intersection_update(self.fill_include_names)             # <<<<<<<<<<<<<<
 *             if self.fill_exclude_names is not None:
 *                 self.fill_names.difference_update(self.fill_exclude_names)
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->fill_names, __pyx_n_s_intersection_update); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 555, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_self->fill_include_names) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self->fill_include_names);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 555, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "astropy/io/ascii/cparser.pyx":554
 *         if self.fill_names is None:
 *             self.fill_names = set(self.names)
 *             if self.fill_include_names is not None:             # <<<<<<<<<<<<<<
 *                 self.fill_names.intersection_update(self.fill_include_names)
 *             if self.fill_exclude_names is not None:
 */
    }

    /* "astropy/io/ascii/cparser.pyx":556
 *             if self.fill_include_names is not None:
 *                 self.fill_names.intersection_update(self.fill_include_names)
 *             if self.fill_exclude_names is not None:             # <<<<<<<<<<<<<<
 *                 self.fill_names.difference_update(self.fill_exclude_names)
 *         self.fill_values, self.fill_empty = get_fill_values(self.fill_values)
 */
    __pyx_t_1 = (__pyx_v_self->fill_exclude_names != Py_None);
    __pyx_t_2 = (__pyx_t_1 != 0);
    if (__pyx_t_2) {

      /* "astropy/io/ascii/cparser.pyx":557
 *                 self.fill_names.intersection_update(self.fill_include_names)
 *             if self.fill_exclude_names is not None:
 *                 self.fill_names.difference_update(self.fill_exclude_names)             # <<<<<<<<<<<<<<
 *         self.fill_values, self.fill_empty = get_fill_values(self.fill_values)
 * 
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->fill_names, __pyx_n_s_difference_update); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 557, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_self->fill_exclude_names) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self->fill_exclude_names);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 557, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "astropy/io/ascii/cparser.pyx":556
 *             if self.fill_include_names is not None:
 *                 self.fill_names.intersection_update(self.fill_include_names)
 *             if self.fill_exclude_names is not None:             # <<<<<<<<<<<<<<
 *                 self.fill_names.difference_update(self.fill_exclude_names)
 *         self.fill_values, self.fill_empty = get_fill_values(self.fill_values)
 */
    }

    /* "astropy/io/ascii/cparser.pyx":552
 * 
 *     cdef _set_fill_values(self):
 *         if self.fill_names is None:             # <<<<<<<<<<<<<<
 *             self.fill_names = set(self.names)
 *             if self.fill_include_names is not None:
 */
  }

  /* "astropy/io/ascii/cparser.pyx":558
 *             if self.fill_exclude_names is not None:
 *                 self.fill_names.difference_update(self.fill_exclude_names)
 *         self.fill_values, self.fill_empty = get_fill_values(self.fill_values)             # <<<<<<<<<<<<<<
 * 
 *     cdef _get_comments(self, tokenizer_t *t):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_get_fill_values); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_self->fill_values) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self->fill_values);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
    PyObject* sequence = __pyx_t_3;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 558, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    #else
    __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 558, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 558, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_6 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 558, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_7 = Py_TYPE(__pyx_t_6)->tp_iternext;
    index = 0; __pyx_t_4 = __pyx_t_7(__pyx_t_6); if (unlikely(!__pyx_t_4)) goto __pyx_L6_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_4);
    index = 1; __pyx_t_5 = __pyx_t_7(__pyx_t_6); if (unlikely(!__pyx_t_5)) goto __pyx_L6_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_5);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_6), 2) < 0) __PYX_ERR(0, 558, __pyx_L1_error)
    __pyx_t_7 = NULL;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L7_unpacking_done;
    __pyx_L6_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 558, __pyx_L1_error)
    __pyx_L7_unpacking_done:;
  }
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_GOTREF(__pyx_v_self->fill_values);
  __Pyx_DECREF(__pyx_v_self->fill_values);
  __pyx_v_self->fill_values = __pyx_t_4;
  __pyx_t_4 = 0;
  __Pyx_GIVEREF(__pyx_t_5);
  __Pyx_GOTREF(__pyx_v_self->fill_empty);
  __Pyx_DECREF(__pyx_v_self->fill_empty);
  __pyx_v_self->fill_empty = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "astropy/io/ascii/cparser.pyx":551
 *         return ret, line_comments
 * 
 *     cdef _set_fill_values(self):             # <<<<<<<<<<<<<<
 *         if self.fill_names is None:
 *             self.fill_names = set(self.names)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("astropy.io.ascii.cparser.CParser._set_fill_values", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "astropy/io/ascii/cparser.pyx":560
 *         self.fill_values, self.fill_empty = get_fill_values(self.fill_values)
 * 
 *     cdef _get_comments(self, tokenizer_t *t):             # <<<<<<<<<<<<<<
 *         line_comments = []
 *         comment = ''
 */

static PyObject *__pyx_f_7astropy_2io_5ascii_7cparser_7CParser__get_comments(CYTHON_UNUSED struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self, tokenizer_t *__pyx_v_t) {
  PyObject *__pyx_v_line_comments = NULL;
  PyObject *__pyx_v_comment = NULL;
  int __pyx_v_i;
  PyObject *__pyx_v_c = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  __Pyx_RefNannySetupContext("_get_comments", 0);

  /* "astropy/io/ascii/cparser.pyx":561
 * 
 *     cdef _get_comments(self, tokenizer_t *t):
 *         line_comments = []             # <<<<<<<<<<<<<<
 *         comment = ''
 *         for i in range(t.comment_pos):
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_line_comments = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "astropy/io/ascii/cparser.pyx":562
 *     cdef _get_comments(self, tokenizer_t *t):
 *         line_comments = []
 *         comment = ''             # <<<<<<<<<<<<<<
 *         for i in range(t.comment_pos):
 *             c = t.comment_lines[i] # next char in comment string
 */
  __Pyx_INCREF(__pyx_kp_s__5);
  __pyx_v_comment = __pyx_kp_s__5;

  /* "astropy/io/ascii/cparser.pyx":563
 *         line_comments = []
 *         comment = ''
 *         for i in range(t.comment_pos):             # <<<<<<<<<<<<<<
 *             c = t.comment_lines[i] # next char in comment string
 *             if not c: # zero byte -- line terminator
 */
  __pyx_t_2 = __pyx_v_t->comment_pos;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "astropy/io/ascii/cparser.pyx":564
 *         comment = ''
 *         for i in range(t.comment_pos):
 *             c = t.comment_lines[i] # next char in comment string             # <<<<<<<<<<<<<<
 *             if not c: # zero byte -- line terminator
 *                 # replace empty placeholder with ''
 */
    __pyx_t_1 = __Pyx_PyInt_From_char((__pyx_v_t->comment_lines[__pyx_v_i])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_c, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "astropy/io/ascii/cparser.pyx":565
 *         for i in range(t.comment_pos):
 *             c = t.comment_lines[i] # next char in comment string
 *             if not c: # zero byte -- line terminator             # <<<<<<<<<<<<<<
 *                 # replace empty placeholder with ''
 *                 line_comments.append(comment.replace('\x01', '').strip())
 */
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_c); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 565, __pyx_L1_error)
    __pyx_t_6 = ((!__pyx_t_5) != 0);
    if (__pyx_t_6) {

      /* "astropy/io/ascii/cparser.pyx":567
 *             if not c: # zero byte -- line terminator
 *                 # replace empty placeholder with ''
 *                 line_comments.append(comment.replace('\x01', '').strip())             # <<<<<<<<<<<<<<
 *                 comment = ''
 *             else:
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_comment, __pyx_n_s_replace); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 567, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__18, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 567, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_strip); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 567, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 567, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_line_comments, __pyx_t_1); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 567, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "astropy/io/ascii/cparser.pyx":568
 *                 # replace empty placeholder with ''
 *                 line_comments.append(comment.replace('\x01', '').strip())
 *                 comment = ''             # <<<<<<<<<<<<<<
 *             else:
 *                 comment += chr(c)
 */
      __Pyx_INCREF(__pyx_kp_s__5);
      __Pyx_DECREF_SET(__pyx_v_comment, __pyx_kp_s__5);

      /* "astropy/io/ascii/cparser.pyx":565
 *         for i in range(t.comment_pos):
 *             c = t.comment_lines[i] # next char in comment string
 *             if not c: # zero byte -- line terminator             # <<<<<<<<<<<<<<
 *                 # replace empty placeholder with ''
 *                 line_comments.append(comment.replace('\x01', '').strip())
 */
      goto __pyx_L5;
    }

    /* "astropy/io/ascii/cparser.pyx":570
 *                 comment = ''
 *             else:
 *                 comment += chr(c)             # <<<<<<<<<<<<<<
 *         return line_comments
 * 
 */
    /*else*/ {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_chr, __pyx_v_c); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 570, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = PyNumber_InPlaceAdd(__pyx_v_comment, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 570, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF_SET(__pyx_v_comment, __pyx_t_7);
      __pyx_t_7 = 0;
    }
    __pyx_L5:;
  }

  /* "astropy/io/ascii/cparser.pyx":571
 *             else:
 *                 comment += chr(c)
 *         return line_comments             # <<<<<<<<<<<<<<
 * 
 *     cdef _convert_data(self, tokenizer_t *t, try_int, try_float, try_string, num_rows):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_line_comments);
  __pyx_r = __pyx_v_line_comments;
  goto __pyx_L0;

  /* "astropy/io/ascii/cparser.pyx":560
 *         self.fill_values, self.fill_empty = get_fill_values(self.fill_values)
 * 
 *     cdef _get_comments(self, tokenizer_t *t):             # <<<<<<<<<<<<<<
 *         line_comments = []
 *         comment = ''
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("astropy.io.ascii.cparser.CParser._get_comments", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_line_comments);
  __Pyx_XDECREF(__pyx_v_comment);
  __Pyx_XDECREF(__pyx_v_c);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "astropy/io/ascii/cparser.pyx":573
 *         return line_comments
 * 
 *     cdef _convert_data(self, tokenizer_t *t, try_int, try_float, try_string, num_rows):             # <<<<<<<<<<<<<<
 *         cols = {}
 * 
 */

static PyObject *__pyx_f_7astropy_2io_5ascii_7cparser_7CParser__convert_data(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self, tokenizer_t *__pyx_v_t, PyObject *__pyx_v_try_int, PyObject *__pyx_v_try_float, PyObject *__pyx_v_try_string, PyObject *__pyx_v_num_rows) {
  PyObject *__pyx_v_cols = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  __Pyx_RefNannySetupContext("_convert_data", 0);

  /* "astropy/io/ascii/cparser.pyx":574
 * 
 *     cdef _convert_data(self, tokenizer_t *t, try_int, try_float, try_string, num_rows):
 *         cols = {}             # <<<<<<<<<<<<<<
 * 
 *         for i, name in enumerate(self.names):
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_cols = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "astropy/io/ascii/cparser.pyx":576
 *         cols = {}
 * 
 *         for i, name in enumerate(self.names):             # <<<<<<<<<<<<<<
 *             if name not in self.use_cols:
 *                 continue
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_1 = __pyx_int_0;
  if (likely(PyList_CheckExact(__pyx_v_self->names)) || PyTuple_CheckExact(__pyx_v_self->names)) {
    __pyx_t_2 = __pyx_v_self->names; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_self->names); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 576, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 576, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 576, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 576, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 576, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 576, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 576, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_1);
    __pyx_t_5 = __Pyx_PyInt_AddObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 576, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1);
    __pyx_t_1 = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "astropy/io/ascii/cparser.pyx":577
 * 
 *         for i, name in enumerate(self.names):
 *             if name not in self.use_cols:             # <<<<<<<<<<<<<<
 *                 continue
 *             # Try int first, then float, then string
 */
    if (unlikely(__pyx_v_self->use_cols == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 577, __pyx_L1_error)
    }
    __pyx_t_6 = (__Pyx_PySet_ContainsTF(__pyx_v_name, __pyx_v_self->use_cols, Py_NE)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 577, __pyx_L1_error)
    __pyx_t_7 = (__pyx_t_6 != 0);
    if (__pyx_t_7) {

      /* "astropy/io/ascii/cparser.pyx":578
 *         for i, name in enumerate(self.names):
 *             if name not in self.use_cols:
 *                 continue             # <<<<<<<<<<<<<<
 *             # Try int first, then float, then string
 *             try:
 */
      goto __pyx_L3_continue;

      /* "astropy/io/ascii/cparser.pyx":577
 * 
 *         for i, name in enumerate(self.names):
 *             if name not in self.use_cols:             # <<<<<<<<<<<<<<
 *                 continue
 *             # Try int first, then float, then string
 */
    }

    /* "astropy/io/ascii/cparser.pyx":580
 *                 continue
 *             # Try int first, then float, then string
 *             try:             # <<<<<<<<<<<<<<
 *                 if try_int and not try_int[name]:
 *                     raise ValueError()
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "astropy/io/ascii/cparser.pyx":581
 *             # Try int first, then float, then string
 *             try:
 *                 if try_int and not try_int[name]:             # <<<<<<<<<<<<<<
 *                     raise ValueError()
 *                 cols[name] = self._convert_int(t, i, num_rows)
 */
        __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_try_int); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 581, __pyx_L6_error)
        if (__pyx_t_6) {
        } else {
          __pyx_t_7 = __pyx_t_6;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_try_int, __pyx_v_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 581, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 581, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_11 = ((!__pyx_t_6) != 0);
        __pyx_t_7 = __pyx_t_11;
        __pyx_L15_bool_binop_done:;
        if (unlikely(__pyx_t_7)) {

          /* "astropy/io/ascii/cparser.pyx":582
 *             try:
 *                 if try_int and not try_int[name]:
 *                     raise ValueError()             # <<<<<<<<<<<<<<
 *                 cols[name] = self._convert_int(t, i, num_rows)
 *             except ValueError:
 */
          __pyx_t_5 = __Pyx_PyObject_CallNoArg(__pyx_builtin_ValueError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 582, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_Raise(__pyx_t_5, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __PYX_ERR(0, 582, __pyx_L6_error)

          /* "astropy/io/ascii/cparser.pyx":581
 *             # Try int first, then float, then string
 *             try:
 *                 if try_int and not try_int[name]:             # <<<<<<<<<<<<<<
 *                     raise ValueError()
 *                 cols[name] = self._convert_int(t, i, num_rows)
 */
        }

        /* "astropy/io/ascii/cparser.pyx":583
 *                 if try_int and not try_int[name]:
 *                     raise ValueError()
 *                 cols[name] = self._convert_int(t, i, num_rows)             # <<<<<<<<<<<<<<
 *             except ValueError:
 *                 try:
 */
        __pyx_t_12 = __Pyx_PyInt_As_int(__pyx_v_i); if (unlikely((__pyx_t_12 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 583, __pyx_L6_error)
        __pyx_t_13 = __Pyx_PyInt_As_int(__pyx_v_num_rows); if (unlikely((__pyx_t_13 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 583, __pyx_L6_error)
        __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_7astropy_2io_5ascii_7cparser_CParser *)__pyx_v_self->__pyx_vtab)->_convert_int(__pyx_v_self, __pyx_v_t, __pyx_t_12, __pyx_t_13)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 583, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_5);
        if (unlikely(PyDict_SetItem(__pyx_v_cols, __pyx_v_name, __pyx_t_5) < 0)) __PYX_ERR(0, 583, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "astropy/io/ascii/cparser.pyx":580
 *                 continue
 *             # Try int first, then float, then string
 *             try:             # <<<<<<<<<<<<<<
 *                 if try_int and not try_int[name]:
 *                     raise ValueError()
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "astropy/io/ascii/cparser.pyx":584
 *                     raise ValueError()
 *                 cols[name] = self._convert_int(t, i, num_rows)
 *             except ValueError:             # <<<<<<<<<<<<<<
 *                 try:
 *                     if t.code == OVERFLOW_ERROR:
 */
      __pyx_t_13 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError);
      if (__pyx_t_13) {
        __Pyx_AddTraceback("astropy.io.ascii.cparser.CParser._convert_data", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_14, &__pyx_t_15) < 0) __PYX_ERR(0, 584, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_GOTREF(__pyx_t_15);

        /* "astropy/io/ascii/cparser.pyx":585
 *                 cols[name] = self._convert_int(t, i, num_rows)
 *             except ValueError:
 *                 try:             # <<<<<<<<<<<<<<
 *                     if t.code == OVERFLOW_ERROR:
 *                         # Overflow during int conversion (extending range)
 */
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          /*try:*/ {

            /* "astropy/io/ascii/cparser.pyx":586
 *             except ValueError:
 *                 try:
 *                     if t.code == OVERFLOW_ERROR:             # <<<<<<<<<<<<<<
 *                         # Overflow during int conversion (extending range)
 *                         warnings.warn("OverflowError converting to {0} in column {1}, reverting to String."
 */
            __pyx_t_7 = ((__pyx_v_t->code == OVERFLOW_ERROR) != 0);
            if (__pyx_t_7) {

              /* "astropy/io/ascii/cparser.pyx":588
 *                     if t.code == OVERFLOW_ERROR:
 *                         # Overflow during int conversion (extending range)
 *                         warnings.warn("OverflowError converting to {0} in column {1}, reverting to String."             # <<<<<<<<<<<<<<
 *                                   .format('IntType', name), AstropyWarning)
 *                         if try_string and not try_string[name]:
 */
              __Pyx_GetModuleGlobalName(__pyx_t_20, __pyx_n_s_warnings); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 588, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_20);
              __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_t_20, __pyx_n_s_warn); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 588, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_21);
              __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

              /* "astropy/io/ascii/cparser.pyx":589
 *                         # Overflow during int conversion (extending range)
 *                         warnings.warn("OverflowError converting to {0} in column {1}, reverting to String."
 *                                   .format('IntType', name), AstropyWarning)             # <<<<<<<<<<<<<<
 *                         if try_string and not try_string[name]:
 *                             raise ValueError('Column {0} failed to convert'.format(name))
 */
              __pyx_t_22 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_OverflowError_converting_to_0_in, __pyx_n_s_format); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 589, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_22);
              __pyx_t_23 = NULL;
              __pyx_t_13 = 0;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_22))) {
                __pyx_t_23 = PyMethod_GET_SELF(__pyx_t_22);
                if (likely(__pyx_t_23)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_22);
                  __Pyx_INCREF(__pyx_t_23);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_22, function);
                  __pyx_t_13 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_22)) {
                PyObject *__pyx_temp[3] = {__pyx_t_23, __pyx_n_s_IntType, __pyx_v_name};
                __pyx_t_20 = __Pyx_PyFunction_FastCall(__pyx_t_22, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 589, __pyx_L19_error)
                __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
                __Pyx_GOTREF(__pyx_t_20);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_22)) {
                PyObject *__pyx_temp[3] = {__pyx_t_23, __pyx_n_s_IntType, __pyx_v_name};
                __pyx_t_20 = __Pyx_PyCFunction_FastCall(__pyx_t_22, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 589, __pyx_L19_error)
                __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
                __Pyx_GOTREF(__pyx_t_20);
              } else
              #endif
              {
                __pyx_t_24 = PyTuple_New(2+__pyx_t_13); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 589, __pyx_L19_error)
                __Pyx_GOTREF(__pyx_t_24);
                if (__pyx_t_23) {
                  __Pyx_GIVEREF(__pyx_t_23); PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_23); __pyx_t_23 = NULL;
                }
                __Pyx_INCREF(__pyx_n_s_IntType);
                __Pyx_GIVEREF(__pyx_n_s_IntType);
                PyTuple_SET_ITEM(__pyx_t_24, 0+__pyx_t_13, __pyx_n_s_IntType);
                __Pyx_INCREF(__pyx_v_name);
                __Pyx_GIVEREF(__pyx_v_name);
                PyTuple_SET_ITEM(__pyx_t_24, 1+__pyx_t_13, __pyx_v_name);
                __pyx_t_20 = __Pyx_PyObject_Call(__pyx_t_22, __pyx_t_24, NULL); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 589, __pyx_L19_error)
                __Pyx_GOTREF(__pyx_t_20);
                __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
              }
              __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
              __Pyx_GetModuleGlobalName(__pyx_t_22, __pyx_n_s_AstropyWarning); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 589, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_22);
              __pyx_t_24 = NULL;
              __pyx_t_13 = 0;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_21))) {
                __pyx_t_24 = PyMethod_GET_SELF(__pyx_t_21);
                if (likely(__pyx_t_24)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_21);
                  __Pyx_INCREF(__pyx_t_24);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_21, function);
                  __pyx_t_13 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_21)) {
                PyObject *__pyx_temp[3] = {__pyx_t_24, __pyx_t_20, __pyx_t_22};
                __pyx_t_19 = __Pyx_PyFunction_FastCall(__pyx_t_21, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 588, __pyx_L19_error)
                __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
                __Pyx_GOTREF(__pyx_t_19);
                __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_21)) {
                PyObject *__pyx_temp[3] = {__pyx_t_24, __pyx_t_20, __pyx_t_22};
                __pyx_t_19 = __Pyx_PyCFunction_FastCall(__pyx_t_21, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 588, __pyx_L19_error)
                __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
                __Pyx_GOTREF(__pyx_t_19);
                __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
              } else
              #endif
              {
                __pyx_t_23 = PyTuple_New(2+__pyx_t_13); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 588, __pyx_L19_error)
                __Pyx_GOTREF(__pyx_t_23);
                if (__pyx_t_24) {
                  __Pyx_GIVEREF(__pyx_t_24); PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_t_24); __pyx_t_24 = NULL;
                }
                __Pyx_GIVEREF(__pyx_t_20);
                PyTuple_SET_ITEM(__pyx_t_23, 0+__pyx_t_13, __pyx_t_20);
                __Pyx_GIVEREF(__pyx_t_22);
                PyTuple_SET_ITEM(__pyx_t_23, 1+__pyx_t_13, __pyx_t_22);
                __pyx_t_20 = 0;
                __pyx_t_22 = 0;
                __pyx_t_19 = __Pyx_PyObject_Call(__pyx_t_21, __pyx_t_23, NULL); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 588, __pyx_L19_error)
                __Pyx_GOTREF(__pyx_t_19);
                __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
              }
              __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
              __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

              /* "astropy/io/ascii/cparser.pyx":590
 *                         warnings.warn("OverflowError converting to {0} in column {1}, reverting to String."
 *                                   .format('IntType', name), AstropyWarning)
 *                         if try_string and not try_string[name]:             # <<<<<<<<<<<<<<
 *                             raise ValueError('Column {0} failed to convert'.format(name))
 *                         t.code = NO_ERROR
 */
              __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_v_try_string); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 590, __pyx_L19_error)
              if (__pyx_t_11) {
              } else {
                __pyx_t_7 = __pyx_t_11;
                goto __pyx_L29_bool_binop_done;
              }
              __pyx_t_19 = __Pyx_PyObject_GetItem(__pyx_v_try_string, __pyx_v_name); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 590, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_19);
              __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_19); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 590, __pyx_L19_error)
              __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
              __pyx_t_6 = ((!__pyx_t_11) != 0);
              __pyx_t_7 = __pyx_t_6;
              __pyx_L29_bool_binop_done:;
              if (unlikely(__pyx_t_7)) {

                /* "astropy/io/ascii/cparser.pyx":591
 *                                   .format('IntType', name), AstropyWarning)
 *                         if try_string and not try_string[name]:
 *                             raise ValueError('Column {0} failed to convert'.format(name))             # <<<<<<<<<<<<<<
 *                         t.code = NO_ERROR
 *                         cols[name] = self._convert_str(t, i, num_rows)
 */
                __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_Column_0_failed_to_convert, __pyx_n_s_format); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 591, __pyx_L19_error)
                __Pyx_GOTREF(__pyx_t_21);
                __pyx_t_23 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_21))) {
                  __pyx_t_23 = PyMethod_GET_SELF(__pyx_t_21);
                  if (likely(__pyx_t_23)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_21);
                    __Pyx_INCREF(__pyx_t_23);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_21, function);
                  }
                }
                __pyx_t_19 = (__pyx_t_23) ? __Pyx_PyObject_Call2Args(__pyx_t_21, __pyx_t_23, __pyx_v_name) : __Pyx_PyObject_CallOneArg(__pyx_t_21, __pyx_v_name);
                __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
                if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 591, __pyx_L19_error)
                __Pyx_GOTREF(__pyx_t_19);
                __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
                __pyx_t_21 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_19); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 591, __pyx_L19_error)
                __Pyx_GOTREF(__pyx_t_21);
                __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
                __Pyx_Raise(__pyx_t_21, 0, 0, 0);
                __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
                __PYX_ERR(0, 591, __pyx_L19_error)

                /* "astropy/io/ascii/cparser.pyx":590
 *                         warnings.warn("OverflowError converting to {0} in column {1}, reverting to String."
 *                                   .format('IntType', name), AstropyWarning)
 *                         if try_string and not try_string[name]:             # <<<<<<<<<<<<<<
 *                             raise ValueError('Column {0} failed to convert'.format(name))
 *                         t.code = NO_ERROR
 */
              }

              /* "astropy/io/ascii/cparser.pyx":592
 *                         if try_string and not try_string[name]:
 *                             raise ValueError('Column {0} failed to convert'.format(name))
 *                         t.code = NO_ERROR             # <<<<<<<<<<<<<<
 *                         cols[name] = self._convert_str(t, i, num_rows)
 *                     else:
 */
              __pyx_v_t->code = NO_ERROR;

              /* "astropy/io/ascii/cparser.pyx":593
 *                             raise ValueError('Column {0} failed to convert'.format(name))
 *                         t.code = NO_ERROR
 *                         cols[name] = self._convert_str(t, i, num_rows)             # <<<<<<<<<<<<<<
 *                     else:
 *                         if try_float and not try_float[name]:
 */
              __pyx_t_13 = __Pyx_PyInt_As_int(__pyx_v_i); if (unlikely((__pyx_t_13 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 593, __pyx_L19_error)
              __pyx_t_12 = __Pyx_PyInt_As_int(__pyx_v_num_rows); if (unlikely((__pyx_t_12 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 593, __pyx_L19_error)
              __pyx_t_21 = ((struct __pyx_vtabstruct_7astropy_2io_5ascii_7cparser_CParser *)__pyx_v_self->__pyx_vtab)->_convert_str(__pyx_v_self, __pyx_v_t, __pyx_t_13, __pyx_t_12); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 593, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_21);
              if (unlikely(PyDict_SetItem(__pyx_v_cols, __pyx_v_name, __pyx_t_21) < 0)) __PYX_ERR(0, 593, __pyx_L19_error)
              __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

              /* "astropy/io/ascii/cparser.pyx":586
 *             except ValueError:
 *                 try:
 *                     if t.code == OVERFLOW_ERROR:             # <<<<<<<<<<<<<<
 *                         # Overflow during int conversion (extending range)
 *                         warnings.warn("OverflowError converting to {0} in column {1}, reverting to String."
 */
              goto __pyx_L27;
            }

            /* "astropy/io/ascii/cparser.pyx":595
 *                         cols[name] = self._convert_str(t, i, num_rows)
 *                     else:
 *                         if try_float and not try_float[name]:             # <<<<<<<<<<<<<<
 *                             raise ValueError()
 *                         t.code = NO_ERROR
 */
            /*else*/ {
              __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_try_float); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 595, __pyx_L19_error)
              if (__pyx_t_6) {
              } else {
                __pyx_t_7 = __pyx_t_6;
                goto __pyx_L32_bool_binop_done;
              }
              __pyx_t_21 = __Pyx_PyObject_GetItem(__pyx_v_try_float, __pyx_v_name); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 595, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_21);
              __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_21); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 595, __pyx_L19_error)
              __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
              __pyx_t_11 = ((!__pyx_t_6) != 0);
              __pyx_t_7 = __pyx_t_11;
              __pyx_L32_bool_binop_done:;
              if (unlikely(__pyx_t_7)) {

                /* "astropy/io/ascii/cparser.pyx":596
 *                     else:
 *                         if try_float and not try_float[name]:
 *                             raise ValueError()             # <<<<<<<<<<<<<<
 *                         t.code = NO_ERROR
 *                         cols[name] = self._convert_float(t, i, num_rows)
 */
                __pyx_t_21 = __Pyx_PyObject_CallNoArg(__pyx_builtin_ValueError); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 596, __pyx_L19_error)
                __Pyx_GOTREF(__pyx_t_21);
                __Pyx_Raise(__pyx_t_21, 0, 0, 0);
                __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
                __PYX_ERR(0, 596, __pyx_L19_error)

                /* "astropy/io/ascii/cparser.pyx":595
 *                         cols[name] = self._convert_str(t, i, num_rows)
 *                     else:
 *                         if try_float and not try_float[name]:             # <<<<<<<<<<<<<<
 *                             raise ValueError()
 *                         t.code = NO_ERROR
 */
              }

              /* "astropy/io/ascii/cparser.pyx":597
 *                         if try_float and not try_float[name]:
 *                             raise ValueError()
 *                         t.code = NO_ERROR             # <<<<<<<<<<<<<<
 *                         cols[name] = self._convert_float(t, i, num_rows)
 *                         if t.code == OVERFLOW_ERROR:
 */
              __pyx_v_t->code = NO_ERROR;

              /* "astropy/io/ascii/cparser.pyx":598
 *                             raise ValueError()
 *                         t.code = NO_ERROR
 *                         cols[name] = self._convert_float(t, i, num_rows)             # <<<<<<<<<<<<<<
 *                         if t.code == OVERFLOW_ERROR:
 *                             # Overflow during float conversion (extending range)
 */
              __pyx_t_12 = __Pyx_PyInt_As_int(__pyx_v_i); if (unlikely((__pyx_t_12 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 598, __pyx_L19_error)
              __pyx_t_13 = __Pyx_PyInt_As_int(__pyx_v_num_rows); if (unlikely((__pyx_t_13 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 598, __pyx_L19_error)
              __pyx_t_21 = ((PyObject *)((struct __pyx_vtabstruct_7astropy_2io_5ascii_7cparser_CParser *)__pyx_v_self->__pyx_vtab)->_convert_float(__pyx_v_self, __pyx_v_t, __pyx_t_12, __pyx_t_13)); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 598, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_21);
              if (unlikely(PyDict_SetItem(__pyx_v_cols, __pyx_v_name, __pyx_t_21) < 0)) __PYX_ERR(0, 598, __pyx_L19_error)
              __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

              /* "astropy/io/ascii/cparser.pyx":599
 *                         t.code = NO_ERROR
 *                         cols[name] = self._convert_float(t, i, num_rows)
 *                         if t.code == OVERFLOW_ERROR:             # <<<<<<<<<<<<<<
 *                             # Overflow during float conversion (extending range)
 *                             warnings.warn("OverflowError converting to {0} in column {1}, possibly resulting in degraded precision."
 */
              __pyx_t_7 = ((__pyx_v_t->code == OVERFLOW_ERROR) != 0);
              if (__pyx_t_7) {

                /* "astropy/io/ascii/cparser.pyx":601
 *                         if t.code == OVERFLOW_ERROR:
 *                             # Overflow during float conversion (extending range)
 *                             warnings.warn("OverflowError converting to {0} in column {1}, possibly resulting in degraded precision."             # <<<<<<<<<<<<<<
 *                                           .format('FloatType', name), AstropyWarning)
 *                             t.code = NO_ERROR
 */
                __Pyx_GetModuleGlobalName(__pyx_t_19, __pyx_n_s_warnings); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 601, __pyx_L19_error)
                __Pyx_GOTREF(__pyx_t_19);
                __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_t_19, __pyx_n_s_warn); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 601, __pyx_L19_error)
                __Pyx_GOTREF(__pyx_t_23);
                __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

                /* "astropy/io/ascii/cparser.pyx":602
 *                             # Overflow during float conversion (extending range)
 *                             warnings.warn("OverflowError converting to {0} in column {1}, possibly resulting in degraded precision."
 *                                           .format('FloatType', name), AstropyWarning)             # <<<<<<<<<<<<<<
 *                             t.code = NO_ERROR
 *                 except ValueError:
 */
                __pyx_t_22 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_OverflowError_converting_to_0_in_2, __pyx_n_s_format); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 602, __pyx_L19_error)
                __Pyx_GOTREF(__pyx_t_22);
                __pyx_t_20 = NULL;
                __pyx_t_13 = 0;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_22))) {
                  __pyx_t_20 = PyMethod_GET_SELF(__pyx_t_22);
                  if (likely(__pyx_t_20)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_22);
                    __Pyx_INCREF(__pyx_t_20);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_22, function);
                    __pyx_t_13 = 1;
                  }
                }
                #if CYTHON_FAST_PYCALL
                if (PyFunction_Check(__pyx_t_22)) {
                  PyObject *__pyx_temp[3] = {__pyx_t_20, __pyx_n_s_FloatType, __pyx_v_name};
                  __pyx_t_19 = __Pyx_PyFunction_FastCall(__pyx_t_22, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 602, __pyx_L19_error)
                  __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
                  __Pyx_GOTREF(__pyx_t_19);
                } else
                #endif
                #if CYTHON_FAST_PYCCALL
                if (__Pyx_PyFastCFunction_Check(__pyx_t_22)) {
                  PyObject *__pyx_temp[3] = {__pyx_t_20, __pyx_n_s_FloatType, __pyx_v_name};
                  __pyx_t_19 = __Pyx_PyCFunction_FastCall(__pyx_t_22, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 602, __pyx_L19_error)
                  __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
                  __Pyx_GOTREF(__pyx_t_19);
                } else
                #endif
                {
                  __pyx_t_24 = PyTuple_New(2+__pyx_t_13); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 602, __pyx_L19_error)
                  __Pyx_GOTREF(__pyx_t_24);
                  if (__pyx_t_20) {
                    __Pyx_GIVEREF(__pyx_t_20); PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_20); __pyx_t_20 = NULL;
                  }
                  __Pyx_INCREF(__pyx_n_s_FloatType);
                  __Pyx_GIVEREF(__pyx_n_s_FloatType);
                  PyTuple_SET_ITEM(__pyx_t_24, 0+__pyx_t_13, __pyx_n_s_FloatType);
                  __Pyx_INCREF(__pyx_v_name);
                  __Pyx_GIVEREF(__pyx_v_name);
                  PyTuple_SET_ITEM(__pyx_t_24, 1+__pyx_t_13, __pyx_v_name);
                  __pyx_t_19 = __Pyx_PyObject_Call(__pyx_t_22, __pyx_t_24, NULL); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 602, __pyx_L19_error)
                  __Pyx_GOTREF(__pyx_t_19);
                  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
                }
                __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
                __Pyx_GetModuleGlobalName(__pyx_t_22, __pyx_n_s_AstropyWarning); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 602, __pyx_L19_error)
                __Pyx_GOTREF(__pyx_t_22);
                __pyx_t_24 = NULL;
                __pyx_t_13 = 0;
                if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_23))) {
                  __pyx_t_24 = PyMethod_GET_SELF(__pyx_t_23);
                  if (likely(__pyx_t_24)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_23);
                    __Pyx_INCREF(__pyx_t_24);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_23, function);
                    __pyx_t_13 = 1;
                  }
                }
                #if CYTHON_FAST_PYCALL
                if (PyFunction_Check(__pyx_t_23)) {
                  PyObject *__pyx_temp[3] = {__pyx_t_24, __pyx_t_19, __pyx_t_22};
                  __pyx_t_21 = __Pyx_PyFunction_FastCall(__pyx_t_23, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 601, __pyx_L19_error)
                  __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
                  __Pyx_GOTREF(__pyx_t_21);
                  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
                  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
                } else
                #endif
                #if CYTHON_FAST_PYCCALL
                if (__Pyx_PyFastCFunction_Check(__pyx_t_23)) {
                  PyObject *__pyx_temp[3] = {__pyx_t_24, __pyx_t_19, __pyx_t_22};
                  __pyx_t_21 = __Pyx_PyCFunction_FastCall(__pyx_t_23, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 601, __pyx_L19_error)
                  __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
                  __Pyx_GOTREF(__pyx_t_21);
                  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
                  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
                } else
                #endif
                {
                  __pyx_t_20 = PyTuple_New(2+__pyx_t_13); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 601, __pyx_L19_error)
                  __Pyx_GOTREF(__pyx_t_20);
                  if (__pyx_t_24) {
                    __Pyx_GIVEREF(__pyx_t_24); PyTuple_SET_ITEM(__pyx_t_20, 0, __pyx_t_24); __pyx_t_24 = NULL;
                  }
                  __Pyx_GIVEREF(__pyx_t_19);
                  PyTuple_SET_ITEM(__pyx_t_20, 0+__pyx_t_13, __pyx_t_19);
                  __Pyx_GIVEREF(__pyx_t_22);
                  PyTuple_SET_ITEM(__pyx_t_20, 1+__pyx_t_13, __pyx_t_22);
                  __pyx_t_19 = 0;
                  __pyx_t_22 = 0;
                  __pyx_t_21 = __Pyx_PyObject_Call(__pyx_t_23, __pyx_t_20, NULL); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 601, __pyx_L19_error)
                  __Pyx_GOTREF(__pyx_t_21);
                  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                }
                __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
                __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

                /* "astropy/io/ascii/cparser.pyx":603
 *                             warnings.warn("OverflowError converting to {0} in column {1}, possibly resulting in degraded precision."
 *                                           .format('FloatType', name), AstropyWarning)
 *                             t.code = NO_ERROR             # <<<<<<<<<<<<<<
 *                 except ValueError:
 *                     if try_string and not try_string[name]:
 */
                __pyx_v_t->code = NO_ERROR;

                /* "astropy/io/ascii/cparser.pyx":599
 *                         t.code = NO_ERROR
 *                         cols[name] = self._convert_float(t, i, num_rows)
 *                         if t.code == OVERFLOW_ERROR:             # <<<<<<<<<<<<<<
 *                             # Overflow during float conversion (extending range)
 *                             warnings.warn("OverflowError converting to {0} in column {1}, possibly resulting in degraded precision."
 */
              }
            }
            __pyx_L27:;

            /* "astropy/io/ascii/cparser.pyx":585
 *                 cols[name] = self._convert_int(t, i, num_rows)
 *             except ValueError:
 *                 try:             # <<<<<<<<<<<<<<
 *                     if t.code == OVERFLOW_ERROR:
 *                         # Overflow during int conversion (extending range)
 */
          }
          __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
          __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
          __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
          goto __pyx_L26_try_end;
          __pyx_L19_error:;
          __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
          __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
          __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
          __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
          __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
          __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;

          /* "astropy/io/ascii/cparser.pyx":604
 *                                           .format('FloatType', name), AstropyWarning)
 *                             t.code = NO_ERROR
 *                 except ValueError:             # <<<<<<<<<<<<<<
 *                     if try_string and not try_string[name]:
 *                         raise ValueError('Column {0} failed to convert'.format(name))
 */
          __pyx_t_13 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError);
          if (__pyx_t_13) {
            __Pyx_AddTraceback("astropy.io.ascii.cparser.CParser._convert_data", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_21, &__pyx_t_23, &__pyx_t_20) < 0) __PYX_ERR(0, 604, __pyx_L21_except_error)
            __Pyx_GOTREF(__pyx_t_21);
            __Pyx_GOTREF(__pyx_t_23);
            __Pyx_GOTREF(__pyx_t_20);

            /* "astropy/io/ascii/cparser.pyx":605
 *                             t.code = NO_ERROR
 *                 except ValueError:
 *                     if try_string and not try_string[name]:             # <<<<<<<<<<<<<<
 *                         raise ValueError('Column {0} failed to convert'.format(name))
 *                     cols[name] = self._convert_str(t, i, num_rows)
 */
            __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_v_try_string); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 605, __pyx_L21_except_error)
            if (__pyx_t_11) {
            } else {
              __pyx_t_7 = __pyx_t_11;
              goto __pyx_L38_bool_binop_done;
            }
            __pyx_t_22 = __Pyx_PyObject_GetItem(__pyx_v_try_string, __pyx_v_name); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 605, __pyx_L21_except_error)
            __Pyx_GOTREF(__pyx_t_22);
            __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_22); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 605, __pyx_L21_except_error)
            __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
            __pyx_t_6 = ((!__pyx_t_11) != 0);
            __pyx_t_7 = __pyx_t_6;
            __pyx_L38_bool_binop_done:;
            if (unlikely(__pyx_t_7)) {

              /* "astropy/io/ascii/cparser.pyx":606
 *                 except ValueError:
 *                     if try_string and not try_string[name]:
 *                         raise ValueError('Column {0} failed to convert'.format(name))             # <<<<<<<<<<<<<<
 *                     cols[name] = self._convert_str(t, i, num_rows)
 * 
 */
              __pyx_t_19 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_Column_0_failed_to_convert, __pyx_n_s_format); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 606, __pyx_L21_except_error)
              __Pyx_GOTREF(__pyx_t_19);
              __pyx_t_24 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_19))) {
                __pyx_t_24 = PyMethod_GET_SELF(__pyx_t_19);
                if (likely(__pyx_t_24)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_19);
                  __Pyx_INCREF(__pyx_t_24);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_19, function);
                }
              }
              __pyx_t_22 = (__pyx_t_24) ? __Pyx_PyObject_Call2Args(__pyx_t_19, __pyx_t_24, __pyx_v_name) : __Pyx_PyObject_CallOneArg(__pyx_t_19, __pyx_v_name);
              __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
              if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 606, __pyx_L21_except_error)
              __Pyx_GOTREF(__pyx_t_22);
              __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
              __pyx_t_19 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_22); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 606, __pyx_L21_except_error)
              __Pyx_GOTREF(__pyx_t_19);
              __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
              __Pyx_Raise(__pyx_t_19, 0, 0, 0);
              __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
              __PYX_ERR(0, 606, __pyx_L21_except_error)

              /* "astropy/io/ascii/cparser.pyx":605
 *                             t.code = NO_ERROR
 *                 except ValueError:
 *                     if try_string and not try_string[name]:             # <<<<<<<<<<<<<<
 *                         raise ValueError('Column {0} failed to convert'.format(name))
 *                     cols[name] = self._convert_str(t, i, num_rows)
 */
            }

            /* "astropy/io/ascii/cparser.pyx":607
 *                     if try_string and not try_string[name]:
 *                         raise ValueError('Column {0} failed to convert'.format(name))
 *                     cols[name] = self._convert_str(t, i, num_rows)             # <<<<<<<<<<<<<<
 * 
 *         return cols, self._get_comments(t)
 */
            __pyx_t_13 = __Pyx_PyInt_As_int(__pyx_v_i); if (unlikely((__pyx_t_13 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 607, __pyx_L21_except_error)
            __pyx_t_12 = __Pyx_PyInt_As_int(__pyx_v_num_rows); if (unlikely((__pyx_t_12 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 607, __pyx_L21_except_error)
            __pyx_t_19 = ((struct __pyx_vtabstruct_7astropy_2io_5ascii_7cparser_CParser *)__pyx_v_self->__pyx_vtab)->_convert_str(__pyx_v_self, __pyx_v_t, __pyx_t_13, __pyx_t_12); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 607, __pyx_L21_except_error)
            __Pyx_GOTREF(__pyx_t_19);
            if (unlikely(PyDict_SetItem(__pyx_v_cols, __pyx_v_name, __pyx_t_19) < 0)) __PYX_ERR(0, 607, __pyx_L21_except_error)
            __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
            __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
            __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
            __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
            goto __pyx_L20_exception_handled;
          }
          goto __pyx_L21_except_error;
          __pyx_L21_except_error:;

          /* "astropy/io/ascii/cparser.pyx":585
 *                 cols[name] = self._convert_int(t, i, num_rows)
 *             except ValueError:
 *                 try:             # <<<<<<<<<<<<<<
 *                     if t.code == OVERFLOW_ERROR:
 *                         # Overflow during int conversion (extending range)
 */
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_XGIVEREF(__pyx_t_17);
          __Pyx_XGIVEREF(__pyx_t_18);
          __Pyx_ExceptionReset(__pyx_t_16, __pyx_t_17, __pyx_t_18);
          goto __pyx_L8_except_error;
          __pyx_L20_exception_handled:;
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_XGIVEREF(__pyx_t_17);
          __Pyx_XGIVEREF(__pyx_t_18);
          __Pyx_ExceptionReset(__pyx_t_16, __pyx_t_17, __pyx_t_18);
          __pyx_L26_try_end:;
        }
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "astropy/io/ascii/cparser.pyx":580
 *                 continue
 *             # Try int first, then float, then string
 *             try:             # <<<<<<<<<<<<<<
 *                 if try_int and not try_int[name]:
 *                     raise ValueError()
 */
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L1_error;
      __pyx_L7_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }

    /* "astropy/io/ascii/cparser.pyx":576
 *         cols = {}
 * 
 *         for i, name in enumerate(self.names):             # <<<<<<<<<<<<<<
 *             if name not in self.use_cols:
 *                 continue
 */
    __pyx_L3_continue:;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "astropy/io/ascii/cparser.pyx":609
 *                     cols[name] = self._convert_str(t, i, num_rows)
 * 
 *         return cols, self._get_comments(t)             # <<<<<<<<<<<<<<
 * 
 *     cdef np.ndarray _convert_int(self, tokenizer_t *t, int i, int nrows):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_7astropy_2io_5ascii_7cparser_CParser *)__pyx_v_self->__pyx_vtab)->_get_comments(__pyx_v_self, __pyx_v_t); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_cols);
  __Pyx_GIVEREF(__pyx_v_cols);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_cols);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "astropy/io/ascii/cparser.pyx":573
 *         return line_comments
 * 
 *     cdef _convert_data(self, tokenizer_t *t, try_int, try_float, try_string, num_rows):             # <<<<<<<<<<<<<<
 *         cols = {}
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_XDECREF(__pyx_t_22);
  __Pyx_XDECREF(__pyx_t_23);
  __Pyx_XDECREF(__pyx_t_24);
  __Pyx_AddTraceback("astropy.io.ascii.cparser.CParser._convert_data", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_cols);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "astropy/io/ascii/cparser.pyx":611
 *         return cols, self._get_comments(t)
 * 
 *     cdef np.ndarray _convert_int(self, tokenizer_t *t, int i, int nrows):             # <<<<<<<<<<<<<<
 *         cdef int num_rows = t.num_rows
 *         if nrows != -1:
 */

static PyArrayObject *__pyx_f_7astropy_2io_5ascii_7cparser_7CParser__convert_int(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self, tokenizer_t *__pyx_v_t, int __pyx_v_i, int __pyx_v_nrows) {
  int __pyx_v_num_rows;
  PyArrayObject *__pyx_v_col = 0;
  long __pyx_v_converted;
  int __pyx_v_row;
  long *__pyx_v_data;
  char *__pyx_v_field;
  char *__pyx_v_empty_field;
  PyObject *__pyx_v_new_value = 0;
  PyObject *__pyx_v_mask = NULL;
  PyObject *__pyx_v_replace_info = NULL;
  PyArrayObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  char *__pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  __Pyx_RefNannySetupContext("_convert_int", 0);

  /* "astropy/io/ascii/cparser.pyx":612
 * 
 *     cdef np.ndarray _convert_int(self, tokenizer_t *t, int i, int nrows):
 *         cdef int num_rows = t.num_rows             # <<<<<<<<<<<<<<
 *         if nrows != -1:
 *             num_rows = nrows
 */
  __pyx_t_1 = __pyx_v_t->num_rows;
  __pyx_v_num_rows = __pyx_t_1;

  /* "astropy/io/ascii/cparser.pyx":613
 *     cdef np.ndarray _convert_int(self, tokenizer_t *t, int i, int nrows):
 *         cdef int num_rows = t.num_rows
 *         if nrows != -1:             # <<<<<<<<<<<<<<
 *             num_rows = nrows
 *         # intialize ndarray
 */
  __pyx_t_2 = ((__pyx_v_nrows != -1L) != 0);
  if (__pyx_t_2) {

    /* "astropy/io/ascii/cparser.pyx":614
 *         cdef int num_rows = t.num_rows
 *         if nrows != -1:
 *             num_rows = nrows             # <<<<<<<<<<<<<<
 *         # intialize ndarray
 *         cdef np.ndarray col = np.empty(num_rows, dtype=np.int_)
 */
    __pyx_v_num_rows = __pyx_v_nrows;

    /* "astropy/io/ascii/cparser.pyx":613
 *     cdef np.ndarray _convert_int(self, tokenizer_t *t, int i, int nrows):
 *         cdef int num_rows = t.num_rows
 *         if nrows != -1:             # <<<<<<<<<<<<<<
 *             num_rows = nrows
 *         # intialize ndarray
 */
  }

  /* "astropy/io/ascii/cparser.pyx":616
 *             num_rows = nrows
 *         # intialize ndarray
 *         cdef np.ndarray col = np.empty(num_rows, dtype=np.int_)             # <<<<<<<<<<<<<<
 *         cdef long converted
 *         cdef int row = 0
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_empty); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_num_rows); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_int); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_7) < 0) __PYX_ERR(0, 616, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(((__pyx_t_7) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_7, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 616, __pyx_L1_error)
  __pyx_v_col = ((PyArrayObject *)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "astropy/io/ascii/cparser.pyx":618
 *         cdef np.ndarray col = np.empty(num_rows, dtype=np.int_)
 *         cdef long converted
 *         cdef int row = 0             # <<<<<<<<<<<<<<
 *         cdef long *data = <long *> col.data # pointer to raw data
 *         cdef char *field
 */
  __pyx_v_row = 0;

  /* "astropy/io/ascii/cparser.pyx":619
 *         cdef long converted
 *         cdef int row = 0
 *         cdef long *data = <long *> col.data # pointer to raw data             # <<<<<<<<<<<<<<
 *         cdef char *field
 *         cdef char *empty_field = t.buf # memory address of designated empty buffer
 */
  __pyx_v_data = ((long *)__pyx_v_col->data);

  /* "astropy/io/ascii/cparser.pyx":621
 *         cdef long *data = <long *> col.data # pointer to raw data
 *         cdef char *field
 *         cdef char *empty_field = t.buf # memory address of designated empty buffer             # <<<<<<<<<<<<<<
 *         cdef bytes new_value
 *         mask = set() # set of indices for masked values
 */
  __pyx_t_8 = __pyx_v_t->buf;
  __pyx_v_empty_field = __pyx_t_8;

  /* "astropy/io/ascii/cparser.pyx":623
 *         cdef char *empty_field = t.buf # memory address of designated empty buffer
 *         cdef bytes new_value
 *         mask = set() # set of indices for masked values             # <<<<<<<<<<<<<<
 *         start_iteration(t, i) # begin the iteration process in C
 * 
 */
  __pyx_t_7 = PySet_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_v_mask = ((PyObject*)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "astropy/io/ascii/cparser.pyx":624
 *         cdef bytes new_value
 *         mask = set() # set of indices for masked values
 *         start_iteration(t, i) # begin the iteration process in C             # <<<<<<<<<<<<<<
 * 
 *         for row in range(num_rows):
 */
  start_iteration(__pyx_v_t, __pyx_v_i);

  /* "astropy/io/ascii/cparser.pyx":626
 *         start_iteration(t, i) # begin the iteration process in C
 * 
 *         for row in range(num_rows):             # <<<<<<<<<<<<<<
 *             # retrieve the next field as a C pointer
 *             field = next_field(t, <int *>0)
 */
  __pyx_t_1 = __pyx_v_num_rows;
  __pyx_t_9 = __pyx_t_1;
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
    __pyx_v_row = __pyx_t_10;

    /* "astropy/io/ascii/cparser.pyx":628
 *         for row in range(num_rows):
 *             # retrieve the next field as a C pointer
 *             field = next_field(t, <int *>0)             # <<<<<<<<<<<<<<
 *             replace_info = None
 * 
 */
    __pyx_v_field = next_field(__pyx_v_t, ((int *)0));

    /* "astropy/io/ascii/cparser.pyx":629
 *             # retrieve the next field as a C pointer
 *             field = next_field(t, <int *>0)
 *             replace_info = None             # <<<<<<<<<<<<<<
 * 
 *             if field == empty_field and self.fill_empty:
 */
    __Pyx_INCREF(Py_None);
    __Pyx_XDECREF_SET(__pyx_v_replace_info, Py_None);

    /* "astropy/io/ascii/cparser.pyx":631
 *             replace_info = None
 * 
 *             if field == empty_field and self.fill_empty:             # <<<<<<<<<<<<<<
 *                 replace_info = self.fill_empty
 *             # hopefully this implicit char * -> byte conversion for fill values
 */
    __pyx_t_11 = ((__pyx_v_field == __pyx_v_empty_field) != 0);
    if (__pyx_t_11) {
    } else {
      __pyx_t_2 = __pyx_t_11;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_v_self->fill_empty); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 631, __pyx_L1_error)
    __pyx_t_2 = __pyx_t_11;
    __pyx_L7_bool_binop_done:;
    if (__pyx_t_2) {

      /* "astropy/io/ascii/cparser.pyx":632
 * 
 *             if field == empty_field and self.fill_empty:
 *                 replace_info = self.fill_empty             # <<<<<<<<<<<<<<
 *             # hopefully this implicit char * -> byte conversion for fill values
 *             # checking can be avoided in most cases, since self.fill_values will
 */
      __pyx_t_7 = __pyx_v_self->fill_empty;
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_DECREF_SET(__pyx_v_replace_info, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "astropy/io/ascii/cparser.pyx":631
 *             replace_info = None
 * 
 *             if field == empty_field and self.fill_empty:             # <<<<<<<<<<<<<<
 *                 replace_info = self.fill_empty
 *             # hopefully this implicit char * -> byte conversion for fill values
 */
      goto __pyx_L6;
    }

    /* "astropy/io/ascii/cparser.pyx":637
 *             # be empty in the default case (self.fill_empty will do the work
 *             # instead)
 *             elif field != empty_field and self.fill_values and field in self.fill_values:             # <<<<<<<<<<<<<<
 *                 replace_info = self.fill_values[field]
 * 
 */
    __pyx_t_11 = ((__pyx_v_field != __pyx_v_empty_field) != 0);
    if (__pyx_t_11) {
    } else {
      __pyx_t_2 = __pyx_t_11;
      goto __pyx_L9_bool_binop_done;
    }
    __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_v_self->fill_values); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 637, __pyx_L1_error)
    if (__pyx_t_11) {
    } else {
      __pyx_t_2 = __pyx_t_11;
      goto __pyx_L9_bool_binop_done;
    }
    __pyx_t_7 = __Pyx_PyBytes_FromString(__pyx_v_field); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 637, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_11 = (__Pyx_PySequence_ContainsTF(__pyx_t_7, __pyx_v_self->fill_values, Py_EQ)); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 637, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_12 = (__pyx_t_11 != 0);
    __pyx_t_2 = __pyx_t_12;
    __pyx_L9_bool_binop_done:;
    if (__pyx_t_2) {

      /* "astropy/io/ascii/cparser.pyx":638
 *             # instead)
 *             elif field != empty_field and self.fill_values and field in self.fill_values:
 *                 replace_info = self.fill_values[field]             # <<<<<<<<<<<<<<
 * 
 *             if replace_info is not None:
 */
      __pyx_t_7 = __Pyx_PyBytes_FromString(__pyx_v_field); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 638, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_self->fill_values, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 638, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF_SET(__pyx_v_replace_info, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "astropy/io/ascii/cparser.pyx":637
 *             # be empty in the default case (self.fill_empty will do the work
 *             # instead)
 *             elif field != empty_field and self.fill_values and field in self.fill_values:             # <<<<<<<<<<<<<<
 *                 replace_info = self.fill_values[field]
 * 
 */
    }
    __pyx_L6:;

    /* "astropy/io/ascii/cparser.pyx":640
 *                 replace_info = self.fill_values[field]
 * 
 *             if replace_info is not None:             # <<<<<<<<<<<<<<
 *                 # Either this column applies to the field as specified in the
 *                 # fill_values parameter, or no specific columns are specified
 */
    __pyx_t_2 = (__pyx_v_replace_info != Py_None);
    __pyx_t_12 = (__pyx_t_2 != 0);
    if (__pyx_t_12) {

      /* "astropy/io/ascii/cparser.pyx":644
 *                 # fill_values parameter, or no specific columns are specified
 *                 # and this column should apply fill_values.
 *                 if (len(replace_info) > 1 and self.names[i] in replace_info[1:]) \             # <<<<<<<<<<<<<<
 *                    or (len(replace_info) == 1 and self.names[i] in self.fill_names):
 *                     mask.add(row)
 */
      __pyx_t_13 = PyObject_Length(__pyx_v_replace_info); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 644, __pyx_L1_error)
      __pyx_t_2 = ((__pyx_t_13 > 1) != 0);
      if (!__pyx_t_2) {
        goto __pyx_L15_next_or;
      } else {
      }
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_self->names, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 644, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = __Pyx_PyObject_GetSlice(__pyx_v_replace_info, 1, 0, NULL, NULL, &__pyx_slice__19, 1, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 644, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_3, __pyx_t_7, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 644, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_11 = (__pyx_t_2 != 0);
      if (!__pyx_t_11) {
      } else {
        __pyx_t_12 = __pyx_t_11;
        goto __pyx_L14_bool_binop_done;
      }
      __pyx_L15_next_or:;

      /* "astropy/io/ascii/cparser.pyx":645
 *                 # and this column should apply fill_values.
 *                 if (len(replace_info) > 1 and self.names[i] in replace_info[1:]) \
 *                    or (len(replace_info) == 1 and self.names[i] in self.fill_names):             # <<<<<<<<<<<<<<
 *                     mask.add(row)
 *                     new_value = str(replace_info[0]).encode('ascii')
 */
      __pyx_t_13 = PyObject_Length(__pyx_v_replace_info); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 645, __pyx_L1_error)
      __pyx_t_11 = ((__pyx_t_13 == 1) != 0);
      if (__pyx_t_11) {
      } else {
        __pyx_t_12 = __pyx_t_11;
        goto __pyx_L14_bool_binop_done;
      }
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_self->names, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 645, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_11 = (__Pyx_PySequence_ContainsTF(__pyx_t_7, __pyx_v_self->fill_names, Py_EQ)); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 645, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_2 = (__pyx_t_11 != 0);
      __pyx_t_12 = __pyx_t_2;
      __pyx_L14_bool_binop_done:;

      /* "astropy/io/ascii/cparser.pyx":644
 *                 # fill_values parameter, or no specific columns are specified
 *                 # and this column should apply fill_values.
 *                 if (len(replace_info) > 1 and self.names[i] in replace_info[1:]) \             # <<<<<<<<<<<<<<
 *                    or (len(replace_info) == 1 and self.names[i] in self.fill_names):
 *                     mask.add(row)
 */
      if (__pyx_t_12) {

        /* "astropy/io/ascii/cparser.pyx":646
 *                 if (len(replace_info) > 1 and self.names[i] in replace_info[1:]) \
 *                    or (len(replace_info) == 1 and self.names[i] in self.fill_names):
 *                     mask.add(row)             # <<<<<<<<<<<<<<
 *                     new_value = str(replace_info[0]).encode('ascii')
 *                     # try converting the new value
 */
        __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_row); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 646, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_14 = PySet_Add(__pyx_v_mask, __pyx_t_7); if (unlikely(__pyx_t_14 == ((int)-1))) __PYX_ERR(0, 646, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "astropy/io/ascii/cparser.pyx":647
 *                    or (len(replace_info) == 1 and self.names[i] in self.fill_names):
 *                     mask.add(row)
 *                     new_value = str(replace_info[0]).encode('ascii')             # <<<<<<<<<<<<<<
 *                     # try converting the new value
 *                     converted = str_to_long(t, new_value)
 */
        __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_replace_info, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 647, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_5 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 647, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_encode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 647, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_7 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_n_s_ascii) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_n_s_ascii);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 647, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(PyBytes_CheckExact(__pyx_t_7))||((__pyx_t_7) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_7)->tp_name), 0))) __PYX_ERR(0, 647, __pyx_L1_error)
        __Pyx_XDECREF_SET(__pyx_v_new_value, ((PyObject*)__pyx_t_7));
        __pyx_t_7 = 0;

        /* "astropy/io/ascii/cparser.pyx":649
 *                     new_value = str(replace_info[0]).encode('ascii')
 *                     # try converting the new value
 *                     converted = str_to_long(t, new_value)             # <<<<<<<<<<<<<<
 *                 else:
 *                     converted = str_to_long(t, field)
 */
        if (unlikely(__pyx_v_new_value == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
          __PYX_ERR(0, 649, __pyx_L1_error)
        }
        __pyx_t_8 = __Pyx_PyBytes_AsWritableString(__pyx_v_new_value); if (unlikely((!__pyx_t_8) && PyErr_Occurred())) __PYX_ERR(0, 649, __pyx_L1_error)
        __pyx_v_converted = str_to_long(__pyx_v_t, __pyx_t_8);

        /* "astropy/io/ascii/cparser.pyx":644
 *                 # fill_values parameter, or no specific columns are specified
 *                 # and this column should apply fill_values.
 *                 if (len(replace_info) > 1 and self.names[i] in replace_info[1:]) \             # <<<<<<<<<<<<<<
 *                    or (len(replace_info) == 1 and self.names[i] in self.fill_names):
 *                     mask.add(row)
 */
        goto __pyx_L13;
      }

      /* "astropy/io/ascii/cparser.pyx":651
 *                     converted = str_to_long(t, new_value)
 *                 else:
 *                     converted = str_to_long(t, field)             # <<<<<<<<<<<<<<
 *             else:
 *                 # convert the field to long (widest integer type)
 */
      /*else*/ {
        __pyx_v_converted = str_to_long(__pyx_v_t, __pyx_v_field);
      }
      __pyx_L13:;

      /* "astropy/io/ascii/cparser.pyx":640
 *                 replace_info = self.fill_values[field]
 * 
 *             if replace_info is not None:             # <<<<<<<<<<<<<<
 *                 # Either this column applies to the field as specified in the
 *                 # fill_values parameter, or no specific columns are specified
 */
      goto __pyx_L12;
    }

    /* "astropy/io/ascii/cparser.pyx":654
 *             else:
 *                 # convert the field to long (widest integer type)
 *                 converted = str_to_long(t, field)             # <<<<<<<<<<<<<<
 * 
 *             if t.code in (CONVERSION_ERROR, OVERFLOW_ERROR):
 */
    /*else*/ {
      __pyx_v_converted = str_to_long(__pyx_v_t, __pyx_v_field);
    }
    __pyx_L12:;

    /* "astropy/io/ascii/cparser.pyx":656
 *                 converted = str_to_long(t, field)
 * 
 *             if t.code in (CONVERSION_ERROR, OVERFLOW_ERROR):             # <<<<<<<<<<<<<<
 *                 # no dice
 *                 if t.code == CONVERSION_ERROR:
 */
    switch (__pyx_v_t->code) {
      case CONVERSION_ERROR:
      case OVERFLOW_ERROR:

      /* "astropy/io/ascii/cparser.pyx":658
 *             if t.code in (CONVERSION_ERROR, OVERFLOW_ERROR):
 *                 # no dice
 *                 if t.code == CONVERSION_ERROR:             # <<<<<<<<<<<<<<
 *                     t.code = NO_ERROR
 *                 raise ValueError()
 */
      __pyx_t_12 = ((__pyx_v_t->code == CONVERSION_ERROR) != 0);
      if (__pyx_t_12) {

        /* "astropy/io/ascii/cparser.pyx":659
 *                 # no dice
 *                 if t.code == CONVERSION_ERROR:
 *                     t.code = NO_ERROR             # <<<<<<<<<<<<<<
 *                 raise ValueError()
 * 
 */
        __pyx_v_t->code = NO_ERROR;

        /* "astropy/io/ascii/cparser.pyx":658
 *             if t.code in (CONVERSION_ERROR, OVERFLOW_ERROR):
 *                 # no dice
 *                 if t.code == CONVERSION_ERROR:             # <<<<<<<<<<<<<<
 *                     t.code = NO_ERROR
 *                 raise ValueError()
 */
      }

      /* "astropy/io/ascii/cparser.pyx":660
 *                 if t.code == CONVERSION_ERROR:
 *                     t.code = NO_ERROR
 *                 raise ValueError()             # <<<<<<<<<<<<<<
 * 
 *             data[row] = converted
 */
      __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_builtin_ValueError); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 660, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_Raise(__pyx_t_7, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __PYX_ERR(0, 660, __pyx_L1_error)

      /* "astropy/io/ascii/cparser.pyx":656
 *                 converted = str_to_long(t, field)
 * 
 *             if t.code in (CONVERSION_ERROR, OVERFLOW_ERROR):             # <<<<<<<<<<<<<<
 *                 # no dice
 *                 if t.code == CONVERSION_ERROR:
 */
      break;
      default: break;
    }

    /* "astropy/io/ascii/cparser.pyx":662
 *                 raise ValueError()
 * 
 *             data[row] = converted             # <<<<<<<<<<<<<<
 *             row += 1
 * 
 */
    (__pyx_v_data[__pyx_v_row]) = __pyx_v_converted;

    /* "astropy/io/ascii/cparser.pyx":663
 * 
 *             data[row] = converted
 *             row += 1             # <<<<<<<<<<<<<<
 * 
 *         if mask:
 */
    __pyx_v_row = (__pyx_v_row + 1);
  }

  /* "astropy/io/ascii/cparser.pyx":665
 *             row += 1
 * 
 *         if mask:             # <<<<<<<<<<<<<<
 *             # convert to masked_array
 *             return ma.masked_array(col, mask=[1 if i in mask else 0 for i in
 */
  __pyx_t_12 = (PySet_GET_SIZE(__pyx_v_mask) != 0);
  if (__pyx_t_12) {

    /* "astropy/io/ascii/cparser.pyx":667
 *         if mask:
 *             # convert to masked_array
 *             return ma.masked_array(col, mask=[1 if i in mask else 0 for i in             # <<<<<<<<<<<<<<
 *                                               range(row)])
 *         else:
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_ma); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 667, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_masked_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 667, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 667, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(((PyObject *)__pyx_v_col));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_col));
    PyTuple_SET_ITEM(__pyx_t_7, 0, ((PyObject *)__pyx_v_col));
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 667, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 667, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "astropy/io/ascii/cparser.pyx":668
 *             # convert to masked_array
 *             return ma.masked_array(col, mask=[1 if i in mask else 0 for i in
 *                                               range(row)])             # <<<<<<<<<<<<<<
 *         else:
 *             return col
 */
    __pyx_t_1 = __pyx_v_row;

    /* "astropy/io/ascii/cparser.pyx":667
 *         if mask:
 *             # convert to masked_array
 *             return ma.masked_array(col, mask=[1 if i in mask else 0 for i in             # <<<<<<<<<<<<<<
 *                                               range(row)])
 *         else:
 */
    __pyx_t_9 = __pyx_t_1;
    for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
      __pyx_v_i = __pyx_t_10;
      __pyx_t_15 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 667, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __pyx_t_12 = (__Pyx_PySet_ContainsTF(__pyx_t_15, __pyx_v_mask, Py_EQ)); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 667, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      if ((__pyx_t_12 != 0)) {
        __Pyx_INCREF(__pyx_int_1);
        __pyx_t_6 = __pyx_int_1;
      } else {
        __Pyx_INCREF(__pyx_int_0);
        __pyx_t_6 = __pyx_int_0;
      }
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 667, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_mask, __pyx_t_4) < 0) __PYX_ERR(0, 667, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 667, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 667, __pyx_L1_error)
    __pyx_r = ((PyArrayObject *)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "astropy/io/ascii/cparser.pyx":665
 *             row += 1
 * 
 *         if mask:             # <<<<<<<<<<<<<<
 *             # convert to masked_array
 *             return ma.masked_array(col, mask=[1 if i in mask else 0 for i in
 */
  }

  /* "astropy/io/ascii/cparser.pyx":670
 *                                               range(row)])
 *         else:
 *             return col             # <<<<<<<<<<<<<<
 * 
 *     cdef np.ndarray _convert_float(self, tokenizer_t *t, int i, int nrows):
 */
  /*else*/ {
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_INCREF(((PyObject *)__pyx_v_col));
    __pyx_r = __pyx_v_col;
    goto __pyx_L0;
  }

  /* "astropy/io/ascii/cparser.pyx":611
 *         return cols, self._get_comments(t)
 * 
 *     cdef np.ndarray _convert_int(self, tokenizer_t *t, int i, int nrows):             # <<<<<<<<<<<<<<
 *         cdef int num_rows = t.num_rows
 *         if nrows != -1:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("astropy.io.ascii.cparser.CParser._convert_int", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_col);
  __Pyx_XDECREF(__pyx_v_new_value);
  __Pyx_XDECREF(__pyx_v_mask);
  __Pyx_XDECREF(__pyx_v_replace_info);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "astropy/io/ascii/cparser.pyx":672
 *             return col
 * 
 *     cdef np.ndarray _convert_float(self, tokenizer_t *t, int i, int nrows):             # <<<<<<<<<<<<<<
 *         # very similar to _convert_int()
 *         cdef int num_rows = t.num_rows
 */

static PyArrayObject *__pyx_f_7astropy_2io_5ascii_7cparser_7CParser__convert_float(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self, tokenizer_t *__pyx_v_t, int __pyx_v_i, int __pyx_v_nrows) {
  int __pyx_v_num_rows;
  PyArrayObject *__pyx_v_col = 0;
  double __pyx_v_converted;
  int __pyx_v_row;
  double *__pyx_v_data;
  char *__pyx_v_field;
  char *__pyx_v_empty_field;
  PyObject *__pyx_v_new_value = 0;
  CYTHON_UNUSED int __pyx_v_replacing;
  err_code __pyx_v_overflown;
  PyObject *__pyx_v_mask = NULL;
  PyObject *__pyx_v_replace_info = NULL;
  PyArrayObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  char *__pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  __Pyx_RefNannySetupContext("_convert_float", 0);

  /* "astropy/io/ascii/cparser.pyx":674
 *     cdef np.ndarray _convert_float(self, tokenizer_t *t, int i, int nrows):
 *         # very similar to _convert_int()
 *         cdef int num_rows = t.num_rows             # <<<<<<<<<<<<<<
 *         if nrows != -1:
 *             num_rows = nrows
 */
  __pyx_t_1 = __pyx_v_t->num_rows;
  __pyx_v_num_rows = __pyx_t_1;

  /* "astropy/io/ascii/cparser.pyx":675
 *         # very similar to _convert_int()
 *         cdef int num_rows = t.num_rows
 *         if nrows != -1:             # <<<<<<<<<<<<<<
 *             num_rows = nrows
 * 
 */
  __pyx_t_2 = ((__pyx_v_nrows != -1L) != 0);
  if (__pyx_t_2) {

    /* "astropy/io/ascii/cparser.pyx":676
 *         cdef int num_rows = t.num_rows
 *         if nrows != -1:
 *             num_rows = nrows             # <<<<<<<<<<<<<<
 * 
 *         cdef np.ndarray col = np.empty(num_rows, dtype=np.float_)
 */
    __pyx_v_num_rows = __pyx_v_nrows;

    /* "astropy/io/ascii/cparser.pyx":675
 *         # very similar to _convert_int()
 *         cdef int num_rows = t.num_rows
 *         if nrows != -1:             # <<<<<<<<<<<<<<
 *             num_rows = nrows
 * 
 */
  }

  /* "astropy/io/ascii/cparser.pyx":678
 *             num_rows = nrows
 * 
 *         cdef np.ndarray col = np.empty(num_rows, dtype=np.float_)             # <<<<<<<<<<<<<<
 *         cdef double converted
 *         cdef int row = 0
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_empty); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_num_rows); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_float); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_7) < 0) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(((__pyx_t_7) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_7, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 678, __pyx_L1_error)
  __pyx_v_col = ((PyArrayObject *)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "astropy/io/ascii/cparser.pyx":680
 *         cdef np.ndarray col = np.empty(num_rows, dtype=np.float_)
 *         cdef double converted
 *         cdef int row = 0             # <<<<<<<<<<<<<<
 *         cdef double *data = <double *> col.data
 *         cdef char *field
 */
  __pyx_v_row = 0;

  /* "astropy/io/ascii/cparser.pyx":681
 *         cdef double converted
 *         cdef int row = 0
 *         cdef double *data = <double *> col.data             # <<<<<<<<<<<<<<
 *         cdef char *field
 *         cdef char *empty_field = t.buf
 */
  __pyx_v_data = ((double *)__pyx_v_col->data);

  /* "astropy/io/ascii/cparser.pyx":683
 *         cdef double *data = <double *> col.data
 *         cdef char *field
 *         cdef char *empty_field = t.buf             # <<<<<<<<<<<<<<
 *         cdef bytes new_value
 *         cdef int replacing
 */
  __pyx_t_8 = __pyx_v_t->buf;
  __pyx_v_empty_field = __pyx_t_8;

  /* "astropy/io/ascii/cparser.pyx":686
 *         cdef bytes new_value
 *         cdef int replacing
 *         cdef err_code overflown = NO_ERROR # store any OVERFLOW to raise warning             # <<<<<<<<<<<<<<
 *         mask = set()
 * 
 */
  __pyx_v_overflown = NO_ERROR;

  /* "astropy/io/ascii/cparser.pyx":687
 *         cdef int replacing
 *         cdef err_code overflown = NO_ERROR # store any OVERFLOW to raise warning
 *         mask = set()             # <<<<<<<<<<<<<<
 * 
 *         start_iteration(t, i)
 */
  __pyx_t_7 = PySet_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_v_mask = ((PyObject*)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "astropy/io/ascii/cparser.pyx":689
 *         mask = set()
 * 
 *         start_iteration(t, i)             # <<<<<<<<<<<<<<
 *         for row in range(num_rows):
 *             field = next_field(t, <int *>0)
 */
  start_iteration(__pyx_v_t, __pyx_v_i);

  /* "astropy/io/ascii/cparser.pyx":690
 * 
 *         start_iteration(t, i)
 *         for row in range(num_rows):             # <<<<<<<<<<<<<<
 *             field = next_field(t, <int *>0)
 *             replace_info = None
 */
  __pyx_t_1 = __pyx_v_num_rows;
  __pyx_t_9 = __pyx_t_1;
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
    __pyx_v_row = __pyx_t_10;

    /* "astropy/io/ascii/cparser.pyx":691
 *         start_iteration(t, i)
 *         for row in range(num_rows):
 *             field = next_field(t, <int *>0)             # <<<<<<<<<<<<<<
 *             replace_info = None
 *             replacing = False
 */
    __pyx_v_field = next_field(__pyx_v_t, ((int *)0));

    /* "astropy/io/ascii/cparser.pyx":692
 *         for row in range(num_rows):
 *             field = next_field(t, <int *>0)
 *             replace_info = None             # <<<<<<<<<<<<<<
 *             replacing = False
 * 
 */
    __Pyx_INCREF(Py_None);
    __Pyx_XDECREF_SET(__pyx_v_replace_info, Py_None);

    /* "astropy/io/ascii/cparser.pyx":693
 *             field = next_field(t, <int *>0)
 *             replace_info = None
 *             replacing = False             # <<<<<<<<<<<<<<
 * 
 *             if field == empty_field and self.fill_empty:
 */
    __pyx_v_replacing = 0;

    /* "astropy/io/ascii/cparser.pyx":695
 *             replacing = False
 * 
 *             if field == empty_field and self.fill_empty:             # <<<<<<<<<<<<<<
 *                 replace_info = self.fill_empty
 * 
 */
    __pyx_t_11 = ((__pyx_v_field == __pyx_v_empty_field) != 0);
    if (__pyx_t_11) {
    } else {
      __pyx_t_2 = __pyx_t_11;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_v_self->fill_empty); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 695, __pyx_L1_error)
    __pyx_t_2 = __pyx_t_11;
    __pyx_L7_bool_binop_done:;
    if (__pyx_t_2) {

      /* "astropy/io/ascii/cparser.pyx":696
 * 
 *             if field == empty_field and self.fill_empty:
 *                 replace_info = self.fill_empty             # <<<<<<<<<<<<<<
 * 
 *             elif field != empty_field and self.fill_values and field in self.fill_values:
 */
      __pyx_t_7 = __pyx_v_self->fill_empty;
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_DECREF_SET(__pyx_v_replace_info, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "astropy/io/ascii/cparser.pyx":695
 *             replacing = False
 * 
 *             if field == empty_field and self.fill_empty:             # <<<<<<<<<<<<<<
 *                 replace_info = self.fill_empty
 * 
 */
      goto __pyx_L6;
    }

    /* "astropy/io/ascii/cparser.pyx":698
 *                 replace_info = self.fill_empty
 * 
 *             elif field != empty_field and self.fill_values and field in self.fill_values:             # <<<<<<<<<<<<<<
 *                 replace_info = self.fill_values[field]
 * 
 */
    __pyx_t_11 = ((__pyx_v_field != __pyx_v_empty_field) != 0);
    if (__pyx_t_11) {
    } else {
      __pyx_t_2 = __pyx_t_11;
      goto __pyx_L9_bool_binop_done;
    }
    __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_v_self->fill_values); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 698, __pyx_L1_error)
    if (__pyx_t_11) {
    } else {
      __pyx_t_2 = __pyx_t_11;
      goto __pyx_L9_bool_binop_done;
    }
    __pyx_t_7 = __Pyx_PyBytes_FromString(__pyx_v_field); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 698, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_11 = (__Pyx_PySequence_ContainsTF(__pyx_t_7, __pyx_v_self->fill_values, Py_EQ)); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 698, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_12 = (__pyx_t_11 != 0);
    __pyx_t_2 = __pyx_t_12;
    __pyx_L9_bool_binop_done:;
    if (__pyx_t_2) {

      /* "astropy/io/ascii/cparser.pyx":699
 * 
 *             elif field != empty_field and self.fill_values and field in self.fill_values:
 *                 replace_info = self.fill_values[field]             # <<<<<<<<<<<<<<
 * 
 *             if replace_info is not None:
 */
      __pyx_t_7 = __Pyx_PyBytes_FromString(__pyx_v_field); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 699, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_self->fill_values, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 699, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF_SET(__pyx_v_replace_info, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "astropy/io/ascii/cparser.pyx":698
 *                 replace_info = self.fill_empty
 * 
 *             elif field != empty_field and self.fill_values and field in self.fill_values:             # <<<<<<<<<<<<<<
 *                 replace_info = self.fill_values[field]
 * 
 */
    }
    __pyx_L6:;

    /* "astropy/io/ascii/cparser.pyx":701
 *                 replace_info = self.fill_values[field]
 * 
 *             if replace_info is not None:             # <<<<<<<<<<<<<<
 *                 if (len(replace_info) > 1 and self.names[i] in replace_info[1:]) \
 *                    or (len(replace_info) == 1 and self.names[i] in self.fill_names):
 */
    __pyx_t_2 = (__pyx_v_replace_info != Py_None);
    __pyx_t_12 = (__pyx_t_2 != 0);
    if (__pyx_t_12) {

      /* "astropy/io/ascii/cparser.pyx":702
 * 
 *             if replace_info is not None:
 *                 if (len(replace_info) > 1 and self.names[i] in replace_info[1:]) \             # <<<<<<<<<<<<<<
 *                    or (len(replace_info) == 1 and self.names[i] in self.fill_names):
 *                     mask.add(row)
 */
      __pyx_t_13 = PyObject_Length(__pyx_v_replace_info); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 702, __pyx_L1_error)
      __pyx_t_2 = ((__pyx_t_13 > 1) != 0);
      if (!__pyx_t_2) {
        goto __pyx_L15_next_or;
      } else {
      }
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_self->names, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 702, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = __Pyx_PyObject_GetSlice(__pyx_v_replace_info, 1, 0, NULL, NULL, &__pyx_slice__19, 1, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 702, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_3, __pyx_t_7, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 702, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_11 = (__pyx_t_2 != 0);
      if (!__pyx_t_11) {
      } else {
        __pyx_t_12 = __pyx_t_11;
        goto __pyx_L14_bool_binop_done;
      }
      __pyx_L15_next_or:;

      /* "astropy/io/ascii/cparser.pyx":703
 *             if replace_info is not None:
 *                 if (len(replace_info) > 1 and self.names[i] in replace_info[1:]) \
 *                    or (len(replace_info) == 1 and self.names[i] in self.fill_names):             # <<<<<<<<<<<<<<
 *                     mask.add(row)
 *                     new_value = str(replace_info[0]).encode('ascii')
 */
      __pyx_t_13 = PyObject_Length(__pyx_v_replace_info); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 703, __pyx_L1_error)
      __pyx_t_11 = ((__pyx_t_13 == 1) != 0);
      if (__pyx_t_11) {
      } else {
        __pyx_t_12 = __pyx_t_11;
        goto __pyx_L14_bool_binop_done;
      }
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_self->names, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 703, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_11 = (__Pyx_PySequence_ContainsTF(__pyx_t_7, __pyx_v_self->fill_names, Py_EQ)); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 703, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_2 = (__pyx_t_11 != 0);
      __pyx_t_12 = __pyx_t_2;
      __pyx_L14_bool_binop_done:;

      /* "astropy/io/ascii/cparser.pyx":702
 * 
 *             if replace_info is not None:
 *                 if (len(replace_info) > 1 and self.names[i] in replace_info[1:]) \             # <<<<<<<<<<<<<<
 *                    or (len(replace_info) == 1 and self.names[i] in self.fill_names):
 *                     mask.add(row)
 */
      if (__pyx_t_12) {

        /* "astropy/io/ascii/cparser.pyx":704
 *                 if (len(replace_info) > 1 and self.names[i] in replace_info[1:]) \
 *                    or (len(replace_info) == 1 and self.names[i] in self.fill_names):
 *                     mask.add(row)             # <<<<<<<<<<<<<<
 *                     new_value = str(replace_info[0]).encode('ascii')
 *                     replacing = True
 */
        __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_row); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 704, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_14 = PySet_Add(__pyx_v_mask, __pyx_t_7); if (unlikely(__pyx_t_14 == ((int)-1))) __PYX_ERR(0, 704, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "astropy/io/ascii/cparser.pyx":705
 *                    or (len(replace_info) == 1 and self.names[i] in self.fill_names):
 *                     mask.add(row)
 *                     new_value = str(replace_info[0]).encode('ascii')             # <<<<<<<<<<<<<<
 *                     replacing = True
 *                     converted = str_to_double(t, new_value)
 */
        __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_replace_info, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 705, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_5 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 705, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_encode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 705, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_7 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_n_s_ascii) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_n_s_ascii);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 705, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(PyBytes_CheckExact(__pyx_t_7))||((__pyx_t_7) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_7)->tp_name), 0))) __PYX_ERR(0, 705, __pyx_L1_error)
        __Pyx_XDECREF_SET(__pyx_v_new_value, ((PyObject*)__pyx_t_7));
        __pyx_t_7 = 0;

        /* "astropy/io/ascii/cparser.pyx":706
 *                     mask.add(row)
 *                     new_value = str(replace_info[0]).encode('ascii')
 *                     replacing = True             # <<<<<<<<<<<<<<
 *                     converted = str_to_double(t, new_value)
 *                 else:
 */
        __pyx_v_replacing = 1;

        /* "astropy/io/ascii/cparser.pyx":707
 *                     new_value = str(replace_info[0]).encode('ascii')
 *                     replacing = True
 *                     converted = str_to_double(t, new_value)             # <<<<<<<<<<<<<<
 *                 else:
 *                     converted = str_to_double(t, field)
 */
        if (unlikely(__pyx_v_new_value == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
          __PYX_ERR(0, 707, __pyx_L1_error)
        }
        __pyx_t_8 = __Pyx_PyBytes_AsWritableString(__pyx_v_new_value); if (unlikely((!__pyx_t_8) && PyErr_Occurred())) __PYX_ERR(0, 707, __pyx_L1_error)
        __pyx_v_converted = str_to_double(__pyx_v_t, __pyx_t_8);

        /* "astropy/io/ascii/cparser.pyx":702
 * 
 *             if replace_info is not None:
 *                 if (len(replace_info) > 1 and self.names[i] in replace_info[1:]) \             # <<<<<<<<<<<<<<
 *                    or (len(replace_info) == 1 and self.names[i] in self.fill_names):
 *                     mask.add(row)
 */
        goto __pyx_L13;
      }

      /* "astropy/io/ascii/cparser.pyx":709
 *                     converted = str_to_double(t, new_value)
 *                 else:
 *                     converted = str_to_double(t, field)             # <<<<<<<<<<<<<<
 *             else:
 *                 converted = str_to_double(t, field)
 */
      /*else*/ {
        __pyx_v_converted = str_to_double(__pyx_v_t, __pyx_v_field);
      }
      __pyx_L13:;

      /* "astropy/io/ascii/cparser.pyx":701
 *                 replace_info = self.fill_values[field]
 * 
 *             if replace_info is not None:             # <<<<<<<<<<<<<<
 *                 if (len(replace_info) > 1 and self.names[i] in replace_info[1:]) \
 *                    or (len(replace_info) == 1 and self.names[i] in self.fill_names):
 */
      goto __pyx_L12;
    }

    /* "astropy/io/ascii/cparser.pyx":711
 *                     converted = str_to_double(t, field)
 *             else:
 *                 converted = str_to_double(t, field)             # <<<<<<<<<<<<<<
 * 
 *             if t.code == CONVERSION_ERROR:
 */
    /*else*/ {
      __pyx_v_converted = str_to_double(__pyx_v_t, __pyx_v_field);
    }
    __pyx_L12:;

    /* "astropy/io/ascii/cparser.pyx":713
 *                 converted = str_to_double(t, field)
 * 
 *             if t.code == CONVERSION_ERROR:             # <<<<<<<<<<<<<<
 *                 t.code = NO_ERROR
 *                 raise ValueError()
 */
    __pyx_t_12 = ((__pyx_v_t->code == CONVERSION_ERROR) != 0);
    if (unlikely(__pyx_t_12)) {

      /* "astropy/io/ascii/cparser.pyx":714
 * 
 *             if t.code == CONVERSION_ERROR:
 *                 t.code = NO_ERROR             # <<<<<<<<<<<<<<
 *                 raise ValueError()
 *             else:
 */
      __pyx_v_t->code = NO_ERROR;

      /* "astropy/io/ascii/cparser.pyx":715
 *             if t.code == CONVERSION_ERROR:
 *                 t.code = NO_ERROR
 *                 raise ValueError()             # <<<<<<<<<<<<<<
 *             else:
 *                 data[row] = converted
 */
      __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_builtin_ValueError); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 715, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_Raise(__pyx_t_7, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __PYX_ERR(0, 715, __pyx_L1_error)

      /* "astropy/io/ascii/cparser.pyx":713
 *                 converted = str_to_double(t, field)
 * 
 *             if t.code == CONVERSION_ERROR:             # <<<<<<<<<<<<<<
 *                 t.code = NO_ERROR
 *                 raise ValueError()
 */
    }

    /* "astropy/io/ascii/cparser.pyx":717
 *                 raise ValueError()
 *             else:
 *                 data[row] = converted             # <<<<<<<<<<<<<<
 *             if t.code == OVERFLOW_ERROR:
 *                 t.code = NO_ERROR
 */
    /*else*/ {
      (__pyx_v_data[__pyx_v_row]) = __pyx_v_converted;
    }

    /* "astropy/io/ascii/cparser.pyx":718
 *             else:
 *                 data[row] = converted
 *             if t.code == OVERFLOW_ERROR:             # <<<<<<<<<<<<<<
 *                 t.code = NO_ERROR
 *                 overflown = OVERFLOW_ERROR
 */
    __pyx_t_12 = ((__pyx_v_t->code == OVERFLOW_ERROR) != 0);
    if (__pyx_t_12) {

      /* "astropy/io/ascii/cparser.pyx":719
 *                 data[row] = converted
 *             if t.code == OVERFLOW_ERROR:
 *                 t.code = NO_ERROR             # <<<<<<<<<<<<<<
 *                 overflown = OVERFLOW_ERROR
 *             row += 1
 */
      __pyx_v_t->code = NO_ERROR;

      /* "astropy/io/ascii/cparser.pyx":720
 *             if t.code == OVERFLOW_ERROR:
 *                 t.code = NO_ERROR
 *                 overflown = OVERFLOW_ERROR             # <<<<<<<<<<<<<<
 *             row += 1
 *         t.code = overflown
 */
      __pyx_v_overflown = OVERFLOW_ERROR;

      /* "astropy/io/ascii/cparser.pyx":718
 *             else:
 *                 data[row] = converted
 *             if t.code == OVERFLOW_ERROR:             # <<<<<<<<<<<<<<
 *                 t.code = NO_ERROR
 *                 overflown = OVERFLOW_ERROR
 */
    }

    /* "astropy/io/ascii/cparser.pyx":721
 *                 t.code = NO_ERROR
 *                 overflown = OVERFLOW_ERROR
 *             row += 1             # <<<<<<<<<<<<<<
 *         t.code = overflown
 * 
 */
    __pyx_v_row = (__pyx_v_row + 1);
  }

  /* "astropy/io/ascii/cparser.pyx":722
 *                 overflown = OVERFLOW_ERROR
 *             row += 1
 *         t.code = overflown             # <<<<<<<<<<<<<<
 * 
 *         if mask:
 */
  __pyx_v_t->code = __pyx_v_overflown;

  /* "astropy/io/ascii/cparser.pyx":724
 *         t.code = overflown
 * 
 *         if mask:             # <<<<<<<<<<<<<<
 *             return ma.masked_array(col, mask=[1 if i in mask else 0 for i in
 *                                               range(row)])
 */
  __pyx_t_12 = (PySet_GET_SIZE(__pyx_v_mask) != 0);
  if (__pyx_t_12) {

    /* "astropy/io/ascii/cparser.pyx":725
 * 
 *         if mask:
 *             return ma.masked_array(col, mask=[1 if i in mask else 0 for i in             # <<<<<<<<<<<<<<
 *                                               range(row)])
 *         else:
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_ma); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 725, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_masked_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 725, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 725, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(((PyObject *)__pyx_v_col));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_col));
    PyTuple_SET_ITEM(__pyx_t_7, 0, ((PyObject *)__pyx_v_col));
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 725, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 725, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "astropy/io/ascii/cparser.pyx":726
 *         if mask:
 *             return ma.masked_array(col, mask=[1 if i in mask else 0 for i in
 *                                               range(row)])             # <<<<<<<<<<<<<<
 *         else:
 *             return col
 */
    __pyx_t_1 = __pyx_v_row;

    /* "astropy/io/ascii/cparser.pyx":725
 * 
 *         if mask:
 *             return ma.masked_array(col, mask=[1 if i in mask else 0 for i in             # <<<<<<<<<<<<<<
 *                                               range(row)])
 *         else:
 */
    __pyx_t_9 = __pyx_t_1;
    for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
      __pyx_v_i = __pyx_t_10;
      __pyx_t_15 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 725, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __pyx_t_12 = (__Pyx_PySet_ContainsTF(__pyx_t_15, __pyx_v_mask, Py_EQ)); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 725, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      if ((__pyx_t_12 != 0)) {
        __Pyx_INCREF(__pyx_int_1);
        __pyx_t_6 = __pyx_int_1;
      } else {
        __Pyx_INCREF(__pyx_int_0);
        __pyx_t_6 = __pyx_int_0;
      }
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 725, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_mask, __pyx_t_4) < 0) __PYX_ERR(0, 725, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 725, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 725, __pyx_L1_error)
    __pyx_r = ((PyArrayObject *)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "astropy/io/ascii/cparser.pyx":724
 *         t.code = overflown
 * 
 *         if mask:             # <<<<<<<<<<<<<<
 *             return ma.masked_array(col, mask=[1 if i in mask else 0 for i in
 *                                               range(row)])
 */
  }

  /* "astropy/io/ascii/cparser.pyx":728
 *                                               range(row)])
 *         else:
 *             return col             # <<<<<<<<<<<<<<
 * 
 *     cdef _convert_str(self, tokenizer_t *t, int i, int nrows):
 */
  /*else*/ {
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_INCREF(((PyObject *)__pyx_v_col));
    __pyx_r = __pyx_v_col;
    goto __pyx_L0;
  }

  /* "astropy/io/ascii/cparser.pyx":672
 *             return col
 * 
 *     cdef np.ndarray _convert_float(self, tokenizer_t *t, int i, int nrows):             # <<<<<<<<<<<<<<
 *         # very similar to _convert_int()
 *         cdef int num_rows = t.num_rows
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("astropy.io.ascii.cparser.CParser._convert_float", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_col);
  __Pyx_XDECREF(__pyx_v_new_value);
  __Pyx_XDECREF(__pyx_v_mask);
  __Pyx_XDECREF(__pyx_v_replace_info);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "astropy/io/ascii/cparser.pyx":730
 *             return col
 * 
 *     cdef _convert_str(self, tokenizer_t *t, int i, int nrows):             # <<<<<<<<<<<<<<
 *         # similar to _convert_int, but no actual conversion
 *         cdef int num_rows = t.num_rows
 */

static PyObject *__pyx_f_7astropy_2io_5ascii_7cparser_7CParser__convert_str(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self, tokenizer_t *__pyx_v_t, int __pyx_v_i, int __pyx_v_nrows) {
  int __pyx_v_num_rows;
  int __pyx_v_row;
  PyObject *__pyx_v_field = 0;
  int __pyx_v_field_len;
  int __pyx_v_max_len;
  PyObject *__pyx_v_fields_list = 0;
  PyObject *__pyx_v_mask = NULL;
  PyObject *__pyx_v_replace_info = NULL;
  PyObject *__pyx_v_el = NULL;
  PyArrayObject *__pyx_v_col = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  __Pyx_RefNannySetupContext("_convert_str", 0);

  /* "astropy/io/ascii/cparser.pyx":732
 *     cdef _convert_str(self, tokenizer_t *t, int i, int nrows):
 *         # similar to _convert_int, but no actual conversion
 *         cdef int num_rows = t.num_rows             # <<<<<<<<<<<<<<
 *         if nrows != -1:
 *             num_rows = nrows
 */
  __pyx_t_1 = __pyx_v_t->num_rows;
  __pyx_v_num_rows = __pyx_t_1;

  /* "astropy/io/ascii/cparser.pyx":733
 *         # similar to _convert_int, but no actual conversion
 *         cdef int num_rows = t.num_rows
 *         if nrows != -1:             # <<<<<<<<<<<<<<
 *             num_rows = nrows
 * 
 */
  __pyx_t_2 = ((__pyx_v_nrows != -1L) != 0);
  if (__pyx_t_2) {

    /* "astropy/io/ascii/cparser.pyx":734
 *         cdef int num_rows = t.num_rows
 *         if nrows != -1:
 *             num_rows = nrows             # <<<<<<<<<<<<<<
 * 
 *         cdef int row = 0
 */
    __pyx_v_num_rows = __pyx_v_nrows;

    /* "astropy/io/ascii/cparser.pyx":733
 *         # similar to _convert_int, but no actual conversion
 *         cdef int num_rows = t.num_rows
 *         if nrows != -1:             # <<<<<<<<<<<<<<
 *             num_rows = nrows
 * 
 */
  }

  /* "astropy/io/ascii/cparser.pyx":736
 *             num_rows = nrows
 * 
 *         cdef int row = 0             # <<<<<<<<<<<<<<
 *         cdef bytes field
 *         cdef int field_len
 */
  __pyx_v_row = 0;

  /* "astropy/io/ascii/cparser.pyx":739
 *         cdef bytes field
 *         cdef int field_len
 *         cdef int max_len = 0             # <<<<<<<<<<<<<<
 *         cdef list fields_list = []
 *         mask = set()
 */
  __pyx_v_max_len = 0;

  /* "astropy/io/ascii/cparser.pyx":740
 *         cdef int field_len
 *         cdef int max_len = 0
 *         cdef list fields_list = []             # <<<<<<<<<<<<<<
 *         mask = set()
 * 
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 740, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_fields_list = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "astropy/io/ascii/cparser.pyx":741
 *         cdef int max_len = 0
 *         cdef list fields_list = []
 *         mask = set()             # <<<<<<<<<<<<<<
 * 
 *         start_iteration(t, i)
 */
  __pyx_t_3 = PySet_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 741, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_mask = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "astropy/io/ascii/cparser.pyx":743
 *         mask = set()
 * 
 *         start_iteration(t, i)             # <<<<<<<<<<<<<<
 *         for row in range(num_rows):
 *             field = next_field(t, &field_len)
 */
  start_iteration(__pyx_v_t, __pyx_v_i);

  /* "astropy/io/ascii/cparser.pyx":744
 * 
 *         start_iteration(t, i)
 *         for row in range(num_rows):             # <<<<<<<<<<<<<<
 *             field = next_field(t, &field_len)
 *             replace_info = None
 */
  __pyx_t_1 = __pyx_v_num_rows;
  __pyx_t_4 = __pyx_t_1;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_row = __pyx_t_5;

    /* "astropy/io/ascii/cparser.pyx":745
 *         start_iteration(t, i)
 *         for row in range(num_rows):
 *             field = next_field(t, &field_len)             # <<<<<<<<<<<<<<
 *             replace_info = None
 * 
 */
    __pyx_t_3 = __Pyx_PyBytes_FromString(next_field(__pyx_v_t, (&__pyx_v_field_len))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 745, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_field, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "astropy/io/ascii/cparser.pyx":746
 *         for row in range(num_rows):
 *             field = next_field(t, &field_len)
 *             replace_info = None             # <<<<<<<<<<<<<<
 * 
 *             if field_len == 0 and self.fill_empty:
 */
    __Pyx_INCREF(Py_None);
    __Pyx_XDECREF_SET(__pyx_v_replace_info, Py_None);

    /* "astropy/io/ascii/cparser.pyx":748
 *             replace_info = None
 * 
 *             if field_len == 0 and self.fill_empty:             # <<<<<<<<<<<<<<
 *                 replace_info = self.fill_empty
 * 
 */
    __pyx_t_6 = ((__pyx_v_field_len == 0) != 0);
    if (__pyx_t_6) {
    } else {
      __pyx_t_2 = __pyx_t_6;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_self->fill_empty); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 748, __pyx_L1_error)
    __pyx_t_2 = __pyx_t_6;
    __pyx_L7_bool_binop_done:;
    if (__pyx_t_2) {

      /* "astropy/io/ascii/cparser.pyx":749
 * 
 *             if field_len == 0 and self.fill_empty:
 *                 replace_info = self.fill_empty             # <<<<<<<<<<<<<<
 * 
 *             elif field_len > 0 and self.fill_values and field in self.fill_values:
 */
      __pyx_t_3 = __pyx_v_self->fill_empty;
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_DECREF_SET(__pyx_v_replace_info, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "astropy/io/ascii/cparser.pyx":748
 *             replace_info = None
 * 
 *             if field_len == 0 and self.fill_empty:             # <<<<<<<<<<<<<<
 *                 replace_info = self.fill_empty
 * 
 */
      goto __pyx_L6;
    }

    /* "astropy/io/ascii/cparser.pyx":751
 *                 replace_info = self.fill_empty
 * 
 *             elif field_len > 0 and self.fill_values and field in self.fill_values:             # <<<<<<<<<<<<<<
 *                 replace_info = self.fill_values[field]
 * 
 */
    __pyx_t_6 = ((__pyx_v_field_len > 0) != 0);
    if (__pyx_t_6) {
    } else {
      __pyx_t_2 = __pyx_t_6;
      goto __pyx_L9_bool_binop_done;
    }
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_self->fill_values); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 751, __pyx_L1_error)
    if (__pyx_t_6) {
    } else {
      __pyx_t_2 = __pyx_t_6;
      goto __pyx_L9_bool_binop_done;
    }
    __pyx_t_6 = (__Pyx_PySequence_ContainsTF(__pyx_v_field, __pyx_v_self->fill_values, Py_EQ)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 751, __pyx_L1_error)
    __pyx_t_7 = (__pyx_t_6 != 0);
    __pyx_t_2 = __pyx_t_7;
    __pyx_L9_bool_binop_done:;
    if (__pyx_t_2) {

      /* "astropy/io/ascii/cparser.pyx":752
 * 
 *             elif field_len > 0 and self.fill_values and field in self.fill_values:
 *                 replace_info = self.fill_values[field]             # <<<<<<<<<<<<<<
 * 
 *             if replace_info is not None:
 */
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_self->fill_values, __pyx_v_field); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 752, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF_SET(__pyx_v_replace_info, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "astropy/io/ascii/cparser.pyx":751
 *                 replace_info = self.fill_empty
 * 
 *             elif field_len > 0 and self.fill_values and field in self.fill_values:             # <<<<<<<<<<<<<<
 *                 replace_info = self.fill_values[field]
 * 
 */
    }
    __pyx_L6:;

    /* "astropy/io/ascii/cparser.pyx":754
 *                 replace_info = self.fill_values[field]
 * 
 *             if replace_info is not None:             # <<<<<<<<<<<<<<
 *                 el = replace_info[0].encode('ascii')
 *                 if (len(replace_info) > 1 and self.names[i] in replace_info[1:]) \
 */
    __pyx_t_2 = (__pyx_v_replace_info != Py_None);
    __pyx_t_7 = (__pyx_t_2 != 0);
    if (__pyx_t_7) {

      /* "astropy/io/ascii/cparser.pyx":755
 * 
 *             if replace_info is not None:
 *                 el = replace_info[0].encode('ascii')             # <<<<<<<<<<<<<<
 *                 if (len(replace_info) > 1 and self.names[i] in replace_info[1:]) \
 *                    or (len(replace_info) == 1 and self.names[i] in self.fill_names):
 */
      __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_replace_info, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 755, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_encode); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 755, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
        }
      }
      __pyx_t_3 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_8, __pyx_n_s_ascii) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_n_s_ascii);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 755, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF_SET(__pyx_v_el, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "astropy/io/ascii/cparser.pyx":756
 *             if replace_info is not None:
 *                 el = replace_info[0].encode('ascii')
 *                 if (len(replace_info) > 1 and self.names[i] in replace_info[1:]) \             # <<<<<<<<<<<<<<
 *                    or (len(replace_info) == 1 and self.names[i] in self.fill_names):
 *                     mask.add(row)
 */
      __pyx_t_10 = PyObject_Length(__pyx_v_replace_info); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 756, __pyx_L1_error)
      __pyx_t_2 = ((__pyx_t_10 > 1) != 0);
      if (!__pyx_t_2) {
        goto __pyx_L15_next_or;
      } else {
      }
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_self->names, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 756, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_9 = __Pyx_PyObject_GetSlice(__pyx_v_replace_info, 1, 0, NULL, NULL, &__pyx_slice__19, 1, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 756, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_3, __pyx_t_9, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 756, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_6 = (__pyx_t_2 != 0);
      if (!__pyx_t_6) {
      } else {
        __pyx_t_7 = __pyx_t_6;
        goto __pyx_L14_bool_binop_done;
      }
      __pyx_L15_next_or:;

      /* "astropy/io/ascii/cparser.pyx":757
 *                 el = replace_info[0].encode('ascii')
 *                 if (len(replace_info) > 1 and self.names[i] in replace_info[1:]) \
 *                    or (len(replace_info) == 1 and self.names[i] in self.fill_names):             # <<<<<<<<<<<<<<
 *                     mask.add(row)
 *                     field = el
 */
      __pyx_t_10 = PyObject_Length(__pyx_v_replace_info); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 757, __pyx_L1_error)
      __pyx_t_6 = ((__pyx_t_10 == 1) != 0);
      if (__pyx_t_6) {
      } else {
        __pyx_t_7 = __pyx_t_6;
        goto __pyx_L14_bool_binop_done;
      }
      __pyx_t_9 = __Pyx_GetItemInt(__pyx_v_self->names, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 757, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_6 = (__Pyx_PySequence_ContainsTF(__pyx_t_9, __pyx_v_self->fill_names, Py_EQ)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 757, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_2 = (__pyx_t_6 != 0);
      __pyx_t_7 = __pyx_t_2;
      __pyx_L14_bool_binop_done:;

      /* "astropy/io/ascii/cparser.pyx":756
 *             if replace_info is not None:
 *                 el = replace_info[0].encode('ascii')
 *                 if (len(replace_info) > 1 and self.names[i] in replace_info[1:]) \             # <<<<<<<<<<<<<<
 *                    or (len(replace_info) == 1 and self.names[i] in self.fill_names):
 *                     mask.add(row)
 */
      if (__pyx_t_7) {

        /* "astropy/io/ascii/cparser.pyx":758
 *                 if (len(replace_info) > 1 and self.names[i] in replace_info[1:]) \
 *                    or (len(replace_info) == 1 and self.names[i] in self.fill_names):
 *                     mask.add(row)             # <<<<<<<<<<<<<<
 *                     field = el
 * 
 */
        __pyx_t_9 = __Pyx_PyInt_From_int(__pyx_v_row); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 758, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_11 = PySet_Add(__pyx_v_mask, __pyx_t_9); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 758, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "astropy/io/ascii/cparser.pyx":759
 *                    or (len(replace_info) == 1 and self.names[i] in self.fill_names):
 *                     mask.add(row)
 *                     field = el             # <<<<<<<<<<<<<<
 * 
 *             fields_list.append(field)
 */
        if (!(likely(PyBytes_CheckExact(__pyx_v_el))||((__pyx_v_el) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_v_el)->tp_name), 0))) __PYX_ERR(0, 759, __pyx_L1_error)
        __pyx_t_9 = __pyx_v_el;
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_DECREF_SET(__pyx_v_field, ((PyObject*)__pyx_t_9));
        __pyx_t_9 = 0;

        /* "astropy/io/ascii/cparser.pyx":756
 *             if replace_info is not None:
 *                 el = replace_info[0].encode('ascii')
 *                 if (len(replace_info) > 1 and self.names[i] in replace_info[1:]) \             # <<<<<<<<<<<<<<
 *                    or (len(replace_info) == 1 and self.names[i] in self.fill_names):
 *                     mask.add(row)
 */
      }

      /* "astropy/io/ascii/cparser.pyx":754
 *                 replace_info = self.fill_values[field]
 * 
 *             if replace_info is not None:             # <<<<<<<<<<<<<<
 *                 el = replace_info[0].encode('ascii')
 *                 if (len(replace_info) > 1 and self.names[i] in replace_info[1:]) \
 */
    }

    /* "astropy/io/ascii/cparser.pyx":761
 *                     field = el
 * 
 *             fields_list.append(field)             # <<<<<<<<<<<<<<
 *             if field_len > max_len:
 *                 max_len = field_len
 */
    __pyx_t_11 = __Pyx_PyList_Append(__pyx_v_fields_list, __pyx_v_field); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 761, __pyx_L1_error)

    /* "astropy/io/ascii/cparser.pyx":762
 * 
 *             fields_list.append(field)
 *             if field_len > max_len:             # <<<<<<<<<<<<<<
 *                 max_len = field_len
 *             row += 1
 */
    __pyx_t_7 = ((__pyx_v_field_len > __pyx_v_max_len) != 0);
    if (__pyx_t_7) {

      /* "astropy/io/ascii/cparser.pyx":763
 *             fields_list.append(field)
 *             if field_len > max_len:
 *                 max_len = field_len             # <<<<<<<<<<<<<<
 *             row += 1
 * 
 */
      __pyx_v_max_len = __pyx_v_field_len;

      /* "astropy/io/ascii/cparser.pyx":762
 * 
 *             fields_list.append(field)
 *             if field_len > max_len:             # <<<<<<<<<<<<<<
 *                 max_len = field_len
 *             row += 1
 */
    }

    /* "astropy/io/ascii/cparser.pyx":764
 *             if field_len > max_len:
 *                 max_len = field_len
 *             row += 1             # <<<<<<<<<<<<<<
 * 
 *         cdef np.ndarray col = np.array(fields_list, dtype=(np.str, max_len))
 */
    __pyx_v_row = (__pyx_v_row + 1);
  }

  /* "astropy/io/ascii/cparser.pyx":766
 *             row += 1
 * 
 *         cdef np.ndarray col = np.array(fields_list, dtype=(np.str, max_len))             # <<<<<<<<<<<<<<
 * 
 *         if mask:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_np); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(__pyx_v_fields_list);
  __Pyx_GIVEREF(__pyx_v_fields_list);
  PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v_fields_list);
  __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_np); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_str); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_12 = __Pyx_PyInt_From_int(__pyx_v_max_len); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_12);
  __pyx_t_13 = 0;
  __pyx_t_12 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_dtype, __pyx_t_14) < 0) __PYX_ERR(0, 766, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_9, __pyx_t_8); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (!(likely(((__pyx_t_14) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_14, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 766, __pyx_L1_error)
  __pyx_v_col = ((PyArrayObject *)__pyx_t_14);
  __pyx_t_14 = 0;

  /* "astropy/io/ascii/cparser.pyx":768
 *         cdef np.ndarray col = np.array(fields_list, dtype=(np.str, max_len))
 * 
 *         if mask:             # <<<<<<<<<<<<<<
 *             return ma.masked_array(col, mask=[1 if i in mask else 0 for i in
 *                                               range(row)])
 */
  __pyx_t_7 = (PySet_GET_SIZE(__pyx_v_mask) != 0);
  if (__pyx_t_7) {

    /* "astropy/io/ascii/cparser.pyx":769
 * 
 *         if mask:
 *             return ma.masked_array(col, mask=[1 if i in mask else 0 for i in             # <<<<<<<<<<<<<<
 *                                               range(row)])
 *         else:
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_ma); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 769, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_masked_array); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 769, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_14 = PyTuple_New(1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 769, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_INCREF(((PyObject *)__pyx_v_col));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_col));
    PyTuple_SET_ITEM(__pyx_t_14, 0, ((PyObject *)__pyx_v_col));
    __pyx_t_9 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 769, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 769, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "astropy/io/ascii/cparser.pyx":770
 *         if mask:
 *             return ma.masked_array(col, mask=[1 if i in mask else 0 for i in
 *                                               range(row)])             # <<<<<<<<<<<<<<
 *         else:
 *             return col
 */
    __pyx_t_1 = __pyx_v_row;

    /* "astropy/io/ascii/cparser.pyx":769
 * 
 *         if mask:
 *             return ma.masked_array(col, mask=[1 if i in mask else 0 for i in             # <<<<<<<<<<<<<<
 *                                               range(row)])
 *         else:
 */
    __pyx_t_4 = __pyx_t_1;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;
      __pyx_t_13 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 769, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_7 = (__Pyx_PySet_ContainsTF(__pyx_t_13, __pyx_v_mask, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 769, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      if ((__pyx_t_7 != 0)) {
        __Pyx_INCREF(__pyx_int_1);
        __pyx_t_12 = __pyx_int_1;
      } else {
        __Pyx_INCREF(__pyx_int_0);
        __pyx_t_12 = __pyx_int_0;
      }
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_12))) __PYX_ERR(0, 769, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    }
    if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_mask, __pyx_t_3) < 0) __PYX_ERR(0, 769, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_14, __pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 769, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "astropy/io/ascii/cparser.pyx":768
 *         cdef np.ndarray col = np.array(fields_list, dtype=(np.str, max_len))
 * 
 *         if mask:             # <<<<<<<<<<<<<<
 *             return ma.masked_array(col, mask=[1 if i in mask else 0 for i in
 *                                               range(row)])
 */
  }

  /* "astropy/io/ascii/cparser.pyx":772
 *                                               range(row)])
 *         else:
 *             return col             # <<<<<<<<<<<<<<
 * 
 *     def get_names(self):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_col));
    __pyx_r = ((PyObject *)__pyx_v_col);
    goto __pyx_L0;
  }

  /* "astropy/io/ascii/cparser.pyx":730
 *             return col
 * 
 *     cdef _convert_str(self, tokenizer_t *t, int i, int nrows):             # <<<<<<<<<<<<<<
 *         # similar to _convert_int, but no actual conversion
 *         cdef int num_rows = t.num_rows
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("astropy.io.ascii.cparser.CParser._convert_str", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_field);
  __Pyx_XDECREF(__pyx_v_fields_list);
  __Pyx_XDECREF(__pyx_v_mask);
  __Pyx_XDECREF(__pyx_v_replace_info);
  __Pyx_XDECREF(__pyx_v_el);
  __Pyx_XDECREF((PyObject *)__pyx_v_col);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "astropy/io/ascii/cparser.pyx":774
 *             return col
 * 
 *     def get_names(self):             # <<<<<<<<<<<<<<
 *         # ignore excluded columns
 *         return [name for name in self.names if name in self.use_cols]
 */

/* Python wrapper */
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_13get_names(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_13get_names(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_names (wrapper)", 0);
  __pyx_r = __pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_12get_names(((struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_12get_names(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self) {
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("get_names", 0);

  /* "astropy/io/ascii/cparser.pyx":776
 *     def get_names(self):
 *         # ignore excluded columns
 *         return [name for name in self.names if name in self.use_cols]             # <<<<<<<<<<<<<<
 * 
 *     def set_names(self, names):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_v_self->names)) || PyTuple_CheckExact(__pyx_v_self->names)) {
    __pyx_t_2 = __pyx_v_self->names; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_self->names); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 776, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 776, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 776, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 776, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 776, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 776, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 776, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_5);
    __pyx_t_5 = 0;
    if (unlikely(__pyx_v_self->use_cols == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 776, __pyx_L1_error)
    }
    __pyx_t_6 = (__Pyx_PySet_ContainsTF(__pyx_v_name, __pyx_v_self->use_cols, Py_EQ)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 776, __pyx_L1_error)
    __pyx_t_7 = (__pyx_t_6 != 0);
    if (__pyx_t_7) {
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_v_name))) __PYX_ERR(0, 776, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "astropy/io/ascii/cparser.pyx":774
 *             return col
 * 
 *     def get_names(self):             # <<<<<<<<<<<<<<
 *         # ignore excluded columns
 *         return [name for name in self.names if name in self.use_cols]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("astropy.io.ascii.cparser.CParser.get_names", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "astropy/io/ascii/cparser.pyx":778
 *         return [name for name in self.names if name in self.use_cols]
 * 
 *     def set_names(self, names):             # <<<<<<<<<<<<<<
 *         self.names = names
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_15set_names(PyObject *__pyx_v_self, PyObject *__pyx_v_names); /*proto*/
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_15set_names(PyObject *__pyx_v_self, PyObject *__pyx_v_names) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_names (wrapper)", 0);
  __pyx_r = __pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_14set_names(((struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *)__pyx_v_self), ((PyObject *)__pyx_v_names));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_14set_names(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self, PyObject *__pyx_v_names) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_names", 0);

  /* "astropy/io/ascii/cparser.pyx":779
 * 
 *     def set_names(self, names):
 *         self.names = names             # <<<<<<<<<<<<<<
 * 
 *     def get_header_names(self):
 */
  __Pyx_INCREF(__pyx_v_names);
  __Pyx_GIVEREF(__pyx_v_names);
  __Pyx_GOTREF(__pyx_v_self->names);
  __Pyx_DECREF(__pyx_v_self->names);
  __pyx_v_self->names = __pyx_v_names;

  /* "astropy/io/ascii/cparser.pyx":778
 *         return [name for name in self.names if name in self.use_cols]
 * 
 *     def set_names(self, names):             # <<<<<<<<<<<<<<
 *         self.names = names
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "astropy/io/ascii/cparser.pyx":781
 *         self.names = names
 * 
 *     def get_header_names(self):             # <<<<<<<<<<<<<<
 *         return self.header_names
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_17get_header_names(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_17get_header_names(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_header_names (wrapper)", 0);
  __pyx_r = __pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_16get_header_names(((struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_16get_header_names(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_header_names", 0);

  /* "astropy/io/ascii/cparser.pyx":782
 * 
 *     def get_header_names(self):
 *         return self.header_names             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->header_names);
  __pyx_r = __pyx_v_self->header_names;
  goto __pyx_L0;

  /* "astropy/io/ascii/cparser.pyx":781
 *         self.names = names
 * 
 *     def get_header_names(self):             # <<<<<<<<<<<<<<
 *         return self.header_names
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "astropy/io/ascii/cparser.pyx":784
 *         return self.header_names
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         cdef bytes source_ptr = self.source_ptr if self.source_ptr else b''
 *         return (_copy_cparser, (source_ptr, self.source_bytes, self.use_cols, self.fill_names,
 */

/* Python wrapper */
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_19__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_19__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_18__reduce__(((struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_18__reduce__(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self) {
  PyObject *__pyx_v_source_ptr = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "astropy/io/ascii/cparser.pyx":785
 * 
 *     def __reduce__(self):
 *         cdef bytes source_ptr = self.source_ptr if self.source_ptr else b''             # <<<<<<<<<<<<<<
 *         return (_copy_cparser, (source_ptr, self.source_bytes, self.use_cols, self.fill_names,
 *                                 self.fill_values, self.tokenizer.strip_whitespace_lines,
 */
  if ((__pyx_v_self->source_ptr != 0)) {
    __pyx_t_2 = __Pyx_PyBytes_FromString(__pyx_v_self->source_ptr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 785, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __pyx_t_2;
    __pyx_t_2 = 0;
  } else {
    __Pyx_INCREF(__pyx_kp_b__5);
    __pyx_t_1 = __pyx_kp_b__5;
  }
  __pyx_v_source_ptr = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "astropy/io/ascii/cparser.pyx":786
 *     def __reduce__(self):
 *         cdef bytes source_ptr = self.source_ptr if self.source_ptr else b''
 *         return (_copy_cparser, (source_ptr, self.source_bytes, self.use_cols, self.fill_names,             # <<<<<<<<<<<<<<
 *                                 self.fill_values, self.tokenizer.strip_whitespace_lines,
 *                                 self.tokenizer.strip_whitespace_fields,
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_copy_cparser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "astropy/io/ascii/cparser.pyx":787
 *         cdef bytes source_ptr = self.source_ptr if self.source_ptr else b''
 *         return (_copy_cparser, (source_ptr, self.source_bytes, self.use_cols, self.fill_names,
 *                                 self.fill_values, self.tokenizer.strip_whitespace_lines,             # <<<<<<<<<<<<<<
 *                                 self.tokenizer.strip_whitespace_fields,
 *                                 dict(delimiter=chr(self.tokenizer.delimiter),
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->tokenizer->strip_whitespace_lines); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "astropy/io/ascii/cparser.pyx":788
 *         return (_copy_cparser, (source_ptr, self.source_bytes, self.use_cols, self.fill_names,
 *                                 self.fill_values, self.tokenizer.strip_whitespace_lines,
 *                                 self.tokenizer.strip_whitespace_fields,             # <<<<<<<<<<<<<<
 *                                 dict(delimiter=chr(self.tokenizer.delimiter),
 *                                 comment=chr(self.tokenizer.comment),
 */
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->tokenizer->strip_whitespace_fields); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 788, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "astropy/io/ascii/cparser.pyx":789
 *                                 self.fill_values, self.tokenizer.strip_whitespace_lines,
 *                                 self.tokenizer.strip_whitespace_fields,
 *                                 dict(delimiter=chr(self.tokenizer.delimiter),             # <<<<<<<<<<<<<<
 *                                 comment=chr(self.tokenizer.comment),
 *                                 quotechar=chr(self.tokenizer.quotechar),
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(16); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 789, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_char(__pyx_v_self->tokenizer->delimiter); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 789, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_chr, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 789, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_delimiter, __pyx_t_6) < 0) __PYX_ERR(0, 789, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "astropy/io/ascii/cparser.pyx":790
 *                                 self.tokenizer.strip_whitespace_fields,
 *                                 dict(delimiter=chr(self.tokenizer.delimiter),
 *                                 comment=chr(self.tokenizer.comment),             # <<<<<<<<<<<<<<
 *                                 quotechar=chr(self.tokenizer.quotechar),
 *                                 expchar=chr(self.tokenizer.expchar),
 */
  __pyx_t_6 = __Pyx_PyInt_From_char(__pyx_v_self->tokenizer->comment); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_chr, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_comment, __pyx_t_5) < 0) __PYX_ERR(0, 789, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "astropy/io/ascii/cparser.pyx":791
 *                                 dict(delimiter=chr(self.tokenizer.delimiter),
 *                                 comment=chr(self.tokenizer.comment),
 *                                 quotechar=chr(self.tokenizer.quotechar),             # <<<<<<<<<<<<<<
 *                                 expchar=chr(self.tokenizer.expchar),
 *                                 header_start=self.header_start,
 */
  __pyx_t_5 = __Pyx_PyInt_From_char(__pyx_v_self->tokenizer->quotechar); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_chr, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_quotechar, __pyx_t_6) < 0) __PYX_ERR(0, 789, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "astropy/io/ascii/cparser.pyx":792
 *                                 comment=chr(self.tokenizer.comment),
 *                                 quotechar=chr(self.tokenizer.quotechar),
 *                                 expchar=chr(self.tokenizer.expchar),             # <<<<<<<<<<<<<<
 *                                 header_start=self.header_start,
 *                                 data_start=self.data_start,
 */
  __pyx_t_6 = __Pyx_PyInt_From_char(__pyx_v_self->tokenizer->expchar); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_chr, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_expchar, __pyx_t_5) < 0) __PYX_ERR(0, 789, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "astropy/io/ascii/cparser.pyx":793
 *                                 quotechar=chr(self.tokenizer.quotechar),
 *                                 expchar=chr(self.tokenizer.expchar),
 *                                 header_start=self.header_start,             # <<<<<<<<<<<<<<
 *                                 data_start=self.data_start,
 *                                 data_end=self.data_end,
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_header_start, __pyx_v_self->header_start) < 0) __PYX_ERR(0, 789, __pyx_L1_error)

  /* "astropy/io/ascii/cparser.pyx":794
 *                                 expchar=chr(self.tokenizer.expchar),
 *                                 header_start=self.header_start,
 *                                 data_start=self.data_start,             # <<<<<<<<<<<<<<
 *                                 data_end=self.data_end,
 *                                 names=self.names,
 */
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_self->data_start); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 794, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_data_start, __pyx_t_5) < 0) __PYX_ERR(0, 789, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "astropy/io/ascii/cparser.pyx":795
 *                                 header_start=self.header_start,
 *                                 data_start=self.data_start,
 *                                 data_end=self.data_end,             # <<<<<<<<<<<<<<
 *                                 names=self.names,
 *                                 include_names=self.include_names,
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_data_end, __pyx_v_self->data_end) < 0) __PYX_ERR(0, 789, __pyx_L1_error)

  /* "astropy/io/ascii/cparser.pyx":796
 *                                 data_start=self.data_start,
 *                                 data_end=self.data_end,
 *                                 names=self.names,             # <<<<<<<<<<<<<<
 *                                 include_names=self.include_names,
 *                                 exclude_names=self.exclude_names,
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_names, __pyx_v_self->names) < 0) __PYX_ERR(0, 789, __pyx_L1_error)

  /* "astropy/io/ascii/cparser.pyx":797
 *                                 data_end=self.data_end,
 *                                 names=self.names,
 *                                 include_names=self.include_names,             # <<<<<<<<<<<<<<
 *                                 exclude_names=self.exclude_names,
 *                                 fill_values=None,
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_include_names, __pyx_v_self->include_names) < 0) __PYX_ERR(0, 789, __pyx_L1_error)

  /* "astropy/io/ascii/cparser.pyx":798
 *                                 names=self.names,
 *                                 include_names=self.include_names,
 *                                 exclude_names=self.exclude_names,             # <<<<<<<<<<<<<<
 *                                 fill_values=None,
 *                                 fill_include_names=self.fill_include_names,
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_exclude_names, __pyx_v_self->exclude_names) < 0) __PYX_ERR(0, 789, __pyx_L1_error)

  /* "astropy/io/ascii/cparser.pyx":799
 *                                 include_names=self.include_names,
 *                                 exclude_names=self.exclude_names,
 *                                 fill_values=None,             # <<<<<<<<<<<<<<
 *                                 fill_include_names=self.fill_include_names,
 *                                 fill_exclude_names=self.fill_exclude_names,
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_fill_values, Py_None) < 0) __PYX_ERR(0, 789, __pyx_L1_error)

  /* "astropy/io/ascii/cparser.pyx":800
 *                                 exclude_names=self.exclude_names,
 *                                 fill_values=None,
 *                                 fill_include_names=self.fill_include_names,             # <<<<<<<<<<<<<<
 *                                 fill_exclude_names=self.fill_exclude_names,
 *                                 fill_extra_cols=self.tokenizer.fill_extra_cols,
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_fill_include_names, __pyx_v_self->fill_include_names) < 0) __PYX_ERR(0, 789, __pyx_L1_error)

  /* "astropy/io/ascii/cparser.pyx":801
 *                                 fill_values=None,
 *                                 fill_include_names=self.fill_include_names,
 *                                 fill_exclude_names=self.fill_exclude_names,             # <<<<<<<<<<<<<<
 *                                 fill_extra_cols=self.tokenizer.fill_extra_cols,
 *                                 use_fast_converter=self.tokenizer.use_fast_converter,
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_fill_exclude_names, __pyx_v_self->fill_exclude_names) < 0) __PYX_ERR(0, 789, __pyx_L1_error)

  /* "astropy/io/ascii/cparser.pyx":802
 *                                 fill_include_names=self.fill_include_names,
 *                                 fill_exclude_names=self.fill_exclude_names,
 *                                 fill_extra_cols=self.tokenizer.fill_extra_cols,             # <<<<<<<<<<<<<<
 *                                 use_fast_converter=self.tokenizer.use_fast_converter,
 *                                 parallel=False)))
 */
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_self->tokenizer->fill_extra_cols); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 802, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_fill_extra_cols, __pyx_t_5) < 0) __PYX_ERR(0, 789, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "astropy/io/ascii/cparser.pyx":803
 *                                 fill_exclude_names=self.fill_exclude_names,
 *                                 fill_extra_cols=self.tokenizer.fill_extra_cols,
 *                                 use_fast_converter=self.tokenizer.use_fast_converter,             # <<<<<<<<<<<<<<
 *                                 parallel=False)))
 * 
 */
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_self->tokenizer->use_fast_converter); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 803, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_use_fast_converter, __pyx_t_5) < 0) __PYX_ERR(0, 789, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "astropy/io/ascii/cparser.pyx":804
 *                                 fill_extra_cols=self.tokenizer.fill_extra_cols,
 *                                 use_fast_converter=self.tokenizer.use_fast_converter,
 *                                 parallel=False)))             # <<<<<<<<<<<<<<
 * 
 * def _copy_cparser(bytes src_ptr, bytes source_bytes, use_cols, fill_names, fill_values,
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_parallel, Py_False) < 0) __PYX_ERR(0, 789, __pyx_L1_error)

  /* "astropy/io/ascii/cparser.pyx":786
 *     def __reduce__(self):
 *         cdef bytes source_ptr = self.source_ptr if self.source_ptr else b''
 *         return (_copy_cparser, (source_ptr, self.source_bytes, self.use_cols, self.fill_names,             # <<<<<<<<<<<<<<
 *                                 self.fill_values, self.tokenizer.strip_whitespace_lines,
 *                                 self.tokenizer.strip_whitespace_fields,
 */
  __pyx_t_5 = PyTuple_New(8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_source_ptr);
  __Pyx_GIVEREF(__pyx_v_source_ptr);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_source_ptr);
  __Pyx_INCREF(__pyx_v_self->source_bytes);
  __Pyx_GIVEREF(__pyx_v_self->source_bytes);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_self->source_bytes);
  __Pyx_INCREF(__pyx_v_self->use_cols);
  __Pyx_GIVEREF(__pyx_v_self->use_cols);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_self->use_cols);
  __Pyx_INCREF(__pyx_v_self->fill_names);
  __Pyx_GIVEREF(__pyx_v_self->fill_names);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_v_self->fill_names);
  __Pyx_INCREF(__pyx_v_self->fill_values);
  __Pyx_GIVEREF(__pyx_v_self->fill_values);
  PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_v_self->fill_values);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 5, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 6, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 7, __pyx_t_4);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5);
  __pyx_t_1 = 0;
  __pyx_t_5 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "astropy/io/ascii/cparser.pyx":784
 *         return self.header_names
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         cdef bytes source_ptr = self.source_ptr if self.source_ptr else b''
 *         return (_copy_cparser, (source_ptr, self.source_bytes, self.use_cols, self.fill_names,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("astropy.io.ascii.cparser.CParser.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_source_ptr);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "astropy/io/ascii/cparser.pyx":191
 * 
 *     cdef public:
 *         int width             # <<<<<<<<<<<<<<
 *         object source
 *         object header_start
 */

/* Python wrapper */
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_5width_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_5width_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_5width___get__(((struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_5width___get__(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->width); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("astropy.io.ascii.cparser.CParser.width.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_5width_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_5width_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_5width_2__set__(((struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_5width_2__set__(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 191, __pyx_L1_error)
  __pyx_v_self->width = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("astropy.io.ascii.cparser.CParser.width.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "astropy/io/ascii/cparser.pyx":192
 *     cdef public:
 *         int width
 *         object source             # <<<<<<<<<<<<<<
 *         object header_start
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_6source_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_6source_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_6source___get__(((struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_6source___get__(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->source);
  __pyx_r = __pyx_v_self->source;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_6source_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_6source_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_6source_2__set__(((struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_6source_2__set__(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->source);
  __Pyx_DECREF(__pyx_v_self->source);
  __pyx_v_self->source = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_6source_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_6source_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_6source_4__del__(((struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_6source_4__del__(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->source);
  __Pyx_DECREF(__pyx_v_self->source);
  __pyx_v_self->source = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "astropy/io/ascii/cparser.pyx":193
 *         int width
 *         object source
 *         object header_start             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self, source, strip_line_whitespace, strip_line_fields,
 */

/* Python wrapper */
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_12header_start_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_12header_start_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_12header_start___get__(((struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_12header_start___get__(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->header_start);
  __pyx_r = __pyx_v_self->header_start;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_12header_start_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_12header_start_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_12header_start_2__set__(((struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_12header_start_2__set__(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->header_start);
  __Pyx_DECREF(__pyx_v_self->header_start);
  __pyx_v_self->header_start = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_12header_start_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_12header_start_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_12header_start_4__del__(((struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7astropy_2io_5ascii_7cparser_7CParser_12header_start_4__del__(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->header_start);
  __Pyx_DECREF(__pyx_v_self->header_start);
  __pyx_v_self->header_start = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "astropy/io/ascii/cparser.pyx":806
 *                                 parallel=False)))
 * 
 * def _copy_cparser(bytes src_ptr, bytes source_bytes, use_cols, fill_names, fill_values,             # <<<<<<<<<<<<<<
 *                   strip_whitespace_lines, strip_whitespace_fields, kwargs):
 *     parser = CParser(None, strip_whitespace_lines, strip_whitespace_fields, **kwargs)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_1_copy_cparser(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_7astropy_2io_5ascii_7cparser_1_copy_cparser = {"_copy_cparser", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7astropy_2io_5ascii_7cparser_1_copy_cparser, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_1_copy_cparser(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_src_ptr = 0;
  PyObject *__pyx_v_source_bytes = 0;
  PyObject *__pyx_v_use_cols = 0;
  PyObject *__pyx_v_fill_names = 0;
  PyObject *__pyx_v_fill_values = 0;
  PyObject *__pyx_v_strip_whitespace_lines = 0;
  PyObject *__pyx_v_strip_whitespace_fields = 0;
  PyObject *__pyx_v_kwargs = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_copy_cparser (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_src_ptr,&__pyx_n_s_source_bytes,&__pyx_n_s_use_cols,&__pyx_n_s_fill_names,&__pyx_n_s_fill_values,&__pyx_n_s_strip_whitespace_lines,&__pyx_n_s_strip_whitespace_fields,&__pyx_n_s_kwargs,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_src_ptr)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_source_bytes)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_copy_cparser", 1, 8, 8, 1); __PYX_ERR(0, 806, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_use_cols)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_copy_cparser", 1, 8, 8, 2); __PYX_ERR(0, 806, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fill_names)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_copy_cparser", 1, 8, 8, 3); __PYX_ERR(0, 806, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fill_values)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_copy_cparser", 1, 8, 8, 4); __PYX_ERR(0, 806, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strip_whitespace_lines)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_copy_cparser", 1, 8, 8, 5); __PYX_ERR(0, 806, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strip_whitespace_fields)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_copy_cparser", 1, 8, 8, 6); __PYX_ERR(0, 806, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_copy_cparser", 1, 8, 8, 7); __PYX_ERR(0, 806, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_copy_cparser") < 0)) __PYX_ERR(0, 806, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 8) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
    }
    __pyx_v_src_ptr = ((PyObject*)values[0]);
    __pyx_v_source_bytes = ((PyObject*)values[1]);
    __pyx_v_use_cols = values[2];
    __pyx_v_fill_names = values[3];
    __pyx_v_fill_values = values[4];
    __pyx_v_strip_whitespace_lines = values[5];
    __pyx_v_strip_whitespace_fields = values[6];
    __pyx_v_kwargs = values[7];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_copy_cparser", 1, 8, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 806, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("astropy.io.ascii.cparser._copy_cparser", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_src_ptr), (&PyBytes_Type), 1, "src_ptr", 1))) __PYX_ERR(0, 806, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_source_bytes), (&PyBytes_Type), 1, "source_bytes", 1))) __PYX_ERR(0, 806, __pyx_L1_error)
  __pyx_r = __pyx_pf_7astropy_2io_5ascii_7cparser__copy_cparser(__pyx_self, __pyx_v_src_ptr, __pyx_v_source_bytes, __pyx_v_use_cols, __pyx_v_fill_names, __pyx_v_fill_values, __pyx_v_strip_whitespace_lines, __pyx_v_strip_whitespace_fields, __pyx_v_kwargs);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser__copy_cparser(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_src_ptr, PyObject *__pyx_v_source_bytes, PyObject *__pyx_v_use_cols, PyObject *__pyx_v_fill_names, PyObject *__pyx_v_fill_values, PyObject *__pyx_v_strip_whitespace_lines, PyObject *__pyx_v_strip_whitespace_fields, PyObject *__pyx_v_kwargs) {
  struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_parser = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  char *__pyx_t_5;
  __Pyx_RefNannySetupContext("_copy_cparser", 0);

  /* "astropy/io/ascii/cparser.pyx":808
 * def _copy_cparser(bytes src_ptr, bytes source_bytes, use_cols, fill_names, fill_values,
 *                   strip_whitespace_lines, strip_whitespace_fields, kwargs):
 *     parser = CParser(None, strip_whitespace_lines, strip_whitespace_fields, **kwargs)             # <<<<<<<<<<<<<<
 *     parser.use_cols = use_cols
 *     parser.fill_names = fill_names
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_1, 0, Py_None);
  __Pyx_INCREF(__pyx_v_strip_whitespace_lines);
  __Pyx_GIVEREF(__pyx_v_strip_whitespace_lines);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_strip_whitespace_lines);
  __Pyx_INCREF(__pyx_v_strip_whitespace_fields);
  __Pyx_GIVEREF(__pyx_v_strip_whitespace_fields);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_strip_whitespace_fields);
  if (unlikely(__pyx_v_kwargs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "argument after ** must be a mapping, not NoneType");
    __PYX_ERR(0, 808, __pyx_L1_error)
  }
  if (likely(PyDict_CheckExact(__pyx_v_kwargs))) {
    __pyx_t_2 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 808, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  } else {
    __pyx_t_2 = PyObject_CallFunctionObjArgs((PyObject*)&PyDict_Type, __pyx_v_kwargs, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 808, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7astropy_2io_5ascii_7cparser_CParser), __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_parser = ((struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "astropy/io/ascii/cparser.pyx":809
 *                   strip_whitespace_lines, strip_whitespace_fields, kwargs):
 *     parser = CParser(None, strip_whitespace_lines, strip_whitespace_fields, **kwargs)
 *     parser.use_cols = use_cols             # <<<<<<<<<<<<<<
 *     parser.fill_names = fill_names
 *     parser.fill_values = fill_values
 */
  if (!(likely(PySet_CheckExact(__pyx_v_use_cols))||((__pyx_v_use_cols) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "set", Py_TYPE(__pyx_v_use_cols)->tp_name), 0))) __PYX_ERR(0, 809, __pyx_L1_error)
  __pyx_t_3 = __pyx_v_use_cols;
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_parser->use_cols);
  __Pyx_DECREF(__pyx_v_parser->use_cols);
  __pyx_v_parser->use_cols = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "astropy/io/ascii/cparser.pyx":810
 *     parser = CParser(None, strip_whitespace_lines, strip_whitespace_fields, **kwargs)
 *     parser.use_cols = use_cols
 *     parser.fill_names = fill_names             # <<<<<<<<<<<<<<
 *     parser.fill_values = fill_values
 * 
 */
  __Pyx_INCREF(__pyx_v_fill_names);
  __Pyx_GIVEREF(__pyx_v_fill_names);
  __Pyx_GOTREF(__pyx_v_parser->fill_names);
  __Pyx_DECREF(__pyx_v_parser->fill_names);
  __pyx_v_parser->fill_names = __pyx_v_fill_names;

  /* "astropy/io/ascii/cparser.pyx":811
 *     parser.use_cols = use_cols
 *     parser.fill_names = fill_names
 *     parser.fill_values = fill_values             # <<<<<<<<<<<<<<
 * 
 *     if src_ptr:
 */
  __Pyx_INCREF(__pyx_v_fill_values);
  __Pyx_GIVEREF(__pyx_v_fill_values);
  __Pyx_GOTREF(__pyx_v_parser->fill_values);
  __Pyx_DECREF(__pyx_v_parser->fill_values);
  __pyx_v_parser->fill_values = __pyx_v_fill_values;

  /* "astropy/io/ascii/cparser.pyx":813
 *     parser.fill_values = fill_values
 * 
 *     if src_ptr:             # <<<<<<<<<<<<<<
 *         parser.tokenizer.source = src_ptr
 *     else:
 */
  __pyx_t_4 = (__pyx_v_src_ptr != Py_None)&&(PyBytes_GET_SIZE(__pyx_v_src_ptr) != 0);
  if (__pyx_t_4) {

    /* "astropy/io/ascii/cparser.pyx":814
 * 
 *     if src_ptr:
 *         parser.tokenizer.source = src_ptr             # <<<<<<<<<<<<<<
 *     else:
 *         parser.tokenizer.source = source_bytes
 */
    if (unlikely(__pyx_v_src_ptr == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
      __PYX_ERR(0, 814, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_PyBytes_AsWritableString(__pyx_v_src_ptr); if (unlikely((!__pyx_t_5) && PyErr_Occurred())) __PYX_ERR(0, 814, __pyx_L1_error)
    __pyx_v_parser->tokenizer->source = __pyx_t_5;

    /* "astropy/io/ascii/cparser.pyx":813
 *     parser.fill_values = fill_values
 * 
 *     if src_ptr:             # <<<<<<<<<<<<<<
 *         parser.tokenizer.source = src_ptr
 *     else:
 */
    goto __pyx_L3;
  }

  /* "astropy/io/ascii/cparser.pyx":816
 *         parser.tokenizer.source = src_ptr
 *     else:
 *         parser.tokenizer.source = source_bytes             # <<<<<<<<<<<<<<
 *     return parser
 * 
 */
  /*else*/ {
    if (unlikely(__pyx_v_source_bytes == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
      __PYX_ERR(0, 816, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_PyBytes_AsWritableString(__pyx_v_source_bytes); if (unlikely((!__pyx_t_5) && PyErr_Occurred())) __PYX_ERR(0, 816, __pyx_L1_error)
    __pyx_v_parser->tokenizer->source = __pyx_t_5;
  }
  __pyx_L3:;

  /* "astropy/io/ascii/cparser.pyx":817
 *     else:
 *         parser.tokenizer.source = source_bytes
 *     return parser             # <<<<<<<<<<<<<<
 * 
 * def _read_chunk(CParser self, start, end, try_int,
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_parser));
  __pyx_r = ((PyObject *)__pyx_v_parser);
  goto __pyx_L0;

  /* "astropy/io/ascii/cparser.pyx":806
 *                                 parallel=False)))
 * 
 * def _copy_cparser(bytes src_ptr, bytes source_bytes, use_cols, fill_names, fill_values,             # <<<<<<<<<<<<<<
 *                   strip_whitespace_lines, strip_whitespace_fields, kwargs):
 *     parser = CParser(None, strip_whitespace_lines, strip_whitespace_fields, **kwargs)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("astropy.io.ascii.cparser._copy_cparser", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_parser);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "astropy/io/ascii/cparser.pyx":819
 *     return parser
 * 
 * def _read_chunk(CParser self, start, end, try_int,             # <<<<<<<<<<<<<<
 *                 try_float, try_string, queue, reconvert_queue, i):
 *     cdef tokenizer_t *chunk_tokenizer = self.tokenizer
 */

/* Python wrapper */
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_3_read_chunk(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_7astropy_2io_5ascii_7cparser_3_read_chunk = {"_read_chunk", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7astropy_2io_5ascii_7cparser_3_read_chunk, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_3_read_chunk(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self = 0;
  PyObject *__pyx_v_start = 0;
  PyObject *__pyx_v_end = 0;
  PyObject *__pyx_v_try_int = 0;
  PyObject *__pyx_v_try_float = 0;
  PyObject *__pyx_v_try_string = 0;
  PyObject *__pyx_v_queue = 0;
  PyObject *__pyx_v_reconvert_queue = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_read_chunk (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_start,&__pyx_n_s_end,&__pyx_n_s_try_int,&__pyx_n_s_try_float,&__pyx_n_s_try_string,&__pyx_n_s_queue,&__pyx_n_s_reconvert_queue,&__pyx_n_s_i,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_start)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_read_chunk", 1, 9, 9, 1); __PYX_ERR(0, 819, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_end)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_read_chunk", 1, 9, 9, 2); __PYX_ERR(0, 819, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_try_int)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_read_chunk", 1, 9, 9, 3); __PYX_ERR(0, 819, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_try_float)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_read_chunk", 1, 9, 9, 4); __PYX_ERR(0, 819, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_try_string)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_read_chunk", 1, 9, 9, 5); __PYX_ERR(0, 819, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_queue)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_read_chunk", 1, 9, 9, 6); __PYX_ERR(0, 819, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_reconvert_queue)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_read_chunk", 1, 9, 9, 7); __PYX_ERR(0, 819, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_i)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_read_chunk", 1, 9, 9, 8); __PYX_ERR(0, 819, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_read_chunk") < 0)) __PYX_ERR(0, 819, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 9) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
    }
    __pyx_v_self = ((struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *)values[0]);
    __pyx_v_start = values[1];
    __pyx_v_end = values[2];
    __pyx_v_try_int = values[3];
    __pyx_v_try_float = values[4];
    __pyx_v_try_string = values[5];
    __pyx_v_queue = values[6];
    __pyx_v_reconvert_queue = values[7];
    __pyx_v_i = values[8];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_read_chunk", 1, 9, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 819, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("astropy.io.ascii.cparser._read_chunk", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_self), __pyx_ptype_7astropy_2io_5ascii_7cparser_CParser, 1, "self", 0))) __PYX_ERR(0, 819, __pyx_L1_error)
  __pyx_r = __pyx_pf_7astropy_2io_5ascii_7cparser_2_read_chunk(__pyx_self, __pyx_v_self, __pyx_v_start, __pyx_v_end, __pyx_v_try_int, __pyx_v_try_float, __pyx_v_try_string, __pyx_v_queue, __pyx_v_reconvert_queue, __pyx_v_i);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_7astropy_2io_5ascii_7cparser_11_read_chunk_2generator4(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "astropy/io/ascii/cparser.pyx":832
 *         err = (chunk_tokenizer.code, chunk_tokenizer.num_rows)
 *     if chunk_tokenizer.num_rows == 0: # no data
 *         data = dict((name, np.array([], np.int_)) for name in self.get_names())             # <<<<<<<<<<<<<<
 *         line_comments = self._get_comments(chunk_tokenizer)
 *     else:
 */

static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_11_read_chunk_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr *)__pyx_tp_new_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr(__pyx_ptype_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 832, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_7astropy_2io_5ascii_7cparser_11_read_chunk_2generator4, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_read_chunk_locals_genexpr, __pyx_n_s_astropy_io_ascii_cparser); if (unlikely(!gen)) __PYX_ERR(0, 832, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("astropy.io.ascii.cparser._read_chunk.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_7astropy_2io_5ascii_7cparser_11_read_chunk_2generator4(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr *__pyx_cur_scope = ((struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 832, __pyx_L1_error)
  __pyx_r = PyDict_New(); if (unlikely(!__pyx_r)) __PYX_ERR(0, 832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 832, __pyx_L1_error) }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self), __pyx_n_s_get_names); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 832, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 832, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 832, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 832, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 832, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 832, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_5(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 832, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_name);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_name, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 832, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 832, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 832, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 832, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_int); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 832, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_3, __pyx_t_8};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 832, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_3, __pyx_t_8};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 832, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 832, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_8);
      __pyx_t_3 = 0;
      __pyx_t_8 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 832, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(PyDict_SetItem(__pyx_r, (PyObject*)__pyx_cur_scope->__pyx_v_name, (PyObject*)__pyx_t_1))) __PYX_ERR(0, 832, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "astropy/io/ascii/cparser.pyx":819
 *     return parser
 * 
 * def _read_chunk(CParser self, start, end, try_int,             # <<<<<<<<<<<<<<
 *                 try_float, try_string, queue, reconvert_queue, i):
 *     cdef tokenizer_t *chunk_tokenizer = self.tokenizer
 */

static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_2_read_chunk(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *__pyx_v_self, PyObject *__pyx_v_start, PyObject *__pyx_v_end, PyObject *__pyx_v_try_int, PyObject *__pyx_v_try_float, PyObject *__pyx_v_try_string, PyObject *__pyx_v_queue, PyObject *__pyx_v_reconvert_queue, PyObject *__pyx_v_i) {
  struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk *__pyx_cur_scope;
  tokenizer_t *__pyx_v_chunk_tokenizer;
  PyObject *__pyx_v_data = NULL;
  PyObject *__pyx_v_err = NULL;
  PyObject *__pyx_v_line_comments = NULL;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_v_reconvert_cols = NULL;
  PyObject *__pyx_v_col = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  tokenizer_t *__pyx_t_1;
  size_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *(*__pyx_t_12)(PyObject *);
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *(*__pyx_t_17)(PyObject *);
  __Pyx_RefNannySetupContext("_read_chunk", 0);
  __pyx_cur_scope = (struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk *)__pyx_tp_new_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk(__pyx_ptype_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 819, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);

  /* "astropy/io/ascii/cparser.pyx":821
 * def _read_chunk(CParser self, start, end, try_int,
 *                 try_float, try_string, queue, reconvert_queue, i):
 *     cdef tokenizer_t *chunk_tokenizer = self.tokenizer             # <<<<<<<<<<<<<<
 *     chunk_tokenizer.source_len = end
 *     chunk_tokenizer.source_pos = start
 */
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_self->tokenizer;
  __pyx_v_chunk_tokenizer = __pyx_t_1;

  /* "astropy/io/ascii/cparser.pyx":822
 *                 try_float, try_string, queue, reconvert_queue, i):
 *     cdef tokenizer_t *chunk_tokenizer = self.tokenizer
 *     chunk_tokenizer.source_len = end             # <<<<<<<<<<<<<<
 *     chunk_tokenizer.source_pos = start
 *     reset_comments(chunk_tokenizer)
 */
  __pyx_t_2 = __Pyx_PyInt_As_size_t(__pyx_v_end); if (unlikely((__pyx_t_2 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 822, __pyx_L1_error)
  __pyx_v_chunk_tokenizer->source_len = __pyx_t_2;

  /* "astropy/io/ascii/cparser.pyx":823
 *     cdef tokenizer_t *chunk_tokenizer = self.tokenizer
 *     chunk_tokenizer.source_len = end
 *     chunk_tokenizer.source_pos = start             # <<<<<<<<<<<<<<
 *     reset_comments(chunk_tokenizer)
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_As_size_t(__pyx_v_start); if (unlikely((__pyx_t_2 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 823, __pyx_L1_error)
  __pyx_v_chunk_tokenizer->source_pos = __pyx_t_2;

  /* "astropy/io/ascii/cparser.pyx":824
 *     chunk_tokenizer.source_len = end
 *     chunk_tokenizer.source_pos = start
 *     reset_comments(chunk_tokenizer)             # <<<<<<<<<<<<<<
 * 
 *     data = None
 */
  reset_comments(__pyx_v_chunk_tokenizer);

  /* "astropy/io/ascii/cparser.pyx":826
 *     reset_comments(chunk_tokenizer)
 * 
 *     data = None             # <<<<<<<<<<<<<<
 *     err = None
 * 
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_data = Py_None;

  /* "astropy/io/ascii/cparser.pyx":827
 * 
 *     data = None
 *     err = None             # <<<<<<<<<<<<<<
 * 
 *     if tokenize(chunk_tokenizer, -1, 0, <int>len(self.names)) != 0:
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_err = Py_None;

  /* "astropy/io/ascii/cparser.pyx":829
 *     err = None
 * 
 *     if tokenize(chunk_tokenizer, -1, 0, <int>len(self.names)) != 0:             # <<<<<<<<<<<<<<
 *         err = (chunk_tokenizer.code, chunk_tokenizer.num_rows)
 *     if chunk_tokenizer.num_rows == 0: # no data
 */
  __pyx_t_3 = __pyx_cur_scope->__pyx_v_self->names;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = PyObject_Length(__pyx_t_3); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 829, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = ((tokenize(__pyx_v_chunk_tokenizer, -1, 0, ((int)__pyx_t_4)) != 0) != 0);
  if (__pyx_t_5) {

    /* "astropy/io/ascii/cparser.pyx":830
 * 
 *     if tokenize(chunk_tokenizer, -1, 0, <int>len(self.names)) != 0:
 *         err = (chunk_tokenizer.code, chunk_tokenizer.num_rows)             # <<<<<<<<<<<<<<
 *     if chunk_tokenizer.num_rows == 0: # no data
 *         data = dict((name, np.array([], np.int_)) for name in self.get_names())
 */
    __pyx_t_3 = __Pyx_PyInt_From_err_code(__pyx_v_chunk_tokenizer->code); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 830, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_chunk_tokenizer->num_rows); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 830, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 830, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_6);
    __pyx_t_3 = 0;
    __pyx_t_6 = 0;
    __Pyx_DECREF_SET(__pyx_v_err, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "astropy/io/ascii/cparser.pyx":829
 *     err = None
 * 
 *     if tokenize(chunk_tokenizer, -1, 0, <int>len(self.names)) != 0:             # <<<<<<<<<<<<<<
 *         err = (chunk_tokenizer.code, chunk_tokenizer.num_rows)
 *     if chunk_tokenizer.num_rows == 0: # no data
 */
  }

  /* "astropy/io/ascii/cparser.pyx":831
 *     if tokenize(chunk_tokenizer, -1, 0, <int>len(self.names)) != 0:
 *         err = (chunk_tokenizer.code, chunk_tokenizer.num_rows)
 *     if chunk_tokenizer.num_rows == 0: # no data             # <<<<<<<<<<<<<<
 *         data = dict((name, np.array([], np.int_)) for name in self.get_names())
 *         line_comments = self._get_comments(chunk_tokenizer)
 */
  __pyx_t_5 = ((__pyx_v_chunk_tokenizer->num_rows == 0) != 0);
  if (__pyx_t_5) {

    /* "astropy/io/ascii/cparser.pyx":832
 *         err = (chunk_tokenizer.code, chunk_tokenizer.num_rows)
 *     if chunk_tokenizer.num_rows == 0: # no data
 *         data = dict((name, np.array([], np.int_)) for name in self.get_names())             # <<<<<<<<<<<<<<
 *         line_comments = self._get_comments(chunk_tokenizer)
 *     else:
 */
    __pyx_t_7 = __pyx_pf_7astropy_2io_5ascii_7cparser_11_read_chunk_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 832, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_Generator_Next(__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 832, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF_SET(__pyx_v_data, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "astropy/io/ascii/cparser.pyx":833
 *     if chunk_tokenizer.num_rows == 0: # no data
 *         data = dict((name, np.array([], np.int_)) for name in self.get_names())
 *         line_comments = self._get_comments(chunk_tokenizer)             # <<<<<<<<<<<<<<
 *     else:
 *         try:
 */
    __pyx_t_6 = ((struct __pyx_vtabstruct_7astropy_2io_5ascii_7cparser_CParser *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_get_comments(__pyx_cur_scope->__pyx_v_self, __pyx_v_chunk_tokenizer); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 833, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_v_line_comments = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "astropy/io/ascii/cparser.pyx":831
 *     if tokenize(chunk_tokenizer, -1, 0, <int>len(self.names)) != 0:
 *         err = (chunk_tokenizer.code, chunk_tokenizer.num_rows)
 *     if chunk_tokenizer.num_rows == 0: # no data             # <<<<<<<<<<<<<<
 *         data = dict((name, np.array([], np.int_)) for name in self.get_names())
 *         line_comments = self._get_comments(chunk_tokenizer)
 */
    goto __pyx_L4;
  }

  /* "astropy/io/ascii/cparser.pyx":835
 *         line_comments = self._get_comments(chunk_tokenizer)
 *     else:
 *         try:             # <<<<<<<<<<<<<<
 *             data, line_comments = self._convert_data(chunk_tokenizer,
 *                                       try_int, try_float, try_string, -1)
 */
  /*else*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "astropy/io/ascii/cparser.pyx":836
 *     else:
 *         try:
 *             data, line_comments = self._convert_data(chunk_tokenizer,             # <<<<<<<<<<<<<<
 *                                       try_int, try_float, try_string, -1)
 *         except Exception as e:
 */
        __pyx_t_6 = ((struct __pyx_vtabstruct_7astropy_2io_5ascii_7cparser_CParser *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_convert_data(__pyx_cur_scope->__pyx_v_self, __pyx_v_chunk_tokenizer, __pyx_v_try_int, __pyx_v_try_float, __pyx_v_try_string, __pyx_int_neg_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 836, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_6);
        if ((likely(PyTuple_CheckExact(__pyx_t_6))) || (PyList_CheckExact(__pyx_t_6))) {
          PyObject* sequence = __pyx_t_6;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 836, __pyx_L5_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_7 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_7 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_3);
          #else
          __pyx_t_7 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 836, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 836, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_11 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 836, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_12 = Py_TYPE(__pyx_t_11)->tp_iternext;
          index = 0; __pyx_t_7 = __pyx_t_12(__pyx_t_11); if (unlikely(!__pyx_t_7)) goto __pyx_L11_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_7);
          index = 1; __pyx_t_3 = __pyx_t_12(__pyx_t_11); if (unlikely(!__pyx_t_3)) goto __pyx_L11_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_3);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_11), 2) < 0) __PYX_ERR(0, 836, __pyx_L5_error)
          __pyx_t_12 = NULL;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          goto __pyx_L12_unpacking_done;
          __pyx_L11_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_12 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 836, __pyx_L5_error)
          __pyx_L12_unpacking_done:;
        }
        __Pyx_DECREF_SET(__pyx_v_data, __pyx_t_7);
        __pyx_t_7 = 0;
        __pyx_v_line_comments = __pyx_t_3;
        __pyx_t_3 = 0;

        /* "astropy/io/ascii/cparser.pyx":835
 *         line_comments = self._get_comments(chunk_tokenizer)
 *     else:
 *         try:             # <<<<<<<<<<<<<<
 *             data, line_comments = self._convert_data(chunk_tokenizer,
 *                                       try_int, try_float, try_string, -1)
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L10_try_end;
      __pyx_L5_error:;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "astropy/io/ascii/cparser.pyx":838
 *             data, line_comments = self._convert_data(chunk_tokenizer,
 *                                       try_int, try_float, try_string, -1)
 *         except Exception as e:             # <<<<<<<<<<<<<<
 *             delete_tokenizer(chunk_tokenizer)
 *             queue.put((None, e, i))
 */
      __pyx_t_13 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
      if (__pyx_t_13) {
        __Pyx_AddTraceback("astropy.io.ascii.cparser._read_chunk", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_3, &__pyx_t_7) < 0) __PYX_ERR(0, 838, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_3);
        __pyx_v_e = __pyx_t_3;

        /* "astropy/io/ascii/cparser.pyx":839
 *                                       try_int, try_float, try_string, -1)
 *         except Exception as e:
 *             delete_tokenizer(chunk_tokenizer)             # <<<<<<<<<<<<<<
 *             queue.put((None, e, i))
 *             return
 */
        delete_tokenizer(__pyx_v_chunk_tokenizer);

        /* "astropy/io/ascii/cparser.pyx":840
 *         except Exception as e:
 *             delete_tokenizer(chunk_tokenizer)
 *             queue.put((None, e, i))             # <<<<<<<<<<<<<<
 *             return
 * 
 */
        __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_queue, __pyx_n_s_put); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 840, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_14);
        __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 840, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_INCREF(Py_None);
        __Pyx_GIVEREF(Py_None);
        PyTuple_SET_ITEM(__pyx_t_15, 0, Py_None);
        __Pyx_INCREF(__pyx_v_e);
        __Pyx_GIVEREF(__pyx_v_e);
        PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_v_e);
        __Pyx_INCREF(__pyx_v_i);
        __Pyx_GIVEREF(__pyx_v_i);
        PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_v_i);
        __pyx_t_16 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_14))) {
          __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_14);
          if (likely(__pyx_t_16)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
            __Pyx_INCREF(__pyx_t_16);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_14, function);
          }
        }
        __pyx_t_11 = (__pyx_t_16) ? __Pyx_PyObject_Call2Args(__pyx_t_14, __pyx_t_16, __pyx_t_15) : __Pyx_PyObject_CallOneArg(__pyx_t_14, __pyx_t_15);
        __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 840, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

        /* "astropy/io/ascii/cparser.pyx":841
 *             delete_tokenizer(chunk_tokenizer)
 *             queue.put((None, e, i))
 *             return             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L8_except_return;
      }
      goto __pyx_L7_except_error;
      __pyx_L7_except_error:;

      /* "astropy/io/ascii/cparser.pyx":835
 *         line_comments = self._get_comments(chunk_tokenizer)
 *     else:
 *         try:             # <<<<<<<<<<<<<<
 *             data, line_comments = self._convert_data(chunk_tokenizer,
 *                                       try_int, try_float, try_string, -1)
 */
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L1_error;
      __pyx_L8_except_return:;
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L0;
      __pyx_L10_try_end:;
    }
  }
  __pyx_L4:;

  /* "astropy/io/ascii/cparser.pyx":843
 *             return
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         queue.put(((line_comments, data), err, i))
 *     except Queue.Full as e:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_10);
    __Pyx_XGOTREF(__pyx_t_9);
    __Pyx_XGOTREF(__pyx_t_8);
    /*try:*/ {

      /* "astropy/io/ascii/cparser.pyx":844
 * 
 *     try:
 *         queue.put(((line_comments, data), err, i))             # <<<<<<<<<<<<<<
 *     except Queue.Full as e:
 *         # hopefully this shouldn't happen
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_queue, __pyx_n_s_put); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 844, __pyx_L15_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 844, __pyx_L15_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_v_line_comments);
      __Pyx_GIVEREF(__pyx_v_line_comments);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_line_comments);
      __Pyx_INCREF(__pyx_v_data);
      __Pyx_GIVEREF(__pyx_v_data);
      PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_data);
      __pyx_t_11 = PyTuple_New(3); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 844, __pyx_L15_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_err);
      __Pyx_GIVEREF(__pyx_v_err);
      PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_v_err);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_11, 2, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_7 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_6, __pyx_t_11) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_11);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 844, __pyx_L15_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "astropy/io/ascii/cparser.pyx":843
 *             return
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         queue.put(((line_comments, data), err, i))
 *     except Queue.Full as e:
 */
    }
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    goto __pyx_L20_try_end;
    __pyx_L15_error:;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "astropy/io/ascii/cparser.pyx":845
 *     try:
 *         queue.put(((line_comments, data), err, i))
 *     except Queue.Full as e:             # <<<<<<<<<<<<<<
 *         # hopefully this shouldn't happen
 *         delete_tokenizer(chunk_tokenizer)
 */
    __Pyx_ErrFetch(&__pyx_t_7, &__pyx_t_3, &__pyx_t_11);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_Queue); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 845, __pyx_L17_except_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_Full); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 845, __pyx_L17_except_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_13 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_7, __pyx_t_14);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_ErrRestore(__pyx_t_7, __pyx_t_3, __pyx_t_11);
    __pyx_t_7 = 0; __pyx_t_3 = 0; __pyx_t_11 = 0;
    if (__pyx_t_13) {
      __Pyx_AddTraceback("astropy.io.ascii.cparser._read_chunk", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_11, &__pyx_t_3, &__pyx_t_7) < 0) __PYX_ERR(0, 845, __pyx_L17_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_3);
      __pyx_v_e = __pyx_t_3;

      /* "astropy/io/ascii/cparser.pyx":847
 *     except Queue.Full as e:
 *         # hopefully this shouldn't happen
 *         delete_tokenizer(chunk_tokenizer)             # <<<<<<<<<<<<<<
 *         queue.pop()
 *         queue.put((None, e, i))
 */
      delete_tokenizer(__pyx_v_chunk_tokenizer);

      /* "astropy/io/ascii/cparser.pyx":848
 *         # hopefully this shouldn't happen
 *         delete_tokenizer(chunk_tokenizer)
 *         queue.pop()             # <<<<<<<<<<<<<<
 *         queue.put((None, e, i))
 *     reconvert_cols = reconvert_queue.get()
 */
      __pyx_t_14 = __Pyx_PyObject_Pop(__pyx_v_queue); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 848, __pyx_L17_except_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

      /* "astropy/io/ascii/cparser.pyx":849
 *         delete_tokenizer(chunk_tokenizer)
 *         queue.pop()
 *         queue.put((None, e, i))             # <<<<<<<<<<<<<<
 *     reconvert_cols = reconvert_queue.get()
 *     for col in reconvert_cols:
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_queue, __pyx_n_s_put); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 849, __pyx_L17_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 849, __pyx_L17_except_error)
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      PyTuple_SET_ITEM(__pyx_t_15, 0, Py_None);
      __Pyx_INCREF(__pyx_v_e);
      __Pyx_GIVEREF(__pyx_v_e);
      PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_v_e);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_v_i);
      __pyx_t_16 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_16)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_16);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_14 = (__pyx_t_16) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_16, __pyx_t_15) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_15);
      __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 849, __pyx_L17_except_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L16_exception_handled;
    }
    goto __pyx_L17_except_error;
    __pyx_L17_except_error:;

    /* "astropy/io/ascii/cparser.pyx":843
 *             return
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         queue.put(((line_comments, data), err, i))
 *     except Queue.Full as e:
 */
    __Pyx_XGIVEREF(__pyx_t_10);
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_9, __pyx_t_8);
    goto __pyx_L1_error;
    __pyx_L16_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_10);
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_9, __pyx_t_8);
    __pyx_L20_try_end:;
  }

  /* "astropy/io/ascii/cparser.pyx":850
 *         queue.pop()
 *         queue.put((None, e, i))
 *     reconvert_cols = reconvert_queue.get()             # <<<<<<<<<<<<<<
 *     for col in reconvert_cols:
 *         queue.put((self._convert_str(chunk_tokenizer, col, -1), i, col))
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_reconvert_queue, __pyx_n_s_get); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 850, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_11 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_11)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_11);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_7 = (__pyx_t_11) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_11) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 850, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_reconvert_cols = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "astropy/io/ascii/cparser.pyx":851
 *         queue.put((None, e, i))
 *     reconvert_cols = reconvert_queue.get()
 *     for col in reconvert_cols:             # <<<<<<<<<<<<<<
 *         queue.put((self._convert_str(chunk_tokenizer, col, -1), i, col))
 *     delete_tokenizer(chunk_tokenizer)
 */
  if (likely(PyList_CheckExact(__pyx_v_reconvert_cols)) || PyTuple_CheckExact(__pyx_v_reconvert_cols)) {
    __pyx_t_7 = __pyx_v_reconvert_cols; __Pyx_INCREF(__pyx_t_7); __pyx_t_4 = 0;
    __pyx_t_17 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_v_reconvert_cols); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 851, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_17 = Py_TYPE(__pyx_t_7)->tp_iternext; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 851, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_17)) {
      if (likely(PyList_CheckExact(__pyx_t_7))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_7)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 851, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_7, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 851, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_7)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 851, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_7, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 851, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_17(__pyx_t_7);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 851, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_col, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "astropy/io/ascii/cparser.pyx":852
 *     reconvert_cols = reconvert_queue.get()
 *     for col in reconvert_cols:
 *         queue.put((self._convert_str(chunk_tokenizer, col, -1), i, col))             # <<<<<<<<<<<<<<
 *     delete_tokenizer(chunk_tokenizer)
 *     reconvert_queue.put(reconvert_cols) # return to the queue for other processes
 */
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_queue, __pyx_n_s_put); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 852, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_13 = __Pyx_PyInt_As_int(__pyx_v_col); if (unlikely((__pyx_t_13 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 852, __pyx_L1_error)
    __pyx_t_14 = ((struct __pyx_vtabstruct_7astropy_2io_5ascii_7cparser_CParser *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_convert_str(__pyx_cur_scope->__pyx_v_self, __pyx_v_chunk_tokenizer, __pyx_t_13, -1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 852, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 852, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_14);
    __Pyx_INCREF(__pyx_v_i);
    __Pyx_GIVEREF(__pyx_v_i);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_i);
    __Pyx_INCREF(__pyx_v_col);
    __Pyx_GIVEREF(__pyx_v_col);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_v_col);
    __pyx_t_14 = 0;
    __pyx_t_14 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_14)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_14);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_3 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_14, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 852, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "astropy/io/ascii/cparser.pyx":851
 *         queue.put((None, e, i))
 *     reconvert_cols = reconvert_queue.get()
 *     for col in reconvert_cols:             # <<<<<<<<<<<<<<
 *         queue.put((self._convert_str(chunk_tokenizer, col, -1), i, col))
 *     delete_tokenizer(chunk_tokenizer)
 */
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "astropy/io/ascii/cparser.pyx":853
 *     for col in reconvert_cols:
 *         queue.put((self._convert_str(chunk_tokenizer, col, -1), i, col))
 *     delete_tokenizer(chunk_tokenizer)             # <<<<<<<<<<<<<<
 *     reconvert_queue.put(reconvert_cols) # return to the queue for other processes
 * 
 */
  delete_tokenizer(__pyx_v_chunk_tokenizer);

  /* "astropy/io/ascii/cparser.pyx":854
 *         queue.put((self._convert_str(chunk_tokenizer, col, -1), i, col))
 *     delete_tokenizer(chunk_tokenizer)
 *     reconvert_queue.put(reconvert_cols) # return to the queue for other processes             # <<<<<<<<<<<<<<
 * 
 * cdef class FastWriter:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_reconvert_queue, __pyx_n_s_put); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 854, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_11 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_11)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_11);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_7 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_11, __pyx_v_reconvert_cols) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_reconvert_cols);
  __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 854, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "astropy/io/ascii/cparser.pyx":819
 *     return parser
 * 
 * def _read_chunk(CParser self, start, end, try_int,             # <<<<<<<<<<<<<<
 *                 try_float, try_string, queue, reconvert_queue, i):
 *     cdef tokenizer_t *chunk_tokenizer = self.tokenizer
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("astropy.io.ascii.cparser._read_chunk", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XDECREF(__pyx_v_line_comments);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF(__pyx_v_reconvert_cols);
  __Pyx_XDECREF(__pyx_v_col);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "astropy/io/ascii/cparser.pyx":878
 *         object comment
 * 
 *     def __cinit__(self, table,             # <<<<<<<<<<<<<<
 *                   delimiter=',',
 *                   comment='# ',
 */

/* Python wrapper */
static int __pyx_pw_7astropy_2io_5ascii_7cparser_10FastWriter_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7astropy_2io_5ascii_7cparser_10FastWriter_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_table = 0;
  PyObject *__pyx_v_delimiter = 0;
  PyObject *__pyx_v_comment = 0;
  PyObject *__pyx_v_quotechar = 0;
  CYTHON_UNUSED PyObject *__pyx_v_expchar = 0;
  PyObject *__pyx_v_formats = 0;
  PyObject *__pyx_v_strip_whitespace = 0;
  CYTHON_UNUSED PyObject *__pyx_v_names = 0;
  PyObject *__pyx_v_include_names = 0;
  PyObject *__pyx_v_exclude_names = 0;
  PyObject *__pyx_v_fill_values = 0;
  PyObject *__pyx_v_fill_include_names = 0;
  PyObject *__pyx_v_fill_exclude_names = 0;
  PyObject *__pyx_v_fast_writer = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_table,&__pyx_n_s_delimiter,&__pyx_n_s_comment,&__pyx_n_s_quotechar,&__pyx_n_s_expchar,&__pyx_n_s_formats,&__pyx_n_s_strip_whitespace,&__pyx_n_s_names,&__pyx_n_s_include_names,&__pyx_n_s_exclude_names,&__pyx_n_s_fill_values,&__pyx_n_s_fill_include_names,&__pyx_n_s_fill_exclude_names,&__pyx_n_s_fast_writer,0};
    PyObject* values[14] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    values[1] = ((PyObject *)__pyx_kp_s__3);
    values[2] = ((PyObject *)__pyx_kp_s__20);
    values[3] = ((PyObject *)__pyx_kp_s__4);
    values[4] = ((PyObject *)__pyx_n_s_e);

    /* "astropy/io/ascii/cparser.pyx":883
 *                   quotechar='"',
 *                   expchar='e',
 *                   formats=None,             # <<<<<<<<<<<<<<
 *                   strip_whitespace=True,
 *                   names=None, # ignore, already used in _get_writer
 */
    values[5] = ((PyObject *)Py_None);

    /* "astropy/io/ascii/cparser.pyx":884
 *                   expchar='e',
 *                   formats=None,
 *                   strip_whitespace=True,             # <<<<<<<<<<<<<<
 *                   names=None, # ignore, already used in _get_writer
 *                   include_names=None,
 */
    values[6] = ((PyObject *)Py_True);

    /* "astropy/io/ascii/cparser.pyx":885
 *                   formats=None,
 *                   strip_whitespace=True,
 *                   names=None, # ignore, already used in _get_writer             # <<<<<<<<<<<<<<
 *                   include_names=None,
 *                   exclude_names=None,
 */
    values[7] = ((PyObject *)Py_None);

    /* "astropy/io/ascii/cparser.pyx":886
 *                   strip_whitespace=True,
 *                   names=None, # ignore, already used in _get_writer
 *                   include_names=None,             # <<<<<<<<<<<<<<
 *                   exclude_names=None,
 *                   fill_values=[],
 */
    values[8] = ((PyObject *)Py_None);

    /* "astropy/io/ascii/cparser.pyx":887
 *                   names=None, # ignore, already used in _get_writer
 *                   include_names=None,
 *                   exclude_names=None,             # <<<<<<<<<<<<<<
 *                   fill_values=[],
 *                   fill_include_names=None,
 */
    values[9] = ((PyObject *)Py_None);
    values[10] = __pyx_k__21;

    /* "astropy/io/ascii/cparser.pyx":889
 *                   exclude_names=None,
 *                   fill_values=[],
 *                   fill_include_names=None,             # <<<<<<<<<<<<<<
 *                   fill_exclude_names=None,
 *                   fast_writer=True):
 */
    values[11] = ((PyObject *)Py_None);

    /* "astropy/io/ascii/cparser.pyx":890
 *                   fill_values=[],
 *                   fill_include_names=None,
 *                   fill_exclude_names=None,             # <<<<<<<<<<<<<<
 *                   fast_writer=True):
 * 
 */
    values[12] = ((PyObject *)Py_None);

    /* "astropy/io/ascii/cparser.pyx":891
 *                   fill_include_names=None,
 *                   fill_exclude_names=None,
 *                   fast_writer=True):             # <<<<<<<<<<<<<<
 * 
 *         if fast_writer is True:
 */
    values[13] = ((PyObject *)Py_True);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_table)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_delimiter);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_comment);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_quotechar);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_expchar);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_formats);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strip_whitespace);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_names);
          if (value) { values[7] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_include_names);
          if (value) { values[8] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_exclude_names);
          if (value) { values[9] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fill_values);
          if (value) { values[10] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fill_include_names);
          if (value) { values[11] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fill_exclude_names);
          if (value) { values[12] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fast_writer);
          if (value) { values[13] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 878, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_table = values[0];
    __pyx_v_delimiter = values[1];
    __pyx_v_comment = values[2];
    __pyx_v_quotechar = values[3];
    __pyx_v_expchar = values[4];
    __pyx_v_formats = values[5];
    __pyx_v_strip_whitespace = values[6];
    __pyx_v_names = values[7];
    __pyx_v_include_names = values[8];
    __pyx_v_exclude_names = values[9];
    __pyx_v_fill_values = values[10];
    __pyx_v_fill_include_names = values[11];
    __pyx_v_fill_exclude_names = values[12];
    __pyx_v_fast_writer = values[13];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 1, 14, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 878, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("astropy.io.ascii.cparser.FastWriter.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7astropy_2io_5ascii_7cparser_10FastWriter___cinit__(((struct __pyx_obj_7astropy_2io_5ascii_7cparser_FastWriter *)__pyx_v_self), __pyx_v_table, __pyx_v_delimiter, __pyx_v_comment, __pyx_v_quotechar, __pyx_v_expchar, __pyx_v_formats, __pyx_v_strip_whitespace, __pyx_v_names, __pyx_v_include_names, __pyx_v_exclude_names, __pyx_v_fill_values, __pyx_v_fill_include_names, __pyx_v_fill_exclude_names, __pyx_v_fast_writer);

  /* "astropy/io/ascii/cparser.pyx":878
 *         object comment
 * 
 *     def __cinit__(self, table,             # <<<<<<<<<<<<<<
 *                   delimiter=',',
 *                   comment='# ',
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7astropy_2io_5ascii_7cparser_10FastWriter___cinit__(struct __pyx_obj_7astropy_2io_5ascii_7cparser_FastWriter *__pyx_v_self, PyObject *__pyx_v_table, PyObject *__pyx_v_delimiter, PyObject *__pyx_v_comment, PyObject *__pyx_v_quotechar, CYTHON_UNUSED PyObject *__pyx_v_expchar, PyObject *__pyx_v_formats, PyObject *__pyx_v_strip_whitespace, CYTHON_UNUSED PyObject *__pyx_v_names, PyObject *__pyx_v_include_names, PyObject *__pyx_v_exclude_names, PyObject *__pyx_v_fill_values, PyObject *__pyx_v_fill_include_names, PyObject *__pyx_v_fill_exclude_names, PyObject *__pyx_v_fast_writer) {
  PyObject *__pyx_v_use_names = NULL;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_v_fill_names = NULL;
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_v_col = NULL;
  PyObject *__pyx_v_x = NULL;
  PyObject *__pyx_v_i = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *(*__pyx_t_11)(PyObject *);
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  int __pyx_t_17;
  PyObject *__pyx_t_18 = NULL;
  __Pyx_RefNannySetupContext("__cinit__", 0);
  __Pyx_INCREF(__pyx_v_fill_values);
  __Pyx_INCREF(__pyx_v_fast_writer);

  /* "astropy/io/ascii/cparser.pyx":893
 *                   fast_writer=True):
 * 
 *         if fast_writer is True:             # <<<<<<<<<<<<<<
 *             fast_writer = {}
 *         # fast_writer might contain custom writing options
 */
  __pyx_t_1 = (__pyx_v_fast_writer == Py_True);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "astropy/io/ascii/cparser.pyx":894
 * 
 *         if fast_writer is True:
 *             fast_writer = {}             # <<<<<<<<<<<<<<
 *         # fast_writer might contain custom writing options
 * 
 */
    __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 894, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_fast_writer, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "astropy/io/ascii/cparser.pyx":893
 *                   fast_writer=True):
 * 
 *         if fast_writer is True:             # <<<<<<<<<<<<<<
 *             fast_writer = {}
 *         # fast_writer might contain custom writing options
 */
  }

  /* "astropy/io/ascii/cparser.pyx":897
 *         # fast_writer might contain custom writing options
 * 
 *         self.table = table             # <<<<<<<<<<<<<<
 *         self.comment = comment
 *         self.strip_whitespace = strip_whitespace
 */
  __Pyx_INCREF(__pyx_v_table);
  __Pyx_GIVEREF(__pyx_v_table);
  __Pyx_GOTREF(__pyx_v_self->table);
  __Pyx_DECREF(__pyx_v_self->table);
  __pyx_v_self->table = __pyx_v_table;

  /* "astropy/io/ascii/cparser.pyx":898
 * 
 *         self.table = table
 *         self.comment = comment             # <<<<<<<<<<<<<<
 *         self.strip_whitespace = strip_whitespace
 *         use_names = set(table.colnames)
 */
  __Pyx_INCREF(__pyx_v_comment);
  __Pyx_GIVEREF(__pyx_v_comment);
  __Pyx_GOTREF(__pyx_v_self->comment);
  __Pyx_DECREF(__pyx_v_self->comment);
  __pyx_v_self->comment = __pyx_v_comment;

  /* "astropy/io/ascii/cparser.pyx":899
 *         self.table = table
 *         self.comment = comment
 *         self.strip_whitespace = strip_whitespace             # <<<<<<<<<<<<<<
 *         use_names = set(table.colnames)
 * 
 */
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_strip_whitespace); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 899, __pyx_L1_error)
  __pyx_v_self->strip_whitespace = __pyx_t_4;

  /* "astropy/io/ascii/cparser.pyx":900
 *         self.comment = comment
 *         self.strip_whitespace = strip_whitespace
 *         use_names = set(table.colnames)             # <<<<<<<<<<<<<<
 * 
 *         # Apply include_names before exclude_names
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_table, __pyx_n_s_colnames); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 900, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = PySet_New(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 900, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_use_names = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "astropy/io/ascii/cparser.pyx":903
 * 
 *         # Apply include_names before exclude_names
 *         if include_names is not None:             # <<<<<<<<<<<<<<
 *             use_names.intersection_update(include_names)
 *         if exclude_names is not None:
 */
  __pyx_t_2 = (__pyx_v_include_names != Py_None);
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "astropy/io/ascii/cparser.pyx":904
 *         # Apply include_names before exclude_names
 *         if include_names is not None:
 *             use_names.intersection_update(include_names)             # <<<<<<<<<<<<<<
 *         if exclude_names is not None:
 *             use_names.difference_update(exclude_names)
 */
    __pyx_t_5 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PySet_Type_intersection_update, __pyx_v_use_names, __pyx_v_include_names); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 904, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "astropy/io/ascii/cparser.pyx":903
 * 
 *         # Apply include_names before exclude_names
 *         if include_names is not None:             # <<<<<<<<<<<<<<
 *             use_names.intersection_update(include_names)
 *         if exclude_names is not None:
 */
  }

  /* "astropy/io/ascii/cparser.pyx":905
 *         if include_names is not None:
 *             use_names.intersection_update(include_names)
 *         if exclude_names is not None:             # <<<<<<<<<<<<<<
 *             use_names.difference_update(exclude_names)
 *         # preserve column ordering via list
 */
  __pyx_t_1 = (__pyx_v_exclude_names != Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "astropy/io/ascii/cparser.pyx":906
 *             use_names.intersection_update(include_names)
 *         if exclude_names is not None:
 *             use_names.difference_update(exclude_names)             # <<<<<<<<<<<<<<
 *         # preserve column ordering via list
 *         self.use_names = [x for x in table.colnames if x in use_names]
 */
    __pyx_t_5 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PySet_Type_difference_update, __pyx_v_use_names, __pyx_v_exclude_names); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 906, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "astropy/io/ascii/cparser.pyx":905
 *         if include_names is not None:
 *             use_names.intersection_update(include_names)
 *         if exclude_names is not None:             # <<<<<<<<<<<<<<
 *             use_names.difference_update(exclude_names)
 *         # preserve column ordering via list
 */
  }

  /* "astropy/io/ascii/cparser.pyx":908
 *             use_names.difference_update(exclude_names)
 *         # preserve column ordering via list
 *         self.use_names = [x for x in table.colnames if x in use_names]             # <<<<<<<<<<<<<<
 * 
 *         fill_values = get_fill_values(fill_values, False)
 */
  __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 908, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_table, __pyx_n_s_colnames); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 908, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_6 = __pyx_t_3; __Pyx_INCREF(__pyx_t_6); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 908, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 908, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_6))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_6)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 908, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 908, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 908, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 908, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_8(__pyx_t_6);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 908, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_x, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_2 = (__Pyx_PySet_ContainsTF(__pyx_v_x, __pyx_v_use_names, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 908, __pyx_L1_error)
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_5, (PyObject*)__pyx_v_x))) __PYX_ERR(0, 908, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_GIVEREF(__pyx_t_5);
  __Pyx_GOTREF(__pyx_v_self->use_names);
  __Pyx_DECREF(__pyx_v_self->use_names);
  __pyx_v_self->use_names = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "astropy/io/ascii/cparser.pyx":910
 *         self.use_names = [x for x in table.colnames if x in use_names]
 * 
 *         fill_values = get_fill_values(fill_values, False)             # <<<<<<<<<<<<<<
 *         self.fill_values = fill_values.copy()
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_get_fill_values); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 910, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_fill_values, Py_False};
    __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 910, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_5);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_fill_values, Py_False};
    __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 910, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_5);
  } else
  #endif
  {
    __pyx_t_9 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 910, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_fill_values);
    __Pyx_GIVEREF(__pyx_v_fill_values);
    PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_4, __pyx_v_fill_values);
    __Pyx_INCREF(Py_False);
    __Pyx_GIVEREF(Py_False);
    PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_4, Py_False);
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_9, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 910, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF_SET(__pyx_v_fill_values, __pyx_t_5);
  __pyx_t_5 = 0;

  /* "astropy/io/ascii/cparser.pyx":911
 * 
 *         fill_values = get_fill_values(fill_values, False)
 *         self.fill_values = fill_values.copy()             # <<<<<<<<<<<<<<
 * 
 *         # Add int/float versions of each fill value (if applicable)
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_fill_values, __pyx_n_s_copy); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 911, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_9 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_5 = (__pyx_t_9) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_9) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 911, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (!(likely(PyDict_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_5)->tp_name), 0))) __PYX_ERR(0, 911, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_5);
  __Pyx_GOTREF(__pyx_v_self->fill_values);
  __Pyx_DECREF(__pyx_v_self->fill_values);
  __pyx_v_self->fill_values = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "astropy/io/ascii/cparser.pyx":917
 *         # to call unicode() on every value, which is a major
 *         # performance hit.
 *         for key, val in six.iteritems(fill_values):             # <<<<<<<<<<<<<<
 *             try:
 *                 self.fill_values[int(key)] = val
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_six); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_iteritems); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_9);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_9, function);
    }
  }
  __pyx_t_5 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_6, __pyx_v_fill_values) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_v_fill_values);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
    __pyx_t_9 = __pyx_t_5; __Pyx_INCREF(__pyx_t_9); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_9 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 917, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_8 = Py_TYPE(__pyx_t_9)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 917, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_9))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_9)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_9, __pyx_t_7); __Pyx_INCREF(__pyx_t_5); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 917, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_9, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 917, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_9)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_9, __pyx_t_7); __Pyx_INCREF(__pyx_t_5); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 917, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_9, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 917, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_8(__pyx_t_9);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 917, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_5))) || (PyList_CheckExact(__pyx_t_5))) {
      PyObject* sequence = __pyx_t_5;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 917, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_6 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_3);
      #else
      __pyx_t_6 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 917, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 917, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      #endif
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_10 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 917, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_11 = Py_TYPE(__pyx_t_10)->tp_iternext;
      index = 0; __pyx_t_6 = __pyx_t_11(__pyx_t_10); if (unlikely(!__pyx_t_6)) goto __pyx_L11_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      index = 1; __pyx_t_3 = __pyx_t_11(__pyx_t_10); if (unlikely(!__pyx_t_3)) goto __pyx_L11_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_3);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_10), 2) < 0) __PYX_ERR(0, 917, __pyx_L1_error)
      __pyx_t_11 = NULL;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L12_unpacking_done;
      __pyx_L11_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_11 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 917, __pyx_L1_error)
      __pyx_L12_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "astropy/io/ascii/cparser.pyx":918
 *         # performance hit.
 *         for key, val in six.iteritems(fill_values):
 *             try:             # <<<<<<<<<<<<<<
 *                 self.fill_values[int(key)] = val
 *                 self.fill_values[float(key)] = val
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_14);
      /*try:*/ {

        /* "astropy/io/ascii/cparser.pyx":919
 *         for key, val in six.iteritems(fill_values):
 *             try:
 *                 self.fill_values[int(key)] = val             # <<<<<<<<<<<<<<
 *                 self.fill_values[float(key)] = val
 *             except (ValueError, np.ma.MaskError):
 */
        if (unlikely(__pyx_v_self->fill_values == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 919, __pyx_L13_error)
        }
        __pyx_t_5 = __Pyx_PyNumber_Int(__pyx_v_key); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 919, __pyx_L13_error)
        __Pyx_GOTREF(__pyx_t_5);
        if (unlikely(PyDict_SetItem(__pyx_v_self->fill_values, __pyx_t_5, __pyx_v_val) < 0)) __PYX_ERR(0, 919, __pyx_L13_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "astropy/io/ascii/cparser.pyx":920
 *             try:
 *                 self.fill_values[int(key)] = val
 *                 self.fill_values[float(key)] = val             # <<<<<<<<<<<<<<
 *             except (ValueError, np.ma.MaskError):
 *                 pass
 */
        if (unlikely(__pyx_v_self->fill_values == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 920, __pyx_L13_error)
        }
        __pyx_t_5 = __Pyx_PyNumber_Float(__pyx_v_key); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 920, __pyx_L13_error)
        __Pyx_GOTREF(__pyx_t_5);
        if (unlikely(PyDict_SetItem(__pyx_v_self->fill_values, __pyx_t_5, __pyx_v_val) < 0)) __PYX_ERR(0, 920, __pyx_L13_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "astropy/io/ascii/cparser.pyx":918
 *         # performance hit.
 *         for key, val in six.iteritems(fill_values):
 *             try:             # <<<<<<<<<<<<<<
 *                 self.fill_values[int(key)] = val
 *                 self.fill_values[float(key)] = val
 */
      }
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      goto __pyx_L20_try_end;
      __pyx_L13_error:;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "astropy/io/ascii/cparser.pyx":921
 *                 self.fill_values[int(key)] = val
 *                 self.fill_values[float(key)] = val
 *             except (ValueError, np.ma.MaskError):             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
      __Pyx_ErrFetch(&__pyx_t_5, &__pyx_t_3, &__pyx_t_6);
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_np); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 921, __pyx_L15_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_ma); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 921, __pyx_L15_except_error)
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_MaskError); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 921, __pyx_L15_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      __pyx_t_4 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_5, __pyx_builtin_ValueError) || __Pyx_PyErr_GivenExceptionMatches(__pyx_t_5, __pyx_t_10);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_ErrRestore(__pyx_t_5, __pyx_t_3, __pyx_t_6);
      __pyx_t_5 = 0; __pyx_t_3 = 0; __pyx_t_6 = 0;
      if (__pyx_t_4) {
        __Pyx_ErrRestore(0,0,0);
        goto __pyx_L14_exception_handled;
      }
      goto __pyx_L15_except_error;
      __pyx_L15_except_error:;

      /* "astropy/io/ascii/cparser.pyx":918
 *         # performance hit.
 *         for key, val in six.iteritems(fill_values):
 *             try:             # <<<<<<<<<<<<<<
 *                 self.fill_values[int(key)] = val
 *                 self.fill_values[float(key)] = val
 */
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_XGIVEREF(__pyx_t_14);
      __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
      goto __pyx_L1_error;
      __pyx_L14_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_XGIVEREF(__pyx_t_14);
      __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
      __pyx_L20_try_end:;
    }

    /* "astropy/io/ascii/cparser.pyx":917
 *         # to call unicode() on every value, which is a major
 *         # performance hit.
 *         for key, val in six.iteritems(fill_values):             # <<<<<<<<<<<<<<
 *             try:
 *                 self.fill_values[int(key)] = val
 */
  }
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "astropy/io/ascii/cparser.pyx":924
 *                 pass
 * 
 *         fill_names = set(self.use_names)             # <<<<<<<<<<<<<<
 *         # Apply fill_include_names before fill_exclude_names
 *         if fill_include_names is not None:
 */
  __pyx_t_9 = PySet_New(__pyx_v_self->use_names); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 924, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_v_fill_names = ((PyObject*)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "astropy/io/ascii/cparser.pyx":926
 *         fill_names = set(self.use_names)
 *         # Apply fill_include_names before fill_exclude_names
 *         if fill_include_names is not None:             # <<<<<<<<<<<<<<
 *             fill_names.intersection_update(fill_include_names)
 *         if fill_exclude_names is not None:
 */
  __pyx_t_1 = (__pyx_v_fill_include_names != Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "astropy/io/ascii/cparser.pyx":927
 *         # Apply fill_include_names before fill_exclude_names
 *         if fill_include_names is not None:
 *             fill_names.intersection_update(fill_include_names)             # <<<<<<<<<<<<<<
 *         if fill_exclude_names is not None:
 *             fill_names.difference_update(fill_exclude_names)
 */
    __pyx_t_9 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PySet_Type_intersection_update, __pyx_v_fill_names, __pyx_v_fill_include_names); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 927, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "astropy/io/ascii/cparser.pyx":926
 *         fill_names = set(self.use_names)
 *         # Apply fill_include_names before fill_exclude_names
 *         if fill_include_names is not None:             # <<<<<<<<<<<<<<
 *             fill_names.intersection_update(fill_include_names)
 *         if fill_exclude_names is not None:
 */
  }

  /* "astropy/io/ascii/cparser.pyx":928
 *         if fill_include_names is not None:
 *             fill_names.intersection_update(fill_include_names)
 *         if fill_exclude_names is not None:             # <<<<<<<<<<<<<<
 *             fill_names.difference_update(fill_exclude_names)
 *         # Preserve column ordering
 */
  __pyx_t_2 = (__pyx_v_fill_exclude_names != Py_None);
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "astropy/io/ascii/cparser.pyx":929
 *             fill_names.intersection_update(fill_include_names)
 *         if fill_exclude_names is not None:
 *             fill_names.difference_update(fill_exclude_names)             # <<<<<<<<<<<<<<
 *         # Preserve column ordering
 *         self.fill_cols = set([i for i, name in enumerate(self.use_names) if
 */
    __pyx_t_9 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PySet_Type_difference_update, __pyx_v_fill_names, __pyx_v_fill_exclude_names); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 929, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "astropy/io/ascii/cparser.pyx":928
 *         if fill_include_names is not None:
 *             fill_names.intersection_update(fill_include_names)
 *         if fill_exclude_names is not None:             # <<<<<<<<<<<<<<
 *             fill_names.difference_update(fill_exclude_names)
 *         # Preserve column ordering
 */
  }

  /* "astropy/io/ascii/cparser.pyx":931
 *             fill_names.difference_update(fill_exclude_names)
 *         # Preserve column ordering
 *         self.fill_cols = set([i for i, name in enumerate(self.use_names) if             # <<<<<<<<<<<<<<
 *                               name in fill_names])
 * 
 */
  __pyx_t_9 = PyList_New(0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 931, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_6 = __pyx_int_0;
  __pyx_t_3 = __pyx_v_self->use_names; __Pyx_INCREF(__pyx_t_3); __pyx_t_7 = 0;
  for (;;) {
    if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_3)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_5 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_5); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 931, __pyx_L1_error)
    #else
    __pyx_t_5 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 931, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_t_6);
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_6);
    __pyx_t_5 = __Pyx_PyInt_AddObjC(__pyx_t_6, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 931, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6);
    __pyx_t_6 = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "astropy/io/ascii/cparser.pyx":932
 *         # Preserve column ordering
 *         self.fill_cols = set([i for i, name in enumerate(self.use_names) if
 *                               name in fill_names])             # <<<<<<<<<<<<<<
 * 
 *         # formats in user-specified dict should override
 */
    __pyx_t_1 = (__Pyx_PySet_ContainsTF(__pyx_v_name, __pyx_v_fill_names, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 932, __pyx_L1_error)
    __pyx_t_2 = (__pyx_t_1 != 0);

    /* "astropy/io/ascii/cparser.pyx":931
 *             fill_names.difference_update(fill_exclude_names)
 *         # Preserve column ordering
 *         self.fill_cols = set([i for i, name in enumerate(self.use_names) if             # <<<<<<<<<<<<<<
 *                               name in fill_names])
 * 
 */
    if (__pyx_t_2) {
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_9, (PyObject*)__pyx_v_i))) __PYX_ERR(0, 931, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = PySet_New(__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 931, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_GIVEREF(__pyx_t_6);
  __Pyx_GOTREF(__pyx_v_self->fill_cols);
  __Pyx_DECREF(__pyx_v_self->fill_cols);
  __pyx_v_self->fill_cols = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "astropy/io/ascii/cparser.pyx":936
 *         # formats in user-specified dict should override
 *         # existing column formats
 *         if formats is not None:             # <<<<<<<<<<<<<<
 *             for name in self.use_names:
 *                 if name in formats:
 */
  __pyx_t_2 = (__pyx_v_formats != Py_None);
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "astropy/io/ascii/cparser.pyx":937
 *         # existing column formats
 *         if formats is not None:
 *             for name in self.use_names:             # <<<<<<<<<<<<<<
 *                 if name in formats:
 *                     self.table[name].format = formats[name]
 */
    if (unlikely(__pyx_v_self->use_names == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 937, __pyx_L1_error)
    }
    __pyx_t_6 = __pyx_v_self->use_names; __Pyx_INCREF(__pyx_t_6); __pyx_t_7 = 0;
    for (;;) {
      if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_6)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_9 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 937, __pyx_L1_error)
      #else
      __pyx_t_9 = PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 937, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_9);
      __pyx_t_9 = 0;

      /* "astropy/io/ascii/cparser.pyx":938
 *         if formats is not None:
 *             for name in self.use_names:
 *                 if name in formats:             # <<<<<<<<<<<<<<
 *                     self.table[name].format = formats[name]
 * 
 */
      __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_v_name, __pyx_v_formats, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 938, __pyx_L1_error)
      __pyx_t_2 = (__pyx_t_1 != 0);
      if (__pyx_t_2) {

        /* "astropy/io/ascii/cparser.pyx":939
 *             for name in self.use_names:
 *                 if name in formats:
 *                     self.table[name].format = formats[name]             # <<<<<<<<<<<<<<
 * 
 *         self.col_iters = []
 */
        __pyx_t_9 = __Pyx_PyObject_GetItem(__pyx_v_formats, __pyx_v_name); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 939, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_self->table, __pyx_v_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 939, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (__Pyx_PyObject_SetAttrStr(__pyx_t_3, __pyx_n_s_format, __pyx_t_9) < 0) __PYX_ERR(0, 939, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "astropy/io/ascii/cparser.pyx":938
 *         if formats is not None:
 *             for name in self.use_names:
 *                 if name in formats:             # <<<<<<<<<<<<<<
 *                     self.table[name].format = formats[name]
 * 
 */
      }

      /* "astropy/io/ascii/cparser.pyx":937
 *         # existing column formats
 *         if formats is not None:
 *             for name in self.use_names:             # <<<<<<<<<<<<<<
 *                 if name in formats:
 *                     self.table[name].format = formats[name]
 */
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "astropy/io/ascii/cparser.pyx":936
 *         # formats in user-specified dict should override
 *         # existing column formats
 *         if formats is not None:             # <<<<<<<<<<<<<<
 *             for name in self.use_names:
 *                 if name in formats:
 */
  }

  /* "astropy/io/ascii/cparser.pyx":941
 *                     self.table[name].format = formats[name]
 * 
 *         self.col_iters = []             # <<<<<<<<<<<<<<
 *         self.formats = []
 *         self.format_funcs = []
 */
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 941, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  __Pyx_GOTREF(__pyx_v_self->col_iters);
  __Pyx_DECREF(__pyx_v_self->col_iters);
  __pyx_v_self->col_iters = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "astropy/io/ascii/cparser.pyx":942
 * 
 *         self.col_iters = []
 *         self.formats = []             # <<<<<<<<<<<<<<
 *         self.format_funcs = []
 *         self.line_comments = table.meta.get('comments', [])
 */
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  __Pyx_GOTREF(__pyx_v_self->formats);
  __Pyx_DECREF(__pyx_v_self->formats);
  __pyx_v_self->formats = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "astropy/io/ascii/cparser.pyx":943
 *         self.col_iters = []
 *         self.formats = []
 *         self.format_funcs = []             # <<<<<<<<<<<<<<
 *         self.line_comments = table.meta.get('comments', [])
 * 
 */
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 943, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  __Pyx_GOTREF(__pyx_v_self->format_funcs);
  __Pyx_DECREF(__pyx_v_self->format_funcs);
  __pyx_v_self->format_funcs = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "astropy/io/ascii/cparser.pyx":944
 *         self.formats = []
 *         self.format_funcs = []
 *         self.line_comments = table.meta.get('comments', [])             # <<<<<<<<<<<<<<
 * 
 *         for col in six.itervalues(table.columns):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_table, __pyx_n_s_meta); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 944, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_get); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 944, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 944, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_9);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_9, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_9)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_n_s_comments, __pyx_t_3};
    __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 944, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_9)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_n_s_comments, __pyx_t_3};
    __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 944, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  {
    __pyx_t_10 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 944, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_n_s_comments);
    __Pyx_GIVEREF(__pyx_n_s_comments);
    PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_4, __pyx_n_s_comments);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_4, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_10, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 944, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  }
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (!(likely(PyList_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_6)->tp_name), 0))) __PYX_ERR(0, 944, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_6);
  __Pyx_GOTREF(__pyx_v_self->line_comments);
  __Pyx_DECREF(__pyx_v_self->line_comments);
  __pyx_v_self->line_comments = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "astropy/io/ascii/cparser.pyx":946
 *         self.line_comments = table.meta.get('comments', [])
 * 
 *         for col in six.itervalues(table.columns):             # <<<<<<<<<<<<<<
 *             if col.name in self.use_names: # iterate over included columns
 *                 # If col.format is None then don't use any formatter to improve
 */
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_six); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_itervalues); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_table, __pyx_n_s_columns); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
    }
  }
  __pyx_t_6 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_3, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_9);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
    __pyx_t_10 = __pyx_t_6; __Pyx_INCREF(__pyx_t_10); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_10 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 946, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_8 = Py_TYPE(__pyx_t_10)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 946, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_10))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_10)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_6 = PyList_GET_ITEM(__pyx_t_10, __pyx_t_7); __Pyx_INCREF(__pyx_t_6); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 946, __pyx_L1_error)
        #else
        __pyx_t_6 = PySequence_ITEM(__pyx_t_10, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 946, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_10)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_10, __pyx_t_7); __Pyx_INCREF(__pyx_t_6); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 946, __pyx_L1_error)
        #else
        __pyx_t_6 = PySequence_ITEM(__pyx_t_10, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 946, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
      }
    } else {
      __pyx_t_6 = __pyx_t_8(__pyx_t_10);
      if (unlikely(!__pyx_t_6)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 946, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_6);
    }
    __Pyx_XDECREF_SET(__pyx_v_col, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "astropy/io/ascii/cparser.pyx":947
 * 
 *         for col in six.itervalues(table.columns):
 *             if col.name in self.use_names: # iterate over included columns             # <<<<<<<<<<<<<<
 *                 # If col.format is None then don't use any formatter to improve
 *                 # speed.  However, if the column is a byte string and this
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_col, __pyx_n_s_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 947, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_6, __pyx_v_self->use_names, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 947, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "astropy/io/ascii/cparser.pyx":952
 *                 # is Py3, then use the default formatter (which in this case
 *                 # does val.decode('utf-8')) in order to avoid a leading 'b'.
 *                 if col.format is None and not (six.PY3 and col.dtype.kind == 'S'):             # <<<<<<<<<<<<<<
 *                     self.format_funcs.append(None)
 *                 else:
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_col, __pyx_n_s_format); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 952, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_2 = (__pyx_t_6 == Py_None);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_16 = (__pyx_t_2 != 0);
      if (__pyx_t_16) {
      } else {
        __pyx_t_1 = __pyx_t_16;
        goto __pyx_L34_bool_binop_done;
      }
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_six); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 952, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_PY3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 952, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 952, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (__pyx_t_2) {
      } else {
        __pyx_t_16 = __pyx_t_2;
        goto __pyx_L36_bool_binop_done;
      }
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_col, __pyx_n_s_dtype); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 952, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_kind); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 952, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_t_6, __pyx_n_s_S, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 952, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_16 = __pyx_t_2;
      __pyx_L36_bool_binop_done:;
      __pyx_t_2 = ((!__pyx_t_16) != 0);
      __pyx_t_1 = __pyx_t_2;
      __pyx_L34_bool_binop_done:;
      if (__pyx_t_1) {

        /* "astropy/io/ascii/cparser.pyx":953
 *                 # does val.decode('utf-8')) in order to avoid a leading 'b'.
 *                 if col.format is None and not (six.PY3 and col.dtype.kind == 'S'):
 *                     self.format_funcs.append(None)             # <<<<<<<<<<<<<<
 *                 else:
 *                     self.format_funcs.append(col.info._format_funcs.get(
 */
        if (unlikely(__pyx_v_self->format_funcs == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
          __PYX_ERR(0, 953, __pyx_L1_error)
        }
        __pyx_t_17 = __Pyx_PyList_Append(__pyx_v_self->format_funcs, Py_None); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 953, __pyx_L1_error)

        /* "astropy/io/ascii/cparser.pyx":952
 *                 # is Py3, then use the default formatter (which in this case
 *                 # does val.decode('utf-8')) in order to avoid a leading 'b'.
 *                 if col.format is None and not (six.PY3 and col.dtype.kind == 'S'):             # <<<<<<<<<<<<<<
 *                     self.format_funcs.append(None)
 *                 else:
 */
        goto __pyx_L33;
      }

      /* "astropy/io/ascii/cparser.pyx":955
 *                     self.format_funcs.append(None)
 *                 else:
 *                     self.format_funcs.append(col.info._format_funcs.get(             # <<<<<<<<<<<<<<
 *                         col.format, pprint.get_auto_format_func(col)))
 *                 # col is a numpy.ndarray, so we convert it to
 */
      /*else*/ {
        if (unlikely(__pyx_v_self->format_funcs == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
          __PYX_ERR(0, 955, __pyx_L1_error)
        }
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_col, __pyx_n_s_info); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 955, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_format_funcs); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 955, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_get); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 955, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "astropy/io/ascii/cparser.pyx":956
 *                 else:
 *                     self.format_funcs.append(col.info._format_funcs.get(
 *                         col.format, pprint.get_auto_format_func(col)))             # <<<<<<<<<<<<<<
 *                 # col is a numpy.ndarray, so we convert it to
 *                 # an ordinary list because csv.writer will call
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_col, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 956, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_pprint); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 956, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_get_auto_format_func); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 956, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_18);
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        __pyx_t_15 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_18))) {
          __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_18);
          if (likely(__pyx_t_15)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_18);
            __Pyx_INCREF(__pyx_t_15);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_18, function);
          }
        }
        __pyx_t_5 = (__pyx_t_15) ? __Pyx_PyObject_Call2Args(__pyx_t_18, __pyx_t_15, __pyx_v_col) : __Pyx_PyObject_CallOneArg(__pyx_t_18, __pyx_v_col);
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 956, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
        __pyx_t_18 = NULL;
        __pyx_t_4 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
          __pyx_t_18 = PyMethod_GET_SELF(__pyx_t_9);
          if (likely(__pyx_t_18)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
            __Pyx_INCREF(__pyx_t_18);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_9, function);
            __pyx_t_4 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_9)) {
          PyObject *__pyx_temp[3] = {__pyx_t_18, __pyx_t_3, __pyx_t_5};
          __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 955, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_9)) {
          PyObject *__pyx_temp[3] = {__pyx_t_18, __pyx_t_3, __pyx_t_5};
          __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 955, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        } else
        #endif
        {
          __pyx_t_15 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 955, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_15);
          if (__pyx_t_18) {
            __Pyx_GIVEREF(__pyx_t_18); PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_18); __pyx_t_18 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_15, 0+__pyx_t_4, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_5);
          PyTuple_SET_ITEM(__pyx_t_15, 1+__pyx_t_4, __pyx_t_5);
          __pyx_t_3 = 0;
          __pyx_t_5 = 0;
          __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_15, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 955, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        }
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "astropy/io/ascii/cparser.pyx":955
 *                     self.format_funcs.append(None)
 *                 else:
 *                     self.format_funcs.append(col.info._format_funcs.get(             # <<<<<<<<<<<<<<
 *                         col.format, pprint.get_auto_format_func(col)))
 *                 # col is a numpy.ndarray, so we convert it to
 */
        __pyx_t_17 = __Pyx_PyList_Append(__pyx_v_self->format_funcs, __pyx_t_6); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 955, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __pyx_L33:;

      /* "astropy/io/ascii/cparser.pyx":961
 *                 # np.array_str() on each numpy value, which is
 *                 # very inefficient
 *                 self.col_iters.append(iter(col.tolist()))             # <<<<<<<<<<<<<<
 *                 self.formats.append(col.format)
 * 
 */
      if (unlikely(__pyx_v_self->col_iters == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
        __PYX_ERR(0, 961, __pyx_L1_error)
      }
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_col, __pyx_n_s_tolist); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 961, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_15 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_15)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_15);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
        }
      }
      __pyx_t_6 = (__pyx_t_15) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_15) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 961, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 961, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_17 = __Pyx_PyList_Append(__pyx_v_self->col_iters, __pyx_t_9); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 961, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "astropy/io/ascii/cparser.pyx":962
 *                 # very inefficient
 *                 self.col_iters.append(iter(col.tolist()))
 *                 self.formats.append(col.format)             # <<<<<<<<<<<<<<
 * 
 *         self.quotechar = None if quotechar is None else str(quotechar)
 */
      if (unlikely(__pyx_v_self->formats == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
        __PYX_ERR(0, 962, __pyx_L1_error)
      }
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_col, __pyx_n_s_format); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 962, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_17 = __Pyx_PyList_Append(__pyx_v_self->formats, __pyx_t_9); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 962, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "astropy/io/ascii/cparser.pyx":947
 * 
 *         for col in six.itervalues(table.columns):
 *             if col.name in self.use_names: # iterate over included columns             # <<<<<<<<<<<<<<
 *                 # If col.format is None then don't use any formatter to improve
 *                 # speed.  However, if the column is a byte string and this
 */
    }

    /* "astropy/io/ascii/cparser.pyx":946
 *         self.line_comments = table.meta.get('comments', [])
 * 
 *         for col in six.itervalues(table.columns):             # <<<<<<<<<<<<<<
 *             if col.name in self.use_names: # iterate over included columns
 *                 # If col.format is None then don't use any formatter to improve
 */
  }
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "astropy/io/ascii/cparser.pyx":964
 *                 self.formats.append(col.format)
 * 
 *         self.quotechar = None if quotechar is None else str(quotechar)             # <<<<<<<<<<<<<<
 *         self.delimiter = ' ' if delimiter is None else str(delimiter)
 *         # 'S' for string types, 'N' for numeric types
 */
  __pyx_t_1 = (__pyx_v_quotechar == Py_None);
  if ((__pyx_t_1 != 0)) {
    __Pyx_INCREF(Py_None);
    __pyx_t_10 = Py_None;
  } else {
    __pyx_t_9 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_v_quotechar); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 964, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (!(likely(PyString_CheckExact(__pyx_t_9))||((__pyx_t_9) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_9)->tp_name), 0))) __PYX_ERR(0, 964, __pyx_L1_error)
    __pyx_t_10 = __pyx_t_9;
    __pyx_t_9 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_10);
  __Pyx_GOTREF(__pyx_v_self->quotechar);
  __Pyx_DECREF(__pyx_v_self->quotechar);
  __pyx_v_self->quotechar = ((PyObject*)__pyx_t_10);
  __pyx_t_10 = 0;

  /* "astropy/io/ascii/cparser.pyx":965
 * 
 *         self.quotechar = None if quotechar is None else str(quotechar)
 *         self.delimiter = ' ' if delimiter is None else str(delimiter)             # <<<<<<<<<<<<<<
 *         # 'S' for string types, 'N' for numeric types
 *         self.types = ['S' if self.table[name].dtype.kind in ('S', 'U') else 'N'
 */
  __pyx_t_1 = (__pyx_v_delimiter == Py_None);
  if ((__pyx_t_1 != 0)) {
    __Pyx_INCREF(__pyx_kp_s__22);
    __pyx_t_10 = __pyx_kp_s__22;
  } else {
    __pyx_t_9 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_v_delimiter); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 965, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (!(likely(PyString_CheckExact(__pyx_t_9))||((__pyx_t_9) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_9)->tp_name), 0))) __PYX_ERR(0, 965, __pyx_L1_error)
    __pyx_t_10 = __pyx_t_9;
    __pyx_t_9 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_10);
  __Pyx_GOTREF(__pyx_v_self->delimiter);
  __Pyx_DECREF(__pyx_v_self->delimiter);
  __pyx_v_self->delimiter = ((PyObject*)__pyx_t_10);
  __pyx_t_10 = 0;

  /* "astropy/io/ascii/cparser.pyx":967
 *         self.delimiter = ' ' if delimiter is None else str(delimiter)
 *         # 'S' for string types, 'N' for numeric types
 *         self.types = ['S' if self.table[name].dtype.kind in ('S', 'U') else 'N'             # <<<<<<<<<<<<<<
 *                       for name in self.use_names]
 * 
 */
  __pyx_t_10 = PyList_New(0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 967, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);

  /* "astropy/io/ascii/cparser.pyx":968
 *         # 'S' for string types, 'N' for numeric types
 *         self.types = ['S' if self.table[name].dtype.kind in ('S', 'U') else 'N'
 *                       for name in self.use_names]             # <<<<<<<<<<<<<<
 * 
 *     cdef _write_comments(self, output):
 */
  if (unlikely(__pyx_v_self->use_names == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 968, __pyx_L1_error)
  }
  __pyx_t_9 = __pyx_v_self->use_names; __Pyx_INCREF(__pyx_t_9); __pyx_t_7 = 0;
  for (;;) {
    if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_9)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_6 = PyList_GET_ITEM(__pyx_t_9, __pyx_t_7); __Pyx_INCREF(__pyx_t_6); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 968, __pyx_L1_error)
    #else
    __pyx_t_6 = PySequence_ITEM(__pyx_t_9, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 968, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "astropy/io/ascii/cparser.pyx":967
 *         self.delimiter = ' ' if delimiter is None else str(delimiter)
 *         # 'S' for string types, 'N' for numeric types
 *         self.types = ['S' if self.table[name].dtype.kind in ('S', 'U') else 'N'             # <<<<<<<<<<<<<<
 *                       for name in self.use_names]
 * 
 */
    __pyx_t_15 = __Pyx_PyObject_GetItem(__pyx_v_self->table, __pyx_v_name); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 967, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_dtype); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 967, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_kind); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 967, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_t_15, __pyx_n_s_S, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 967, __pyx_L1_error)
    if (!__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L40_bool_binop_done;
    }
    __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_t_15, __pyx_n_s_U, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 967, __pyx_L1_error)
    __pyx_t_1 = __pyx_t_2;
    __pyx_L40_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    if ((__pyx_t_1 != 0)) {
      __Pyx_INCREF(__pyx_n_s_S);
      __pyx_t_6 = __pyx_n_s_S;
    } else {
      __Pyx_INCREF(__pyx_n_s_N);
      __pyx_t_6 = __pyx_n_s_N;
    }
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_10, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 967, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "astropy/io/ascii/cparser.pyx":968
 *         # 'S' for string types, 'N' for numeric types
 *         self.types = ['S' if self.table[name].dtype.kind in ('S', 'U') else 'N'
 *                       for name in self.use_names]             # <<<<<<<<<<<<<<
 * 
 *     cdef _write_comments(self, output):
 */
  }
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "astropy/io/ascii/cparser.pyx":967
 *         self.delimiter = ' ' if delimiter is None else str(delimiter)
 *         # 'S' for string types, 'N' for numeric types
 *         self.types = ['S' if self.table[name].dtype.kind in ('S', 'U') else 'N'             # <<<<<<<<<<<<<<
 *                       for name in self.use_names]
 * 
 */
  __Pyx_GIVEREF(__pyx_t_10);
  __Pyx_GOTREF(__pyx_v_self->types);
  __Pyx_DECREF(__pyx_v_self->types);
  __pyx_v_self->types = ((PyObject*)__pyx_t_10);
  __pyx_t_10 = 0;

  /* "astropy/io/ascii/cparser.pyx":878
 *         object comment
 * 
 *     def __cinit__(self, table,             # <<<<<<<<<<<<<<
 *                   delimiter=',',
 *                   comment='# ',
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_AddTraceback("astropy.io.ascii.cparser.FastWriter.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_use_names);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XDECREF(__pyx_v_fill_names);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_col);
  __Pyx_XDECREF(__pyx_v_x);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_fill_values);
  __Pyx_XDECREF(__pyx_v_fast_writer);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "astropy/io/ascii/cparser.pyx":970
 *                       for name in self.use_names]
 * 
 *     cdef _write_comments(self, output):             # <<<<<<<<<<<<<<
 *         if self.comment is not False:
 *             for comment_line in self.line_comments:
 */

static PyObject *__pyx_f_7astropy_2io_5ascii_7cparser_10FastWriter__write_comments(struct __pyx_obj_7astropy_2io_5ascii_7cparser_FastWriter *__pyx_v_self, PyObject *__pyx_v_output) {
  PyObject *__pyx_v_comment_line = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("_write_comments", 0);

  /* "astropy/io/ascii/cparser.pyx":971
 * 
 *     cdef _write_comments(self, output):
 *         if self.comment is not False:             # <<<<<<<<<<<<<<
 *             for comment_line in self.line_comments:
 *                 output.write(self.comment + comment_line + '\n')
 */
  __pyx_t_1 = (__pyx_v_self->comment != Py_False);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "astropy/io/ascii/cparser.pyx":972
 *     cdef _write_comments(self, output):
 *         if self.comment is not False:
 *             for comment_line in self.line_comments:             # <<<<<<<<<<<<<<
 *                 output.write(self.comment + comment_line + '\n')
 * 
 */
    if (unlikely(__pyx_v_self->line_comments == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 972, __pyx_L1_error)
    }
    __pyx_t_3 = __pyx_v_self->line_comments; __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;
    for (;;) {
      if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_5 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_5); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 972, __pyx_L1_error)
      #else
      __pyx_t_5 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 972, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      __Pyx_XDECREF_SET(__pyx_v_comment_line, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "astropy/io/ascii/cparser.pyx":973
 *         if self.comment is not False:
 *             for comment_line in self.line_comments:
 *                 output.write(self.comment + comment_line + '\n')             # <<<<<<<<<<<<<<
 * 
 *     def _write_header(self, output, writer, header_output, output_types):
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_output, __pyx_n_s_write); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 973, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = PyNumber_Add(__pyx_v_self->comment, __pyx_v_comment_line); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 973, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = PyNumber_Add(__pyx_t_7, __pyx_kp_s__13); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 973, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_5 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_7, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 973, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "astropy/io/ascii/cparser.pyx":972
 *     cdef _write_comments(self, output):
 *         if self.comment is not False:
 *             for comment_line in self.line_comments:             # <<<<<<<<<<<<<<
 *                 output.write(self.comment + comment_line + '\n')
 * 
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "astropy/io/ascii/cparser.pyx":971
 * 
 *     cdef _write_comments(self, output):
 *         if self.comment is not False:             # <<<<<<<<<<<<<<
 *             for comment_line in self.line_comments:
 *                 output.write(self.comment + comment_line + '\n')
 */
  }

  /* "astropy/io/ascii/cparser.pyx":970
 *                       for name in self.use_names]
 * 
 *     cdef _write_comments(self, output):             # <<<<<<<<<<<<<<
 *         if self.comment is not False:
 *             for comment_line in self.line_comments:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("astropy.io.ascii.cparser.FastWriter._write_comments", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_comment_line);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "astropy/io/ascii/cparser.pyx":975
 *                 output.write(self.comment + comment_line + '\n')
 * 
 *     def _write_header(self, output, writer, header_output, output_types):             # <<<<<<<<<<<<<<
 *         if header_output is not None and header_output == 'comment':
 *             output.write(self.comment)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_10FastWriter_3_write_header(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_10FastWriter_3_write_header(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_output = 0;
  PyObject *__pyx_v_writer = 0;
  PyObject *__pyx_v_header_output = 0;
  PyObject *__pyx_v_output_types = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_write_header (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_output,&__pyx_n_s_writer,&__pyx_n_s_header_output,&__pyx_n_s_output_types,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_output)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_writer)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_write_header", 1, 4, 4, 1); __PYX_ERR(0, 975, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_header_output)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_write_header", 1, 4, 4, 2); __PYX_ERR(0, 975, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_output_types)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_write_header", 1, 4, 4, 3); __PYX_ERR(0, 975, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_write_header") < 0)) __PYX_ERR(0, 975, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_output = values[0];
    __pyx_v_writer = values[1];
    __pyx_v_header_output = values[2];
    __pyx_v_output_types = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_write_header", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 975, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("astropy.io.ascii.cparser.FastWriter._write_header", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7astropy_2io_5ascii_7cparser_10FastWriter_2_write_header(((struct __pyx_obj_7astropy_2io_5ascii_7cparser_FastWriter *)__pyx_v_self), __pyx_v_output, __pyx_v_writer, __pyx_v_header_output, __pyx_v_output_types);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_10FastWriter_2_write_header(struct __pyx_obj_7astropy_2io_5ascii_7cparser_FastWriter *__pyx_v_self, PyObject *__pyx_v_output, PyObject *__pyx_v_writer, PyObject *__pyx_v_header_output, PyObject *__pyx_v_output_types) {
  PyObject *__pyx_v_x = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  __Pyx_RefNannySetupContext("_write_header", 0);

  /* "astropy/io/ascii/cparser.pyx":976
 * 
 *     def _write_header(self, output, writer, header_output, output_types):
 *         if header_output is not None and header_output == 'comment':             # <<<<<<<<<<<<<<
 *             output.write(self.comment)
 *             writer.writerow([x.strip() for x in self.use_names] if
 */
  __pyx_t_2 = (__pyx_v_header_output != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_v_header_output, __pyx_n_s_comment, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 976, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "astropy/io/ascii/cparser.pyx":977
 *     def _write_header(self, output, writer, header_output, output_types):
 *         if header_output is not None and header_output == 'comment':
 *             output.write(self.comment)             # <<<<<<<<<<<<<<
 *             writer.writerow([x.strip() for x in self.use_names] if
 *                             self.strip_whitespace else self.use_names)
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_output, __pyx_n_s_write); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 977, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_v_self->comment) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_self->comment);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 977, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "astropy/io/ascii/cparser.pyx":978
 *         if header_output is not None and header_output == 'comment':
 *             output.write(self.comment)
 *             writer.writerow([x.strip() for x in self.use_names] if             # <<<<<<<<<<<<<<
 *                             self.strip_whitespace else self.use_names)
 *             self._write_comments(output)
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_writerow); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 978, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "astropy/io/ascii/cparser.pyx":979
 *             output.write(self.comment)
 *             writer.writerow([x.strip() for x in self.use_names] if
 *                             self.strip_whitespace else self.use_names)             # <<<<<<<<<<<<<<
 *             self._write_comments(output)
 *         else:
 */
    if ((__pyx_v_self->strip_whitespace != 0)) {

      /* "astropy/io/ascii/cparser.pyx":978
 *         if header_output is not None and header_output == 'comment':
 *             output.write(self.comment)
 *             writer.writerow([x.strip() for x in self.use_names] if             # <<<<<<<<<<<<<<
 *                             self.strip_whitespace else self.use_names)
 *             self._write_comments(output)
 */
      __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 978, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (unlikely(__pyx_v_self->use_names == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
        __PYX_ERR(0, 978, __pyx_L1_error)
      }
      __pyx_t_8 = __pyx_v_self->use_names; __Pyx_INCREF(__pyx_t_8); __pyx_t_9 = 0;
      for (;;) {
        if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_8)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_10 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_9); __Pyx_INCREF(__pyx_t_10); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 978, __pyx_L1_error)
        #else
        __pyx_t_10 = PySequence_ITEM(__pyx_t_8, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 978, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        #endif
        __Pyx_XDECREF_SET(__pyx_v_x, __pyx_t_10);
        __pyx_t_10 = 0;
        __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_x, __pyx_n_s_strip); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 978, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_12 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
          __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
          if (likely(__pyx_t_12)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
            __Pyx_INCREF(__pyx_t_12);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_11, function);
          }
        }
        __pyx_t_10 = (__pyx_t_12) ? __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_12) : __Pyx_PyObject_CallNoArg(__pyx_t_11);
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 978, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_7, (PyObject*)__pyx_t_10))) __PYX_ERR(0, 978, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_6 = __pyx_t_7;
      __pyx_t_7 = 0;
    } else {

      /* "astropy/io/ascii/cparser.pyx":979
 *             output.write(self.comment)
 *             writer.writerow([x.strip() for x in self.use_names] if
 *                             self.strip_whitespace else self.use_names)             # <<<<<<<<<<<<<<
 *             self._write_comments(output)
 *         else:
 */
      __Pyx_INCREF(__pyx_v_self->use_names);
      __pyx_t_6 = __pyx_v_self->use_names;
    }
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_4 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_7, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 978, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "astropy/io/ascii/cparser.pyx":980
 *             writer.writerow([x.strip() for x in self.use_names] if
 *                             self.strip_whitespace else self.use_names)
 *             self._write_comments(output)             # <<<<<<<<<<<<<<
 *         else:
 *             self._write_comments(output)
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_7astropy_2io_5ascii_7cparser_FastWriter *)__pyx_v_self->__pyx_vtab)->_write_comments(__pyx_v_self, __pyx_v_output); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 980, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "astropy/io/ascii/cparser.pyx":976
 * 
 *     def _write_header(self, output, writer, header_output, output_types):
 *         if header_output is not None and header_output == 'comment':             # <<<<<<<<<<<<<<
 *             output.write(self.comment)
 *             writer.writerow([x.strip() for x in self.use_names] if
 */
    goto __pyx_L3;
  }

  /* "astropy/io/ascii/cparser.pyx":982
 *             self._write_comments(output)
 *         else:
 *             self._write_comments(output)             # <<<<<<<<<<<<<<
 *             if header_output is not None:
 *                 writer.writerow([x.strip() for x in self.use_names] if
 */
  /*else*/ {
    __pyx_t_4 = ((struct __pyx_vtabstruct_7astropy_2io_5ascii_7cparser_FastWriter *)__pyx_v_self->__pyx_vtab)->_write_comments(__pyx_v_self, __pyx_v_output); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 982, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "astropy/io/ascii/cparser.pyx":983
 *         else:
 *             self._write_comments(output)
 *             if header_output is not None:             # <<<<<<<<<<<<<<
 *                 writer.writerow([x.strip() for x in self.use_names] if
 *                             self.strip_whitespace else self.use_names)
 */
    __pyx_t_1 = (__pyx_v_header_output != Py_None);
    __pyx_t_3 = (__pyx_t_1 != 0);
    if (__pyx_t_3) {

      /* "astropy/io/ascii/cparser.pyx":984
 *             self._write_comments(output)
 *             if header_output is not None:
 *                 writer.writerow([x.strip() for x in self.use_names] if             # <<<<<<<<<<<<<<
 *                             self.strip_whitespace else self.use_names)
 *         if output_types:
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_writerow); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 984, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);

      /* "astropy/io/ascii/cparser.pyx":985
 *             if header_output is not None:
 *                 writer.writerow([x.strip() for x in self.use_names] if
 *                             self.strip_whitespace else self.use_names)             # <<<<<<<<<<<<<<
 *         if output_types:
 *             writer.writerow(self.types)
 */
      if ((__pyx_v_self->strip_whitespace != 0)) {

        /* "astropy/io/ascii/cparser.pyx":984
 *             self._write_comments(output)
 *             if header_output is not None:
 *                 writer.writerow([x.strip() for x in self.use_names] if             # <<<<<<<<<<<<<<
 *                             self.strip_whitespace else self.use_names)
 *         if output_types:
 */
        __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 984, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        if (unlikely(__pyx_v_self->use_names == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
          __PYX_ERR(0, 984, __pyx_L1_error)
        }
        __pyx_t_8 = __pyx_v_self->use_names; __Pyx_INCREF(__pyx_t_8); __pyx_t_9 = 0;
        for (;;) {
          if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_8)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_10 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_9); __Pyx_INCREF(__pyx_t_10); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 984, __pyx_L1_error)
          #else
          __pyx_t_10 = PySequence_ITEM(__pyx_t_8, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 984, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          #endif
          __Pyx_XDECREF_SET(__pyx_v_x, __pyx_t_10);
          __pyx_t_10 = 0;
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_x, __pyx_n_s_strip); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 984, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_12 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_12)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_12);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
            }
          }
          __pyx_t_10 = (__pyx_t_12) ? __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_12) : __Pyx_PyObject_CallNoArg(__pyx_t_11);
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 984, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_7, (PyObject*)__pyx_t_10))) __PYX_ERR(0, 984, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_6 = __pyx_t_7;
        __pyx_t_7 = 0;
      } else {

        /* "astropy/io/ascii/cparser.pyx":985
 *             if header_output is not None:
 *                 writer.writerow([x.strip() for x in self.use_names] if
 *                             self.strip_whitespace else self.use_names)             # <<<<<<<<<<<<<<
 *         if output_types:
 *             writer.writerow(self.types)
 */
        __Pyx_INCREF(__pyx_v_self->use_names);
        __pyx_t_6 = __pyx_v_self->use_names;
      }
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_4 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_7, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 984, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "astropy/io/ascii/cparser.pyx":983
 *         else:
 *             self._write_comments(output)
 *             if header_output is not None:             # <<<<<<<<<<<<<<
 *                 writer.writerow([x.strip() for x in self.use_names] if
 *                             self.strip_whitespace else self.use_names)
 */
    }
  }
  __pyx_L3:;

  /* "astropy/io/ascii/cparser.pyx":986
 *                 writer.writerow([x.strip() for x in self.use_names] if
 *                             self.strip_whitespace else self.use_names)
 *         if output_types:             # <<<<<<<<<<<<<<
 *             writer.writerow(self.types)
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_output_types); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 986, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "astropy/io/ascii/cparser.pyx":987
 *                             self.strip_whitespace else self.use_names)
 *         if output_types:
 *             writer.writerow(self.types)             # <<<<<<<<<<<<<<
 * 
 *     def write(self, output, header_output, output_types):
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_writerow); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 987, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_v_self->types) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_self->types);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 987, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "astropy/io/ascii/cparser.pyx":986
 *                 writer.writerow([x.strip() for x in self.use_names] if
 *                             self.strip_whitespace else self.use_names)
 *         if output_types:             # <<<<<<<<<<<<<<
 *             writer.writerow(self.types)
 * 
 */
  }

  /* "astropy/io/ascii/cparser.pyx":975
 *                 output.write(self.comment + comment_line + '\n')
 * 
 *     def _write_header(self, output, writer, header_output, output_types):             # <<<<<<<<<<<<<<
 *         if header_output is not None and header_output == 'comment':
 *             output.write(self.comment)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("astropy.io.ascii.cparser.FastWriter._write_header", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_x);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "astropy/io/ascii/cparser.pyx":989
 *             writer.writerow(self.types)
 * 
 *     def write(self, output, header_output, output_types):             # <<<<<<<<<<<<<<
 *         opened_file = False
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_10FastWriter_5write(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_10FastWriter_5write(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_output = 0;
  PyObject *__pyx_v_header_output = 0;
  PyObject *__pyx_v_output_types = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("write (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_output,&__pyx_n_s_header_output,&__pyx_n_s_output_types,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_output)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_header_output)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("write", 1, 3, 3, 1); __PYX_ERR(0, 989, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_output_types)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("write", 1, 3, 3, 2); __PYX_ERR(0, 989, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "write") < 0)) __PYX_ERR(0, 989, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_output = values[0];
    __pyx_v_header_output = values[1];
    __pyx_v_output_types = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("write", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 989, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("astropy.io.ascii.cparser.FastWriter.write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7astropy_2io_5ascii_7cparser_10FastWriter_4write(((struct __pyx_obj_7astropy_2io_5ascii_7cparser_FastWriter *)__pyx_v_self), __pyx_v_output, __pyx_v_header_output, __pyx_v_output_types);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_10FastWriter_4write(struct __pyx_obj_7astropy_2io_5ascii_7cparser_FastWriter *__pyx_v_self, PyObject *__pyx_v_output, PyObject *__pyx_v_header_output, PyObject *__pyx_v_output_types) {
  int __pyx_v_opened_file;
  PyObject *__pyx_v_writer = NULL;
  int __pyx_v_N;
  int __pyx_v_num_cols;
  int __pyx_v_num_rows;
  PyObject *__pyx_v_string_rows = 0;
  PyObject *__pyx_v_rows = 0;
  PyObject *__pyx_v_i = NULL;
  int __pyx_v_j;
  PyObject *__pyx_v_orig_field = NULL;
  PyObject *__pyx_v_str_val = NULL;
  PyObject *__pyx_v_field = NULL;
  PyObject *__pyx_v_new_val = NULL;
  PyObject *__pyx_v_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *(*__pyx_t_10)(PyObject *);
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  Py_ssize_t __pyx_t_15;
  int __pyx_t_16;
  __Pyx_RefNannySetupContext("write", 0);
  __Pyx_INCREF(__pyx_v_output);

  /* "astropy/io/ascii/cparser.pyx":990
 * 
 *     def write(self, output, header_output, output_types):
 *         opened_file = False             # <<<<<<<<<<<<<<
 * 
 *         if not hasattr(output, 'write'): # output is a filename
 */
  __pyx_v_opened_file = 0;

  /* "astropy/io/ascii/cparser.pyx":992
 *         opened_file = False
 * 
 *         if not hasattr(output, 'write'): # output is a filename             # <<<<<<<<<<<<<<
 *             output = open(output, 'w')
 *             opened_file = True # remember to close file afterwards
 */
  __pyx_t_1 = __Pyx_HasAttr(__pyx_v_output, __pyx_n_s_write); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 992, __pyx_L1_error)
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "astropy/io/ascii/cparser.pyx":993
 * 
 *         if not hasattr(output, 'write'): # output is a filename
 *             output = open(output, 'w')             # <<<<<<<<<<<<<<
 *             opened_file = True # remember to close file afterwards
 *         writer = core.CsvWriter(output,
 */
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 993, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_output);
    __Pyx_GIVEREF(__pyx_v_output);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_output);
    __Pyx_INCREF(__pyx_n_s_w);
    __Pyx_GIVEREF(__pyx_n_s_w);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_n_s_w);
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_open, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 993, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_output, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "astropy/io/ascii/cparser.pyx":994
 *         if not hasattr(output, 'write'): # output is a filename
 *             output = open(output, 'w')
 *             opened_file = True # remember to close file afterwards             # <<<<<<<<<<<<<<
 *         writer = core.CsvWriter(output,
 *                                 delimiter=self.delimiter,
 */
    __pyx_v_opened_file = 1;

    /* "astropy/io/ascii/cparser.pyx":992
 *         opened_file = False
 * 
 *         if not hasattr(output, 'write'): # output is a filename             # <<<<<<<<<<<<<<
 *             output = open(output, 'w')
 *             opened_file = True # remember to close file afterwards
 */
  }

  /* "astropy/io/ascii/cparser.pyx":995
 *             output = open(output, 'w')
 *             opened_file = True # remember to close file afterwards
 *         writer = core.CsvWriter(output,             # <<<<<<<<<<<<<<
 *                                 delimiter=self.delimiter,
 *                                 doublequote=True,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_core); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_CsvWriter); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_output);
  __Pyx_GIVEREF(__pyx_v_output);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_output);

  /* "astropy/io/ascii/cparser.pyx":996
 *             opened_file = True # remember to close file afterwards
 *         writer = core.CsvWriter(output,
 *                                 delimiter=self.delimiter,             # <<<<<<<<<<<<<<
 *                                 doublequote=True,
 *                                 escapechar=None,
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 996, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_delimiter, __pyx_v_self->delimiter) < 0) __PYX_ERR(0, 996, __pyx_L1_error)

  /* "astropy/io/ascii/cparser.pyx":997
 *         writer = core.CsvWriter(output,
 *                                 delimiter=self.delimiter,
 *                                 doublequote=True,             # <<<<<<<<<<<<<<
 *                                 escapechar=None,
 *                                 quotechar=self.quotechar,
 */
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_doublequote, Py_True) < 0) __PYX_ERR(0, 996, __pyx_L1_error)

  /* "astropy/io/ascii/cparser.pyx":998
 *                                 delimiter=self.delimiter,
 *                                 doublequote=True,
 *                                 escapechar=None,             # <<<<<<<<<<<<<<
 *                                 quotechar=self.quotechar,
 *                                 quoting=csv.QUOTE_MINIMAL,
 */
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_escapechar, Py_None) < 0) __PYX_ERR(0, 996, __pyx_L1_error)

  /* "astropy/io/ascii/cparser.pyx":999
 *                                 doublequote=True,
 *                                 escapechar=None,
 *                                 quotechar=self.quotechar,             # <<<<<<<<<<<<<<
 *                                 quoting=csv.QUOTE_MINIMAL,
 *                                 lineterminator=os.linesep)
 */
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_quotechar, __pyx_v_self->quotechar) < 0) __PYX_ERR(0, 996, __pyx_L1_error)

  /* "astropy/io/ascii/cparser.pyx":1000
 *                                 escapechar=None,
 *                                 quotechar=self.quotechar,
 *                                 quoting=csv.QUOTE_MINIMAL,             # <<<<<<<<<<<<<<
 *                                 lineterminator=os.linesep)
 *         self._write_header(output, writer, header_output, output_types)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_csv); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1000, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_QUOTE_MINIMAL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1000, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_quoting, __pyx_t_7) < 0) __PYX_ERR(0, 996, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "astropy/io/ascii/cparser.pyx":1001
 *                                 quotechar=self.quotechar,
 *                                 quoting=csv.QUOTE_MINIMAL,
 *                                 lineterminator=os.linesep)             # <<<<<<<<<<<<<<
 *         self._write_header(output, writer, header_output, output_types)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_os); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1001, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_linesep); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1001, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_lineterminator, __pyx_t_6) < 0) __PYX_ERR(0, 996, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "astropy/io/ascii/cparser.pyx":995
 *             output = open(output, 'w')
 *             opened_file = True # remember to close file afterwards
 *         writer = core.CsvWriter(output,             # <<<<<<<<<<<<<<
 *                                 delimiter=self.delimiter,
 *                                 doublequote=True,
 */
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_writer = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "astropy/io/ascii/cparser.pyx":1002
 *                                 quoting=csv.QUOTE_MINIMAL,
 *                                 lineterminator=os.linesep)
 *         self._write_header(output, writer, header_output, output_types)             # <<<<<<<<<<<<<<
 * 
 *         # Split rows into N-sized chunks, since we don't want to
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_write_header); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1002, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[5] = {__pyx_t_4, __pyx_v_output, __pyx_v_writer, __pyx_v_header_output, __pyx_v_output_types};
    __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 4+__pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1002, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_6);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[5] = {__pyx_t_4, __pyx_v_output, __pyx_v_writer, __pyx_v_header_output, __pyx_v_output_types};
    __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 4+__pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1002, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_6);
  } else
  #endif
  {
    __pyx_t_3 = PyTuple_New(4+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1002, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_output);
    __Pyx_GIVEREF(__pyx_v_output);
    PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_8, __pyx_v_output);
    __Pyx_INCREF(__pyx_v_writer);
    __Pyx_GIVEREF(__pyx_v_writer);
    PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_8, __pyx_v_writer);
    __Pyx_INCREF(__pyx_v_header_output);
    __Pyx_GIVEREF(__pyx_v_header_output);
    PyTuple_SET_ITEM(__pyx_t_3, 2+__pyx_t_8, __pyx_v_header_output);
    __Pyx_INCREF(__pyx_v_output_types);
    __Pyx_GIVEREF(__pyx_v_output_types);
    PyTuple_SET_ITEM(__pyx_t_3, 3+__pyx_t_8, __pyx_v_output_types);
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_3, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1002, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "astropy/io/ascii/cparser.pyx":1008
 *         # or fail to take advantage of the speed boost of writerows()
 *         # over writerow().
 *         cdef int N = 100             # <<<<<<<<<<<<<<
 *         cdef int num_cols = <int>len(self.use_names)
 *         cdef int num_rows = <int>len(self.table)
 */
  __pyx_v_N = 0x64;

  /* "astropy/io/ascii/cparser.pyx":1009
 *         # over writerow().
 *         cdef int N = 100
 *         cdef int num_cols = <int>len(self.use_names)             # <<<<<<<<<<<<<<
 *         cdef int num_rows = <int>len(self.table)
 *         # cache string columns beforehand
 */
  __pyx_t_6 = __pyx_v_self->use_names;
  __Pyx_INCREF(__pyx_t_6);
  if (unlikely(__pyx_t_6 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 1009, __pyx_L1_error)
  }
  __pyx_t_9 = PyList_GET_SIZE(__pyx_t_6); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1009, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_num_cols = ((int)__pyx_t_9);

  /* "astropy/io/ascii/cparser.pyx":1010
 *         cdef int N = 100
 *         cdef int num_cols = <int>len(self.use_names)
 *         cdef int num_rows = <int>len(self.table)             # <<<<<<<<<<<<<<
 *         # cache string columns beforehand
 *         cdef set string_rows = set([i for i, type in enumerate(self.types) if
 */
  __pyx_t_6 = __pyx_v_self->table;
  __Pyx_INCREF(__pyx_t_6);
  __pyx_t_9 = PyObject_Length(__pyx_t_6); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1010, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_num_rows = ((int)__pyx_t_9);

  /* "astropy/io/ascii/cparser.pyx":1012
 *         cdef int num_rows = <int>len(self.table)
 *         # cache string columns beforehand
 *         cdef set string_rows = set([i for i, type in enumerate(self.types) if             # <<<<<<<<<<<<<<
 *                                     type == 'S'])
 *         cdef list rows = [[None] * num_cols for i in range(N)]
 */
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1012, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_5 = __pyx_int_0;
  __pyx_t_3 = __pyx_v_self->types; __Pyx_INCREF(__pyx_t_3); __pyx_t_9 = 0;
  for (;;) {
    if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_3)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_4 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_9); __Pyx_INCREF(__pyx_t_4); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 1012, __pyx_L1_error)
    #else
    __pyx_t_4 = PySequence_ITEM(__pyx_t_3, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1012, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_type, __pyx_t_4);
    __pyx_t_4 = 0;
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_5);
    __pyx_t_4 = __Pyx_PyInt_AddObjC(__pyx_t_5, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1012, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5);
    __pyx_t_5 = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "astropy/io/ascii/cparser.pyx":1013
 *         # cache string columns beforehand
 *         cdef set string_rows = set([i for i, type in enumerate(self.types) if
 *                                     type == 'S'])             # <<<<<<<<<<<<<<
 *         cdef list rows = [[None] * num_cols for i in range(N)]
 * 
 */
    __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_v_type, __pyx_n_s_S, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1013, __pyx_L1_error)

    /* "astropy/io/ascii/cparser.pyx":1012
 *         cdef int num_rows = <int>len(self.table)
 *         # cache string columns beforehand
 *         cdef set string_rows = set([i for i, type in enumerate(self.types) if             # <<<<<<<<<<<<<<
 *                                     type == 'S'])
 *         cdef list rows = [[None] * num_cols for i in range(N)]
 */
    if (__pyx_t_2) {
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_6, (PyObject*)__pyx_v_i))) __PYX_ERR(0, 1012, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PySet_New(__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1012, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_string_rows = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "astropy/io/ascii/cparser.pyx":1014
 *         cdef set string_rows = set([i for i, type in enumerate(self.types) if
 *                                     type == 'S'])
 *         cdef list rows = [[None] * num_cols for i in range(N)]             # <<<<<<<<<<<<<<
 * 
 *         for i in range(num_rows):
 */
  __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_6 = __pyx_t_3; __Pyx_INCREF(__pyx_t_6); __pyx_t_9 = 0;
    __pyx_t_10 = NULL;
  } else {
    __pyx_t_9 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1014, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_10 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1014, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_10)) {
      if (likely(PyList_CheckExact(__pyx_t_6))) {
        if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_6)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_9); __Pyx_INCREF(__pyx_t_3); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 1014, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_6, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1014, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_9 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_9); __Pyx_INCREF(__pyx_t_3); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 1014, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_6, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1014, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_10(__pyx_t_6);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1014, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = PyList_New(1 * ((__pyx_v_num_cols<0) ? 0:__pyx_v_num_cols)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1014, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    { Py_ssize_t __pyx_temp;
      for (__pyx_temp=0; __pyx_temp < __pyx_v_num_cols; __pyx_temp++) {
        __Pyx_INCREF(Py_None);
        __Pyx_GIVEREF(Py_None);
        PyList_SET_ITEM(__pyx_t_3, __pyx_temp, Py_None);
      }
    }
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_5, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 1014, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_rows = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "astropy/io/ascii/cparser.pyx":1016
 *         cdef list rows = [[None] * num_cols for i in range(N)]
 * 
 *         for i in range(num_rows):             # <<<<<<<<<<<<<<
 *             for j in range(num_cols):
 *                 orig_field = next(self.col_iters[j]) # get field
 */
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_num_rows); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1016, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1016, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
    __pyx_t_5 = __pyx_t_6; __Pyx_INCREF(__pyx_t_5); __pyx_t_9 = 0;
    __pyx_t_10 = NULL;
  } else {
    __pyx_t_9 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1016, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_10 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1016, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  for (;;) {
    if (likely(!__pyx_t_10)) {
      if (likely(PyList_CheckExact(__pyx_t_5))) {
        if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_6 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_9); __Pyx_INCREF(__pyx_t_6); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 1016, __pyx_L1_error)
        #else
        __pyx_t_6 = PySequence_ITEM(__pyx_t_5, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1016, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
      } else {
        if (__pyx_t_9 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_9); __Pyx_INCREF(__pyx_t_6); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 1016, __pyx_L1_error)
        #else
        __pyx_t_6 = PySequence_ITEM(__pyx_t_5, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1016, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
      }
    } else {
      __pyx_t_6 = __pyx_t_10(__pyx_t_5);
      if (unlikely(!__pyx_t_6)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1016, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_6);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "astropy/io/ascii/cparser.pyx":1017
 * 
 *         for i in range(num_rows):
 *             for j in range(num_cols):             # <<<<<<<<<<<<<<
 *                 orig_field = next(self.col_iters[j]) # get field
 *                 # str_val monitors whether we should check if the field
 */
    __pyx_t_8 = __pyx_v_num_cols;
    __pyx_t_11 = __pyx_t_8;
    for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
      __pyx_v_j = __pyx_t_12;

      /* "astropy/io/ascii/cparser.pyx":1018
 *         for i in range(num_rows):
 *             for j in range(num_cols):
 *                 orig_field = next(self.col_iters[j]) # get field             # <<<<<<<<<<<<<<
 *                 # str_val monitors whether we should check if the field
 *                 # should be stripped
 */
      if (unlikely(__pyx_v_self->col_iters == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 1018, __pyx_L1_error)
      }
      __pyx_t_6 = __Pyx_GetItemInt_List(__pyx_v_self->col_iters, __pyx_v_j, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1018, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_3 = __Pyx_PyIter_Next(__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1018, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF_SET(__pyx_v_orig_field, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "astropy/io/ascii/cparser.pyx":1021
 *                 # str_val monitors whether we should check if the field
 *                 # should be stripped
 *                 str_val = True             # <<<<<<<<<<<<<<
 * 
 *                 if orig_field is None: # tolist() converts ma.masked to None
 */
      __Pyx_INCREF(Py_True);
      __Pyx_XDECREF_SET(__pyx_v_str_val, Py_True);

      /* "astropy/io/ascii/cparser.pyx":1023
 *                 str_val = True
 * 
 *                 if orig_field is None: # tolist() converts ma.masked to None             # <<<<<<<<<<<<<<
 *                     field = core.masked
 *                     rows[i % N][j] = ''
 */
      __pyx_t_2 = (__pyx_v_orig_field == Py_None);
      __pyx_t_1 = (__pyx_t_2 != 0);
      if (__pyx_t_1) {

        /* "astropy/io/ascii/cparser.pyx":1024
 * 
 *                 if orig_field is None: # tolist() converts ma.masked to None
 *                     field = core.masked             # <<<<<<<<<<<<<<
 *                     rows[i % N][j] = ''
 * 
 */
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_core); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1024, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_masked); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1024, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF_SET(__pyx_v_field, __pyx_t_6);
        __pyx_t_6 = 0;

        /* "astropy/io/ascii/cparser.pyx":1025
 *                 if orig_field is None: # tolist() converts ma.masked to None
 *                     field = core.masked
 *                     rows[i % N][j] = ''             # <<<<<<<<<<<<<<
 * 
 *                 elif self.format_funcs[j] is not None:
 */
        __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1025, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_3 = PyNumber_Remainder(__pyx_v_i, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1025, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_v_rows, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1025, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(__Pyx_SetItemInt(__pyx_t_6, __pyx_v_j, __pyx_kp_s__5, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 1025, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "astropy/io/ascii/cparser.pyx":1023
 *                 str_val = True
 * 
 *                 if orig_field is None: # tolist() converts ma.masked to None             # <<<<<<<<<<<<<<
 *                     field = core.masked
 *                     rows[i % N][j] = ''
 */
        goto __pyx_L13;
      }

      /* "astropy/io/ascii/cparser.pyx":1027
 *                     rows[i % N][j] = ''
 * 
 *                 elif self.format_funcs[j] is not None:             # <<<<<<<<<<<<<<
 *                     field = self.format_funcs[j](self.formats[j], orig_field)
 *                     rows[i % N][j] = field
 */
      if (unlikely(__pyx_v_self->format_funcs == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 1027, __pyx_L1_error)
      }
      __pyx_t_6 = __Pyx_GetItemInt_List(__pyx_v_self->format_funcs, __pyx_v_j, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1027, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_1 = (__pyx_t_6 != Py_None);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_2 = (__pyx_t_1 != 0);
      if (__pyx_t_2) {

        /* "astropy/io/ascii/cparser.pyx":1028
 * 
 *                 elif self.format_funcs[j] is not None:
 *                     field = self.format_funcs[j](self.formats[j], orig_field)             # <<<<<<<<<<<<<<
 *                     rows[i % N][j] = field
 * 
 */
        if (unlikely(__pyx_v_self->format_funcs == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 1028, __pyx_L1_error)
        }
        __pyx_t_3 = __Pyx_GetItemInt_List(__pyx_v_self->format_funcs, __pyx_v_j, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1028, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (unlikely(__pyx_v_self->formats == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 1028, __pyx_L1_error)
        }
        __pyx_t_4 = __Pyx_GetItemInt_List(__pyx_v_self->formats, __pyx_v_j, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1028, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_7 = NULL;
        __pyx_t_13 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_13 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_4, __pyx_v_orig_field};
          __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1028, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_4, __pyx_v_orig_field};
          __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1028, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_14 = PyTuple_New(2+__pyx_t_13); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1028, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          if (__pyx_t_7) {
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_7); __pyx_t_7 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_14, 0+__pyx_t_13, __pyx_t_4);
          __Pyx_INCREF(__pyx_v_orig_field);
          __Pyx_GIVEREF(__pyx_v_orig_field);
          PyTuple_SET_ITEM(__pyx_t_14, 1+__pyx_t_13, __pyx_v_orig_field);
          __pyx_t_4 = 0;
          __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_14, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1028, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF_SET(__pyx_v_field, __pyx_t_6);
        __pyx_t_6 = 0;

        /* "astropy/io/ascii/cparser.pyx":1029
 *                 elif self.format_funcs[j] is not None:
 *                     field = self.format_funcs[j](self.formats[j], orig_field)
 *                     rows[i % N][j] = field             # <<<<<<<<<<<<<<
 * 
 *                 else:
 */
        __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1029, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_3 = PyNumber_Remainder(__pyx_v_i, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1029, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_v_rows, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1029, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(__Pyx_SetItemInt(__pyx_t_6, __pyx_v_j, __pyx_v_field, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 1029, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "astropy/io/ascii/cparser.pyx":1027
 *                     rows[i % N][j] = ''
 * 
 *                 elif self.format_funcs[j] is not None:             # <<<<<<<<<<<<<<
 *                     field = self.format_funcs[j](self.formats[j], orig_field)
 *                     rows[i % N][j] = field
 */
        goto __pyx_L13;
      }

      /* "astropy/io/ascii/cparser.pyx":1032
 * 
 *                 else:
 *                     field = orig_field             # <<<<<<<<<<<<<<
 *                     rows[i % N][j] = field
 *                     str_val = j in string_rows
 */
      /*else*/ {
        __Pyx_INCREF(__pyx_v_orig_field);
        __Pyx_XDECREF_SET(__pyx_v_field, __pyx_v_orig_field);

        /* "astropy/io/ascii/cparser.pyx":1033
 *                 else:
 *                     field = orig_field
 *                     rows[i % N][j] = field             # <<<<<<<<<<<<<<
 *                     str_val = j in string_rows
 * 
 */
        __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1033, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_3 = PyNumber_Remainder(__pyx_v_i, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1033, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_v_rows, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1033, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(__Pyx_SetItemInt(__pyx_t_6, __pyx_v_j, __pyx_v_field, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 1033, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "astropy/io/ascii/cparser.pyx":1034
 *                     field = orig_field
 *                     rows[i % N][j] = field
 *                     str_val = j in string_rows             # <<<<<<<<<<<<<<
 * 
 *                 if field in self.fill_values:
 */
        __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_j); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1034, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_2 = (__Pyx_PySet_ContainsTF(__pyx_t_6, __pyx_v_string_rows, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1034, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1034, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF_SET(__pyx_v_str_val, __pyx_t_6);
        __pyx_t_6 = 0;
      }
      __pyx_L13:;

      /* "astropy/io/ascii/cparser.pyx":1036
 *                     str_val = j in string_rows
 * 
 *                 if field in self.fill_values:             # <<<<<<<<<<<<<<
 *                     new_val = self.fill_values[field][0]
 *                     # Either this column applies to the field as specified in
 */
      if (unlikely(__pyx_v_self->fill_values == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
        __PYX_ERR(0, 1036, __pyx_L1_error)
      }
      __pyx_t_2 = (__Pyx_PyDict_ContainsTF(__pyx_v_field, __pyx_v_self->fill_values, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1036, __pyx_L1_error)
      __pyx_t_1 = (__pyx_t_2 != 0);
      if (__pyx_t_1) {

        /* "astropy/io/ascii/cparser.pyx":1037
 * 
 *                 if field in self.fill_values:
 *                     new_val = self.fill_values[field][0]             # <<<<<<<<<<<<<<
 *                     # Either this column applies to the field as specified in
 *                     # the fill_values parameter, or no specific columns are
 */
        if (unlikely(__pyx_v_self->fill_values == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 1037, __pyx_L1_error)
        }
        __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_self->fill_values, __pyx_v_field); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1037, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_6, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1037, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF_SET(__pyx_v_new_val, __pyx_t_3);
        __pyx_t_3 = 0;

        /* "astropy/io/ascii/cparser.pyx":1041
 *                     # the fill_values parameter, or no specific columns are
 *                     # specified and this column should apply fill_values.
 *                     if (len(self.fill_values[field]) > 1 and self.use_names[j] in self.fill_values[field][1:]) \             # <<<<<<<<<<<<<<
 *                        or (len(self.fill_values[field]) == 1 and j in self.fill_cols):
 *                         str_val = True
 */
        if (unlikely(__pyx_v_self->fill_values == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 1041, __pyx_L1_error)
        }
        __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_self->fill_values, __pyx_v_field); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1041, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_15 = PyObject_Length(__pyx_t_3); if (unlikely(__pyx_t_15 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1041, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_2 = ((__pyx_t_15 > 1) != 0);
        if (!__pyx_t_2) {
          goto __pyx_L17_next_or;
        } else {
        }
        if (unlikely(__pyx_v_self->use_names == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 1041, __pyx_L1_error)
        }
        __pyx_t_3 = __Pyx_GetItemInt_List(__pyx_v_self->use_names, __pyx_v_j, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1041, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (unlikely(__pyx_v_self->fill_values == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 1041, __pyx_L1_error)
        }
        __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_self->fill_values, __pyx_v_field); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1041, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_14 = __Pyx_PyObject_GetSlice(__pyx_t_6, 1, 0, NULL, NULL, &__pyx_slice__19, 1, 0, 1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1041, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_3, __pyx_t_14, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1041, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __pyx_t_16 = (__pyx_t_2 != 0);
        if (!__pyx_t_16) {
        } else {
          __pyx_t_1 = __pyx_t_16;
          goto __pyx_L16_bool_binop_done;
        }
        __pyx_L17_next_or:;

        /* "astropy/io/ascii/cparser.pyx":1042
 *                     # specified and this column should apply fill_values.
 *                     if (len(self.fill_values[field]) > 1 and self.use_names[j] in self.fill_values[field][1:]) \
 *                        or (len(self.fill_values[field]) == 1 and j in self.fill_cols):             # <<<<<<<<<<<<<<
 *                         str_val = True
 *                         rows[i % N][j] = new_val
 */
        if (unlikely(__pyx_v_self->fill_values == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 1042, __pyx_L1_error)
        }
        __pyx_t_14 = __Pyx_PyDict_GetItem(__pyx_v_self->fill_values, __pyx_v_field); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1042, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        __pyx_t_15 = PyObject_Length(__pyx_t_14); if (unlikely(__pyx_t_15 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1042, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __pyx_t_16 = ((__pyx_t_15 == 1) != 0);
        if (__pyx_t_16) {
        } else {
          __pyx_t_1 = __pyx_t_16;
          goto __pyx_L16_bool_binop_done;
        }
        __pyx_t_14 = __Pyx_PyInt_From_int(__pyx_v_j); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1042, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        if (unlikely(__pyx_v_self->fill_cols == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
          __PYX_ERR(0, 1042, __pyx_L1_error)
        }
        __pyx_t_16 = (__Pyx_PySet_ContainsTF(__pyx_t_14, __pyx_v_self->fill_cols, Py_EQ)); if (unlikely(__pyx_t_16 < 0)) __PYX_ERR(0, 1042, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __pyx_t_2 = (__pyx_t_16 != 0);
        __pyx_t_1 = __pyx_t_2;
        __pyx_L16_bool_binop_done:;

        /* "astropy/io/ascii/cparser.pyx":1041
 *                     # the fill_values parameter, or no specific columns are
 *                     # specified and this column should apply fill_values.
 *                     if (len(self.fill_values[field]) > 1 and self.use_names[j] in self.fill_values[field][1:]) \             # <<<<<<<<<<<<<<
 *                        or (len(self.fill_values[field]) == 1 and j in self.fill_cols):
 *                         str_val = True
 */
        if (__pyx_t_1) {

          /* "astropy/io/ascii/cparser.pyx":1043
 *                     if (len(self.fill_values[field]) > 1 and self.use_names[j] in self.fill_values[field][1:]) \
 *                        or (len(self.fill_values[field]) == 1 and j in self.fill_cols):
 *                         str_val = True             # <<<<<<<<<<<<<<
 *                         rows[i % N][j] = new_val
 *                         if self.strip_whitespace: # new_val should be a string
 */
          __Pyx_INCREF(Py_True);
          __Pyx_DECREF_SET(__pyx_v_str_val, Py_True);

          /* "astropy/io/ascii/cparser.pyx":1044
 *                        or (len(self.fill_values[field]) == 1 and j in self.fill_cols):
 *                         str_val = True
 *                         rows[i % N][j] = new_val             # <<<<<<<<<<<<<<
 *                         if self.strip_whitespace: # new_val should be a string
 *                             rows[i % N][j] = rows[i % N][j].strip()
 */
          __pyx_t_14 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1044, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_3 = PyNumber_Remainder(__pyx_v_i, __pyx_t_14); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1044, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __pyx_t_14 = __Pyx_PyObject_GetItem(__pyx_v_rows, __pyx_t_3); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1044, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(__Pyx_SetItemInt(__pyx_t_14, __pyx_v_j, __pyx_v_new_val, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 1044, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

          /* "astropy/io/ascii/cparser.pyx":1045
 *                         str_val = True
 *                         rows[i % N][j] = new_val
 *                         if self.strip_whitespace: # new_val should be a string             # <<<<<<<<<<<<<<
 *                             rows[i % N][j] = rows[i % N][j].strip()
 * 
 */
          __pyx_t_1 = (__pyx_v_self->strip_whitespace != 0);
          if (__pyx_t_1) {

            /* "astropy/io/ascii/cparser.pyx":1046
 *                         rows[i % N][j] = new_val
 *                         if self.strip_whitespace: # new_val should be a string
 *                             rows[i % N][j] = rows[i % N][j].strip()             # <<<<<<<<<<<<<<
 * 
 *                 if str_val and self.strip_whitespace:
 */
            __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1046, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_6 = PyNumber_Remainder(__pyx_v_i, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1046, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_rows, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1046, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_3, __pyx_v_j, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1046, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_strip); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1046, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __pyx_t_6 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
              __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
              if (likely(__pyx_t_6)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_6);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_3, function);
              }
            }
            __pyx_t_14 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1046, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_14);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1046, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_6 = PyNumber_Remainder(__pyx_v_i, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1046, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_rows, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1046, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            if (unlikely(__Pyx_SetItemInt(__pyx_t_3, __pyx_v_j, __pyx_t_14, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 1046, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

            /* "astropy/io/ascii/cparser.pyx":1045
 *                         str_val = True
 *                         rows[i % N][j] = new_val
 *                         if self.strip_whitespace: # new_val should be a string             # <<<<<<<<<<<<<<
 *                             rows[i % N][j] = rows[i % N][j].strip()
 * 
 */
          }

          /* "astropy/io/ascii/cparser.pyx":1041
 *                     # the fill_values parameter, or no specific columns are
 *                     # specified and this column should apply fill_values.
 *                     if (len(self.fill_values[field]) > 1 and self.use_names[j] in self.fill_values[field][1:]) \             # <<<<<<<<<<<<<<
 *                        or (len(self.fill_values[field]) == 1 and j in self.fill_cols):
 *                         str_val = True
 */
        }

        /* "astropy/io/ascii/cparser.pyx":1036
 *                     str_val = j in string_rows
 * 
 *                 if field in self.fill_values:             # <<<<<<<<<<<<<<
 *                     new_val = self.fill_values[field][0]
 *                     # Either this column applies to the field as specified in
 */
      }

      /* "astropy/io/ascii/cparser.pyx":1048
 *                             rows[i % N][j] = rows[i % N][j].strip()
 * 
 *                 if str_val and self.strip_whitespace:             # <<<<<<<<<<<<<<
 *                     rows[i % N][j] = rows[i % N][j].strip()
 * 
 */
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_str_val); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1048, __pyx_L1_error)
      if (__pyx_t_2) {
      } else {
        __pyx_t_1 = __pyx_t_2;
        goto __pyx_L22_bool_binop_done;
      }
      __pyx_t_2 = (__pyx_v_self->strip_whitespace != 0);
      __pyx_t_1 = __pyx_t_2;
      __pyx_L22_bool_binop_done:;
      if (__pyx_t_1) {

        /* "astropy/io/ascii/cparser.pyx":1049
 * 
 *                 if str_val and self.strip_whitespace:
 *                     rows[i % N][j] = rows[i % N][j].strip()             # <<<<<<<<<<<<<<
 * 
 *             if i >= N - 1 and i % N == N - 1: # rows is now full
 */
        __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1049, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_6 = PyNumber_Remainder(__pyx_v_i, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1049, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_rows, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1049, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_3, __pyx_v_j, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1049, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_strip); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1049, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_14 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1049, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1049, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_6 = PyNumber_Remainder(__pyx_v_i, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1049, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_rows, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1049, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(__Pyx_SetItemInt(__pyx_t_3, __pyx_v_j, __pyx_t_14, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 1049, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

        /* "astropy/io/ascii/cparser.pyx":1048
 *                             rows[i % N][j] = rows[i % N][j].strip()
 * 
 *                 if str_val and self.strip_whitespace:             # <<<<<<<<<<<<<<
 *                     rows[i % N][j] = rows[i % N][j].strip()
 * 
 */
      }
    }

    /* "astropy/io/ascii/cparser.pyx":1051
 *                     rows[i % N][j] = rows[i % N][j].strip()
 * 
 *             if i >= N - 1 and i % N == N - 1: # rows is now full             # <<<<<<<<<<<<<<
 *                 writer.writerows(rows)
 * 
 */
    __pyx_t_14 = __Pyx_PyInt_From_long((__pyx_v_N - 1)); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1051, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_i, __pyx_t_14, Py_GE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1051, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1051, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L25_bool_binop_done;
    }
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1051, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_14 = PyNumber_Remainder(__pyx_v_i, __pyx_t_3); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1051, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyInt_From_long((__pyx_v_N - 1)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1051, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = PyObject_RichCompare(__pyx_t_14, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1051, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1051, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_1 = __pyx_t_2;
    __pyx_L25_bool_binop_done:;
    if (__pyx_t_1) {

      /* "astropy/io/ascii/cparser.pyx":1052
 * 
 *             if i >= N - 1 and i % N == N - 1: # rows is now full
 *                 writer.writerows(rows)             # <<<<<<<<<<<<<<
 * 
 *         # Write leftover rows not included in previous chunks
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_writerows); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1052, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_14 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_14)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_14);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_6 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_14, __pyx_v_rows) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_rows);
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1052, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "astropy/io/ascii/cparser.pyx":1051
 *                     rows[i % N][j] = rows[i % N][j].strip()
 * 
 *             if i >= N - 1 and i % N == N - 1: # rows is now full             # <<<<<<<<<<<<<<
 *                 writer.writerows(rows)
 * 
 */
    }

    /* "astropy/io/ascii/cparser.pyx":1016
 *         cdef list rows = [[None] * num_cols for i in range(N)]
 * 
 *         for i in range(num_rows):             # <<<<<<<<<<<<<<
 *             for j in range(num_cols):
 *                 orig_field = next(self.col_iters[j]) # get field
 */
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "astropy/io/ascii/cparser.pyx":1055
 * 
 *         # Write leftover rows not included in previous chunks
 *         if i % N != N - 1:             # <<<<<<<<<<<<<<
 *             writer.writerows(rows[:i % N + 1])
 * 
 */
  if (unlikely(!__pyx_v_i)) { __Pyx_RaiseUnboundLocalError("i"); __PYX_ERR(0, 1055, __pyx_L1_error) }
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1055, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyNumber_Remainder(__pyx_v_i, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1055, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_From_long((__pyx_v_N - 1)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1055, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_6, __pyx_t_5, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1055, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1055, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {

    /* "astropy/io/ascii/cparser.pyx":1056
 *         # Write leftover rows not included in previous chunks
 *         if i % N != N - 1:
 *             writer.writerows(rows[:i % N + 1])             # <<<<<<<<<<<<<<
 * 
 *         if opened_file:
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer, __pyx_n_s_writerows); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1056, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(!__pyx_v_i)) { __Pyx_RaiseUnboundLocalError("i"); __PYX_ERR(0, 1056, __pyx_L1_error) }
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1056, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_14 = PyNumber_Remainder(__pyx_v_i, __pyx_t_6); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1056, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyInt_AddObjC(__pyx_t_14, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1056, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_1 = (__pyx_t_6 == Py_None);
    if (__pyx_t_1) {
      __pyx_t_9 = PY_SSIZE_T_MAX;
    } else {
      __pyx_t_15 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_15 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1056, __pyx_L1_error)
      __pyx_t_9 = __pyx_t_15;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyList_GetSlice(__pyx_v_rows, 0, __pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1056, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_14 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_14)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_14);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_14, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1056, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "astropy/io/ascii/cparser.pyx":1055
 * 
 *         # Write leftover rows not included in previous chunks
 *         if i % N != N - 1:             # <<<<<<<<<<<<<<
 *             writer.writerows(rows[:i % N + 1])
 * 
 */
  }

  /* "astropy/io/ascii/cparser.pyx":1058
 *             writer.writerows(rows[:i % N + 1])
 * 
 *         if opened_file:             # <<<<<<<<<<<<<<
 *             output.close()
 * 
 */
  __pyx_t_1 = (__pyx_v_opened_file != 0);
  if (__pyx_t_1) {

    /* "astropy/io/ascii/cparser.pyx":1059
 * 
 *         if opened_file:
 *             output.close()             # <<<<<<<<<<<<<<
 * 
 * def get_fill_values(fill_values, read=True):
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_output, __pyx_n_s_close); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1059, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1059, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "astropy/io/ascii/cparser.pyx":1058
 *             writer.writerows(rows[:i % N + 1])
 * 
 *         if opened_file:             # <<<<<<<<<<<<<<
 *             output.close()
 * 
 */
  }

  /* "astropy/io/ascii/cparser.pyx":989
 *             writer.writerow(self.types)
 * 
 *     def write(self, output, header_output, output_types):             # <<<<<<<<<<<<<<
 *         opened_file = False
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("astropy.io.ascii.cparser.FastWriter.write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_writer);
  __Pyx_XDECREF(__pyx_v_string_rows);
  __Pyx_XDECREF(__pyx_v_rows);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_orig_field);
  __Pyx_XDECREF(__pyx_v_str_val);
  __Pyx_XDECREF(__pyx_v_field);
  __Pyx_XDECREF(__pyx_v_new_val);
  __Pyx_XDECREF(__pyx_v_type);
  __Pyx_XDECREF(__pyx_v_output);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_10FastWriter_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_10FastWriter_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7astropy_2io_5ascii_7cparser_10FastWriter_6__reduce_cython__(((struct __pyx_obj_7astropy_2io_5ascii_7cparser_FastWriter *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_10FastWriter_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7astropy_2io_5ascii_7cparser_FastWriter *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__23, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("astropy.io.ascii.cparser.FastWriter.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_10FastWriter_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_10FastWriter_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7astropy_2io_5ascii_7cparser_10FastWriter_8__setstate_cython__(((struct __pyx_obj_7astropy_2io_5ascii_7cparser_FastWriter *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_10FastWriter_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7astropy_2io_5ascii_7cparser_FastWriter *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__24, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("astropy.io.ascii.cparser.FastWriter.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "astropy/io/ascii/cparser.pyx":1061
 *             output.close()
 * 
 * def get_fill_values(fill_values, read=True):             # <<<<<<<<<<<<<<
 *     if len(fill_values) > 0 and isinstance(fill_values[0], six.string_types):
 *         # e.g. fill_values=('999', '0')
 */

/* Python wrapper */
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_5get_fill_values(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_7astropy_2io_5ascii_7cparser_5get_fill_values = {"get_fill_values", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7astropy_2io_5ascii_7cparser_5get_fill_values, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7astropy_2io_5ascii_7cparser_5get_fill_values(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_fill_values = 0;
  PyObject *__pyx_v_read = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_fill_values (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fill_values,&__pyx_n_s_read,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_True);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fill_values)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_read);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get_fill_values") < 0)) __PYX_ERR(0, 1061, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_fill_values = values[0];
    __pyx_v_read = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_fill_values", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1061, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("astropy.io.ascii.cparser.get_fill_values", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7astropy_2io_5ascii_7cparser_4get_fill_values(__pyx_self, __pyx_v_fill_values, __pyx_v_read);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7astropy_2io_5ascii_7cparser_4get_fill_values(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_fill_values, PyObject *__pyx_v_read) {
  PyObject *__pyx_v_fill_empty = NULL;
  PyObject *__pyx_v_el = NULL;
  PyObject *__pyx_v_l = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  __Pyx_RefNannySetupContext("get_fill_values", 0);
  __Pyx_INCREF(__pyx_v_fill_values);

  /* "astropy/io/ascii/cparser.pyx":1062
 * 
 * def get_fill_values(fill_values, read=True):
 *     if len(fill_values) > 0 and isinstance(fill_values[0], six.string_types):             # <<<<<<<<<<<<<<
 *         # e.g. fill_values=('999', '0')
 *         fill_values = [fill_values]
 */
  __pyx_t_2 = PyObject_Length(__pyx_v_fill_values); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1062, __pyx_L1_error)
  __pyx_t_3 = ((__pyx_t_2 > 0) != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_fill_values, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_six); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_string_types); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = PyObject_IsInstance(__pyx_t_4, __pyx_t_6); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 1062, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_7 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "astropy/io/ascii/cparser.pyx":1064
 *     if len(fill_values) > 0 and isinstance(fill_values[0], six.string_types):
 *         # e.g. fill_values=('999', '0')
 *         fill_values = [fill_values]             # <<<<<<<<<<<<<<
 *     else:
 *         fill_values = fill_values
 */
    __pyx_t_6 = PyList_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1064, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_v_fill_values);
    __Pyx_GIVEREF(__pyx_v_fill_values);
    PyList_SET_ITEM(__pyx_t_6, 0, __pyx_v_fill_values);
    __Pyx_DECREF_SET(__pyx_v_fill_values, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "astropy/io/ascii/cparser.pyx":1062
 * 
 * def get_fill_values(fill_values, read=True):
 *     if len(fill_values) > 0 and isinstance(fill_values[0], six.string_types):             # <<<<<<<<<<<<<<
 *         # e.g. fill_values=('999', '0')
 *         fill_values = [fill_values]
 */
    goto __pyx_L3;
  }

  /* "astropy/io/ascii/cparser.pyx":1066
 *         fill_values = [fill_values]
 *     else:
 *         fill_values = fill_values             # <<<<<<<<<<<<<<
 * 
 *     # look for an empty replacement to cache for speedy conversion
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_v_fill_values);
    __Pyx_DECREF_SET(__pyx_v_fill_values, __pyx_v_fill_values);
  }
  __pyx_L3:;

  /* "astropy/io/ascii/cparser.pyx":1069
 * 
 *     # look for an empty replacement to cache for speedy conversion
 *     fill_empty = None             # <<<<<<<<<<<<<<
 *     for el in fill_values:
 *         if el[0] == '':
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_fill_empty = Py_None;

  /* "astropy/io/ascii/cparser.pyx":1070
 *     # look for an empty replacement to cache for speedy conversion
 *     fill_empty = None
 *     for el in fill_values:             # <<<<<<<<<<<<<<
 *         if el[0] == '':
 *             fill_empty = el[1:]
 */
  if (likely(PyList_CheckExact(__pyx_v_fill_values)) || PyTuple_CheckExact(__pyx_v_fill_values)) {
    __pyx_t_6 = __pyx_v_fill_values; __Pyx_INCREF(__pyx_t_6); __pyx_t_2 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_v_fill_values); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1070, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1070, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_6))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_6)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 1070, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_6, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1070, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 1070, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_6, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1070, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_8(__pyx_t_6);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1070, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_el, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "astropy/io/ascii/cparser.pyx":1071
 *     fill_empty = None
 *     for el in fill_values:
 *         if el[0] == '':             # <<<<<<<<<<<<<<
 *             fill_empty = el[1:]
 *             break
 */
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_el, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1071, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = (__Pyx_PyString_Equals(__pyx_t_4, __pyx_kp_s__5, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1071, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_1) {

      /* "astropy/io/ascii/cparser.pyx":1072
 *     for el in fill_values:
 *         if el[0] == '':
 *             fill_empty = el[1:]             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_v_el, 1, 0, NULL, NULL, &__pyx_slice__19, 1, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1072, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF_SET(__pyx_v_fill_empty, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "astropy/io/ascii/cparser.pyx":1073
 *         if el[0] == '':
 *             fill_empty = el[1:]
 *             break             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
      goto __pyx_L7_break;

      /* "astropy/io/ascii/cparser.pyx":1071
 *     fill_empty = None
 *     for el in fill_values:
 *         if el[0] == '':             # <<<<<<<<<<<<<<
 *             fill_empty = el[1:]
 *             break
 */
    }

    /* "astropy/io/ascii/cparser.pyx":1070
 *     # look for an empty replacement to cache for speedy conversion
 *     fill_empty = None
 *     for el in fill_values:             # <<<<<<<<<<<<<<
 *         if el[0] == '':
 *             fill_empty = el[1:]
 */
  }
  __pyx_L7_break:;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "astropy/io/ascii/cparser.pyx":1075
 *             break
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         # Create a dict with the values to be replaced as keys
 *         if read:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
    __Pyx_XGOTREF(__pyx_t_9);
    __Pyx_XGOTREF(__pyx_t_10);
    __Pyx_XGOTREF(__pyx_t_11);
    /*try:*/ {

      /* "astropy/io/ascii/cparser.pyx":1077
 *     try:
 *         # Create a dict with the values to be replaced as keys
 *         if read:             # <<<<<<<<<<<<<<
 *             fill_values = dict([(l[0].encode('ascii'), l[1:]) for
 *                                 l in fill_values if l[0] != ''])
 */
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_read); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1077, __pyx_L9_error)
      if (__pyx_t_1) {

        /* "astropy/io/ascii/cparser.pyx":1078
 *         # Create a dict with the values to be replaced as keys
 *         if read:
 *             fill_values = dict([(l[0].encode('ascii'), l[1:]) for             # <<<<<<<<<<<<<<
 *                                 l in fill_values if l[0] != ''])
 *         else:
 */
        __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1078, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_6);

        /* "astropy/io/ascii/cparser.pyx":1079
 *         if read:
 *             fill_values = dict([(l[0].encode('ascii'), l[1:]) for
 *                                 l in fill_values if l[0] != ''])             # <<<<<<<<<<<<<<
 *         else:
 *             # don't worry about encoding for writing
 */
        if (likely(PyList_CheckExact(__pyx_v_fill_values)) || PyTuple_CheckExact(__pyx_v_fill_values)) {
          __pyx_t_4 = __pyx_v_fill_values; __Pyx_INCREF(__pyx_t_4); __pyx_t_2 = 0;
          __pyx_t_8 = NULL;
        } else {
          __pyx_t_2 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_fill_values); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1079, __pyx_L9_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_8 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1079, __pyx_L9_error)
        }
        for (;;) {
          if (likely(!__pyx_t_8)) {
            if (likely(PyList_CheckExact(__pyx_t_4))) {
              if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_4)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_5 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_2); __Pyx_INCREF(__pyx_t_5); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 1079, __pyx_L9_error)
              #else
              __pyx_t_5 = PySequence_ITEM(__pyx_t_4, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1079, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_5);
              #endif
            } else {
              if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_2); __Pyx_INCREF(__pyx_t_5); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 1079, __pyx_L9_error)
              #else
              __pyx_t_5 = PySequence_ITEM(__pyx_t_4, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1079, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_5);
              #endif
            }
          } else {
            __pyx_t_5 = __pyx_t_8(__pyx_t_4);
            if (unlikely(!__pyx_t_5)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 1079, __pyx_L9_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_5);
          }
          __Pyx_XDECREF_SET(__pyx_v_l, __pyx_t_5);
          __pyx_t_5 = 0;
          __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_l, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1079, __pyx_L9_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_1 = (__Pyx_PyString_Equals(__pyx_t_5, __pyx_kp_s__5, Py_NE)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1079, __pyx_L9_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (__pyx_t_1) {

            /* "astropy/io/ascii/cparser.pyx":1078
 *         # Create a dict with the values to be replaced as keys
 *         if read:
 *             fill_values = dict([(l[0].encode('ascii'), l[1:]) for             # <<<<<<<<<<<<<<
 *                                 l in fill_values if l[0] != ''])
 *         else:
 */
            __pyx_t_12 = __Pyx_GetItemInt(__pyx_v_l, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1078, __pyx_L9_error)
            __Pyx_GOTREF(__pyx_t_12);
            __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_encode); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1078, __pyx_L9_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            __pyx_t_12 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
              __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_13);
              if (likely(__pyx_t_12)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
                __Pyx_INCREF(__pyx_t_12);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_13, function);
              }
            }
            __pyx_t_5 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_13, __pyx_t_12, __pyx_n_s_ascii) : __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_n_s_ascii);
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1078, __pyx_L9_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            __pyx_t_13 = __Pyx_PyObject_GetSlice(__pyx_v_l, 1, 0, NULL, NULL, &__pyx_slice__19, 1, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1078, __pyx_L9_error)
            __Pyx_GOTREF(__pyx_t_13);
            __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1078, __pyx_L9_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_GIVEREF(__pyx_t_5);
            PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_5);
            __Pyx_GIVEREF(__pyx_t_13);
            PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_t_13);
            __pyx_t_5 = 0;
            __pyx_t_13 = 0;
            if (unlikely(__Pyx_ListComp_Append(__pyx_t_6, (PyObject*)__pyx_t_12))) __PYX_ERR(0, 1078, __pyx_L9_error)
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

            /* "astropy/io/ascii/cparser.pyx":1079
 *         if read:
 *             fill_values = dict([(l[0].encode('ascii'), l[1:]) for
 *                                 l in fill_values if l[0] != ''])             # <<<<<<<<<<<<<<
 *         else:
 *             # don't worry about encoding for writing
 */
          }

          /* "astropy/io/ascii/cparser.pyx":1078
 *         # Create a dict with the values to be replaced as keys
 *         if read:
 *             fill_values = dict([(l[0].encode('ascii'), l[1:]) for             # <<<<<<<<<<<<<<
 *                                 l in fill_values if l[0] != ''])
 *         else:
 */
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1078, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF_SET(__pyx_v_fill_values, __pyx_t_4);
        __pyx_t_4 = 0;

        /* "astropy/io/ascii/cparser.pyx":1077
 *     try:
 *         # Create a dict with the values to be replaced as keys
 *         if read:             # <<<<<<<<<<<<<<
 *             fill_values = dict([(l[0].encode('ascii'), l[1:]) for
 *                                 l in fill_values if l[0] != ''])
 */
        goto __pyx_L15;
      }

      /* "astropy/io/ascii/cparser.pyx":1082
 *         else:
 *             # don't worry about encoding for writing
 *             fill_values = dict([(l[0], l[1:]) for l in fill_values])             # <<<<<<<<<<<<<<
 * 
 *     except IndexError:
 */
      /*else*/ {
        __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1082, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (likely(PyList_CheckExact(__pyx_v_fill_values)) || PyTuple_CheckExact(__pyx_v_fill_values)) {
          __pyx_t_6 = __pyx_v_fill_values; __Pyx_INCREF(__pyx_t_6); __pyx_t_2 = 0;
          __pyx_t_8 = NULL;
        } else {
          __pyx_t_2 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_v_fill_values); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1082, __pyx_L9_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_8 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1082, __pyx_L9_error)
        }
        for (;;) {
          if (likely(!__pyx_t_8)) {
            if (likely(PyList_CheckExact(__pyx_t_6))) {
              if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_6)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_12 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_2); __Pyx_INCREF(__pyx_t_12); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 1082, __pyx_L9_error)
              #else
              __pyx_t_12 = PySequence_ITEM(__pyx_t_6, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1082, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_12);
              #endif
            } else {
              if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_12 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_2); __Pyx_INCREF(__pyx_t_12); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 1082, __pyx_L9_error)
              #else
              __pyx_t_12 = PySequence_ITEM(__pyx_t_6, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1082, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_12);
              #endif
            }
          } else {
            __pyx_t_12 = __pyx_t_8(__pyx_t_6);
            if (unlikely(!__pyx_t_12)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 1082, __pyx_L9_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_12);
          }
          __Pyx_XDECREF_SET(__pyx_v_l, __pyx_t_12);
          __pyx_t_12 = 0;
          __pyx_t_12 = __Pyx_GetItemInt(__pyx_v_l, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1082, __pyx_L9_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_13 = __Pyx_PyObject_GetSlice(__pyx_v_l, 1, 0, NULL, NULL, &__pyx_slice__19, 1, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1082, __pyx_L9_error)
          __Pyx_GOTREF(__pyx_t_13);
          __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1082, __pyx_L9_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_12);
          PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_12);
          __Pyx_GIVEREF(__pyx_t_13);
          PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_13);
          __pyx_t_12 = 0;
          __pyx_t_13 = 0;
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 1082, __pyx_L9_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1082, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF_SET(__pyx_v_fill_values, __pyx_t_6);
        __pyx_t_6 = 0;
      }
      __pyx_L15:;

      /* "astropy/io/ascii/cparser.pyx":1075
 *             break
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         # Create a dict with the values to be replaced as keys
 *         if read:
 */
    }
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    goto __pyx_L14_try_end;
    __pyx_L9_error:;
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "astropy/io/ascii/cparser.pyx":1084
 *             fill_values = dict([(l[0], l[1:]) for l in fill_values])
 * 
 *     except IndexError:             # <<<<<<<<<<<<<<
 *         raise ValueError("Format of fill_values must be "
 *                          "(<bad>, <fill>, <optional col1>, ...)")
 */
    __pyx_t_14 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_IndexError);
    if (__pyx_t_14) {
      __Pyx_AddTraceback("astropy.io.ascii.cparser.get_fill_values", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_4, &__pyx_t_5) < 0) __PYX_ERR(0, 1084, __pyx_L11_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_5);

      /* "astropy/io/ascii/cparser.pyx":1085
 * 
 *     except IndexError:
 *         raise ValueError("Format of fill_values must be "             # <<<<<<<<<<<<<<
 *                          "(<bad>, <fill>, <optional col1>, ...)")
 *     if read:
 */
      __pyx_t_13 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__25, NULL); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1085, __pyx_L11_except_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_Raise(__pyx_t_13, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __PYX_ERR(0, 1085, __pyx_L11_except_error)
    }
    goto __pyx_L11_except_error;
    __pyx_L11_except_error:;

    /* "astropy/io/ascii/cparser.pyx":1075
 *             break
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         # Create a dict with the values to be replaced as keys
 *         if read:
 */
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_XGIVEREF(__pyx_t_10);
    __Pyx_XGIVEREF(__pyx_t_11);
    __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
    goto __pyx_L1_error;
    __pyx_L14_try_end:;
  }

  /* "astropy/io/ascii/cparser.pyx":1087
 *         raise ValueError("Format of fill_values must be "
 *                          "(<bad>, <fill>, <optional col1>, ...)")
 *     if read:             # <<<<<<<<<<<<<<
 *         return (fill_values, fill_empty)
 *     else:
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_read); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1087, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "astropy/io/ascii/cparser.pyx":1088
 *                          "(<bad>, <fill>, <optional col1>, ...)")
 *     if read:
 *         return (fill_values, fill_empty)             # <<<<<<<<<<<<<<
 *     else:
 *         return fill_values # cache for empty values doesn't matter for writing
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1088, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_fill_values);
    __Pyx_GIVEREF(__pyx_v_fill_values);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_fill_values);
    __Pyx_INCREF(__pyx_v_fill_empty);
    __Pyx_GIVEREF(__pyx_v_fill_empty);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_fill_empty);
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "astropy/io/ascii/cparser.pyx":1087
 *         raise ValueError("Format of fill_values must be "
 *                          "(<bad>, <fill>, <optional col1>, ...)")
 *     if read:             # <<<<<<<<<<<<<<
 *         return (fill_values, fill_empty)
 *     else:
 */
  }

  /* "astropy/io/ascii/cparser.pyx":1090
 *         return (fill_values, fill_empty)
 *     else:
 *         return fill_values # cache for empty values doesn't matter for writing             # <<<<<<<<<<<<<<
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_fill_values);
    __pyx_r = __pyx_v_fill_values;
    goto __pyx_L0;
  }

  /* "astropy/io/ascii/cparser.pyx":1061
 *             output.close()
 * 
 * def get_fill_values(fill_values, read=True):             # <<<<<<<<<<<<<<
 *     if len(fill_values) > 0 and isinstance(fill_values[0], six.string_types):
 *         # e.g. fill_values=('999', '0')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("astropy.io.ascii.cparser.get_fill_values", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fill_empty);
  __Pyx_XDECREF(__pyx_v_el);
  __Pyx_XDECREF(__pyx_v_l);
  __Pyx_XDECREF(__pyx_v_fill_values);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":258
 *         # experimental exception made for __getbuffer__ and __releasebuffer__
 *         # -- the details of this may change.
 *         def __getbuffer__(ndarray self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fulfill the PEP.
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5numpy_7ndarray___getbuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_v_i;
  int __pyx_v_ndim;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  int __pyx_v_t;
  char *__pyx_v_f;
  PyArray_Descr *__pyx_v_descr = 0;
  int __pyx_v_offset;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyArray_Descr *__pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  char *__pyx_t_9;
  if (__pyx_v_info == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":265
 * 
 *             cdef int i, ndim
 *             cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 */
  __pyx_v_endian_detector = 1;

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":266
 *             cdef int i, ndim
 *             cdef int endian_detector = 1
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 * 
 *             ndim = PyArray_NDIM(self)
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":268
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 *             ndim = PyArray_NDIM(self)             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 */
  __pyx_v_ndim = PyArray_NDIM(__pyx_v_self);

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":270
 *             ndim = PyArray_NDIM(self)
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  __pyx_t_2 = (((__pyx_v_flags & PyBUF_C_CONTIGUOUS) == PyBUF_C_CONTIGUOUS) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":271
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 */
  __pyx_t_2 = ((!(PyArray_CHKFLAGS(__pyx_v_self, NPY_ARRAY_C_CONTIGUOUS) != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":270
 *             ndim = PyArray_NDIM(self)
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  if (unlikely(__pyx_t_1)) {

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":272
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__26, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 272, __pyx_L1_error)

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":270
 *             ndim = PyArray_NDIM(self)
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  }

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  __pyx_t_2 = (((__pyx_v_flags & PyBUF_F_CONTIGUOUS) == PyBUF_F_CONTIGUOUS) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":275
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 */
  __pyx_t_2 = ((!(PyArray_CHKFLAGS(__pyx_v_self, NPY_ARRAY_F_CONTIGUOUS) != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L7_bool_binop_done:;

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  if (unlikely(__pyx_t_1)) {

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":276
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__27, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 276, __pyx_L1_error)

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  }

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":278
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 *             info.buf = PyArray_DATA(self)             # <<<<<<<<<<<<<<
 *             info.ndim = ndim
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_v_info->buf = PyArray_DATA(__pyx_v_self);

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":279
 * 
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim             # <<<<<<<<<<<<<<
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 # Allocate new buffer for strides and shape info.
 */
  __pyx_v_info->ndim = __pyx_v_ndim;

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":280
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 */
  __pyx_t_1 = (((sizeof(npy_intp)) != (sizeof(Py_ssize_t))) != 0);
  if (__pyx_t_1) {

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":283
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>PyObject_Malloc(sizeof(Py_ssize_t) * 2 * <size_t>ndim)             # <<<<<<<<<<<<<<
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 */
    __pyx_v_info->strides = ((Py_ssize_t *)PyObject_Malloc((((sizeof(Py_ssize_t)) * 2) * ((size_t)__pyx_v_ndim))));

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":284
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>PyObject_Malloc(sizeof(Py_ssize_t) * 2 * <size_t>ndim)
 *                 info.shape = info.strides + ndim             # <<<<<<<<<<<<<<
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 */
    __pyx_v_info->shape = (__pyx_v_info->strides + __pyx_v_ndim);

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":285
 *                 info.strides = <Py_ssize_t*>PyObject_Malloc(sizeof(Py_ssize_t) * 2 * <size_t>ndim)
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):             # <<<<<<<<<<<<<<
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 */
    __pyx_t_4 = __pyx_v_ndim;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":286
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]             # <<<<<<<<<<<<<<
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 */
      (__pyx_v_info->strides[__pyx_v_i]) = (PyArray_STRIDES(__pyx_v_self)[__pyx_v_i]);

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":287
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]             # <<<<<<<<<<<<<<
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 */
      (__pyx_v_info->shape[__pyx_v_i]) = (PyArray_DIMS(__pyx_v_self)[__pyx_v_i]);
    }

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":280
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 */
    goto __pyx_L9;
  }

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":289
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)             # <<<<<<<<<<<<<<
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 */
  /*else*/ {
    __pyx_v_info->strides = ((Py_ssize_t *)PyArray_STRIDES(__pyx_v_self));

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":290
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)             # <<<<<<<<<<<<<<
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 */
    __pyx_v_info->shape = ((Py_ssize_t *)PyArray_DIMS(__pyx_v_self));
  }
  __pyx_L9:;

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":291
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 */
  __pyx_v_info->suboffsets = NULL;

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":292
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)             # <<<<<<<<<<<<<<
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 * 
 */
  __pyx_v_info->itemsize = PyArray_ITEMSIZE(__pyx_v_self);

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":293
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)             # <<<<<<<<<<<<<<
 * 
 *             cdef int t
 */
  __pyx_v_info->readonly = (!(PyArray_ISWRITEABLE(__pyx_v_self) != 0));

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":296
 * 
 *             cdef int t
 *             cdef char* f = NULL             # <<<<<<<<<<<<<<
 *             cdef dtype descr = <dtype>PyArray_DESCR(self)
 *             cdef int offset
 */
  __pyx_v_f = NULL;

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":297
 *             cdef int t
 *             cdef char* f = NULL
 *             cdef dtype descr = <dtype>PyArray_DESCR(self)             # <<<<<<<<<<<<<<
 *             cdef int offset
 * 
 */
  __pyx_t_7 = PyArray_DESCR(__pyx_v_self);
  __pyx_t_3 = ((PyObject *)__pyx_t_7);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_descr = ((PyArray_Descr *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":300
 *             cdef int offset
 * 
 *             info.obj = self             # <<<<<<<<<<<<<<
 * 
 *             if not PyDataType_HASFIELDS(descr):
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":302
 *             info.obj = self
 * 
 *             if not PyDataType_HASFIELDS(descr):             # <<<<<<<<<<<<<<
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 */
  __pyx_t_1 = ((!(PyDataType_HASFIELDS(__pyx_v_descr) != 0)) != 0);
  if (__pyx_t_1) {

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":303
 * 
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num             # <<<<<<<<<<<<<<
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 */
    __pyx_t_4 = __pyx_v_descr->type_num;
    __pyx_v_t = __pyx_t_4;

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":304
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_2 = ((__pyx_v_descr->byteorder == '>') != 0);
    if (!__pyx_t_2) {
      goto __pyx_L15_next_or;
    } else {
    }
    __pyx_t_2 = (__pyx_v_little_endian != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_L15_next_or:;

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":305
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 */
    __pyx_t_2 = ((__pyx_v_descr->byteorder == '<') != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_t_2 = ((!(__pyx_v_little_endian != 0)) != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L14_bool_binop_done:;

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":304
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    if (unlikely(__pyx_t_1)) {

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":306
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__28, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 306, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(2, 306, __pyx_L1_error)

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":304
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    }

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":307
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 */
    switch (__pyx_v_t) {
      case NPY_BYTE:
      __pyx_v_f = ((char *)"b");
      break;
      case NPY_UBYTE:

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":308
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 */
      __pyx_v_f = ((char *)"B");
      break;
      case NPY_SHORT:

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":309
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 */
      __pyx_v_f = ((char *)"h");
      break;
      case NPY_USHORT:

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":310
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 */
      __pyx_v_f = ((char *)"H");
      break;
      case NPY_INT:

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":311
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 */
      __pyx_v_f = ((char *)"i");
      break;
      case NPY_UINT:

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":312
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 */
      __pyx_v_f = ((char *)"I");
      break;
      case NPY_LONG:

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":313
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 */
      __pyx_v_f = ((char *)"l");
      break;
      case NPY_ULONG:

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":314
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 */
      __pyx_v_f = ((char *)"L");
      break;
      case NPY_LONGLONG:

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":315
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 */
      __pyx_v_f = ((char *)"q");
      break;
      case NPY_ULONGLONG:

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":316
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 */
      __pyx_v_f = ((char *)"Q");
      break;
      case NPY_FLOAT:

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":317
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 */
      __pyx_v_f = ((char *)"f");
      break;
      case NPY_DOUBLE:

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":318
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 */
      __pyx_v_f = ((char *)"d");
      break;
      case NPY_LONGDOUBLE:

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":319
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 */
      __pyx_v_f = ((char *)"g");
      break;
      case NPY_CFLOAT:

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":320
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 */
      __pyx_v_f = ((char *)"Zf");
      break;
      case NPY_CDOUBLE:

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":321
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"
 */
      __pyx_v_f = ((char *)"Zd");
      break;
      case NPY_CLONGDOUBLE:

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":322
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 */
      __pyx_v_f = ((char *)"Zg");
      break;
      case NPY_OBJECT:

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":323
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
      __pyx_v_f = ((char *)"O");
      break;
      default:

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":325
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *                 info.format = f
 *                 return
 */
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_t); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = PyUnicode_Format(__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(2, 325, __pyx_L1_error)
      break;
    }

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":326
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f             # <<<<<<<<<<<<<<
 *                 return
 *             else:
 */
    __pyx_v_info->format = __pyx_v_f;

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":327
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f
 *                 return             # <<<<<<<<<<<<<<
 *             else:
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":302
 *             info.obj = self
 * 
 *             if not PyDataType_HASFIELDS(descr):             # <<<<<<<<<<<<<<
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 */
  }

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":329
 *                 return
 *             else:
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)             # <<<<<<<<<<<<<<
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0
 */
  /*else*/ {
    __pyx_v_info->format = ((char *)PyObject_Malloc(0xFF));

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":330
 *             else:
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment             # <<<<<<<<<<<<<<
 *                 offset = 0
 *                 f = _util_dtypestring(descr, info.format + 1,
 */
    (__pyx_v_info->format[0]) = '^';

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":331
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0             # <<<<<<<<<<<<<<
 *                 f = _util_dtypestring(descr, info.format + 1,
 *                                       info.format + _buffer_format_string_len,
 */
    __pyx_v_offset = 0;

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":332
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0
 *                 f = _util_dtypestring(descr, info.format + 1,             # <<<<<<<<<<<<<<
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)
 */
    __pyx_t_9 = __pyx_f_5numpy__util_dtypestring(__pyx_v_descr, (__pyx_v_info->format + 1), (__pyx_v_info->format + 0xFF), (&__pyx_v_offset)); if (unlikely(__pyx_t_9 == ((char *)NULL))) __PYX_ERR(2, 332, __pyx_L1_error)
    __pyx_v_f = __pyx_t_9;

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":335
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)
 *                 f[0] = c'\0' # Terminate format string             # <<<<<<<<<<<<<<
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 */
    (__pyx_v_f[0]) = '\x00';
  }

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":258
 *         # experimental exception made for __getbuffer__ and __releasebuffer__
 *         # -- the details of this may change.
 *         def __getbuffer__(ndarray self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fulfill the PEP.
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.ndarray.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_descr);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":337
 *                 f[0] = c'\0' # Terminate format string
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 */

/* Python wrapper */
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info); /*proto*/
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_pf_5numpy_7ndarray_2__releasebuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__releasebuffer__", 0);

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":338
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_t_1 = (PyArray_HASFIELDS(__pyx_v_self) != 0);
  if (__pyx_t_1) {

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":339
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)             # <<<<<<<<<<<<<<
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 PyObject_Free(info.strides)
 */
    PyObject_Free(__pyx_v_info->format);

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":338
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  }

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":340
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.strides)
 *                 # info.shape was stored after info.strides in the same block
 */
  __pyx_t_1 = (((sizeof(npy_intp)) != (sizeof(Py_ssize_t))) != 0);
  if (__pyx_t_1) {

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":341
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 PyObject_Free(info.strides)             # <<<<<<<<<<<<<<
 *                 # info.shape was stored after info.strides in the same block
 * 
 */
    PyObject_Free(__pyx_v_info->strides);

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":340
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.strides)
 *                 # info.shape was stored after info.strides in the same block
 */
  }

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":337
 *                 f[0] = c'\0' # Terminate format string
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":821
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":822
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":821
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":824
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":825
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":824
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":827
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":828
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":827
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":830
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":831
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":830
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":833
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":834
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":833
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":836
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 0);

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":837
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  __pyx_t_1 = (PyDataType_HASSUBARRAY(__pyx_v_d) != 0);
  if (__pyx_t_1) {

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":838
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject*)__pyx_v_d->subarray->shape));
    __pyx_r = ((PyObject*)__pyx_v_d->subarray->shape);
    goto __pyx_L0;

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":837
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  }

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":840
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_empty_tuple);
    __pyx_r = __pyx_empty_tuple;
    goto __pyx_L0;
  }

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":836
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":842
 *         return ()
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:             # <<<<<<<<<<<<<<
 *     # Recursive utility function used in __getbuffer__ to get format
 *     # string. The new location in the format string is returned.
 */

static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *__pyx_v_descr, char *__pyx_v_f, char *__pyx_v_end, int *__pyx_v_offset) {
  PyArray_Descr *__pyx_v_child = 0;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  PyObject *__pyx_v_fields = 0;
  PyObject *__pyx_v_childname = NULL;
  PyObject *__pyx_v_new_offset = NULL;
  PyObject *__pyx_v_t = NULL;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  long __pyx_t_8;
  char *__pyx_t_9;
  __Pyx_RefNannySetupContext("_util_dtypestring", 0);

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":847
 * 
 *     cdef dtype child
 *     cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 *     cdef tuple fields
 */
  __pyx_v_endian_detector = 1;

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":848
 *     cdef dtype child
 *     cdef int endian_detector = 1
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 *     cdef tuple fields
 * 
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":851
 *     cdef tuple fields
 * 
 *     for childname in descr.names:             # <<<<<<<<<<<<<<
 *         fields = descr.fields[childname]
 *         child, new_offset = fields
 */
  if (unlikely(__pyx_v_descr->names == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(2, 851, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_descr->names; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(2, 851, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 851, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_childname, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":852
 * 
 *     for childname in descr.names:
 *         fields = descr.fields[childname]             # <<<<<<<<<<<<<<
 *         child, new_offset = fields
 * 
 */
    if (unlikely(__pyx_v_descr->fields == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 852, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_descr->fields, __pyx_v_childname); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 852, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(PyTuple_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(2, 852, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_fields, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":853
 *     for childname in descr.names:
 *         fields = descr.fields[childname]
 *         child, new_offset = fields             # <<<<<<<<<<<<<<
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 */
    if (likely(__pyx_v_fields != Py_None)) {
      PyObject* sequence = __pyx_v_fields;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(2, 853, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 853, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 853, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
    } else {
      __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(2, 853, __pyx_L1_error)
    }
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_dtype))))) __PYX_ERR(2, 853, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_child, ((PyArray_Descr *)__pyx_t_3));
    __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_new_offset, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":855
 *         child, new_offset = fields
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 */
    __pyx_t_4 = __Pyx_PyInt_From_int((__pyx_v_offset[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 855, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyNumber_Subtract(__pyx_v_new_offset, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 855, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 855, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = ((((__pyx_v_end - __pyx_v_f) - ((int)__pyx_t_5)) < 15) != 0);
    if (unlikely(__pyx_t_6)) {

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":856
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__29, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 856, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(2, 856, __pyx_L1_error)

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":855
 *         child, new_offset = fields
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 */
    }

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":858
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_7 = ((__pyx_v_child->byteorder == '>') != 0);
    if (!__pyx_t_7) {
      goto __pyx_L8_next_or;
    } else {
    }
    __pyx_t_7 = (__pyx_v_little_endian != 0);
    if (!__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_L8_next_or:;

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":859
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *             raise ValueError(u"Non-native byte order not supported")
 *             # One could encode it in the format string and have Cython
 */
    __pyx_t_7 = ((__pyx_v_child->byteorder == '<') != 0);
    if (__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_7 = ((!(__pyx_v_little_endian != 0)) != 0);
    __pyx_t_6 = __pyx_t_7;
    __pyx_L7_bool_binop_done:;

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":858
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    if (unlikely(__pyx_t_6)) {

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":860
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *             # One could encode it in the format string and have Cython
 *             # complain instead, BUT: < and > in format strings also imply
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__28, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 860, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(2, 860, __pyx_L1_error)

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":858
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    }

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":870
 * 
 *         # Output padding bytes
 *         while offset[0] < new_offset:             # <<<<<<<<<<<<<<
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 */
    while (1) {
      __pyx_t_3 = __Pyx_PyInt_From_int((__pyx_v_offset[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 870, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_t_3, __pyx_v_new_offset, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 870, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 870, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (!__pyx_t_6) break;

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":871
 *         # Output padding bytes
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte             # <<<<<<<<<<<<<<
 *             f += 1
 *             offset[0] += 1
 */
      (__pyx_v_f[0]) = 0x78;

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":872
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte
 *             f += 1             # <<<<<<<<<<<<<<
 *             offset[0] += 1
 * 
 */
      __pyx_v_f = (__pyx_v_f + 1);

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":873
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 *             offset[0] += 1             # <<<<<<<<<<<<<<
 * 
 *         offset[0] += child.itemsize
 */
      __pyx_t_8 = 0;
      (__pyx_v_offset[__pyx_t_8]) = ((__pyx_v_offset[__pyx_t_8]) + 1);
    }

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":875
 *             offset[0] += 1
 * 
 *         offset[0] += child.itemsize             # <<<<<<<<<<<<<<
 * 
 *         if not PyDataType_HASFIELDS(child):
 */
    __pyx_t_8 = 0;
    (__pyx_v_offset[__pyx_t_8]) = ((__pyx_v_offset[__pyx_t_8]) + __pyx_v_child->elsize);

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":877
 *         offset[0] += child.itemsize
 * 
 *         if not PyDataType_HASFIELDS(child):             # <<<<<<<<<<<<<<
 *             t = child.type_num
 *             if end - f < 5:
 */
    __pyx_t_6 = ((!(PyDataType_HASFIELDS(__pyx_v_child) != 0)) != 0);
    if (__pyx_t_6) {

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":878
 * 
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num             # <<<<<<<<<<<<<<
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")
 */
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_child->type_num); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 878, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_t, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":879
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num
 *             if end - f < 5:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(u"Format string allocated too short.")
 * 
 */
      __pyx_t_6 = (((__pyx_v_end - __pyx_v_f) < 5) != 0);
      if (unlikely(__pyx_t_6)) {

        /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":880
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__30, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 880, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(2, 880, __pyx_L1_error)

        /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":879
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num
 *             if end - f < 5:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(u"Format string allocated too short.")
 * 
 */
      }

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":883
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_BYTE); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 883, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 883, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 883, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 98;
        goto __pyx_L15;
      }

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":884
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"             # <<<<<<<<<<<<<<
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_UBYTE); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 884, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 884, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 884, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 66;
        goto __pyx_L15;
      }

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":885
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"             # <<<<<<<<<<<<<<
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_SHORT); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 885, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 885, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 885, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x68;
        goto __pyx_L15;
      }

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":886
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"             # <<<<<<<<<<<<<<
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_USHORT); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 886, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 886, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 886, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 72;
        goto __pyx_L15;
      }

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":887
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_INT); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 887, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 887, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 887, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x69;
        goto __pyx_L15;
      }

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":888
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_UINT); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 888, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 888, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 888, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 73;
        goto __pyx_L15;
      }

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":889
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONG); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 889, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 889, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 889, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x6C;
        goto __pyx_L15;
      }

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":890
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_ULONG); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 890, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 890, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 890, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 76;
        goto __pyx_L15;
      }

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":891
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONGLONG); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 891, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 891, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 891, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x71;
        goto __pyx_L15;
      }

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":892
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_ULONGLONG); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 892, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 892, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 892, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 81;
        goto __pyx_L15;
      }

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":893
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"             # <<<<<<<<<<<<<<
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_FLOAT); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 893, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 893, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 893, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x66;
        goto __pyx_L15;
      }

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":894
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_DOUBLE); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 894, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 894, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 894, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x64;
        goto __pyx_L15;
      }

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":895
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"             # <<<<<<<<<<<<<<
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONGDOUBLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 895, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 895, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 895, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x67;
        goto __pyx_L15;
      }

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":896
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf             # <<<<<<<<<<<<<<
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CFLOAT); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 896, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 896, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 896, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x66;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":897
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd             # <<<<<<<<<<<<<<
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CDOUBLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 897, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 897, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 897, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x64;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":898
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg             # <<<<<<<<<<<<<<
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CLONGDOUBLE); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 898, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 898, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 898, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x67;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":899
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"             # <<<<<<<<<<<<<<
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_OBJECT); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 899, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 899, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 899, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (likely(__pyx_t_6)) {
        (__pyx_v_f[0]) = 79;
        goto __pyx_L15;
      }

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":901
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *             f += 1
 *         else:
 */
      /*else*/ {
        __pyx_t_3 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_v_t); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 901, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 901, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(2, 901, __pyx_L1_error)
      }
      __pyx_L15:;

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":902
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *             f += 1             # <<<<<<<<<<<<<<
 *         else:
 *             # Cython ignores struct boundary information ("T{...}"),
 */
      __pyx_v_f = (__pyx_v_f + 1);

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":877
 *         offset[0] += child.itemsize
 * 
 *         if not PyDataType_HASFIELDS(child):             # <<<<<<<<<<<<<<
 *             t = child.type_num
 *             if end - f < 5:
 */
      goto __pyx_L13;
    }

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":906
 *             # Cython ignores struct boundary information ("T{...}"),
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)             # <<<<<<<<<<<<<<
 *     return f
 * 
 */
    /*else*/ {
      __pyx_t_9 = __pyx_f_5numpy__util_dtypestring(__pyx_v_child, __pyx_v_f, __pyx_v_end, __pyx_v_offset); if (unlikely(__pyx_t_9 == ((char *)NULL))) __PYX_ERR(2, 906, __pyx_L1_error)
      __pyx_v_f = __pyx_t_9;
    }
    __pyx_L13:;

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":851
 *     cdef tuple fields
 * 
 *     for childname in descr.names:             # <<<<<<<<<<<<<<
 *         fields = descr.fields[childname]
 *         child, new_offset = fields
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":907
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)
 *     return f             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_f;
  goto __pyx_L0;

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":842
 *         return ()
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:             # <<<<<<<<<<<<<<
 *     # Recursive utility function used in __getbuffer__ to get format
 *     # string. The new location in the format string is returned.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("numpy._util_dtypestring", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_child);
  __Pyx_XDECREF(__pyx_v_fields);
  __Pyx_XDECREF(__pyx_v_childname);
  __Pyx_XDECREF(__pyx_v_new_offset);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1022
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1023
 * 
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */
  Py_INCREF(__pyx_v_base);

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1024
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  (void)(PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base));

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1022
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1026
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1027
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1028
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  __pyx_t_1 = ((__pyx_v_base == NULL) != 0);
  if (__pyx_t_1) {

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1029
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1028
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  }

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1030
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_base));
  __pyx_r = ((PyObject *)__pyx_v_base);
  goto __pyx_L0;

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1026
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1034
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_array()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_array", 0);

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1035
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1036
 * cdef inline int import_array() except -1:
 *     try:
 *         _import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 1036, __pyx_L3_error)

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1035
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1037
 *     try:
 *         _import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 1037, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1038
 *         _import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__31, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 1038, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 1038, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1035
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1034
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_array()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1040
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_umath", 0);

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1041
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1042
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 1042, __pyx_L3_error)

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1041
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1043
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 1043, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1044
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__32, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 1044, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 1044, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1041
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1040
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1046
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_ufunc", 0);

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1047
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1048
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 1048, __pyx_L3_error)

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1047
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1049
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 1049, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1050
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__32, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 1050, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 1050, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1047
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1046
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_tp_new_7astropy_2io_5ascii_7cparser_FileString(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString *)o);
  p->fhandle = Py_None; Py_INCREF(Py_None);
  p->mmap = Py_None; Py_INCREF(Py_None);
  p->buf.obj = NULL;
  if (unlikely(__pyx_pw_7astropy_2io_5ascii_7cparser_10FileString_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_7astropy_2io_5ascii_7cparser_FileString(PyObject *o) {
  struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString *p = (struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_7astropy_2io_5ascii_7cparser_10FileString_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->fhandle);
  Py_CLEAR(p->mmap);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_7astropy_2io_5ascii_7cparser_FileString(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString *p = (struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString *)o;
  if (p->fhandle) {
    e = (*v)(p->fhandle, a); if (e) return e;
  }
  if (p->mmap) {
    e = (*v)(p->mmap, a); if (e) return e;
  }
  if (p->buf.obj) {
    e = (*v)(p->buf.obj, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7astropy_2io_5ascii_7cparser_FileString(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString *p = (struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString *)o;
  tmp = ((PyObject*)p->fhandle);
  p->fhandle = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->mmap);
  p->mmap = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  Py_CLEAR(p->buf.obj);
  return 0;
}
static PyObject *__pyx_sq_item_7astropy_2io_5ascii_7cparser_FileString(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyMethodDef __pyx_methods_7astropy_2io_5ascii_7cparser_FileString[] = {
  {"splitlines", (PyCFunction)__pyx_pw_7astropy_2io_5ascii_7cparser_10FileString_9splitlines, METH_NOARGS, __pyx_doc_7astropy_2io_5ascii_7cparser_10FileString_8splitlines},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_7astropy_2io_5ascii_7cparser_10FileString_12__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_7astropy_2io_5ascii_7cparser_10FileString_14__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_FileString = {
  __pyx_pw_7astropy_2io_5ascii_7cparser_10FileString_5__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_7astropy_2io_5ascii_7cparser_FileString, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_FileString = {
  __pyx_pw_7astropy_2io_5ascii_7cparser_10FileString_5__len__, /*mp_length*/
  __pyx_pw_7astropy_2io_5ascii_7cparser_10FileString_7__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_7astropy_2io_5ascii_7cparser_FileString = {
  PyVarObject_HEAD_INIT(0, 0)
  "astropy.io.ascii.cparser.FileString", /*tp_name*/
  sizeof(struct __pyx_obj_7astropy_2io_5ascii_7cparser_FileString), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7astropy_2io_5ascii_7cparser_FileString, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_FileString, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_FileString, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "\n    A wrapper class for a memory-mapped file pointer.\n    ", /*tp_doc*/
  __pyx_tp_traverse_7astropy_2io_5ascii_7cparser_FileString, /*tp_traverse*/
  __pyx_tp_clear_7astropy_2io_5ascii_7cparser_FileString, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7astropy_2io_5ascii_7cparser_FileString, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7astropy_2io_5ascii_7cparser_FileString, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_7astropy_2io_5ascii_7cparser_CParser __pyx_vtable_7astropy_2io_5ascii_7cparser_CParser;

static PyObject *__pyx_tp_new_7astropy_2io_5ascii_7cparser_CParser(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *)o);
  p->__pyx_vtab = __pyx_vtabptr_7astropy_2io_5ascii_7cparser_CParser;
  p->names = Py_None; Py_INCREF(Py_None);
  p->header_names = Py_None; Py_INCREF(Py_None);
  p->data_end = Py_None; Py_INCREF(Py_None);
  p->include_names = Py_None; Py_INCREF(Py_None);
  p->exclude_names = Py_None; Py_INCREF(Py_None);
  p->fill_values = Py_None; Py_INCREF(Py_None);
  p->fill_empty = Py_None; Py_INCREF(Py_None);
  p->fill_include_names = Py_None; Py_INCREF(Py_None);
  p->fill_exclude_names = Py_None; Py_INCREF(Py_None);
  p->fill_names = Py_None; Py_INCREF(Py_None);
  p->source_bytes = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->parallel = Py_None; Py_INCREF(Py_None);
  p->use_cols = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->source = Py_None; Py_INCREF(Py_None);
  p->header_start = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_7astropy_2io_5ascii_7cparser_CParser(PyObject *o) {
  struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *p = (struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->names);
  Py_CLEAR(p->header_names);
  Py_CLEAR(p->data_end);
  Py_CLEAR(p->include_names);
  Py_CLEAR(p->exclude_names);
  Py_CLEAR(p->fill_values);
  Py_CLEAR(p->fill_empty);
  Py_CLEAR(p->fill_include_names);
  Py_CLEAR(p->fill_exclude_names);
  Py_CLEAR(p->fill_names);
  Py_CLEAR(p->source_bytes);
  Py_CLEAR(p->parallel);
  Py_CLEAR(p->use_cols);
  Py_CLEAR(p->source);
  Py_CLEAR(p->header_start);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_7astropy_2io_5ascii_7cparser_CParser(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *p = (struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *)o;
  if (p->names) {
    e = (*v)(p->names, a); if (e) return e;
  }
  if (p->header_names) {
    e = (*v)(p->header_names, a); if (e) return e;
  }
  if (p->data_end) {
    e = (*v)(p->data_end, a); if (e) return e;
  }
  if (p->include_names) {
    e = (*v)(p->include_names, a); if (e) return e;
  }
  if (p->exclude_names) {
    e = (*v)(p->exclude_names, a); if (e) return e;
  }
  if (p->fill_values) {
    e = (*v)(p->fill_values, a); if (e) return e;
  }
  if (p->fill_empty) {
    e = (*v)(p->fill_empty, a); if (e) return e;
  }
  if (p->fill_include_names) {
    e = (*v)(p->fill_include_names, a); if (e) return e;
  }
  if (p->fill_exclude_names) {
    e = (*v)(p->fill_exclude_names, a); if (e) return e;
  }
  if (p->fill_names) {
    e = (*v)(p->fill_names, a); if (e) return e;
  }
  if (p->parallel) {
    e = (*v)(p->parallel, a); if (e) return e;
  }
  if (p->use_cols) {
    e = (*v)(p->use_cols, a); if (e) return e;
  }
  if (p->source) {
    e = (*v)(p->source, a); if (e) return e;
  }
  if (p->header_start) {
    e = (*v)(p->header_start, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7astropy_2io_5ascii_7cparser_CParser(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *p = (struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *)o;
  tmp = ((PyObject*)p->names);
  p->names = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->header_names);
  p->header_names = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->data_end);
  p->data_end = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->include_names);
  p->include_names = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->exclude_names);
  p->exclude_names = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->fill_values);
  p->fill_values = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->fill_empty);
  p->fill_empty = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->fill_include_names);
  p->fill_include_names = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->fill_exclude_names);
  p->fill_exclude_names = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->fill_names);
  p->fill_names = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->parallel);
  p->parallel = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->use_cols);
  p->use_cols = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->source);
  p->source = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->header_start);
  p->header_start = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7astropy_2io_5ascii_7cparser_7CParser_width(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_5width_1__get__(o);
}

static int __pyx_setprop_7astropy_2io_5ascii_7cparser_7CParser_width(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_5width_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7astropy_2io_5ascii_7cparser_7CParser_source(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_6source_1__get__(o);
}

static int __pyx_setprop_7astropy_2io_5ascii_7cparser_7CParser_source(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_6source_3__set__(o, v);
  }
  else {
    return __pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_6source_5__del__(o);
  }
}

static PyObject *__pyx_getprop_7astropy_2io_5ascii_7cparser_7CParser_header_start(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_12header_start_1__get__(o);
}

static int __pyx_setprop_7astropy_2io_5ascii_7cparser_7CParser_header_start(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_12header_start_3__set__(o, v);
  }
  else {
    return __pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_12header_start_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_7astropy_2io_5ascii_7cparser_CParser[] = {
  {"setup_tokenizer", (PyCFunction)__pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_5setup_tokenizer, METH_O, 0},
  {"read_header", (PyCFunction)__pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_7read_header, METH_NOARGS, 0},
  {"read", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_9read, METH_VARARGS|METH_KEYWORDS, 0},
  {"_read_parallel", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_11_read_parallel, METH_VARARGS|METH_KEYWORDS, 0},
  {"get_names", (PyCFunction)__pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_13get_names, METH_NOARGS, 0},
  {"set_names", (PyCFunction)__pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_15set_names, METH_O, 0},
  {"get_header_names", (PyCFunction)__pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_17get_header_names, METH_NOARGS, 0},
  {"__reduce__", (PyCFunction)__pyx_pw_7astropy_2io_5ascii_7cparser_7CParser_19__reduce__, METH_NOARGS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7astropy_2io_5ascii_7cparser_CParser[] = {
  {(char *)"width", __pyx_getprop_7astropy_2io_5ascii_7cparser_7CParser_width, __pyx_setprop_7astropy_2io_5ascii_7cparser_7CParser_width, (char *)0, 0},
  {(char *)"source", __pyx_getprop_7astropy_2io_5ascii_7cparser_7CParser_source, __pyx_setprop_7astropy_2io_5ascii_7cparser_7CParser_source, (char *)0, 0},
  {(char *)"header_start", __pyx_getprop_7astropy_2io_5ascii_7cparser_7CParser_header_start, __pyx_setprop_7astropy_2io_5ascii_7cparser_7CParser_header_start, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_7astropy_2io_5ascii_7cparser_CParser = {
  PyVarObject_HEAD_INIT(0, 0)
  "astropy.io.ascii.cparser.CParser", /*tp_name*/
  sizeof(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7astropy_2io_5ascii_7cparser_CParser, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "\n    A fast Cython parser class which uses underlying C code\n    for tokenization.\n    ", /*tp_doc*/
  __pyx_tp_traverse_7astropy_2io_5ascii_7cparser_CParser, /*tp_traverse*/
  __pyx_tp_clear_7astropy_2io_5ascii_7cparser_CParser, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7astropy_2io_5ascii_7cparser_CParser, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7astropy_2io_5ascii_7cparser_CParser, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7astropy_2io_5ascii_7cparser_CParser, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_7astropy_2io_5ascii_7cparser_FastWriter __pyx_vtable_7astropy_2io_5ascii_7cparser_FastWriter;

static PyObject *__pyx_tp_new_7astropy_2io_5ascii_7cparser_FastWriter(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7astropy_2io_5ascii_7cparser_FastWriter *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7astropy_2io_5ascii_7cparser_FastWriter *)o);
  p->__pyx_vtab = __pyx_vtabptr_7astropy_2io_5ascii_7cparser_FastWriter;
  p->table = Py_None; Py_INCREF(Py_None);
  p->use_names = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->fill_values = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->fill_cols = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->col_iters = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->formats = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->format_funcs = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->types = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->line_comments = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->quotechar = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->expchar = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->delimiter = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->comment = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_7astropy_2io_5ascii_7cparser_10FastWriter_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_7astropy_2io_5ascii_7cparser_FastWriter(PyObject *o) {
  struct __pyx_obj_7astropy_2io_5ascii_7cparser_FastWriter *p = (struct __pyx_obj_7astropy_2io_5ascii_7cparser_FastWriter *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->table);
  Py_CLEAR(p->use_names);
  Py_CLEAR(p->fill_values);
  Py_CLEAR(p->fill_cols);
  Py_CLEAR(p->col_iters);
  Py_CLEAR(p->formats);
  Py_CLEAR(p->format_funcs);
  Py_CLEAR(p->types);
  Py_CLEAR(p->line_comments);
  Py_CLEAR(p->quotechar);
  Py_CLEAR(p->expchar);
  Py_CLEAR(p->delimiter);
  Py_CLEAR(p->comment);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_7astropy_2io_5ascii_7cparser_FastWriter(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7astropy_2io_5ascii_7cparser_FastWriter *p = (struct __pyx_obj_7astropy_2io_5ascii_7cparser_FastWriter *)o;
  if (p->table) {
    e = (*v)(p->table, a); if (e) return e;
  }
  if (p->use_names) {
    e = (*v)(p->use_names, a); if (e) return e;
  }
  if (p->fill_values) {
    e = (*v)(p->fill_values, a); if (e) return e;
  }
  if (p->fill_cols) {
    e = (*v)(p->fill_cols, a); if (e) return e;
  }
  if (p->col_iters) {
    e = (*v)(p->col_iters, a); if (e) return e;
  }
  if (p->formats) {
    e = (*v)(p->formats, a); if (e) return e;
  }
  if (p->format_funcs) {
    e = (*v)(p->format_funcs, a); if (e) return e;
  }
  if (p->types) {
    e = (*v)(p->types, a); if (e) return e;
  }
  if (p->line_comments) {
    e = (*v)(p->line_comments, a); if (e) return e;
  }
  if (p->comment) {
    e = (*v)(p->comment, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7astropy_2io_5ascii_7cparser_FastWriter(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7astropy_2io_5ascii_7cparser_FastWriter *p = (struct __pyx_obj_7astropy_2io_5ascii_7cparser_FastWriter *)o;
  tmp = ((PyObject*)p->table);
  p->table = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->use_names);
  p->use_names = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->fill_values);
  p->fill_values = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->fill_cols);
  p->fill_cols = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->col_iters);
  p->col_iters = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->formats);
  p->formats = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->format_funcs);
  p->format_funcs = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->types);
  p->types = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->line_comments);
  p->line_comments = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->comment);
  p->comment = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_7astropy_2io_5ascii_7cparser_FastWriter[] = {
  {"_write_header", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7astropy_2io_5ascii_7cparser_10FastWriter_3_write_header, METH_VARARGS|METH_KEYWORDS, 0},
  {"write", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7astropy_2io_5ascii_7cparser_10FastWriter_5write, METH_VARARGS|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_7astropy_2io_5ascii_7cparser_10FastWriter_7__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_7astropy_2io_5ascii_7cparser_10FastWriter_9__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_7astropy_2io_5ascii_7cparser_FastWriter = {
  PyVarObject_HEAD_INIT(0, 0)
  "astropy.io.ascii.cparser.FastWriter", /*tp_name*/
  sizeof(struct __pyx_obj_7astropy_2io_5ascii_7cparser_FastWriter), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7astropy_2io_5ascii_7cparser_FastWriter, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "\n    A fast Cython writing class for writing tables\n    as ASCII data.\n    ", /*tp_doc*/
  __pyx_tp_traverse_7astropy_2io_5ascii_7cparser_FastWriter, /*tp_traverse*/
  __pyx_tp_clear_7astropy_2io_5ascii_7cparser_FastWriter, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7astropy_2io_5ascii_7cparser_FastWriter, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7astropy_2io_5ascii_7cparser_FastWriter, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};

static struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines *__pyx_freelist_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines[8];
static int __pyx_freecount_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines = 0;

static PyObject *__pyx_tp_new_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines)))) {
    o = (PyObject*)__pyx_freelist_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines[--__pyx_freecount_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines];
    memset(o, 0, sizeof(struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines(PyObject *o) {
  struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines *p = (struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines)))) {
    __pyx_freelist_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines[__pyx_freecount_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines++] = ((struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines *p = (struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines = {
  PyVarObject_HEAD_INIT(0, 0)
  "astropy.io.ascii.cparser.__pyx_scope_struct__splitlines", /*tp_name*/
  sizeof(struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};

static struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel *__pyx_freelist_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel[8];
static int __pyx_freecount_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel = 0;

static PyObject *__pyx_tp_new_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel)))) {
    o = (PyObject*)__pyx_freelist_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel[--__pyx_freecount_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel];
    memset(o, 0, sizeof(struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel(PyObject *o) {
  struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel *p = (struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_col_chunks);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel)))) {
    __pyx_freelist_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel[__pyx_freecount_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel++] = ((struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel *p = (struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel *)o;
  if (p->__pyx_v_col_chunks) {
    e = (*v)(p->__pyx_v_col_chunks, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel *p = (struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel *)o;
  tmp = ((PyObject*)p->__pyx_v_col_chunks);
  p->__pyx_v_col_chunks = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = ((struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel = {
  PyVarObject_HEAD_INIT(0, 0)
  "astropy.io.ascii.cparser.__pyx_scope_struct_1__read_parallel", /*tp_name*/
  sizeof(struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel, /*tp_traverse*/
  __pyx_tp_clear_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};

static struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr *__pyx_freelist_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr[8];
static int __pyx_freecount_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr = 0;

static PyObject *__pyx_tp_new_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr)))) {
    o = (PyObject*)__pyx_freelist_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr[--__pyx_freecount_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr(PyObject *o) {
  struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr *p = (struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_name);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr)))) {
    __pyx_freelist_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr[__pyx_freecount_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr++] = ((struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr *p = (struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_name) {
    e = (*v)(p->__pyx_v_name, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "astropy.io.ascii.cparser.__pyx_scope_struct_2_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};

static struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr *__pyx_freelist_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr[8];
static int __pyx_freecount_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr = 0;

static PyObject *__pyx_tp_new_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr)))) {
    o = (PyObject*)__pyx_freelist_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr[--__pyx_freecount_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr(PyObject *o) {
  struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr *p = (struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_name);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr)))) {
    __pyx_freelist_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr[__pyx_freecount_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr++] = ((struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr *p = (struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_name) {
    e = (*v)(p->__pyx_v_name, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "astropy.io.ascii.cparser.__pyx_scope_struct_3_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};

static struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr *__pyx_freelist_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr[8];
static int __pyx_freecount_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr = 0;

static PyObject *__pyx_tp_new_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr)))) {
    o = (PyObject*)__pyx_freelist_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr[--__pyx_freecount_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr(PyObject *o) {
  struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr *p = (struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_col_chunk);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr)))) {
    __pyx_freelist_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr[__pyx_freecount_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr++] = ((struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr *p = (struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_col_chunk) {
    e = (*v)(p->__pyx_v_col_chunk, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "astropy.io.ascii.cparser.__pyx_scope_struct_4_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};

static struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk *__pyx_freelist_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk[8];
static int __pyx_freecount_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk = 0;

static PyObject *__pyx_tp_new_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk)))) {
    o = (PyObject*)__pyx_freelist_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk[--__pyx_freecount_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk];
    memset(o, 0, sizeof(struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk(PyObject *o) {
  struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk *p = (struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk)))) {
    __pyx_freelist_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk[__pyx_freecount_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk++] = ((struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk *p = (struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk *p = (struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = ((struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk = {
  PyVarObject_HEAD_INIT(0, 0)
  "astropy.io.ascii.cparser.__pyx_scope_struct_5__read_chunk", /*tp_name*/
  sizeof(struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk, /*tp_traverse*/
  __pyx_tp_clear_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};

static struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr *__pyx_freelist_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr[8];
static int __pyx_freecount_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr = 0;

static PyObject *__pyx_tp_new_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr)))) {
    o = (PyObject*)__pyx_freelist_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr[--__pyx_freecount_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr(PyObject *o) {
  struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr *p = (struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_name);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr)))) {
    __pyx_freelist_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr[__pyx_freecount_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr++] = ((struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr *p = (struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_name) {
    e = (*v)(p->__pyx_v_name, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "astropy.io.ascii.cparser.__pyx_scope_struct_6_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_cparser(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_cparser},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "cparser",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_s_0, __pyx_k_0, sizeof(__pyx_k_0), 0, 0, 1, 0},
  {&__pyx_kp_s_0_1, __pyx_k_0_1, sizeof(__pyx_k_0_1), 0, 0, 1, 0},
  {&__pyx_n_s_A, __pyx_k_A, sizeof(__pyx_k_A), 0, 0, 1, 1},
  {&__pyx_n_s_ACCESS_READ, __pyx_k_ACCESS_READ, sizeof(__pyx_k_ACCESS_READ), 0, 0, 1, 1},
  {&__pyx_kp_s_An_instance_of_this_class_is_th, __pyx_k_An_instance_of_this_class_is_th, sizeof(__pyx_k_An_instance_of_this_class_is_th), 0, 0, 1, 0},
  {&__pyx_n_s_AstropyWarning, __pyx_k_AstropyWarning, sizeof(__pyx_k_AstropyWarning), 0, 0, 1, 1},
  {&__pyx_n_s_AttributeError, __pyx_k_AttributeError, sizeof(__pyx_k_AttributeError), 0, 0, 1, 1},
  {&__pyx_n_s_CParser, __pyx_k_CParser, sizeof(__pyx_k_CParser), 0, 0, 1, 1},
  {&__pyx_n_s_CParserError, __pyx_k_CParserError, sizeof(__pyx_k_CParserError), 0, 0, 1, 1},
  {&__pyx_kp_s_Column_0_failed_to_convert, __pyx_k_Column_0_failed_to_convert, sizeof(__pyx_k_Column_0_failed_to_convert), 0, 0, 1, 0},
  {&__pyx_n_s_CsvWriter, __pyx_k_CsvWriter, sizeof(__pyx_k_CsvWriter), 0, 0, 1, 1},
  {&__pyx_n_s_E, __pyx_k_E, sizeof(__pyx_k_E), 0, 0, 1, 1},
  {&__pyx_n_s_ERR_CODES, __pyx_k_ERR_CODES, sizeof(__pyx_k_ERR_CODES), 0, 0, 1, 1},
  {&__pyx_n_s_Empty, __pyx_k_Empty, sizeof(__pyx_k_Empty), 0, 0, 1, 1},
  {&__pyx_n_s_FORT, __pyx_k_FORT, sizeof(__pyx_k_FORT), 0, 0, 1, 1},
  {&__pyx_n_s_FastOptionsError, __pyx_k_FastOptionsError, sizeof(__pyx_k_FastOptionsError), 0, 0, 1, 1},
  {&__pyx_n_s_FastWriter, __pyx_k_FastWriter, sizeof(__pyx_k_FastWriter), 0, 0, 1, 1},
  {&__pyx_n_s_FileString, __pyx_k_FileString, sizeof(__pyx_k_FileString), 0, 0, 1, 1},
  {&__pyx_n_s_FileString_splitlines, __pyx_k_FileString_splitlines, sizeof(__pyx_k_FileString_splitlines), 0, 0, 1, 1},
  {&__pyx_kp_s_File_0_could_not_be_opened, __pyx_k_File_0_could_not_be_opened, sizeof(__pyx_k_File_0_could_not_be_opened), 0, 0, 1, 0},
  {&__pyx_n_s_FloatType, __pyx_k_FloatType, sizeof(__pyx_k_FloatType), 0, 0, 1, 1},
  {&__pyx_kp_s_Format_of_fill_values_must_be_ba, __pyx_k_Format_of_fill_values_must_be_ba, sizeof(__pyx_k_Format_of_fill_values_must_be_ba), 0, 0, 1, 0},
  {&__pyx_kp_u_Format_string_allocated_too_shor, __pyx_k_Format_string_allocated_too_shor, sizeof(__pyx_k_Format_string_allocated_too_shor), 0, 1, 0, 0},
  {&__pyx_kp_u_Format_string_allocated_too_shor_2, __pyx_k_Format_string_allocated_too_shor_2, sizeof(__pyx_k_Format_string_allocated_too_shor_2), 0, 1, 0, 0},
  {&__pyx_n_s_Full, __pyx_k_Full, sizeof(__pyx_k_Full), 0, 0, 1, 1},
  {&__pyx_n_s_IOError, __pyx_k_IOError, sizeof(__pyx_k_IOError), 0, 0, 1, 1},
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_n_s_InconsistentTableError, __pyx_k_InconsistentTableError, sizeof(__pyx_k_InconsistentTableError), 0, 0, 1, 1},
  {&__pyx_n_s_IndexError, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
  {&__pyx_kp_s_Input_table_must_be_a_file_like, __pyx_k_Input_table_must_be_a_file_like, sizeof(__pyx_k_Input_table_must_be_a_file_like), 0, 0, 1, 0},
  {&__pyx_n_s_IntType, __pyx_k_IntType, sizeof(__pyx_k_IntType), 0, 0, 1, 1},
  {&__pyx_kp_s_Invalid_parameter_in_fast_reader, __pyx_k_Invalid_parameter_in_fast_reader, sizeof(__pyx_k_Invalid_parameter_in_fast_reader), 0, 0, 1, 0},
  {&__pyx_n_s_MaskError, __pyx_k_MaskError, sizeof(__pyx_k_MaskError), 0, 0, 1, 1},
  {&__pyx_kp_s_Multiprocessing_is_not_yet_suppo, __pyx_k_Multiprocessing_is_not_yet_suppo, sizeof(__pyx_k_Multiprocessing_is_not_yet_suppo), 0, 0, 1, 0},
  {&__pyx_n_s_N, __pyx_k_N, sizeof(__pyx_k_N), 0, 0, 1, 1},
  {&__pyx_kp_s_No_data_lines_found_C_reader_can, __pyx_k_No_data_lines_found_C_reader_can, sizeof(__pyx_k_No_data_lines_found_C_reader_can), 0, 0, 1, 0},
  {&__pyx_kp_u_Non_native_byte_order_not_suppor, __pyx_k_Non_native_byte_order_not_suppor, sizeof(__pyx_k_Non_native_byte_order_not_suppor), 0, 1, 0, 0},
  {&__pyx_n_s_NotImplementedError, __pyx_k_NotImplementedError, sizeof(__pyx_k_NotImplementedError), 0, 0, 1, 1},
  {&__pyx_n_s_OSError, __pyx_k_OSError, sizeof(__pyx_k_OSError), 0, 0, 1, 1},
  {&__pyx_kp_s_OverflowError_converting_to_0_in, __pyx_k_OverflowError_converting_to_0_in, sizeof(__pyx_k_OverflowError_converting_to_0_in), 0, 0, 1, 0},
  {&__pyx_kp_s_OverflowError_converting_to_0_in_2, __pyx_k_OverflowError_converting_to_0_in_2, sizeof(__pyx_k_OverflowError_converting_to_0_in_2), 0, 0, 1, 0},
  {&__pyx_n_s_PY2, __pyx_k_PY2, sizeof(__pyx_k_PY2), 0, 0, 1, 1},
  {&__pyx_n_s_PY3, __pyx_k_PY3, sizeof(__pyx_k_PY3), 0, 0, 1, 1},
  {&__pyx_n_s_ParameterError, __pyx_k_ParameterError, sizeof(__pyx_k_ParameterError), 0, 0, 1, 1},
  {&__pyx_n_s_Process, __pyx_k_Process, sizeof(__pyx_k_Process), 0, 0, 1, 1},
  {&__pyx_n_s_QUOTE_MINIMAL, __pyx_k_QUOTE_MINIMAL, sizeof(__pyx_k_QUOTE_MINIMAL), 0, 0, 1, 1},
  {&__pyx_n_s_Queue, __pyx_k_Queue, sizeof(__pyx_k_Queue), 0, 0, 1, 1},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_n_s_S, __pyx_k_S, sizeof(__pyx_k_S), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_U, __pyx_k_U, sizeof(__pyx_k_U), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_kp_s__12, __pyx_k__12, sizeof(__pyx_k__12), 0, 0, 1, 0},
  {&__pyx_kp_s__13, __pyx_k__13, sizeof(__pyx_k__13), 0, 0, 1, 0},
  {&__pyx_kp_s__14, __pyx_k__14, sizeof(__pyx_k__14), 0, 0, 1, 0},
  {&__pyx_kp_s__17, __pyx_k__17, sizeof(__pyx_k__17), 0, 0, 1, 0},
  {&__pyx_kp_s__20, __pyx_k__20, sizeof(__pyx_k__20), 0, 0, 1, 0},
  {&__pyx_kp_s__22, __pyx_k__22, sizeof(__pyx_k__22), 0, 0, 1, 0},
  {&__pyx_kp_s__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 0, 1, 0},
  {&__pyx_kp_s__4, __pyx_k__4, sizeof(__pyx_k__4), 0, 0, 1, 0},
  {&__pyx_kp_b__5, __pyx_k__5, sizeof(__pyx_k__5), 0, 0, 0, 0},
  {&__pyx_kp_s__5, __pyx_k__5, sizeof(__pyx_k__5), 0, 0, 1, 0},
  {&__pyx_n_s_access, __pyx_k_access, sizeof(__pyx_k_access), 0, 0, 1, 1},
  {&__pyx_kp_s_an_error_occurred_while_advancin, __pyx_k_an_error_occurred_while_advancin, sizeof(__pyx_k_an_error_occurred_while_advancin), 0, 0, 1, 0},
  {&__pyx_kp_s_an_error_occurred_while_advancin_2, __pyx_k_an_error_occurred_while_advancin_2, sizeof(__pyx_k_an_error_occurred_while_advancin_2), 0, 0, 1, 0},
  {&__pyx_kp_s_an_error_occurred_while_parsing, __pyx_k_an_error_occurred_while_parsing, sizeof(__pyx_k_an_error_occurred_while_parsing), 0, 0, 1, 0},
  {&__pyx_kp_s_an_error_occurred_while_tokenizi, __pyx_k_an_error_occurred_while_tokenizi, sizeof(__pyx_k_an_error_occurred_while_tokenizi), 0, 0, 1, 0},
  {&__pyx_kp_s_an_error_occurred_while_tokenizi_2, __pyx_k_an_error_occurred_while_tokenizi_2, sizeof(__pyx_k_an_error_occurred_while_tokenizi_2), 0, 0, 1, 0},
  {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_array, __pyx_k_array, sizeof(__pyx_k_array), 0, 0, 1, 1},
  {&__pyx_n_s_ascii, __pyx_k_ascii, sizeof(__pyx_k_ascii), 0, 0, 1, 1},
  {&__pyx_n_s_astropy_io_ascii_cparser, __pyx_k_astropy_io_ascii_cparser, sizeof(__pyx_k_astropy_io_ascii_cparser), 0, 0, 1, 1},
  {&__pyx_kp_s_astropy_io_ascii_cparser_pyx, __pyx_k_astropy_io_ascii_cparser_pyx, sizeof(__pyx_k_astropy_io_ascii_cparser_pyx), 0, 0, 1, 0},
  {&__pyx_n_s_ceil, __pyx_k_ceil, sizeof(__pyx_k_ceil), 0, 0, 1, 1},
  {&__pyx_n_s_chr, __pyx_k_chr, sizeof(__pyx_k_chr), 0, 0, 1, 1},
  {&__pyx_n_s_chunk_tokenizer, __pyx_k_chunk_tokenizer, sizeof(__pyx_k_chunk_tokenizer), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_col, __pyx_k_col, sizeof(__pyx_k_col), 0, 0, 1, 1},
  {&__pyx_kp_s_col_0, __pyx_k_col_0, sizeof(__pyx_k_col_0), 0, 0, 1, 0},
  {&__pyx_n_s_colnames, __pyx_k_colnames, sizeof(__pyx_k_colnames), 0, 0, 1, 1},
  {&__pyx_n_s_columns, __pyx_k_columns, sizeof(__pyx_k_columns), 0, 0, 1, 1},
  {&__pyx_n_s_comment, __pyx_k_comment, sizeof(__pyx_k_comment), 0, 0, 1, 1},
  {&__pyx_n_s_comments, __pyx_k_comments, sizeof(__pyx_k_comments), 0, 0, 1, 1},
  {&__pyx_n_s_concatenate, __pyx_k_concatenate, sizeof(__pyx_k_concatenate), 0, 0, 1, 1},
  {&__pyx_n_s_copy, __pyx_k_copy, sizeof(__pyx_k_copy), 0, 0, 1, 1},
  {&__pyx_n_s_copy_cparser, __pyx_k_copy_cparser, sizeof(__pyx_k_copy_cparser), 0, 0, 1, 1},
  {&__pyx_n_s_core, __pyx_k_core, sizeof(__pyx_k_core), 0, 0, 1, 1},
  {&__pyx_n_s_cpu_count, __pyx_k_cpu_count, sizeof(__pyx_k_cpu_count), 0, 0, 1, 1},
  {&__pyx_n_s_csv, __pyx_k_csv, sizeof(__pyx_k_csv), 0, 0, 1, 1},
  {&__pyx_n_s_data, __pyx_k_data, sizeof(__pyx_k_data), 0, 0, 1, 1},
  {&__pyx_n_s_data_end, __pyx_k_data_end, sizeof(__pyx_k_data_end), 0, 0, 1, 1},
  {&__pyx_n_s_data_start, __pyx_k_data_start, sizeof(__pyx_k_data_start), 0, 0, 1, 1},
  {&__pyx_n_s_delimiter, __pyx_k_delimiter, sizeof(__pyx_k_delimiter), 0, 0, 1, 1},
  {&__pyx_n_s_difference_update, __pyx_k_difference_update, sizeof(__pyx_k_difference_update), 0, 0, 1, 1},
  {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
  {&__pyx_n_s_doublequote, __pyx_k_doublequote, sizeof(__pyx_k_doublequote), 0, 0, 1, 1},
  {&__pyx_n_s_dtype, __pyx_k_dtype, sizeof(__pyx_k_dtype), 0, 0, 1, 1},
  {&__pyx_n_s_e, __pyx_k_e, sizeof(__pyx_k_e), 0, 0, 1, 1},
  {&__pyx_n_s_el, __pyx_k_el, sizeof(__pyx_k_el), 0, 0, 1, 1},
  {&__pyx_n_s_empty, __pyx_k_empty, sizeof(__pyx_k_empty), 0, 0, 1, 1},
  {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
  {&__pyx_n_s_end, __pyx_k_end, sizeof(__pyx_k_end), 0, 0, 1, 1},
  {&__pyx_n_s_enter, __pyx_k_enter, sizeof(__pyx_k_enter), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_err, __pyx_k_err, sizeof(__pyx_k_err), 0, 0, 1, 1},
  {&__pyx_n_s_escapechar, __pyx_k_escapechar, sizeof(__pyx_k_escapechar), 0, 0, 1, 1},
  {&__pyx_n_s_exclude_names, __pyx_k_exclude_names, sizeof(__pyx_k_exclude_names), 0, 0, 1, 1},
  {&__pyx_n_s_exit, __pyx_k_exit, sizeof(__pyx_k_exit), 0, 0, 1, 1},
  {&__pyx_n_s_expchar, __pyx_k_expchar, sizeof(__pyx_k_expchar), 0, 0, 1, 1},
  {&__pyx_n_s_exponent_style, __pyx_k_exponent_style, sizeof(__pyx_k_exponent_style), 0, 0, 1, 1},
  {&__pyx_n_s_extern, __pyx_k_extern, sizeof(__pyx_k_extern), 0, 0, 1, 1},
  {&__pyx_n_s_fast_reader, __pyx_k_fast_reader, sizeof(__pyx_k_fast_reader), 0, 0, 1, 1},
  {&__pyx_kp_s_fast_reader_cannot_be_False_for, __pyx_k_fast_reader_cannot_be_False_for, sizeof(__pyx_k_fast_reader_cannot_be_False_for), 0, 0, 1, 0},
  {&__pyx_kp_s_fast_reader_exponent_style_requi, __pyx_k_fast_reader_exponent_style_requi, sizeof(__pyx_k_fast_reader_exponent_style_requi), 0, 0, 1, 0},
  {&__pyx_n_s_fast_writer, __pyx_k_fast_writer, sizeof(__pyx_k_fast_writer), 0, 0, 1, 1},
  {&__pyx_n_s_fileno, __pyx_k_fileno, sizeof(__pyx_k_fileno), 0, 0, 1, 1},
  {&__pyx_n_s_fill_empty, __pyx_k_fill_empty, sizeof(__pyx_k_fill_empty), 0, 0, 1, 1},
  {&__pyx_n_s_fill_exclude_names, __pyx_k_fill_exclude_names, sizeof(__pyx_k_fill_exclude_names), 0, 0, 1, 1},
  {&__pyx_n_s_fill_extra_cols, __pyx_k_fill_extra_cols, sizeof(__pyx_k_fill_extra_cols), 0, 0, 1, 1},
  {&__pyx_n_s_fill_include_names, __pyx_k_fill_include_names, sizeof(__pyx_k_fill_include_names), 0, 0, 1, 1},
  {&__pyx_n_s_fill_names, __pyx_k_fill_names, sizeof(__pyx_k_fill_names), 0, 0, 1, 1},
  {&__pyx_n_s_fill_values, __pyx_k_fill_values, sizeof(__pyx_k_fill_values), 0, 0, 1, 1},
  {&__pyx_n_s_float, __pyx_k_float, sizeof(__pyx_k_float), 0, 0, 1, 1},
  {&__pyx_n_s_fname, __pyx_k_fname, sizeof(__pyx_k_fname), 0, 0, 1, 1},
  {&__pyx_n_s_force, __pyx_k_force, sizeof(__pyx_k_force), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_format_funcs, __pyx_k_format_funcs, sizeof(__pyx_k_format_funcs), 0, 0, 1, 1},
  {&__pyx_n_s_formats, __pyx_k_formats, sizeof(__pyx_k_formats), 0, 0, 1, 1},
  {&__pyx_n_s_genexpr, __pyx_k_genexpr, sizeof(__pyx_k_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
  {&__pyx_n_s_get_auto_format_func, __pyx_k_get_auto_format_func, sizeof(__pyx_k_get_auto_format_func), 0, 0, 1, 1},
  {&__pyx_n_s_get_fill_values, __pyx_k_get_fill_values, sizeof(__pyx_k_get_fill_values), 0, 0, 1, 1},
  {&__pyx_n_s_get_names, __pyx_k_get_names, sizeof(__pyx_k_get_names), 0, 0, 1, 1},
  {&__pyx_n_s_get_readable_fileobj, __pyx_k_get_readable_fileobj, sizeof(__pyx_k_get_readable_fileobj), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_header_output, __pyx_k_header_output, sizeof(__pyx_k_header_output), 0, 0, 1, 1},
  {&__pyx_n_s_header_start, __pyx_k_header_start, sizeof(__pyx_k_header_start), 0, 0, 1, 1},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_include_names, __pyx_k_include_names, sizeof(__pyx_k_include_names), 0, 0, 1, 1},
  {&__pyx_n_s_info, __pyx_k_info, sizeof(__pyx_k_info), 0, 0, 1, 1},
  {&__pyx_n_s_int, __pyx_k_int, sizeof(__pyx_k_int), 0, 0, 1, 1},
  {&__pyx_n_s_intersection_update, __pyx_k_intersection_update, sizeof(__pyx_k_intersection_update), 0, 0, 1, 1},
  {&__pyx_kp_s_invalid_line_supplied, __pyx_k_invalid_line_supplied, sizeof(__pyx_k_invalid_line_supplied), 0, 0, 1, 0},
  {&__pyx_n_s_iteritems, __pyx_k_iteritems, sizeof(__pyx_k_iteritems), 0, 0, 1, 1},
  {&__pyx_n_s_itervalues, __pyx_k_itervalues, sizeof(__pyx_k_itervalues), 0, 0, 1, 1},
  {&__pyx_n_s_join, __pyx_k_join, sizeof(__pyx_k_join), 0, 0, 1, 1},
  {&__pyx_n_s_kind, __pyx_k_kind, sizeof(__pyx_k_kind), 0, 0, 1, 1},
  {&__pyx_n_s_kwargs, __pyx_k_kwargs, sizeof(__pyx_k_kwargs), 0, 0, 1, 1},
  {&__pyx_n_s_l, __pyx_k_l, sizeof(__pyx_k_l), 0, 0, 1, 1},
  {&__pyx_n_s_lambda, __pyx_k_lambda, sizeof(__pyx_k_lambda), 0, 0, 1, 1},
  {&__pyx_n_s_line_comments, __pyx_k_line_comments, sizeof(__pyx_k_line_comments), 0, 0, 1, 1},
  {&__pyx_n_s_linesep, __pyx_k_linesep, sizeof(__pyx_k_linesep), 0, 0, 1, 1},
  {&__pyx_n_s_lineterminator, __pyx_k_lineterminator, sizeof(__pyx_k_lineterminator), 0, 0, 1, 1},
  {&__pyx_n_s_ma, __pyx_k_ma, sizeof(__pyx_k_ma), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_mask, __pyx_k_mask, sizeof(__pyx_k_mask), 0, 0, 1, 1},
  {&__pyx_n_s_masked, __pyx_k_masked, sizeof(__pyx_k_masked), 0, 0, 1, 1},
  {&__pyx_n_s_masked_array, __pyx_k_masked_array, sizeof(__pyx_k_masked_array), 0, 0, 1, 1},
  {&__pyx_n_s_math, __pyx_k_math, sizeof(__pyx_k_math), 0, 0, 1, 1},
  {&__pyx_n_s_meta, __pyx_k_meta, sizeof(__pyx_k_meta), 0, 0, 1, 1},
  {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
  {&__pyx_n_s_mmap, __pyx_k_mmap, sizeof(__pyx_k_mmap), 0, 0, 1, 1},
  {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
  {&__pyx_n_s_multiprocessing, __pyx_k_multiprocessing, sizeof(__pyx_k_multiprocessing), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_n_s_names, __pyx_k_names, sizeof(__pyx_k_names), 0, 0, 1, 1},
  {&__pyx_kp_u_ndarray_is_not_C_contiguous, __pyx_k_ndarray_is_not_C_contiguous, sizeof(__pyx_k_ndarray_is_not_C_contiguous), 0, 1, 0, 0},
  {&__pyx_kp_u_ndarray_is_not_Fortran_contiguou, __pyx_k_ndarray_is_not_Fortran_contiguou, sizeof(__pyx_k_ndarray_is_not_Fortran_contiguou), 0, 1, 0, 0},
  {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
  {&__pyx_kp_s_no_error, __pyx_k_no_error, sizeof(__pyx_k_no_error), 0, 0, 1, 0},
  {&__pyx_kp_s_not_enough_columns_found_in_line, __pyx_k_not_enough_columns_found_in_line, sizeof(__pyx_k_not_enough_columns_found_in_line), 0, 0, 1, 0},
  {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
  {&__pyx_n_s_nt, __pyx_k_nt, sizeof(__pyx_k_nt), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_kp_s_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 0, 1, 0},
  {&__pyx_kp_s_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 0, 1, 0},
  {&__pyx_n_s_open, __pyx_k_open, sizeof(__pyx_k_open), 0, 0, 1, 1},
  {&__pyx_n_s_os, __pyx_k_os, sizeof(__pyx_k_os), 0, 0, 1, 1},
  {&__pyx_n_s_output, __pyx_k_output, sizeof(__pyx_k_output), 0, 0, 1, 1},
  {&__pyx_n_s_output_types, __pyx_k_output_types, sizeof(__pyx_k_output_types), 0, 0, 1, 1},
  {&__pyx_kp_s_overflow_error, __pyx_k_overflow_error, sizeof(__pyx_k_overflow_error), 0, 0, 1, 0},
  {&__pyx_n_s_parallel, __pyx_k_parallel, sizeof(__pyx_k_parallel), 0, 0, 1, 1},
  {&__pyx_n_s_parser, __pyx_k_parser, sizeof(__pyx_k_parser), 0, 0, 1, 1},
  {&__pyx_n_s_pop, __pyx_k_pop, sizeof(__pyx_k_pop), 0, 0, 1, 1},
  {&__pyx_n_s_pprint, __pyx_k_pprint, sizeof(__pyx_k_pprint), 0, 0, 1, 1},
  {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
  {&__pyx_n_s_put, __pyx_k_put, sizeof(__pyx_k_put), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
  {&__pyx_n_s_queue, __pyx_k_queue, sizeof(__pyx_k_queue), 0, 0, 1, 1},
  {&__pyx_n_s_quotechar, __pyx_k_quotechar, sizeof(__pyx_k_quotechar), 0, 0, 1, 1},
  {&__pyx_n_s_quoting, __pyx_k_quoting, sizeof(__pyx_k_quoting), 0, 0, 1, 1},
  {&__pyx_n_s_r, __pyx_k_r, sizeof(__pyx_k_r), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_read, __pyx_k_read, sizeof(__pyx_k_read), 0, 0, 1, 1},
  {&__pyx_n_s_read_chunk, __pyx_k_read_chunk, sizeof(__pyx_k_read_chunk), 0, 0, 1, 1},
  {&__pyx_n_s_read_chunk_locals_genexpr, __pyx_k_read_chunk_locals_genexpr, sizeof(__pyx_k_read_chunk_locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_read_parallel, __pyx_k_read_parallel, sizeof(__pyx_k_read_parallel), 0, 0, 1, 1},
  {&__pyx_n_s_read_parallel_locals_genexpr, __pyx_k_read_parallel_locals_genexpr, sizeof(__pyx_k_read_parallel_locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_reconvert_cols, __pyx_k_reconvert_cols, sizeof(__pyx_k_reconvert_cols), 0, 0, 1, 1},
  {&__pyx_n_s_reconvert_queue, __pyx_k_reconvert_queue, sizeof(__pyx_k_reconvert_queue), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_replace, __pyx_k_replace, sizeof(__pyx_k_replace), 0, 0, 1, 1},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_setup_tokenizer, __pyx_k_setup_tokenizer, sizeof(__pyx_k_setup_tokenizer), 0, 0, 1, 1},
  {&__pyx_kp_s_shared_semaphore_implementation, __pyx_k_shared_semaphore_implementation, sizeof(__pyx_k_shared_semaphore_implementation), 0, 0, 1, 0},
  {&__pyx_n_s_six, __pyx_k_six, sizeof(__pyx_k_six), 0, 0, 1, 1},
  {&__pyx_n_s_source, __pyx_k_source, sizeof(__pyx_k_source), 0, 0, 1, 1},
  {&__pyx_n_s_source_bytes, __pyx_k_source_bytes, sizeof(__pyx_k_source_bytes), 0, 0, 1, 1},
  {&__pyx_n_s_splitlines, __pyx_k_splitlines, sizeof(__pyx_k_splitlines), 0, 0, 1, 1},
  {&__pyx_n_s_src_ptr, __pyx_k_src_ptr, sizeof(__pyx_k_src_ptr), 0, 0, 1, 1},
  {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
  {&__pyx_n_s_startswith, __pyx_k_startswith, sizeof(__pyx_k_startswith), 0, 0, 1, 1},
  {&__pyx_n_s_str, __pyx_k_str, sizeof(__pyx_k_str), 0, 0, 1, 1},
  {&__pyx_n_s_string_types, __pyx_k_string_types, sizeof(__pyx_k_string_types), 0, 0, 1, 1},
  {&__pyx_n_s_strip, __pyx_k_strip, sizeof(__pyx_k_strip), 0, 0, 1, 1},
  {&__pyx_n_s_strip_line_fields, __pyx_k_strip_line_fields, sizeof(__pyx_k_strip_line_fields), 0, 0, 1, 1},
  {&__pyx_n_s_strip_line_whitespace, __pyx_k_strip_line_whitespace, sizeof(__pyx_k_strip_line_whitespace), 0, 0, 1, 1},
  {&__pyx_n_s_strip_whitespace, __pyx_k_strip_whitespace, sizeof(__pyx_k_strip_whitespace), 0, 0, 1, 1},
  {&__pyx_n_s_strip_whitespace_fields, __pyx_k_strip_whitespace_fields, sizeof(__pyx_k_strip_whitespace_fields), 0, 0, 1, 1},
  {&__pyx_n_s_strip_whitespace_lines, __pyx_k_strip_whitespace_lines, sizeof(__pyx_k_strip_whitespace_lines), 0, 0, 1, 1},
  {&__pyx_n_s_table, __pyx_k_table, sizeof(__pyx_k_table), 0, 0, 1, 1},
  {&__pyx_n_s_target, __pyx_k_target, sizeof(__pyx_k_target), 0, 0, 1, 1},
  {&__pyx_n_s_terminate, __pyx_k_terminate, sizeof(__pyx_k_terminate), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_n_s_tolist, __pyx_k_tolist, sizeof(__pyx_k_tolist), 0, 0, 1, 1},
  {&__pyx_kp_s_too_many_columns_found_in_line_0, __pyx_k_too_many_columns_found_in_line_0, sizeof(__pyx_k_too_many_columns_found_in_line_0), 0, 0, 1, 0},
  {&__pyx_n_s_try_float, __pyx_k_try_float, sizeof(__pyx_k_try_float), 0, 0, 1, 1},
  {&__pyx_n_s_try_int, __pyx_k_try_int, sizeof(__pyx_k_try_int), 0, 0, 1, 1},
  {&__pyx_n_s_try_string, __pyx_k_try_string, sizeof(__pyx_k_try_string), 0, 0, 1, 1},
  {&__pyx_kp_s_type_conversion_error, __pyx_k_type_conversion_error, sizeof(__pyx_k_type_conversion_error), 0, 0, 1, 0},
  {&__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_k_unknown_dtype_code_in_numpy_pxd, sizeof(__pyx_k_unknown_dtype_code_in_numpy_pxd), 0, 1, 0, 0},
  {&__pyx_kp_s_unknown_error, __pyx_k_unknown_error, sizeof(__pyx_k_unknown_error), 0, 0, 1, 0},
  {&__pyx_n_s_upper, __pyx_k_upper, sizeof(__pyx_k_upper), 0, 0, 1, 1},
  {&__pyx_n_s_use_cols, __pyx_k_use_cols, sizeof(__pyx_k_use_cols), 0, 0, 1, 1},
  {&__pyx_n_s_use_fast_converter, __pyx_k_use_fast_converter, sizeof(__pyx_k_use_fast_converter), 0, 0, 1, 1},
  {&__pyx_n_s_utils_data, __pyx_k_utils_data, sizeof(__pyx_k_utils_data), 0, 0, 1, 1},
  {&__pyx_n_s_utils_exceptions, __pyx_k_utils_exceptions, sizeof(__pyx_k_utils_exceptions), 0, 0, 1, 1},
  {&__pyx_n_s_w, __pyx_k_w, sizeof(__pyx_k_w), 0, 0, 1, 1},
  {&__pyx_n_s_warn, __pyx_k_warn, sizeof(__pyx_k_warn), 0, 0, 1, 1},
  {&__pyx_n_s_warnings, __pyx_k_warnings, sizeof(__pyx_k_warnings), 0, 0, 1, 1},
  {&__pyx_n_s_write, __pyx_k_write, sizeof(__pyx_k_write), 0, 0, 1, 1},
  {&__pyx_n_s_write_header, __pyx_k_write_header, sizeof(__pyx_k_write_header), 0, 0, 1, 1},
  {&__pyx_n_s_writer, __pyx_k_writer, sizeof(__pyx_k_writer), 0, 0, 1, 1},
  {&__pyx_n_s_writerow, __pyx_k_writerow, sizeof(__pyx_k_writerow), 0, 0, 1, 1},
  {&__pyx_n_s_writerows, __pyx_k_writerows, sizeof(__pyx_k_writerows), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(0, 26, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 107, __pyx_L1_error)
  __pyx_builtin_open = __Pyx_GetBuiltinName(__pyx_n_s_open); if (!__pyx_builtin_open) __PYX_ERR(0, 127, __pyx_L1_error)
  __pyx_builtin_IOError = __Pyx_GetBuiltinName(__pyx_n_s_IOError); if (!__pyx_builtin_IOError) __PYX_ERR(0, 129, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_builtin_NotImplementedError = __Pyx_GetBuiltinName(__pyx_n_s_NotImplementedError); if (!__pyx_builtin_NotImplementedError) __PYX_ERR(0, 232, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 328, __pyx_L1_error)
  __pyx_builtin_chr = __Pyx_GetBuiltinName(__pyx_n_s_chr); if (!__pyx_builtin_chr) __PYX_ERR(0, 338, __pyx_L1_error)
  __pyx_builtin_AttributeError = __Pyx_GetBuiltinName(__pyx_n_s_AttributeError); if (!__pyx_builtin_AttributeError) __PYX_ERR(0, 412, __pyx_L1_error)
  __pyx_builtin_OSError = __Pyx_GetBuiltinName(__pyx_n_s_OSError); if (!__pyx_builtin_OSError) __PYX_ERR(0, 412, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 582, __pyx_L1_error)
  __pyx_builtin_IndexError = __Pyx_GetBuiltinName(__pyx_n_s_IndexError); if (!__pyx_builtin_IndexError) __PYX_ERR(0, 1084, __pyx_L1_error)
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(2, 856, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple_)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "astropy/io/ascii/cparser.pyx":205
 *                   include_names=None,
 *                   exclude_names=None,
 *                   fill_values=('', '0'),             # <<<<<<<<<<<<<<
 *                   fill_include_names=None,
 *                   fill_exclude_names=None,
 */
  __pyx_tuple__6 = PyTuple_Pack(2, __pyx_kp_s__5, __pyx_kp_s_0); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "astropy/io/ascii/cparser.pyx":216
 *         elif fast_reader is False: # shouldn't happen
 *             raise core.ParameterError("fast_reader cannot be False for fast readers")
 *         expchar = fast_reader.pop('exponent_style', 'E').upper()             # <<<<<<<<<<<<<<
 *         # parallel and use_fast_reader are False by default, but only the latter
 *         # supports Fortran double precision notation
 */
  __pyx_tuple__7 = PyTuple_Pack(2, __pyx_n_s_exponent_style, __pyx_n_s_E); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "astropy/io/ascii/cparser.pyx":220
 *         # supports Fortran double precision notation
 *         if expchar == 'E':
 *             use_fast_converter = fast_reader.pop('use_fast_converter', False)             # <<<<<<<<<<<<<<
 *         else:
 *             use_fast_converter = fast_reader.pop('use_fast_converter', True)
 */
  __pyx_tuple__8 = PyTuple_Pack(2, __pyx_n_s_use_fast_converter, Py_False); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "astropy/io/ascii/cparser.pyx":222
 *             use_fast_converter = fast_reader.pop('use_fast_converter', False)
 *         else:
 *             use_fast_converter = fast_reader.pop('use_fast_converter', True)             # <<<<<<<<<<<<<<
 *             if not use_fast_converter:
 *                 raise core.FastOptionsError("fast_reader: exponent_style requires use_fast_converter")
 */
  __pyx_tuple__9 = PyTuple_Pack(2, __pyx_n_s_use_fast_converter, Py_True); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "astropy/io/ascii/cparser.pyx":227
 *             if expchar.startswith('FORT'):
 *                 expchar = 'A'
 *         parallel = fast_reader.pop('parallel', False)             # <<<<<<<<<<<<<<
 * 
 *         if fast_reader:
 */
  __pyx_tuple__10 = PyTuple_Pack(2, __pyx_n_s_parallel, Py_False); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "astropy/io/ascii/cparser.pyx":232
 *             raise core.FastOptionsError("Invalid parameter in fast_reader dict")
 *         if parallel and os.name == 'nt':
 *             raise NotImplementedError("Multiprocessing is not yet supported on Windows")             # <<<<<<<<<<<<<<
 * 
 *         if comment is None:
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_s_Multiprocessing_is_not_yet_suppo); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "astropy/io/ascii/cparser.pyx":294
 *             # Otherwise, source is the actual data so we leave it be
 *         elif hasattr(source, 'read'): # file-like object
 *             with get_readable_fileobj(source) as file_obj:             # <<<<<<<<<<<<<<
 *                 source = file_obj.read()
 *         elif isinstance(source, FileString):
 */
  __pyx_tuple__15 = PyTuple_Pack(3, Py_None, Py_None, Py_None); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(0, 294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "astropy/io/ascii/cparser.pyx":305
 *                 source = '\n'.join(source) # iterable sequence of lines
 *             except TypeError:
 *                 raise TypeError('Input "table" must be a file-like object, a '             # <<<<<<<<<<<<<<
 *                                 'string (filename or data), or an iterable')
 *         # Create a reference to the Python object so its char * pointer remains valid
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_s_Input_table_must_be_a_file_like); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(0, 305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "astropy/io/ascii/cparser.pyx":333
 *                     if name:
 *                         # replace empty placeholder with ''
 *                         self.header_names.append(name.replace('\x01', ''))             # <<<<<<<<<<<<<<
 *                         name = ''
 *                     else:
 */
  __pyx_tuple__18 = PyTuple_Pack(2, __pyx_kp_s__17, __pyx_kp_s__5); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(0, 333, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "astropy/io/ascii/cparser.pyx":644
 *                 # fill_values parameter, or no specific columns are specified
 *                 # and this column should apply fill_values.
 *                 if (len(replace_info) > 1 and self.names[i] in replace_info[1:]) \             # <<<<<<<<<<<<<<
 *                    or (len(replace_info) == 1 and self.names[i] in self.fill_names):
 *                     mask.add(row)
 */
  __pyx_slice__19 = PySlice_New(__pyx_int_1, Py_None, Py_None); if (unlikely(!__pyx_slice__19)) __PYX_ERR(0, 644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__19);
  __Pyx_GIVEREF(__pyx_slice__19);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__23 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__24 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);

  /* "astropy/io/ascii/cparser.pyx":1085
 * 
 *     except IndexError:
 *         raise ValueError("Format of fill_values must be "             # <<<<<<<<<<<<<<
 *                          "(<bad>, <fill>, <optional col1>, ...)")
 *     if read:
 */
  __pyx_tuple__25 = PyTuple_Pack(1, __pyx_kp_s_Format_of_fill_values_must_be_ba); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(0, 1085, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":272
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
  __pyx_tuple__26 = PyTuple_Pack(1, __pyx_kp_u_ndarray_is_not_C_contiguous); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(2, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":276
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
  __pyx_tuple__27 = PyTuple_Pack(1, __pyx_kp_u_ndarray_is_not_Fortran_contiguou); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(2, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":306
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
  __pyx_tuple__28 = PyTuple_Pack(1, __pyx_kp_u_Non_native_byte_order_not_suppor); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(2, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":856
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
  __pyx_tuple__29 = PyTuple_Pack(1, __pyx_kp_u_Format_string_allocated_too_shor); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(2, 856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":880
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
  __pyx_tuple__30 = PyTuple_Pack(1, __pyx_kp_u_Format_string_allocated_too_shor_2); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(2, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1038
 *         _import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
  __pyx_tuple__31 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(2, 1038, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1044
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
  __pyx_tuple__32 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(2, 1044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);

  /* "astropy/io/ascii/cparser.pyx":806
 *                                 parallel=False)))
 * 
 * def _copy_cparser(bytes src_ptr, bytes source_bytes, use_cols, fill_names, fill_values,             # <<<<<<<<<<<<<<
 *                   strip_whitespace_lines, strip_whitespace_fields, kwargs):
 *     parser = CParser(None, strip_whitespace_lines, strip_whitespace_fields, **kwargs)
 */
  __pyx_tuple__33 = PyTuple_Pack(9, __pyx_n_s_src_ptr, __pyx_n_s_source_bytes, __pyx_n_s_use_cols, __pyx_n_s_fill_names, __pyx_n_s_fill_values, __pyx_n_s_strip_whitespace_lines, __pyx_n_s_strip_whitespace_fields, __pyx_n_s_kwargs, __pyx_n_s_parser); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(0, 806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);
  __pyx_codeobj__34 = (PyObject*)__Pyx_PyCode_New(8, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__33, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_astropy_io_ascii_cparser_pyx, __pyx_n_s_copy_cparser, 806, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__34)) __PYX_ERR(0, 806, __pyx_L1_error)

  /* "astropy/io/ascii/cparser.pyx":819
 *     return parser
 * 
 * def _read_chunk(CParser self, start, end, try_int,             # <<<<<<<<<<<<<<
 *                 try_float, try_string, queue, reconvert_queue, i):
 *     cdef tokenizer_t *chunk_tokenizer = self.tokenizer
 */
  __pyx_tuple__35 = PyTuple_Pack(18, __pyx_n_s_self, __pyx_n_s_start, __pyx_n_s_end, __pyx_n_s_try_int, __pyx_n_s_try_float, __pyx_n_s_try_string, __pyx_n_s_queue, __pyx_n_s_reconvert_queue, __pyx_n_s_i, __pyx_n_s_chunk_tokenizer, __pyx_n_s_data, __pyx_n_s_err, __pyx_n_s_line_comments, __pyx_n_s_e, __pyx_n_s_reconvert_cols, __pyx_n_s_col, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(0, 819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);
  __pyx_codeobj__36 = (PyObject*)__Pyx_PyCode_New(9, 0, 18, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__35, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_astropy_io_ascii_cparser_pyx, __pyx_n_s_read_chunk, 819, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__36)) __PYX_ERR(0, 819, __pyx_L1_error)

  /* "astropy/io/ascii/cparser.pyx":1061
 *             output.close()
 * 
 * def get_fill_values(fill_values, read=True):             # <<<<<<<<<<<<<<
 *     if len(fill_values) > 0 and isinstance(fill_values[0], six.string_types):
 *         # e.g. fill_values=('999', '0')
 */
  __pyx_tuple__37 = PyTuple_Pack(5, __pyx_n_s_fill_values, __pyx_n_s_read, __pyx_n_s_fill_empty, __pyx_n_s_el, __pyx_n_s_l); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(0, 1061, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__37);
  __Pyx_GIVEREF(__pyx_tuple__37);
  __pyx_codeobj__38 = (PyObject*)__Pyx_PyCode_New(2, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__37, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_astropy_io_ascii_cparser_pyx, __pyx_n_s_get_fill_values, 1061, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__38)) __PYX_ERR(0, 1061, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  __pyx_umethod_PyList_Type_pop.type = (PyObject*)&PyList_Type;
  __pyx_umethod_PySet_Type_difference_update.type = (PyObject*)&PySet_Type;
  __pyx_umethod_PySet_Type_intersection_update.type = (PyObject*)&PySet_Type;
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  if (PyType_Ready(&__pyx_type_7astropy_2io_5ascii_7cparser_FileString) < 0) __PYX_ERR(0, 116, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_7astropy_2io_5ascii_7cparser_FileString.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7astropy_2io_5ascii_7cparser_FileString.tp_dictoffset && __pyx_type_7astropy_2io_5ascii_7cparser_FileString.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7astropy_2io_5ascii_7cparser_FileString.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_FileString, (PyObject *)&__pyx_type_7astropy_2io_5ascii_7cparser_FileString) < 0) __PYX_ERR(0, 116, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_7astropy_2io_5ascii_7cparser_FileString) < 0) __PYX_ERR(0, 116, __pyx_L1_error)
  __pyx_ptype_7astropy_2io_5ascii_7cparser_FileString = &__pyx_type_7astropy_2io_5ascii_7cparser_FileString;
  __pyx_vtabptr_7astropy_2io_5ascii_7cparser_CParser = &__pyx_vtable_7astropy_2io_5ascii_7cparser_CParser;
  __pyx_vtable_7astropy_2io_5ascii_7cparser_CParser.get_error = (PyObject *(*)(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *, PyObject *, PyObject *, PyObject *))__pyx_f_7astropy_2io_5ascii_7cparser_7CParser_get_error;
  __pyx_vtable_7astropy_2io_5ascii_7cparser_CParser.raise_error = (PyObject *(*)(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *, PyObject *))__pyx_f_7astropy_2io_5ascii_7cparser_7CParser_raise_error;
  __pyx_vtable_7astropy_2io_5ascii_7cparser_CParser.setup_tokenizer = (PyObject *(*)(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *, PyObject *, int __pyx_skip_dispatch))__pyx_f_7astropy_2io_5ascii_7cparser_7CParser_setup_tokenizer;
  __pyx_vtable_7astropy_2io_5ascii_7cparser_CParser._set_fill_values = (PyObject *(*)(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *))__pyx_f_7astropy_2io_5ascii_7cparser_7CParser__set_fill_values;
  __pyx_vtable_7astropy_2io_5ascii_7cparser_CParser._get_comments = (PyObject *(*)(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *, tokenizer_t *))__pyx_f_7astropy_2io_5ascii_7cparser_7CParser__get_comments;
  __pyx_vtable_7astropy_2io_5ascii_7cparser_CParser._convert_data = (PyObject *(*)(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *, tokenizer_t *, PyObject *, PyObject *, PyObject *, PyObject *))__pyx_f_7astropy_2io_5ascii_7cparser_7CParser__convert_data;
  __pyx_vtable_7astropy_2io_5ascii_7cparser_CParser._convert_int = (PyArrayObject *(*)(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *, tokenizer_t *, int, int))__pyx_f_7astropy_2io_5ascii_7cparser_7CParser__convert_int;
  __pyx_vtable_7astropy_2io_5ascii_7cparser_CParser._convert_float = (PyArrayObject *(*)(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *, tokenizer_t *, int, int))__pyx_f_7astropy_2io_5ascii_7cparser_7CParser__convert_float;
  __pyx_vtable_7astropy_2io_5ascii_7cparser_CParser._convert_str = (PyObject *(*)(struct __pyx_obj_7astropy_2io_5ascii_7cparser_CParser *, tokenizer_t *, int, int))__pyx_f_7astropy_2io_5ascii_7cparser_7CParser__convert_str;
  if (PyType_Ready(&__pyx_type_7astropy_2io_5ascii_7cparser_CParser) < 0) __PYX_ERR(0, 165, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_7astropy_2io_5ascii_7cparser_CParser.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7astropy_2io_5ascii_7cparser_CParser.tp_dictoffset && __pyx_type_7astropy_2io_5ascii_7cparser_CParser.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7astropy_2io_5ascii_7cparser_CParser.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_7astropy_2io_5ascii_7cparser_CParser.tp_dict, __pyx_vtabptr_7astropy_2io_5ascii_7cparser_CParser) < 0) __PYX_ERR(0, 165, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CParser, (PyObject *)&__pyx_type_7astropy_2io_5ascii_7cparser_CParser) < 0) __PYX_ERR(0, 165, __pyx_L1_error)
  __pyx_ptype_7astropy_2io_5ascii_7cparser_CParser = &__pyx_type_7astropy_2io_5ascii_7cparser_CParser;
  __pyx_vtabptr_7astropy_2io_5ascii_7cparser_FastWriter = &__pyx_vtable_7astropy_2io_5ascii_7cparser_FastWriter;
  __pyx_vtable_7astropy_2io_5ascii_7cparser_FastWriter._write_comments = (PyObject *(*)(struct __pyx_obj_7astropy_2io_5ascii_7cparser_FastWriter *, PyObject *))__pyx_f_7astropy_2io_5ascii_7cparser_10FastWriter__write_comments;
  if (PyType_Ready(&__pyx_type_7astropy_2io_5ascii_7cparser_FastWriter) < 0) __PYX_ERR(0, 856, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_7astropy_2io_5ascii_7cparser_FastWriter.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7astropy_2io_5ascii_7cparser_FastWriter.tp_dictoffset && __pyx_type_7astropy_2io_5ascii_7cparser_FastWriter.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7astropy_2io_5ascii_7cparser_FastWriter.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_7astropy_2io_5ascii_7cparser_FastWriter.tp_dict, __pyx_vtabptr_7astropy_2io_5ascii_7cparser_FastWriter) < 0) __PYX_ERR(0, 856, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_FastWriter, (PyObject *)&__pyx_type_7astropy_2io_5ascii_7cparser_FastWriter) < 0) __PYX_ERR(0, 856, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_7astropy_2io_5ascii_7cparser_FastWriter) < 0) __PYX_ERR(0, 856, __pyx_L1_error)
  __pyx_ptype_7astropy_2io_5ascii_7cparser_FastWriter = &__pyx_type_7astropy_2io_5ascii_7cparser_FastWriter;
  if (PyType_Ready(&__pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines) < 0) __PYX_ERR(0, 151, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines.tp_dictoffset && __pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines = &__pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct__splitlines;
  if (PyType_Ready(&__pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel) < 0) __PYX_ERR(0, 400, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel.tp_dictoffset && __pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel = &__pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_1__read_parallel;
  if (PyType_Ready(&__pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr) < 0) __PYX_ERR(0, 418, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr.tp_dictoffset && __pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr = &__pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_2_genexpr;
  if (PyType_Ready(&__pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr) < 0) __PYX_ERR(0, 525, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr.tp_dictoffset && __pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr = &__pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_3_genexpr;
  if (PyType_Ready(&__pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr) < 0) __PYX_ERR(0, 541, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr.tp_dictoffset && __pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr = &__pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_4_genexpr;
  if (PyType_Ready(&__pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk) < 0) __PYX_ERR(0, 819, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk.tp_dictoffset && __pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk = &__pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_5__read_chunk;
  if (PyType_Ready(&__pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr) < 0) __PYX_ERR(0, 832, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr.tp_dictoffset && __pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr = &__pyx_type_7astropy_2io_5ascii_7cparser___pyx_scope_struct_6_genexpr;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(2, 206, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(2, 229, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(2, 233, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(2, 242, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(2, 918, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION < 3
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC void
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#else
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initcparser(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initcparser(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_cparser(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_cparser(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_cparser(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'cparser' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_cparser(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("cparser", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_astropy__io__ascii__cparser) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name_2, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "astropy.io.ascii.cparser")) {
      if (unlikely(PyDict_SetItemString(modules, "astropy.io.ascii.cparser", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_type_import_code() != 0)) goto __pyx_L1_error;
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "astropy/io/ascii/cparser.pyx":3
 * # Licensed under a 3-clause BSD style license - see LICENSE.rst
 * 
 * import csv             # <<<<<<<<<<<<<<
 * import os
 * import math
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_csv, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_csv, __pyx_t_1) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "astropy/io/ascii/cparser.pyx":4
 * 
 * import csv
 * import os             # <<<<<<<<<<<<<<
 * import math
 * import multiprocessing
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_os, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_os, __pyx_t_1) < 0) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "astropy/io/ascii/cparser.pyx":5
 * import csv
 * import os
 * import math             # <<<<<<<<<<<<<<
 * import multiprocessing
 * import mmap
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_math, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_math, __pyx_t_1) < 0) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "astropy/io/ascii/cparser.pyx":6
 * import os
 * import math
 * import multiprocessing             # <<<<<<<<<<<<<<
 * import mmap
 * import warnings
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_multiprocessing, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_multiprocessing, __pyx_t_1) < 0) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "astropy/io/ascii/cparser.pyx":7
 * import math
 * import multiprocessing
 * import mmap             # <<<<<<<<<<<<<<
 * import warnings
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_mmap, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_mmap, __pyx_t_1) < 0) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "astropy/io/ascii/cparser.pyx":8
 * import multiprocessing
 * import mmap
 * import warnings             # <<<<<<<<<<<<<<
 * 
 * import numpy as np
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_warnings, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_warnings, __pyx_t_1) < 0) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "astropy/io/ascii/cparser.pyx":10
 * import warnings
 * 
 * import numpy as np             # <<<<<<<<<<<<<<
 * cimport numpy as np
 * from numpy import ma
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_numpy, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_1) < 0) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "astropy/io/ascii/cparser.pyx":12
 * import numpy as np
 * cimport numpy as np
 * from numpy import ma             # <<<<<<<<<<<<<<
 * from libc cimport stdio
 * from cpython.buffer cimport PyBUF_SIMPLE
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_ma);
  __Pyx_GIVEREF(__pyx_n_s_ma);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_ma);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_numpy, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_ma); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ma, __pyx_t_1) < 0) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "astropy/io/ascii/cparser.pyx":18
 * from cpython.buffer cimport PyObject_GetBuffer, PyBuffer_Release
 * 
 * from ...utils.data import get_readable_fileobj             # <<<<<<<<<<<<<<
 * from ...utils.exceptions import AstropyWarning
 * from ...table import pprint
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_get_readable_fileobj);
  __Pyx_GIVEREF(__pyx_n_s_get_readable_fileobj);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_get_readable_fileobj);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_utils_data, __pyx_t_2, 3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_get_readable_fileobj); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_get_readable_fileobj, __pyx_t_2) < 0) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "astropy/io/ascii/cparser.pyx":19
 * 
 * from ...utils.data import get_readable_fileobj
 * from ...utils.exceptions import AstropyWarning             # <<<<<<<<<<<<<<
 * from ...table import pprint
 * from ...extern import six
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_AstropyWarning);
  __Pyx_GIVEREF(__pyx_n_s_AstropyWarning);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_AstropyWarning);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_utils_exceptions, __pyx_t_1, 3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_AstropyWarning); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_AstropyWarning, __pyx_t_1) < 0) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "astropy/io/ascii/cparser.pyx":20
 * from ...utils.data import get_readable_fileobj
 * from ...utils.exceptions import AstropyWarning
 * from ...table import pprint             # <<<<<<<<<<<<<<
 * from ...extern import six
 * from . import core
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_pprint);
  __Pyx_GIVEREF(__pyx_n_s_pprint);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_pprint);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_table, __pyx_t_2, 3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_pprint); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pprint, __pyx_t_2) < 0) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "astropy/io/ascii/cparser.pyx":21
 * from ...utils.exceptions import AstropyWarning
 * from ...table import pprint
 * from ...extern import six             # <<<<<<<<<<<<<<
 * from . import core
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_six);
  __Pyx_GIVEREF(__pyx_n_s_six);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_six);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_extern, __pyx_t_1, 3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_six); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_six, __pyx_t_1) < 0) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "astropy/io/ascii/cparser.pyx":22
 * from ...table import pprint
 * from ...extern import six
 * from . import core             # <<<<<<<<<<<<<<
 * 
 * try:
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_core);
  __Pyx_GIVEREF(__pyx_n_s_core);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_core);
  __pyx_t_1 = __Pyx_Import(__pyx_kp_s__5, __pyx_t_2, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_core); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_core, __pyx_t_2) < 0) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "astropy/io/ascii/cparser.pyx":24
 * from . import core
 * 
 * try:             # <<<<<<<<<<<<<<
 *     import Queue
 * except ImportError: # in python 3, the module is named queue
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    /*try:*/ {

      /* "astropy/io/ascii/cparser.pyx":25
 * 
 * try:
 *     import Queue             # <<<<<<<<<<<<<<
 * except ImportError: # in python 3, the module is named queue
 *     import queue as Queue
 */
      __pyx_t_1 = __Pyx_Import(__pyx_n_s_Queue, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L2_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_Queue, __pyx_t_1) < 0) __PYX_ERR(0, 25, __pyx_L2_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "astropy/io/ascii/cparser.pyx":24
 * from . import core
 * 
 * try:             # <<<<<<<<<<<<<<
 *     import Queue
 * except ImportError: # in python 3, the module is named queue
 */
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L7_try_end;
    __pyx_L2_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "astropy/io/ascii/cparser.pyx":26
 * try:
 *     import Queue
 * except ImportError: # in python 3, the module is named queue             # <<<<<<<<<<<<<<
 *     import queue as Queue
 * 
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ImportError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("astropy.io.ascii.cparser", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_2, &__pyx_t_7) < 0) __PYX_ERR(0, 26, __pyx_L4_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GOTREF(__pyx_t_7);

      /* "astropy/io/ascii/cparser.pyx":27
 *     import Queue
 * except ImportError: # in python 3, the module is named queue
 *     import queue as Queue             # <<<<<<<<<<<<<<
 * 
 * cdef extern from "src/tokenizer.h":
 */
      __pyx_t_8 = __Pyx_Import(__pyx_n_s_queue, 0, -1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 27, __pyx_L4_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_Queue, __pyx_t_8) < 0) __PYX_ERR(0, 27, __pyx_L4_except_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L3_exception_handled;
    }
    goto __pyx_L4_except_error;
    __pyx_L4_except_error:;

    /* "astropy/io/ascii/cparser.pyx":24
 * from . import core
 * 
 * try:             # <<<<<<<<<<<<<<
 *     import Queue
 * except ImportError: # in python 3, the module is named queue
 */
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L1_error;
    __pyx_L3_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    __pyx_L7_try_end:;
  }

  /* "astropy/io/ascii/cparser.pyx":101
 *     int PyObject_AsReadBuffer(object obj, const void **buffer, Py_ssize_t *buffer_len)
 * 
 * class CParserError(Exception):             # <<<<<<<<<<<<<<
 *     """
 *     An instance of this class is thrown when an error occurs
 */
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
  __Pyx_GIVEREF(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
  PyTuple_SET_ITEM(__pyx_t_7, 0, ((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_7, __pyx_n_s_CParserError, __pyx_n_s_CParserError, (PyObject *) NULL, __pyx_n_s_astropy_io_ascii_cparser, __pyx_kp_s_An_instance_of_this_class_is_th); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_CParserError, __pyx_t_7, __pyx_t_1, NULL, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CParserError, __pyx_t_8) < 0) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "astropy/io/ascii/cparser.pyx":110
 *     "no error",
 *     "invalid line supplied",
 *     lambda line: "too many columns found in line {0} of data".format(line),             # <<<<<<<<<<<<<<
 *     lambda line: "not enough columns found in line {0} of data".format(line),
 *     "type conversion error",
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_7astropy_2io_5ascii_7cparser_6lambda, 0, __pyx_n_s_lambda, NULL, __pyx_n_s_astropy_io_ascii_cparser, __pyx_d, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "astropy/io/ascii/cparser.pyx":111
 *     "invalid line supplied",
 *     lambda line: "too many columns found in line {0} of data".format(line),
 *     lambda line: "not enough columns found in line {0} of data".format(line),             # <<<<<<<<<<<<<<
 *     "type conversion error",
 *     "overflow error"
 */
  __pyx_t_2 = __Pyx_CyFunction_NewEx(&__pyx_mdef_7astropy_2io_5ascii_7cparser_7lambda1, 0, __pyx_n_s_lambda, NULL, __pyx_n_s_astropy_io_ascii_cparser, __pyx_d, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "astropy/io/ascii/cparser.pyx":107
 *     """
 * 
 * ERR_CODES = dict(enumerate([             # <<<<<<<<<<<<<<
 *     "no error",
 *     "invalid line supplied",
 */
  __pyx_t_1 = PyList_New(6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_kp_s_no_error);
  __Pyx_GIVEREF(__pyx_kp_s_no_error);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_kp_s_no_error);
  __Pyx_INCREF(__pyx_kp_s_invalid_line_supplied);
  __Pyx_GIVEREF(__pyx_kp_s_invalid_line_supplied);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_kp_s_invalid_line_supplied);
  __Pyx_GIVEREF(__pyx_t_7);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_2);
  PyList_SET_ITEM(__pyx_t_1, 3, __pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_type_conversion_error);
  __Pyx_GIVEREF(__pyx_kp_s_type_conversion_error);
  PyList_SET_ITEM(__pyx_t_1, 4, __pyx_kp_s_type_conversion_error);
  __Pyx_INCREF(__pyx_kp_s_overflow_error);
  __Pyx_GIVEREF(__pyx_kp_s_overflow_error);
  PyList_SET_ITEM(__pyx_t_1, 5, __pyx_kp_s_overflow_error);
  __pyx_t_7 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_enumerate, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ERR_CODES, __pyx_t_1) < 0) __PYX_ERR(0, 107, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "astropy/io/ascii/cparser.pyx":806
 *                                 parallel=False)))
 * 
 * def _copy_cparser(bytes src_ptr, bytes source_bytes, use_cols, fill_names, fill_values,             # <<<<<<<<<<<<<<
 *                   strip_whitespace_lines, strip_whitespace_fields, kwargs):
 *     parser = CParser(None, strip_whitespace_lines, strip_whitespace_fields, **kwargs)
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7astropy_2io_5ascii_7cparser_1_copy_cparser, NULL, __pyx_n_s_astropy_io_ascii_cparser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_copy_cparser, __pyx_t_1) < 0) __PYX_ERR(0, 806, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "astropy/io/ascii/cparser.pyx":819
 *     return parser
 * 
 * def _read_chunk(CParser self, start, end, try_int,             # <<<<<<<<<<<<<<
 *                 try_float, try_string, queue, reconvert_queue, i):
 *     cdef tokenizer_t *chunk_tokenizer = self.tokenizer
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7astropy_2io_5ascii_7cparser_3_read_chunk, NULL, __pyx_n_s_astropy_io_ascii_cparser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_read_chunk, __pyx_t_1) < 0) __PYX_ERR(0, 819, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "astropy/io/ascii/cparser.pyx":888
 *                   include_names=None,
 *                   exclude_names=None,
 *                   fill_values=[],             # <<<<<<<<<<<<<<
 *                   fill_include_names=None,
 *                   fill_exclude_names=None,
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 888, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_k__21 = __pyx_t_1;
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "astropy/io/ascii/cparser.pyx":1061
 *             output.close()
 * 
 * def get_fill_values(fill_values, read=True):             # <<<<<<<<<<<<<<
 *     if len(fill_values) > 0 and isinstance(fill_values[0], six.string_types):
 *         # e.g. fill_values=('999', '0')
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7astropy_2io_5ascii_7cparser_5get_fill_values, NULL, __pyx_n_s_astropy_io_ascii_cparser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1061, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_get_fill_values, __pyx_t_1) < 0) __PYX_ERR(0, 1061, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "astropy/io/ascii/cparser.pyx":1
 * # Licensed under a 3-clause BSD style license - see LICENSE.rst             # <<<<<<<<<<<<<<
 * 
 * import csv
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "../../../../../usr/local/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1046
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init astropy.io.ascii.cparser", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init astropy.io.ascii.cparser");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* decode_c_string */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    Py_ssize_t length;
    if (unlikely((start < 0) | (stop < 0))) {
        size_t slen = strlen(cstring);
        if (unlikely(slen > (size_t) PY_SSIZE_T_MAX)) {
            PyErr_SetString(PyExc_OverflowError,
                            "c-string too long to convert to Python");
            return NULL;
        }
        length = (Py_ssize_t) slen;
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    length = stop - start;
    if (unlikely(length <= 0))
        return PyUnicode_FromUnicode(NULL, 0);
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* UnicodeAsUCS4 */
static CYTHON_INLINE Py_UCS4 __Pyx_PyUnicode_AsPy_UCS4(PyObject* x) {
   Py_ssize_t length;
   #if CYTHON_PEP393_ENABLED
   length = PyUnicode_GET_LENGTH(x);
   if (likely(length == 1)) {
       return PyUnicode_READ_CHAR(x, 0);
   }
   #else
   length = PyUnicode_GET_SIZE(x);
   if (likely(length == 1)) {
       return PyUnicode_AS_UNICODE(x)[0];
   }
   #if Py_UNICODE_SIZE == 2
   else if (PyUnicode_GET_SIZE(x) == 2) {
       Py_UCS4 high_val = PyUnicode_AS_UNICODE(x)[0];
       if (high_val >= 0xD800 && high_val <= 0xDBFF) {
           Py_UCS4 low_val = PyUnicode_AS_UNICODE(x)[1];
           if (low_val >= 0xDC00 && low_val <= 0xDFFF) {
               return 0x10000 + (((high_val & ((1<<10)-1)) << 10) | (low_val & ((1<<10)-1)));
           }
       }
   }
   #endif
   #endif
   PyErr_Format(PyExc_ValueError,
                "only single character unicode strings can be converted to Py_UCS4, "
                "got length %" CYTHON_FORMAT_SSIZE_T "d", length);
   return (Py_UCS4)-1;
}

/* object_ord */
static long __Pyx__PyObject_Ord(PyObject* c) {
    Py_ssize_t size;
    if (PyBytes_Check(c)) {
        size = PyBytes_GET_SIZE(c);
        if (likely(size == 1)) {
            return (unsigned char) PyBytes_AS_STRING(c)[0];
        }
#if PY_MAJOR_VERSION < 3
    } else if (PyUnicode_Check(c)) {
        return (long)__Pyx_PyUnicode_AsPy_UCS4(c);
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    } else if (PyByteArray_Check(c)) {
        size = PyByteArray_GET_SIZE(c);
        if (likely(size == 1)) {
            return (unsigned char) PyByteArray_AS_STRING(c)[0];
        }
#endif
    } else {
        PyErr_Format(PyExc_TypeError,
            "ord() expected string of length 1, but %.200s found", c->ob_type->tp_name);
        return (long)(Py_UCS4)-1;
    }
    PyErr_Format(PyExc_TypeError,
        "ord() expected a character, but string of length %zd found", size);
    return (long)(Py_UCS4)-1;
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* StringJoin */
#if !CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values) {
    return PyObject_CallMethodObjArgs(sep, __pyx_n_s_join, values, NULL);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod1 */
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* append */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x) {
    if (likely(PyList_CheckExact(L))) {
        if (unlikely(__Pyx_PyList_Append(L, x) < 0)) return -1;
    } else {
        PyObject* retval = __Pyx_PyObject_CallMethod1(L, __pyx_n_s_append, x);
        if (unlikely(!retval))
            return -1;
        Py_DECREF(retval);
    }
    return 0;
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a - b);
            if (likely((x^a) >= 0 || (x^~b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
            }
        }
                x = a - b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla - llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("subtract", return NULL)
            result = ((double)a) - (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
}
#endif

/* None */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname) {
    PyErr_Format(PyExc_NameError, "free variable '%s' referenced before assignment in enclosing scope", varname);
}

/* None */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* SliceObject */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object is unsliceable", Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

/* SliceObject */
static CYTHON_INLINE int __Pyx_PyObject_SetSlice(PyObject* obj, PyObject* value,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_ass_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_ass_slice(obj, cstart, cstop, value);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_ass_subscript))
#endif
    {
        int result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_ass_subscript(obj, py_slice, value);
#else
        result = value ? PyObject_SetItem(obj, py_slice, value) : PyObject_DelItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object does not support slice %.10s",
        Py_TYPE(obj)->tp_name, value ? "assignment" : "deletion");
bad:
    return -1;
}

/* pyfrozenset_new */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it) {
    if (it) {
        PyObject* result;
#if CYTHON_COMPILING_IN_PYPY
        PyObject* args;
        args = PyTuple_Pack(1, it);
        if (unlikely(!args))
            return NULL;
        result = PyObject_Call((PyObject*)&PyFrozenSet_Type, args, NULL);
        Py_DECREF(args);
        return result;
#else
        if (PyFrozenSet_CheckExact(it)) {
            Py_INCREF(it);
            return it;
        }
        result = PyFrozenSet_New(it);
        if (unlikely(!result))
            return NULL;
        if (likely(PySet_GET_SIZE(result)))
            return result;
        Py_DECREF(result);
#endif
    }
#if CYTHON_USE_TYPE_SLOTS
    return PyFrozenSet_Type.tp_new(&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#else
    return PyObject_Call((PyObject*)&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#endif
}

/* PySetContains */
static int __Pyx_PySet_ContainsUnhashable(PyObject *set, PyObject *key) {
    int result = -1;
    if (PySet_Check(key) && PyErr_ExceptionMatches(PyExc_TypeError)) {
        PyObject *tmpkey;
        PyErr_Clear();
        tmpkey = __Pyx_PyFrozenSet_New(key);
        if (tmpkey != NULL) {
            result = PySet_Contains(set, tmpkey);
            Py_DECREF(tmpkey);
        }
    }
    return result;
}
static CYTHON_INLINE int __Pyx_PySet_ContainsTF(PyObject* key, PyObject* set, int eq) {
    int result = PySet_Contains(set, key);
    if (unlikely(result < 0)) {
        result = __Pyx_PySet_ContainsUnhashable(set, key);
    }
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* UnpackUnboundCMethod */
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    }
#endif
    return 0;
}

/* CallUnboundCMethod0 */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_ASSUME_SAFE_MACROS
    args = PyTuple_New(1);
    if (unlikely(!args)) goto bad;
    Py_INCREF(self);
    PyTuple_SET_ITEM(args, 0, self);
#else
    args = PyTuple_Pack(1, self);
    if (unlikely(!args)) goto bad;
#endif
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    Py_DECREF(args);
bad:
    return result;
}

/* pop */
static CYTHON_INLINE PyObject* __Pyx__PyObject_Pop(PyObject* L) {
    if (Py_TYPE(L) == &PySet_Type) {
        return PySet_Pop(L);
    }
    return __Pyx_PyObject_CallMethod0(L, __pyx_n_s_pop);
}
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE PyObject* __Pyx_PyList_Pop(PyObject* L) {
    if (likely(PyList_GET_SIZE(L) > (((PyListObject*)L)->allocated >> 1))) {
        Py_SIZE(L) -= 1;
        return PyList_GET_ITEM(L, PyList_GET_SIZE(L));
    }
    return __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyList_Type_pop, L);
}
#endif

/* CallUnboundCMethod1 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg) {
    if (likely(cfunc->func)) {
        int flag = cfunc->flag;
        if (flag == METH_O) {
            return (*(cfunc->func))(self, arg);
        } else if (PY_VERSION_HEX >= 0x030600B1 && flag == METH_FASTCALL) {
            if (PY_VERSION_HEX >= 0x030700A0) {
                return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, &arg, 1);
            } else {
                return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
            }
        } else if (PY_VERSION_HEX >= 0x030700A0 && flag == (METH_FASTCALL | METH_KEYWORDS)) {
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
        }
    }
    return __Pyx__CallUnboundCMethod1(cfunc, self, arg);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 0, arg);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 1, arg);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(2, self, arg);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* IterNext */
static PyObject *__Pyx_PyIter_Next2Default(PyObject* defval) {
    PyObject* exc_type;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    exc_type = __Pyx_PyErr_Occurred();
    if (unlikely(exc_type)) {
        if (!defval || unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
            return NULL;
        __Pyx_PyErr_Clear();
        Py_INCREF(defval);
        return defval;
    }
    if (defval) {
        Py_INCREF(defval);
        return defval;
    }
    __Pyx_PyErr_SetNone(PyExc_StopIteration);
    return NULL;
}
static void __Pyx_PyIter_Next_ErrorNoIterator(PyObject *iterator) {
    PyErr_Format(PyExc_TypeError,
        "%.200s object is not an iterator", Py_TYPE(iterator)->tp_name);
}
static CYTHON_INLINE PyObject *__Pyx_PyIter_Next2(PyObject* iterator, PyObject* defval) {
    PyObject* next;
    iternextfunc iternext = Py_TYPE(iterator)->tp_iternext;
    if (likely(iternext)) {
#if CYTHON_USE_TYPE_SLOTS
        next = iternext(iterator);
        if (likely(next))
            return next;
        #if PY_VERSION_HEX >= 0x02070000
        if (unlikely(iternext == &_PyObject_NextNotImplemented))
            return NULL;
        #endif
#else
        next = PyIter_Next(iterator);
        if (likely(next))
            return next;
#endif
    } else if (CYTHON_USE_TYPE_SLOTS || unlikely(!PyIter_Check(iterator))) {
        __Pyx_PyIter_Next_ErrorNoIterator(iterator);
        return NULL;
    }
#if !CYTHON_USE_TYPE_SLOTS
    else {
        next = PyIter_Next(iterator);
        if (likely(next))
            return next;
    }
#endif
    return __Pyx_PyIter_Next2Default(defval);
}

/* SetItemInt */
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (!j) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return m->sq_ass_item(o, i, v);
        }
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (is_list || (PySequence_Check(o) && !PyDict_Check(o)))
#else
    if (is_list || PySequence_Check(o))
#endif
    {
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* SliceTupleAndList */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_crop_slice(Py_ssize_t* _start, Py_ssize_t* _stop, Py_ssize_t* _length) {
    Py_ssize_t start = *_start, stop = *_stop, length = *_length;
    if (start < 0) {
        start += length;
        if (start < 0)
            start = 0;
    }
    if (stop < 0)
        stop += length;
    else if (stop > length)
        stop = length;
    *_length = stop - start;
    *_start = start;
    *_stop = stop;
}
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject** CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject* __Pyx_PyList_GetSlice(
            PyObject* src, Py_ssize_t start, Py_ssize_t stop) {
    PyObject* dest;
    Py_ssize_t length = PyList_GET_SIZE(src);
    __Pyx_crop_slice(&start, &stop, &length);
    if (unlikely(length <= 0))
        return PyList_New(0);
    dest = PyList_New(length);
    if (unlikely(!dest))
        return NULL;
    __Pyx_copy_object_array(
        ((PyListObject*)src)->ob_item + start,
        ((PyListObject*)dest)->ob_item,
        length);
    return dest;
}
static CYTHON_INLINE PyObject* __Pyx_PyTuple_GetSlice(
            PyObject* src, Py_ssize_t start, Py_ssize_t stop) {
    PyObject* dest;
    Py_ssize_t length = PyTuple_GET_SIZE(src);
    __Pyx_crop_slice(&start, &stop, &length);
    if (unlikely(length <= 0))
        return PyTuple_New(0);
    dest = PyTuple_New(length);
    if (unlikely(!dest))
        return NULL;
    __Pyx_copy_object_array(
        ((PyTupleObject*)src)->ob_item + start,
        ((PyTupleObject*)dest)->ob_item,
        length);
    return dest;
}
#endif

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name_2);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_cython); if (unlikely(!reduce_cython)) goto BAD;
            ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto BAD;
            ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto BAD;
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate_cython); if (unlikely(!setstate_cython)) goto BAD;
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto BAD;
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto GOOD;
BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* SetVTable */
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* CalculateMetaclass */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases = PyTuple_GET_SIZE(bases);
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* Py3ClassCreate */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStr(metaclass, __pyx_n_s_prepare);
        if (prep) {
            PyObject *pargs = PyTuple_Pack(2, name, bases);
            if (unlikely(!pargs)) {
                Py_DECREF(prep);
                return NULL;
            }
            ns = PyObject_Call(prep, pargs, mkw);
            Py_DECREF(prep);
            Py_DECREF(pargs);
        } else {
            if (unlikely(!PyErr_ExceptionMatches(PyExc_AttributeError)))
                return NULL;
            PyErr_Clear();
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result, *margs;
    PyObject *owned_metaclass = NULL;
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    margs = PyTuple_Pack(3, name, bases, dict);
    if (unlikely(!margs)) {
        result = NULL;
    } else {
        result = PyObject_Call(metaclass, margs, mkw);
        Py_DECREF(margs);
    }
    Py_XDECREF(owned_metaclass);
    return result;
}

/* FetchCommonType */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunction */
#include <structmember.h>
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromString(m->func.m_ml->ml_name);
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_New(PyTypeObject *type, PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    __pyx_CyFunctionObject *op = PyObject_GC_New(__pyx_CyFunctionObject, type);
    if (op == NULL)
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    PyObject_GC_Track(op);
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_err_code(err_code value) {
    const err_code neg_one = (err_code) ((err_code) 0 - (err_code) 1), const_zero = (err_code) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(err_code) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(err_code) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(err_code) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(err_code) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(err_code) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(err_code),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_char(char value) {
    const char neg_one = (char) ((char) 0 - (char) 1), const_zero = (char) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(char) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(char) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(char) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(char) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(char) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(char),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* Declarations */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = (float)(1.0) / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = (float)(1.0) / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__NPY_TYPES(enum NPY_TYPES value) {
    const enum NPY_TYPES neg_one = (enum NPY_TYPES) ((enum NPY_TYPES) 0 - (enum NPY_TYPES) 1), const_zero = (enum NPY_TYPES) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum NPY_TYPES) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum NPY_TYPES) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum NPY_TYPES) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum NPY_TYPES) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum NPY_TYPES) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum NPY_TYPES),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
    const size_t neg_one = (size_t) ((size_t) 0 - (size_t) 1), const_zero = (size_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* CoroutineBase */
#include <structmember.h>
#include <frameobject.h>
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (Py_TYPE(ev) == (PyTypeObject*)PyExc_StopIteration) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(CYTHON_UNUSED __pyx_CoroutineObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(CYTHON_UNUSED PyObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(CYTHON_UNUSED PyObject *gen, PyObject *value, CYTHON_UNUSED int closing) {
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_type) {
        #if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
        #else
        if (exc_state->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_state->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            Py_XINCREF(tstate->frame);
            assert(f->f_back == NULL);
            f->f_back = tstate->frame;
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body((PyObject *) self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *exc_tb = exc_state->exc_traceback;
    if (likely(exc_tb)) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
#endif
    }
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(CYTHON_UNUSED PyObject* gen, PyObject *retval) {
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = Py_TYPE(yf)->tp_iternext(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                PyErr_WriteUnraisable(yf);
            }
            PyErr_Clear();
        } else {
            retval = PyObject_CallFunction(meth, NULL);
            Py_DECREF(meth);
            if (!retval)
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = Py_TYPE(yf)->tp_iternext(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, CYTHON_UNUSED PyObject *arg) {
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                    gen->is_running = 0;
                    return NULL;
                }
                PyErr_Clear();
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = PyObject_CallObject(meth, args);
            } else {
                ret = PyObject_CallFunctionObjArgs(meth, typ, val, tb, NULL);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (PyObject_CallFinalizerFromDealloc(self))
#else
        Py_TYPE(gen)->tp_del(self);
        if (self->ob_refcnt > 0)
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    self->ob_refcnt = 1;
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt > 0);
    if (--self->ob_refcnt == 0) {
        return;
    }
    {
        Py_ssize_t refcnt = self->ob_refcnt;
        _Py_NewReference(self);
        self->ob_refcnt = refcnt;
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(self->ob_type) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_name;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_name;
    Py_INCREF(value);
    self->gi_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_qualname;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_qualname;
    Py_INCREF(value);
    self->gi_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
#ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (!module) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {0, 0, 0, 0, 0}
};
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
};
static int __pyx_Generator_init(void) {
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
